#if defined(CYGPKG_KERNEL)
#include <cyg/kernel/kapi.h>
#endif
#include <cyg/io/i2c.h>
#include <cyg/infra/diag.h>
#include <cyg/hal/hal_intr.h>
#include <cyg/hal/iic.h>

#define Trace()  diag_printf("%s: %d\n",__FILE__,__LINE__)

//======================================================================
// Bit-bang driver routines for DM2 I2C serial EEPROM

#define RD_32_DIRECT( ADDR        )  __builtin_ldwio(   (void*)(ADDR) )
#define RD_16_DIRECT( ADDR        )  __builtin_ldhuio(  (void*)(ADDR) )
#define RD_8_DIRECT(  ADDR        )  __builtin_ldbuio(  (void*)(ADDR) )

#define WR_32_DIRECT( ADDR, DATA  )  __builtin_stwio(  (void*)(ADDR), (DATA) )
#define WR_16_DIRECT( ADDR, DATA  )  __builtin_sthio(  (void*)(ADDR), (DATA) )
#define WR_8_DIRECT(  ADDR, DATA  )  __builtin_stbio(  (void*)(ADDR), (DATA) )

// base addresses of NIOS PIO devices (1-bit wide)
#define Sck 0x2001900
#define Sda 0x2001910

// macros to manipulate bit 0 of PIO devices
#define BitIn(base)     WR_32_DIRECT(base+4,0)
#define BitOut(base)    WR_32_DIRECT(base+4,1)
#define BitSet(base)    WR_32_DIRECT(base,1)
#define BitClr(base)    WR_32_DIRECT(base,0)
#define BitRead(base)   (RD_32_DIRECT(base) & 1)
#define WrIntMask(base,val) WR_32_DIRECT(base+8,val)

char* transactionName[] =
{
  "CYG_I2C_BITBANG_INIT",  //                    = 0,
  "CYG_I2C_BITBANG_SCL_HIGH",  //                = 1,
  "CYG_I2C_BITBANG_SCL_LOW",  //                 = 2,
  "CYG_I2C_BITBANG_SCL_HIGH_CLOCKSTRETCH",  //   = 3,
  "CYG_I2C_BITBANG_SCL_LOW_SDA_INPUT",  //       = 4,
  "CYG_I2C_BITBANG_SDA_OUTPUT",  //              = 5,
  "CYG_I2C_BITBANG_SDA_HIGH",  //                = 6,
  "CYG_I2C_BITBANG_SDA_LOW",  //                 = 7,
  "CYG_I2C_BITBANG_SDA_READ"   //                = 8
};

static cyg_bool dm2_i2c_bitbang(cyg_i2c_bus* bus, cyg_i2c_bitbang_op op)
{
  cyg_bool result = 0;
  //diag_printf("bitbang %p: %d %s\n",bus,op,transactionName[op]);
  switch(op) 
    {
     case CYG_I2C_BITBANG_INIT:
      WrIntMask(Sck,0);
      WrIntMask(Sda,0);
      BitSet(Sck);
      BitSet(Sda);
      BitOut(Sck);
      BitOut(Sda);
      diag_printf("dm2 i2c init done\n");
      break;
     case CYG_I2C_BITBANG_SCL_HIGH:
      BitSet(Sck);
      break;
     case CYG_I2C_BITBANG_SCL_LOW:
      BitClr(Sck);
      break;
     case CYG_I2C_BITBANG_SCL_HIGH_CLOCKSTRETCH:
      BitSet(Sck);
      break;
     case CYG_I2C_BITBANG_SCL_LOW_SDA_INPUT:
      BitClr(Sck);
      BitIn(Sda);
      break;
     case CYG_I2C_BITBANG_SDA_OUTPUT:
      BitOut(Sda);
      break;
     case CYG_I2C_BITBANG_SDA_HIGH:
      BitSet(Sda);
      break;
     case CYG_I2C_BITBANG_SDA_LOW:
      BitClr(Sda);
      break;
     case CYG_I2C_BITBANG_SDA_READ:
      result = BitRead(Sda);
      break;
    }
  return result;
}

//======================================================================
// Declare bus and device

// on DMA platform due to overhead in bit-bang drive,
// specifying a delay of 0ns results in a clock period of
// about 50us (20KHz).

#define EepromAddr 0x50

CYG_I2C_BITBANG_BUS(dm2_i2c_bus, &dm2_i2c_bitbang);
CYG_I2C_DEVICE(dm2_i2c_eeprom, &dm2_i2c_bus, EepromAddr, 0x00, 0);

//======================================================================
// High level routines to read/write serial EEPROM

void eepromWait(cyg_uint8 data)
{
  externC cyg_ucount32 cyg_scheduler_read_lock(void);  
  unsigned n,i=100;
  while (i--)
    {
      n = cyg_i2c_transaction_tx(&dm2_i2c_eeprom,true,&data,1,true);
      if (n==1)
        return;
#if !defined(CYGPKG_KERNEL)
      HAL_DELAY_US(1000);
#else      
      if (cyg_scheduler_read_lock()==0)
        cyg_thread_delay(1);
      else
        HAL_DELAY_US(1000);
#endif
    }
  diag_printf("eepromWait: timed out\n");
}

unsigned eepromWrite(unsigned addr, void *p, unsigned count)
{
  cyg_uint8 *buf = (cyg_uint8*)p;
  cyg_uint8 txbuf[2];
  unsigned tries,written;
  
  txbuf[0] = addr;
  
  written = 0;
  tries = 0;
  
  cyg_i2c_transaction_begin(&dm2_i2c_eeprom);
  while (count)
    {
      eepromWait(0);
      txbuf[1] = *buf;
      if (cyg_i2c_transaction_tx(&dm2_i2c_eeprom, true, txbuf, 2, true) == 2)
        {
          // write successful, next byte...
          ++written;
          ++buf;
          ++txbuf[0]; // addr
          --count;
        }
      else
        {
          diag_printf("eepromWrite failure: ");
          ++tries;
          if (tries < 4)
            diag_printf("retrying\n");
          else
            {
              diag_printf("aborting\n");
              break;
            }
        }
    }
  eepromWait(0);
  cyg_i2c_transaction_end(&dm2_i2c_eeprom);
  return written;
}

unsigned eepromRead(unsigned addr, void *p, unsigned count)
{
  cyg_uint8 *buf = (cyg_uint8*)p;
  cyg_uint8 a = addr;
  unsigned n;
  
  cyg_i2c_transaction_begin(&dm2_i2c_eeprom);
  if (cyg_i2c_transaction_tx(&dm2_i2c_eeprom, true, &a, 1, false) != 1)
    diag_printf("eepromRead failed to set addr\n");
  n = cyg_i2c_transaction_rx(&dm2_i2c_eeprom, true, buf, count, true, true);
  if (n != count)
    diag_printf("eepromRead faild to read data\n");
  cyg_i2c_transaction_end(&dm2_i2c_eeprom);
  return n;
}

//======================================================================
// Wrapper functions used to get/set "variables" in EEPROM


//  Macros which define addresses (decimal) and sizes (bytes)
//  of IIC variables 

#define MAC_ADDRESS_ADDR        0
#define SERNUM_PREFIX_DM2_ADDR  6
#define BOARD_REV_ADDR          8
#define MODEL_NUMBER_ADDR       12
#define IP_ADDRESS_ADDR         16
#define NETMASK_ADDR            20
#define GATEWAY_ADDR            24
#define SERNUM_DM2_ADDR         28
#define PASSWORD_ADDR           32
#define AUTH_METHOD_ADDR        48
#define SERVICE_ENABLE_ADDR     49
#define TIMEOUT_ADDR            50
#define TELNET_TIMEOUT_ADDR     52
#define SERNUM_PREFIX_ADDR      56
#define SERNUM_ADDR             58
#define APP_ENABLE_KEY_ADDR     60

#define MAC_ADDRESS_SIZE        6
#define SERNUM_PREFIX_DM2_SIZE  2
#define BOARD_REV_SIZE          4
#define MODEL_NUMBER_SIZE       4
#define IP_ADDRESS_SIZE         4
#define NETMASK_SIZE            4
#define GATEWAY_SIZE            4
#define SERNUM_DM2_SIZE         4
#define PASSWORD_SIZE           16
#define AUTH_METHOD_SIZE        1
#define SERVICE_ENABLE_SIZE     1
#define TIMEOUT_SIZE            1
#define TELNET_TIMEOUT_SIZE     4
#define SERNUM_PREFIX_SIZE      2
#define SERNUM_SIZE             2
#define APP_ENABLE_KEY_SIZE     1

/* Bit map of service enable byte (address: 49) */
#define SERVICE_TELNET          0x01
#define SERVICE_SECURE_DATA     0x02
#define SERVICE_SNMP            0x04
#define SERVICE_SECURE_CONFIG   0x08

void getMacAddress(unsigned char *p)
{
  eepromRead(MAC_ADDRESS_ADDR, p, MAC_ADDRESS_SIZE);
}
void putMacAddress(unsigned char *p)
{
  eepromWrite(MAC_ADDRESS_ADDR,     p,     4);    
  eepromWrite(MAC_ADDRESS_ADDR + 4, p + 4, 2);  
}

void getBoardRev(unsigned long *p)
{
  eepromRead(BOARD_REV_ADDR, p, BOARD_REV_SIZE);
}
void putBoardRev(unsigned long p)
{
  eepromWrite(BOARD_REV_ADDR, &p, BOARD_REV_SIZE); 
}

void getModelNumber(unsigned long *m)
{
  eepromRead(MODEL_NUMBER_ADDR, m, MODEL_NUMBER_SIZE);
}
void putModelNumber(unsigned long m)
{
  eepromWrite(MODEL_NUMBER_ADDR, &m, MODEL_NUMBER_SIZE); 
}

void putIpInfo(unsigned long ipAddr, unsigned long netMask, unsigned long gateway)
{
  eepromWrite(IP_ADDRESS_ADDR, &ipAddr, IP_ADDRESS_SIZE);  
  eepromWrite(NETMASK_ADDR, &netMask, NETMASK_SIZE); 
  eepromWrite(GATEWAY_ADDR, &gateway, GATEWAY_SIZE); 
}
void getIpInfo(unsigned long *ipAddr, unsigned long *netMask, unsigned long *gateway)
{
  eepromRead(IP_ADDRESS_ADDR, ipAddr, IP_ADDRESS_SIZE);
  eepromRead(NETMASK_ADDR, netMask, NETMASK_SIZE);
  eepromRead(GATEWAY_ADDR, gateway, GATEWAY_SIZE);
}

void getPassword(char *p)
{
  eepromRead(PASSWORD_ADDR, p, PASSWORD_SIZE);
  p[15] = '\0';
}
void putPassword(char *p)
{
  p[PASSWORD_SIZE - 1] = '\0';
  eepromWrite(PASSWORD_ADDR,      p,      4); 
  eepromWrite(PASSWORD_ADDR + 4,  p + 4,  4); 
  eepromWrite(PASSWORD_ADDR + 8,  p + 8,  4); 
  eepromWrite(PASSWORD_ADDR + 12, p + 12, 4); 
}

void getAuthMethod(unsigned char *p)
{
  eepromRead(AUTH_METHOD_ADDR, p, AUTH_METHOD_SIZE);
}

void putAuthMethod(unsigned char m)
{
  eepromWrite(AUTH_METHOD_ADDR, &m, AUTH_METHOD_SIZE); 
}

int getServiceEnable(unsigned char service)
{
  unsigned char service_enable;
  eepromRead(SERVICE_ENABLE_ADDR, &service_enable, SERVICE_ENABLE_SIZE);
  return (service_enable & service) != 0;
}
void putServiceEnable(unsigned char service, unsigned char enable)
{
  unsigned char service_enable;
  eepromRead(SERVICE_ENABLE_ADDR, &service_enable, SERVICE_ENABLE_SIZE);
  service_enable = (enable) ? (service_enable | service) :
                              (service_enable & ~service);
  eepromWrite(SERVICE_ENABLE_ADDR, &service_enable, SERVICE_ENABLE_SIZE);
}

void getTelnetEnable(unsigned char *p)
{
  *p = getServiceEnable(SERVICE_TELNET);
}
void putTelnetEnable(unsigned char m)
{
  putServiceEnable(SERVICE_TELNET, m);
}

void getSecureDataEnable(unsigned char *p)
{
  *p = getServiceEnable(SERVICE_SECURE_DATA);
}
void putSecureDataEnable(unsigned char m)
{
  putServiceEnable(SERVICE_SECURE_DATA, m);
}
void getSecureConfigEnable(unsigned char *p)
{
  *p = getServiceEnable(SERVICE_SECURE_CONFIG);
}
void putSecureConfigEnable(unsigned char m)
{
  putServiceEnable(SERVICE_SECURE_CONFIG, m);
}

void getSnmpEnable(unsigned char *p)
{
  *p = getServiceEnable(SERVICE_SNMP);
}
void putSnmpEnable(unsigned char m)
{
  putServiceEnable(SERVICE_SNMP, m);
}

void putRedbootDelay(unsigned char t)
{
  eepromWrite(TIMEOUT_ADDR, &t, TIMEOUT_SIZE); 
}
void getRedbootDelay(unsigned char *p)
{
  eepromRead(TIMEOUT_ADDR, p, TIMEOUT_SIZE);
}

void putTelnetTimeout(unsigned long t)
{
  eepromWrite(TELNET_TIMEOUT_ADDR, &t, TELNET_TIMEOUT_SIZE); 
}
void getTelnetTimeout(unsigned long *p)
{
  eepromRead(TELNET_TIMEOUT_ADDR, p, TELNET_TIMEOUT_SIZE);
}

void putSerNumPrefix(unsigned short p)
{
  eepromWrite(SERNUM_PREFIX_ADDR, &p, SERNUM_PREFIX_SIZE); 
}
void getSerNumPrefix(unsigned short *p)
{
  eepromRead(SERNUM_PREFIX_ADDR, p, SERNUM_PREFIX_SIZE);
}

void putSerNum(unsigned short p)
{
  eepromWrite(SERNUM_ADDR, &p, SERNUM_SIZE); 
}
void getSerNum(unsigned short *p)
{
  eepromRead(SERNUM_ADDR, p, SERNUM_SIZE);
}

void getAppEnableKey(unsigned char *p)
{
  eepromRead(APP_ENABLE_KEY_ADDR, p, APP_ENABLE_KEY_SIZE);
}

#if defined(CYGPKG_HAL_NIOS2)
void putSerNumPrefixDM2(unsigned short p)
{
  eepromWrite(SERNUM_PREFIX_DM2_ADDR, &p, SERNUM_PREFIX_DM2_SIZE); 
}
void getSerNumPrefixDM2(unsigned short *p)
{
  eepromRead(SERNUM_PREFIX_DM2_ADDR, p, SERNUM_PREFIX_DM2_SIZE);
}

void putSerNumDM2(unsigned long p)
{
  eepromWrite(SERNUM_DM2_ADDR, &p, SERNUM_DM2_SIZE);
}
void getSerNumDM2(unsigned long *p)
{
  eepromRead(SERNUM_DM2_ADDR, p, SERNUM_DM2_SIZE);
}
#endif
