#include <cyg/infra/cyg_type.h>
#include <cyg/infra/diag.h>
#if defined(CYGPKG_REDBOOT)
#include <redboot.h>
#include <cyg/hal/hal_intr.h>
#endif

#include "unity.h"

#define Trace()  diag_printf("%s: %d\n",__FILE__,__LINE__)

// AIOPIC channel registers

#define TxFifoSize 512
#define RxFifoSize 512

struct ChannelRegisters
{
  cyg_uint32  data32;
  cyg_uint16  data16;
  cyg_uint16  __unused0;
  union
    {
      cyg_uint16 rxDataStatus;
      cyg_uint8  txData8;
    };
  cyg_uint16  __unused2;
  cyg_uint16  rxFifoCount;
  cyg_uint16  txFifoCount;
  cyg_uint16  chanStatus;
  cyg_uint16  intID;
  union
    {
      cyg_uint32  txRxControl;
      struct 
        {
          cyg_uint16  rxControl;
          cyg_uint16  txControl;
        };
    };
  cyg_uint32  uartControl;
  cyg_uint32  baud;
  cyg_uint32  icTimeout;
  cyg_uint16  priTx16;
  cyg_uint16  __unused3;
  cyg_uint8   priTx8;
  cyg_uint8   __unused4[3];
  cyg_uint32  priorityFifoBypass;
  cyg_uint32  uncondTx;
  cyg_uint32  __unused[3];
} __attribute__((packed));

typedef struct ChannelRegisters tChanReg;

// AIOPIC 16-port memory map

typedef cyg_uint8 tMicrocode[0x80];

struct Aiopic
{
  tChanReg      chan[16];
  tMicrocode    microcode[16];
  cyg_uint32    prescaler;
  cyg_uint32    intStatus;
  cyg_uint32    intMask;
  cyg_uint32    clockEnable;
  cyg_uint32    __unused[(0xD00-0xC10)/4];
} __attribute__((packed));

typedef struct Aiopic tAiopic;

static tAiopic *ap = (tAiopic*)(0x4000000);

const static unsigned char uCode[] = 
{
  0xF6, 0x8B, 0x9D, 0xC4,     /*  0x00 */
  0x13, 0xC4, 0x11, 0x98,     /*  0x04 */
  0x20, 0x0A, 0x21, 0x0A,     /*  0x08 */
  0x8D, 0xFA, 0x86, 0x01,     /*  0x0c */
  0x40, 0x11, 0x19, 0x0A,     /*  0x10 */
  0x00, 0x40, 0x13, 0x19,     /*  0x14 */
  0x0A, 0xFA, 0x83, 0x01,     /*  0x18 */
  0x0A, 0x00, 0x0A, 0x41,     /*  0x1C */
  0xFF, 0x89, 0xC2, 0x66,     /*  0x20 */
  0x86, 0x81, 0x91, 0x40,     /*  0x24 */
  0x65, 0x8E, 0x89, 0xC2,     /*  0x28 */
  0x66, 0x86, 0x81, 0x88,     /*  0x2C */
  0x40, 0x65, 0x85, 0x84,     /*  0x30 */
  0x00, 0x82, 0x0A, 0x08,     /*  0x34 */
  0x0A, 0x0A, 0x0A, 0x0A,     /*  0x38 */
  0x0A, 0x08, 0x0A            /*  0x3C */
};


//======================================================================
// define some commands that manipulate UNITY


#define SETBITW(addr, mask)  WRITEW((READW(addr)|mask), addr)
#define CLRBITW(addr, mask)  WRITEW((READW(addr)&~mask), addr)
#define SETBITL(addr, mask)  WRITEL((READL(addr)|mask), addr)
#define CLRBITL(addr, mask)  WRITEL((READL(addr)&~mask), addr)

#define UnityClock 44236800L
#define UnityPrescale 0
#define UnityBaudDivisor(baud) (((UnityClock/(16*(UnityPrescale+1)))/baud)-1)

//======================================================================
// accessor routines copied/adapted from driver code

static void uSetClockPrescale(unsigned long prescale)
{
  WRITEL(prescale & 0xFF, &ap->prescaler);
}

static void uChannelReset(tChanReg  *cp)
{
  SETBITL(&cp->uartControl, RESET_CHANNEL);
}

static void uSetBaudDivisor(tChanReg  *cp, unsigned long divisor)
{
  WRITEW(divisor, &cp->baud);
}

static void uSetIntercharValue(tChanReg  *cp, unsigned long val)
{
  WRITEL(val, &cp->icTimeout);
}

static void uFlushTxFifo(tChanReg  *cp)
{
  SETBITL(&cp->uartControl, FLUSH_XMIT_FIFO);
  CLRBITL(&cp->uartControl, FLUSH_XMIT_FIFO);
}

static void uFlushRxFifo(tChanReg  *cp)
{
  SETBITL(&cp->uartControl, FLUSH_RCV_FIFO);
  CLRBITL(&cp->uartControl, FLUSH_RCV_FIFO);
}

static void uFlushPriorityFifo(tChanReg  *cp)
{
  SETBITL(&cp->uartControl, FLUSH_PRIORITY_FIFO);
  CLRBITL(&cp->uartControl, FLUSH_PRIORITY_FIFO);
}

unsigned short uGetTxControl(tChanReg  *cp)
{
  return READW(&cp->txControl);
}

static void uSendBreak(tChanReg  *cp)
{
  SETBITL(&cp->txRxControl, BREAK_ON);
}

static void uClrBreak(tChanReg  *cp)
{
  CLRBITL(&cp->txRxControl, BREAK_ON);
}

static void uSetDtr(tChanReg  *cp)
{
  SETBITL(&cp->txRxControl, SET_DTR_ACTIVE);
}

static void uClrDtr(tChanReg  *cp)
{
  CLRBITL(&cp->txRxControl, SET_DTR_ACTIVE);
}

static void uSetRts(tChanReg  *cp)
{
  SETBITL(&cp->txRxControl, SET_RTS_ACTIVE);
}

static void uClrRts(tChanReg  *cp)
{
  CLRBITL(&cp->txRxControl, SET_RTS_ACTIVE);
}

static void uEnCtsFlowCtrl(tChanReg  *cp)
{
  SETBITL(&cp->txRxControl, CTS_FLOW_EN);
}

static void uDisCtsFlowCtrl(tChanReg  *cp)
{
  CLRBITL(&cp->txRxControl, CTS_FLOW_EN);
}

static void uEnRtsFlowCtrl(tChanReg  *cp)
{
  SETBITL(&cp->txRxControl, RTS_FLOW_EN);
}

static void uDisRtsFlowCtrl(tChanReg  *cp)
{
  CLRBITL(&cp->txRxControl, RTS_FLOW_EN);
}

static void uEnParity(tChanReg  *cp)
{
  SETBITL(&cp->txRxControl, PARITY_ON);
}

static void uDisParity(tChanReg  *cp)
{
  CLRBITL(&cp->txRxControl, PARITY_ON);
}

static void uEnForcedParity(tChanReg  *cp)
{
  SETBITL(&cp->txRxControl, FORCE_PARITY);
}

static void uDisForcedParity(tChanReg  *cp)
{
  CLRBITL(&cp->txRxControl, FORCE_PARITY);
}

static void uEnRTSToggle(tChanReg  *cp)
{
  SETBITL(&cp->txRxControl, RTS_TOGGLE);
}

static void uDisRTSToggle(tChanReg  *cp)
{
  CLRBITL(&cp->txRxControl, RTS_TOGGLE);
}


/**********************************************************************
 *  Start/stop the Receive Processor.  Procedure is to have the processor stopped, 
 *  load ucode, then start the processor.  To subsequently disable the recieve 
 *  processor, don't use this bit.  Instead use the uDisRxFIFO()/uEnRxFIFO() functions
 *  which uses the microcode flag.
 *
 *  Background:  Clearing this flag causes the received bytes to be handled by
 *  a 16550 UART core rather than the receive processor.  This only will work for
 *  channel 0, and should never be used by the drivers.  It is for ASIC debug only.
 *  Setting this bit causes received serial bytes to be handled by the receive processor.
 */
static void uStartRxProcessor(tChanReg  *cp)
{
    SETBITL(&cp->txRxControl, RCV_PROC_EN);
}

static void uStopRxProcessor(tChanReg  *cp)
{
    CLRBITL(&cp->txRxControl, RCV_PROC_EN);
}

static void uEnRxFIFO(tChanReg  *cp)
{                                               
    WRITEB(URP_RCV_FIFO_ENABLE, ap->microcode[cp - ap->chan] + URP_RCV_FIFO);
}

static void uDisRxFIFO(tChanReg  *cp)
{
    WRITEB(URP_RCV_FIFO_DISABLE, ap->microcode[cp - ap->chan] + URP_RCV_FIFO);
}

static void uEnTransmit(tChanReg  *cp)
{
    SETBITL(&cp->txRxControl, XMIT_EN);
}

static void uDisTransmit(tChanReg  *cp)
{
    CLRBITL(&cp->txRxControl, XMIT_EN);
}

static void uEnStripNulls(tChanReg  *cp)
{
    WRITEB(URP_STRIP_NULLS_ENABLE, ap->microcode[cp - ap->chan] + URP_STRIP_NULLS);
}

static void uDisStripNulls(tChanReg  *cp)
{
    WRITEB(URP_STRIP_NULLS_DISABLE, ap->microcode[cp - ap->chan] + URP_STRIP_NULLS);
}

static void uEnSoftFlowCtrl(tChanReg  *cp)
{
    WRITEB(URP_TX_SW_FLOW_CONTROL_ENABLE, ap->microcode[cp - ap->chan] + URP_TX_SW_FLOW_CONTROL);
    WRITEB(URP_RX_SW_FLOW_CONTROL_ENABLE, ap->microcode[cp - ap->chan] + URP_RX_SW_FLOW_CONTROL);
}

static void uDisSoftFlowCtrl(tChanReg  *cp)
{
    WRITEB(URP_TX_SW_FLOW_CONTROL_DISABLE, ap->microcode[cp - ap->chan] + URP_TX_SW_FLOW_CONTROL);
    WRITEB(URP_RX_SW_FLOW_CONTROL_DISABLE, ap->microcode[cp - ap->chan] + URP_RX_SW_FLOW_CONTROL);
}

static void uSetMask(tChanReg  *cp, unsigned char mask)
{
    WRITEB(mask, ap->microcode[cp - ap->chan] + URP_MASK);
}

// kdr  Perhaps this should be moved into the application...
/*  Defines used in the Receive Processor Compare function */
#define COMPARE_SAVE_BYTE    0
#define COMPARE_REPLACE_BYTE 1
#define COMPARE_DISCARD_BYTE 2

static void uEnCompare1(tChanReg  *cp, unsigned char compareAction, unsigned char compareChar, unsigned char replaceChar)
{
    switch (compareAction) {
    case COMPARE_DISCARD_BYTE:
	WRITEB(compareChar, ap->microcode[cp - ap->chan] + URP_COMPARE_BYTE1);
	WRITEB(URP_COMPARE_ACTION1_DISCARD, ap->microcode[cp - ap->chan] + URP_COMPARE_ACTION1);
	WRITEB(URP_COMPARE1_ENABLE, ap->microcode[cp - ap->chan] + URP_COMPARE1);
	WRITEB(URP_COMPARE_IRQ1_ENABLE, ap->microcode[cp - ap->chan] + URP_COMPARE_IRQ1);
	break;
    case COMPARE_REPLACE_BYTE:
	WRITEB(compareChar, ap->microcode[cp - ap->chan] + URP_COMPARE_BYTE1);
	WRITEB(replaceChar, ap->microcode[cp - ap->chan] + URP_REPLACE_BYTE1);
	WRITEB(URP_COMPARE_ACTION1_REPLACE, ap->microcode[cp - ap->chan] + URP_COMPARE_ACTION1);
	WRITEB(URP_COMPARE1_ENABLE, ap->microcode[cp - ap->chan] + URP_COMPARE1);
	WRITEB(URP_COMPARE_IRQ1_ENABLE, ap->microcode[cp - ap->chan] + URP_COMPARE_IRQ1);
	break;
    case COMPARE_SAVE_BYTE:
    default: 
	WRITEB(compareChar, ap->microcode[cp - ap->chan] + URP_COMPARE_BYTE1);
	WRITEB(URP_COMPARE_ACTION1_SAVE, ap->microcode[cp - ap->chan] + URP_COMPARE_ACTION1);
	WRITEB(URP_COMPARE1_ENABLE, ap->microcode[cp - ap->chan] + URP_COMPARE1);
	WRITEB(URP_COMPARE_IRQ1_ENABLE, ap->microcode[cp - ap->chan] + URP_COMPARE_IRQ1);
	break;
    }
}

static void uEnCompare2(tChanReg  *cp, unsigned char compareAction, unsigned char compareChar, unsigned char replaceChar)
{
    switch (compareAction) {
    case COMPARE_DISCARD_BYTE:
	WRITEB(compareChar, ap->microcode[cp - ap->chan] + URP_COMPARE_BYTE2);
	WRITEB(URP_COMPARE_ACTION2_DISCARD, ap->microcode[cp - ap->chan] + URP_COMPARE_ACTION2);
	WRITEB(URP_COMPARE2_ENABLE, ap->microcode[cp - ap->chan] + URP_COMPARE2);
	WRITEB(URP_COMPARE_IRQ2_ENABLE, ap->microcode[cp - ap->chan] + URP_COMPARE_IRQ2);
	break;
    case COMPARE_REPLACE_BYTE:
	WRITEB(compareChar, ap->microcode[cp - ap->chan] + URP_COMPARE_BYTE2);
	WRITEB(replaceChar, ap->microcode[cp - ap->chan] + URP_REPLACE_BYTE2);
	WRITEB(URP_COMPARE_ACTION2_REPLACE, ap->microcode[cp - ap->chan] + URP_COMPARE_ACTION2);
	WRITEB(URP_COMPARE2_ENABLE, ap->microcode[cp - ap->chan] + URP_COMPARE2);
	WRITEB(URP_COMPARE_IRQ2_ENABLE, ap->microcode[cp - ap->chan] + URP_COMPARE_IRQ2);
	break;
    case COMPARE_SAVE_BYTE:
    default: 
	WRITEB(compareChar, ap->microcode[cp - ap->chan] + URP_COMPARE_BYTE2);
	WRITEB(URP_COMPARE_ACTION2_SAVE, ap->microcode[cp - ap->chan] + URP_COMPARE_ACTION2);
	WRITEB(URP_COMPARE2_ENABLE, ap->microcode[cp - ap->chan] + URP_COMPARE2);
	WRITEB(URP_COMPARE_IRQ2_ENABLE, ap->microcode[cp - ap->chan] + URP_COMPARE_IRQ2);
	break;
    }
}

static void uDisCompare1(tChanReg  *cp)
{
    WRITEB(URP_COMPARE1_DISABLE, ap->microcode[cp - ap->chan] + URP_COMPARE1);
    WRITEB(URP_COMPARE_IRQ1_DISABLE, ap->microcode[cp - ap->chan] + URP_COMPARE_IRQ1);
}

static void uDisCompare2(tChanReg  *cp)
{
    WRITEB(URP_COMPARE2_DISABLE, ap->microcode[cp - ap->chan] + URP_COMPARE2);
    WRITEB(URP_COMPARE_IRQ2_DISABLE, ap->microcode[cp - ap->chan] + URP_COMPARE_IRQ2);
}

static void uSetTxXOFFChar(tChanReg  *cp, char ch)
{
    WRITEB(ch, ap->microcode[cp - ap->chan] + URP_XMIT_XOFF_CHAR);
}

static void uSetTxXONChar(tChanReg  *cp, char ch)
{
    WRITEB(ch, ap->microcode[cp - ap->chan] + URP_XMIT_XON_CHAR);
}

static void uSetRxXOFFChar(tChanReg  *cp, char ch)
{
    WRITEB(ch, ap->microcode[cp - ap->chan] + URP_RCV_XOFF_CHAR);
}

static void uSetRxXONChar(tChanReg  *cp, char ch)
{
    WRITEB(ch, ap->microcode[cp - ap->chan] + URP_RCV_XON_CHAR);
}

static void uSetStop1(tChanReg  *cp)
{
  CLRBITL(&cp->uartControl, STOP_BITS);
  //diag_printf("in set stop1\n");
}

static void uSetStop2(tChanReg  *cp)
{
  SETBITL(&cp->uartControl, STOP_BITS);
  //diag_printf("in stop2\n");
}


/*************************************************************************
 *  Set TX FIFO trigger level
 *  Inputs:   ChP pointer to channel struct, level sets the number of characters in 
 *            FIFO at which the  interrupt will be generated.  
 *            -  UXMIT_FIFO_TRIG_DISABLED  disabled the XMIT FIFO Trigger
 *            -  UXMIT_FIFO_TRIG_1_8       1/8 full
 *            -  UXMIT_FIFO_TRIG_1_4       1/4 full
 *            -  UXMIT_FIFO_TRIG_1_2       1/2 full
 */
static void uSetTxTrigger(tChanReg  *cp, unsigned long level)
{
    unsigned long aval;

    aval = READL(&cp->txRxControl);  
    aval &= ~UXMIT_FIFO_TRIG_MASK;
    WRITEL(aval | level, &cp->txRxControl);
}

/*************************************************************************
 *  Set RX FIFO trigger level
 *  Inputs:   ChP pointer to channel struct, level sets the number of characters in 
 *            FIFO at which the  interrupt will be generated.  
 *            -  URCV_FIFO_TRIG_DISABLED  disabled the RCV FIFO Trigger
 *            -  URCV_FIFO_TRIG_1_CHAR    1 Character
 *            -  URCV_FIFO_TRIG_1_2       1/2 full
 *            -  URCV_FIFO_TRIG_7_8       7/8 full
 */
static void uSetRxTrigger(tChanReg  *cp, unsigned long level)
{
    unsigned long aval;

    aval = READL(&cp->txRxControl);  
    aval &= ~URCV_FIFO_TRIG_MASK;
    WRITEL(aval | level, &cp->txRxControl);
}

static void uSetEvenParity(tChanReg  *cp)
{
  SETBITL(&cp->txRxControl, EVEN_PARITY);
}

static void uSetOddParity(tChanReg  *cp)
{
  CLRBITL(&cp->txRxControl, EVEN_PARITY);
}

static void uSetDataBits(tChanReg  *cp, unsigned long level)
{
    unsigned long aval;

    aval = READL(&cp->uartControl);  
    aval &= ~DATABITS_MASK;
    WRITEL(aval | level, &cp->uartControl); 
}

static void uSetData5(tChanReg  *cp)
{
    uSetDataBits(cp, DATABITS_5);
}

static void uSetData6(tChanReg  *cp)
{
    uSetDataBits(cp, DATABITS_6);
}

static void uSetData7(tChanReg  *cp)
{
    uSetDataBits(cp, DATABITS_7);
}

static void uSetData8(tChanReg  *cp)
{
    uSetDataBits(cp, DATABITS_8);
}

unsigned short uGetTxCnt(tChanReg  *cp)
{
    return READW(&cp->txFifoCount);
}

unsigned short uGetRxCnt(tChanReg  *cp)
{
    return READW(&cp->rxFifoCount);
}

static void uSetInterfaceMode(tChanReg  *cp, unsigned long mode)
{
    unsigned long aval;

    aval = READL(&cp->uartControl);  
    aval &= ~URS_MASK;
    WRITEL(aval | mode, &cp->uartControl); 
}

static void uEnLoopbackMode(tChanReg  *cp)
{
    SETBITL(&cp->txRxControl, LOCAL_LOOPBACK);
}

static void uDisLoopbackMode(tChanReg  *cp)
{
    CLRBITL(&cp->txRxControl, LOCAL_LOOPBACK);
}

static void uRtsTogglePolarityHigh(tChanReg  *cp)
{
    SETBITL(&cp->txRxControl, RTS_TOGGLE_POLARITY);
}

static void uRtsTogglePolarityLow(tChanReg  *cp)
{
    CLRBITL(&cp->txRxControl, RTS_TOGGLE_POLARITY);
}

static void uEnDsrFlowCtrl(tChanReg  *cp)
{
    SETBITL(&cp->txRxControl, DSR_FLOW_EN);
}

static void uDisDsrFlowCtrl(tChanReg  *cp)
{
    CLRBITL(&cp->txRxControl, DSR_FLOW_EN);
}

static void uEnDtrFlowCtrl(tChanReg  *cp)
{
    SETBITL(&cp->txRxControl, DTR_FLOW_EN);
}

static void uDisDtrFlowCtrl(tChanReg  *cp)
{
    CLRBITL(&cp->txRxControl, DTR_FLOW_EN);
}

static int uGetChanRI(tChanReg *cp)
{
    return (READL(&cp->intID) & CHAN_STATUS_RI_ACTIVE);
}

static void uXmitByte(tChanReg  *cp, unsigned char pval)
{
    WRITEB(pval, &cp->txData8);
}

static void uXmitWord(tChanReg  *cp, unsigned short pval)
{
    WRITEW(pval, &cp->data16);
}

static void uXmitDword(tChanReg  *cp, unsigned long pval)
{
    WRITEL(pval, &cp->data32);
}

static void uXmitBytePriority(tChanReg  *cp, unsigned char pval)
{
    WRITEB(pval, &cp->priTx8);
}

static void uXmitWordPriority(tChanReg  *cp, unsigned short pval)
{
    WRITEW(pval, &cp->priTx16);
}

static void uXmitDwordPriority(tChanReg  *cp, unsigned long pval)
{
    WRITEW((unsigned short)pval, &cp->priTx16);
    WRITEW((unsigned short)(pval>>16), &cp->priTx16);
}

unsigned short uRxByteStatus(tChanReg  *cp)
{
    return READW(&cp->rxDataStatus);
}

unsigned short uRxWord(tChanReg  *cp)
{
    return READW(&cp->data16);
}

unsigned long uRxDword(tChanReg  *cp)
{
    return READL(&cp->data32);
}

static void uEnXmitFifoEmptyInt(tChanReg  *cp)
{
    SETBITL(&cp->txRxControl, XMIT_FIFO_INT_EN);
}

static void uDisXmitFifoEmptyInt(tChanReg  *cp)
{
    CLRBITL(&cp->txRxControl, XMIT_FIFO_INT_EN);
}

static void uEnXmitShiftRegEmptyInt(tChanReg  *cp)
{
    SETBITL(&cp->txRxControl, XMIT_SHIFTREG_INT_EN);
}

static void uDisXmitShiftRegEmptyInt(tChanReg  *cp)
{
    CLRBITL(&cp->txRxControl, XMIT_SHIFTREG_INT_EN);
}

static void uEnInterCharTimeoutInt(tChanReg  *cp)
{
    SETBITL(&cp->txRxControl, INTERCHAR_TIMEOUT_INT_EN);
}

static void uDisInterCharTimeoutInt(tChanReg  *cp)
{
    CLRBITL(&cp->txRxControl, INTERCHAR_TIMEOUT_INT_EN);
}

static void uEnInterCharTimeoutStatus(tChanReg  *cp)
{
    SETBITL(&cp->txRxControl, INTERCHAR_STATUS_EN);
}

static void uDisInterCharTimeoutStatus(tChanReg  *cp)
{
    CLRBITL(&cp->txRxControl, INTERCHAR_STATUS_EN);
}

static void uEnAddRcvdInt(tChanReg  *cp)
{
    SETBITL(&cp->txRxControl, ADDRESS_INT_EN);
}

static void uDisAddRcvdInt(tChanReg  *cp)
{
    CLRBITL(&cp->txRxControl, ADDRESS_INT_EN);
}

static void uEnRcvDataReadyTimeoutInt(tChanReg  *cp)
{
    SETBITL(&cp->txRxControl, RCV_DATA_RDY_INT_EN);
}

static void uDisRcvDataReadyTimeoutInt(tChanReg  *cp)
{
    CLRBITL(&cp->txRxControl, RCV_DATA_RDY_INT_EN);
}

static void uEnSpecialReceiveInt(tChanReg  *cp)
{
    SETBITL(&cp->txRxControl, SPECIAL_RCV_INT_EN);
}

static void uDisSpecialReceiveInt(tChanReg  *cp)
{
    CLRBITL(&cp->txRxControl, SPECIAL_RCV_INT_EN);
}

static void uEnModemChangeInt(tChanReg  *cp)
{
    SETBITL(&cp->txRxControl, MDM_CHANGE_INT_EN);
}

static void uDisModemChangeInt(tChanReg  *cp)
{
    CLRBITL(&cp->txRxControl, MDM_CHANGE_INT_EN);
}

/**************************************************************************
 *  Interrupt identification functions. Bits must be cleared
 * if set by writing a '1' to them.  The special receive is
 * not immediately cleared here, since it has multiple
 * sources. Returns: 1 if yes, else 0 
 */
static int uInterCharTimeoutInt(tChanReg  *cp)
{
    int ret;

    if ((ret = READL(&cp->intID) & INT_INTERCHAR_TIMEOUT))
	WRITEL(INT_INTERCHAR_TIMEOUT, &cp->intID);
    return ret;
}

static int uXmitFifoTriggerInt(tChanReg  *cp)
{
    int ret;

    if ((ret = READL(&cp->intID) & INT_XMIT_FIFO_TRIGGER))
	WRITEL(INT_XMIT_FIFO_TRIGGER, &cp->intID);
    return ret;
}

static int uXmitShiftRegEmptyInt(tChanReg  *cp)
{
    int ret;

    if ((ret = READL(&cp->intID) & INT_XMIT_SHIFTREG_EMPTY))
	WRITEL(INT_XMIT_SHIFTREG_EMPTY, &cp->intID);
    return ret;
}

static int uAddrRcvdInt(tChanReg  *cp)
{
    int ret;

    if ((ret = READL(&cp->intID) & INT_ADDR_RCVD))
	WRITEL(INT_ADDR_RCVD, &cp->intID);
    return ret;
}

static int uRcvDataRdyInt(tChanReg  *cp)
{
    int ret;

    if ((ret = READL(&cp->intID) & INT_RCV_DATA_RDY))
	WRITEL(INT_RCV_DATA_RDY, &cp->intID);
    return ret;
}

static int uIntRiChange(tChanReg  *cp)
{
    int ret;

    if ((ret = READL(&cp->intID) & INT_RI_CHANGE))
	WRITEL(INT_RI_CHANGE, &cp->intID);
    return ret;
}

static int uRcvFifoTriggerInt(tChanReg  *cp)
{
    int ret;

    if ((ret = READL(&cp->intID) & INT_RCV_FIFO_TRIGGER))
	WRITEL(INT_RCV_FIFO_TRIGGER, &cp->intID);
    return ret;
}

static int uXmitFifoEmptyInt(tChanReg  *cp)
{
    int ret;

    if ((ret = READL(&cp->intID) & INT_XMIT_FIFO_EMPTY))
	WRITEL(INT_XMIT_FIFO_EMPTY, &cp->intID);
    return ret;
}

static int uSpecialReceiveInt(tChanReg  *cp)
{
    return (READL(&cp->intID) & INT_SPECIAL_RECEIVE);
}

static void uClrSpecialReceiveInt(tChanReg  *cp)
{
    WRITEL(INT_SPECIAL_RECEIVE, &cp->intID);
}

static int uIntCdChange(tChanReg  *cp)
{
    int ret;

    if ((ret = READL(&cp->intID) & INT_CD_CHANGE))
	WRITEL(INT_CD_CHANGE, &cp->intID);
    return ret;
}

static int uIntCtsChange(tChanReg  *cp)
{
    int ret = 0;

    if ((ret = READL(&cp->intID) & INT_CTS_CHANGE))
	WRITEL(INT_CTS_CHANGE, &cp->intID);
    return ret;
}

static int uIntDsrChange(tChanReg  *cp)
{
    int ret;

    if ((ret = READL(&cp->intID) & INT_DSR_CHANGE))
	WRITEL(INT_DSR_CHANGE, &cp->intID);

    return ret;
}
 
unsigned short uGetChanIntID(tChanReg  *cp)
{
    return READW(&cp->intID);
}

unsigned short uGetChanStatus(tChanReg  *cp)
{
    return READW(&cp->chanStatus);
}

static int uClrChanStatus(tChanReg  *cp)
{
    WRITEL(CHAN_STATUS_RCV_OVERFLOW, &cp->intID);
    WRITEL(CHAN_STATUS_BREAK_DETECTED, &cp->intID);
    WRITEL(CHAN_STATUS_FRAMING_ERROR, &cp->intID);
    WRITEL(CHAN_STATUS_PARITY_ERROR, &cp->intID);
    return 0;
}

static int uPriorityBufferFull(tChanReg  *cp)
{
    return READL(&cp->intID) & CHAN_STATUS_PRIORITY_BUFFER_FULL;
}

static int uChanStatusRcvOverflow(tChanReg  *cp)
{
    int ret;

    if ((ret = READL(&cp->intID) & CHAN_STATUS_RCV_OVERFLOW))
	WRITEL(CHAN_STATUS_RCV_OVERFLOW, &cp->intID);
    return ret;
}

static int uChanStatusCompare2(tChanReg  *cp)
{
    int ret;

    if ((ret = READL(&cp->intID) & CHAN_STATUS_COMPARE2)) 
	WRITEL(CHAN_STATUS_COMPARE2, &cp->intID);
    return ret;
}

static int uChanStatusCompare1(tChanReg  *cp)
{
    int ret;

    if ((ret = READL(&cp->intID) & CHAN_STATUS_COMPARE1))
	WRITEL(CHAN_STATUS_COMPARE1, &cp->intID);
    return ret;
}

static int uChanStatusBreakDetected(tChanReg  *cp)
{
    int ret;

    if ((ret = READL(&cp->intID) & CHAN_STATUS_BREAK_DETECTED))
	WRITEL(CHAN_STATUS_BREAK_DETECTED, &cp->intID);
    return ret;
}

static int uChanStatusFramingError(tChanReg  *cp)
{
    int ret;

    if ((ret = READL(&cp->intID) & CHAN_STATUS_FRAMING_ERROR))
	WRITEL(CHAN_STATUS_FRAMING_ERROR, &cp->intID);
    return ret;
}

static int uChanStatusParityError(tChanReg  *cp)
{
    int ret;

    if ((ret = READL(&cp->intID) & CHAN_STATUS_PARITY_ERROR))
	WRITEL(CHAN_STATUS_PARITY_ERROR, &cp->intID);
    return ret;
}

static int uRiActive(tChanReg  *cp)
{
    return READL(&cp->intID) & CHAN_STATUS_RI_ACTIVE;
}

static int uCtsActive(tChanReg  *cp)
{
    return READL(&cp->intID) & CHAN_STATUS_CTS_ACTIVE;
}

static int uDsrActive(tChanReg  *cp)
{
    return READL(&cp->intID) & CHAN_STATUS_DSR_ACTIVE;
}

static int uCdActive(tChanReg  *cp)
{
    return READL(&cp->intID) & CHAN_STATUS_CD_ACTIVE;
}

static int uXmitFifoEmpty(tChanReg  *cp)
{
    return READL(&cp->intID) &  CHAN_STATUS_XMIT_FIFO_EMPTY;
}

static int uXmitShiftRegEmpty(tChanReg  *cp)
{
    return READL(&cp->intID) &  CHAN_STATUS_XMIT_SHIFT_REG_EMPTY;
}

static int uDtrActive(tChanReg  *cp)
{
    return READL(&cp->txRxControl) & SET_DTR_ACTIVE;
}

static int uRtsActive(tChanReg  *cp)
{
    return READL(&cp->intID) & CHAN_STATUS_RTS_IN_STATUS;
}

static int uRcvDataAvail(tChanReg  *cp)
{
    return READL(&cp->intID) &  CHAN_STATUS_RCV_DATA_AVAIL;
}

static int uGetInterruptStatus(unsigned long port_number)
{
  return READL(&ap->intStatus) & Bit(port_number);
}

static void uEnChannelInterrupt(tChanReg  *cp)
{
  unsigned chan = cp - ap->chan;
  SETBITW(&ap->intMask,Bit(chan));
}

static void uDisChannelInterrupt(tChanReg  *cp)
{
  unsigned chan = cp - ap->chan;
  CLRBITW(&ap->intMask,Bit(chan));
}

/*  ********************************************************************
 *  RCV_DATA_BYTE_STATUS Functions 
 *  Inputs:  The 16 bit byte|status read from RCV_DATA_BYTE_STATUS register (offset 0x0008) 
 *  Returns: 1 if set, else 0
 */
static int uRcvByteParityError(unsigned short bytestatus)
{
    return bytestatus & RCV_BYTE_PARITY_ERROR;
}

static int uRcvByteRcvOverrun(unsigned short bytestatus)
{
    return bytestatus & RCV_BYTE_RCV_OVERRUN;
}

static int uRcvByteFramingError(unsigned short bytestatus)
{
    return bytestatus & RCV_BYTE_FRAMING_ERROR;
}

static int uRcvByteBreakDetected(unsigned short bytestatus)
{
    return bytestatus & RCV_BYTE_BREAK_DETECTED;
}

static int uRcvByteAddressByte(unsigned short bytestatus)
{
    return bytestatus & RCV_BYTE_ADDRESS_BYTE;
}

static int uRcvByteTimeoutDetected(unsigned short bytestatus)
{
    return bytestatus & RCV_BYTE_TIMEOUT_DETECTED;
}

static void uEnXany(tChanReg  *cp)
{
    WRITEB(URP_XANY_ENABLE, ap->microcode[cp - ap->chan] + URP_XANY);
}

static void uDisXany(tChanReg  *cp)
{
    WRITEB(URP_XANY_DISABLE, ap->microcode[cp - ap->chan] + URP_XANY);
}

static void uSetXon(tChanReg  *cp)
{
    SETBITL(&cp->uartControl, RCV_PROC_XMIT_EN);
}

static void uSetXoff(tChanReg  *cp)
{
    CLRBITL(&cp->uartControl, RCV_PROC_XMIT_EN);
}

static int uXonXoffState(tChanReg  *cp)
{
    return READL(&cp->uartControl) & RCV_PROC_XMIT_EN;
}

static void uEnOneStopBit(tChanReg  *cp)
{
    SETBITL(&cp->txRxControl, ONE_STOP_BIT_EN);
}

static void uDisOneStopBit(tChanReg  *cp)
{
    CLRBITL(&cp->txRxControl, ONE_STOP_BIT_EN);
}

static void uGlobalReset(void)
{
  diag_printf("TBD: global reset\n");
  return;
}


static void SetFlowControl(tChanReg *ChP, unsigned flowControl)
{
  switch (flowControl) 
    {
     case FLOW_CONTROL_NONE:
      uDisRtsFlowCtrl(ChP);
      uDisCtsFlowCtrl(ChP);
      uDisDtrFlowCtrl(ChP);
      uDisDsrFlowCtrl(ChP);
      uDisSoftFlowCtrl(ChP);
      break;
          
     case FLOW_CONTROL_RTSCTS:
      uEnRtsFlowCtrl(ChP);
      uEnCtsFlowCtrl(ChP);
      uDisDtrFlowCtrl(ChP);
      uDisDsrFlowCtrl(ChP);
      uDisSoftFlowCtrl(ChP);
      break;
          
     case FLOW_CONTROL_DTRDSR:
      uDisRtsFlowCtrl(ChP);
      uDisCtsFlowCtrl(ChP);
      uEnDtrFlowCtrl(ChP);
      uEnDsrFlowCtrl(ChP);
      uDisSoftFlowCtrl(ChP);
      break;
      
     case FLOW_CONTROL_XONXOFF:
      uDisRtsFlowCtrl(ChP);
      uDisCtsFlowCtrl(ChP);
      uDisDtrFlowCtrl(ChP);
      uDisDsrFlowCtrl(ChP);
      uEnSoftFlowCtrl(ChP);
      break;
    }
}

static void loadMicrocode(unsigned chanIndex)
{
  unsigned j;
  for (j=0; j<sizeof uCode; ++j)
    __builtin_stbio(&ap->microcode[chanIndex][j], uCode[j]);
}


static void initChannel(tChanReg *ChP,unsigned baud)
{
  uChannelReset(ChP);
  uSetBaudDivisor(ChP, UnityBaudDivisor(baud));
  SetFlowControl(ChP, FLOW_CONTROL_NONE);
  uFlushTxFifo(ChP);
  uFlushRxFifo(ChP);
  uFlushPriorityFifo(ChP);
  uEnRxFIFO(ChP);
  uSetData8(ChP);
  uDisParity(ChP);
  uSetStop1(ChP);
  uClrDtr(ChP);
  uClrRts(ChP);
  uSetTxXONChar(ChP,0x11);
  uSetTxXOFFChar(ChP,0x13);
  loadMicrocode(ChP - &ap->chan[0]);
  uStartRxProcessor(ChP);
  uEnTransmit(ChP);

uEnRxFIFO(ChP);
}

static void initChannels(void)
{
  unsigned i;
  uSetClockPrescale(UnityPrescale);
  for (i=0; i<=15; ++i)
    initChannel(&ap->chan[i],57600);
}



void hal_unity_putc(void *__ch_data, char c)
{
  tChanReg *ChP = (tChanReg*)__ch_data;

  CYGARC_HAL_SAVE_GP();
  while ((uGetChanStatus(ChP) & CHAN_STATUS_XMIT_FIFO_EMPTY) == 0)
    ;
  uXmitByte(ChP,c);
  CYGARC_HAL_RESTORE_GP();
}

static void __test_unity_putc(void *__ch_data, char c)
{
  tChanReg *ChP = (tChanReg*)__ch_data;
  while ((uGetChanStatus(ChP) & CHAN_STATUS_XMIT_FIFO_EMPTY) == 0)
    ;
  uXmitByte(ChP,c);
}

// before we start returning data we must see a "wakeup sequence"
// that will enable the RS-232 drivers on the port and set a flag
// to allow subsequent input data to be passed to caller

static cyg_bool hal_unity_getc_nonblock(void* __ch_data, cyg_uint8* p)
{
  static const char wakeupSequence[] = "#!DM";
  static int awake,count,index;
  unsigned char c;
  tChanReg *ChP = (tChanReg*)__ch_data;

  if ((ChP->chanStatus & CHAN_STATUS_RCV_DATA_AVAIL) == 0)
    return false;
  
  c = uRxByteStatus(ChP) & 0xff;
  
  if (awake)
    {
      *p = c;
      return true;
    }
  
  // port not awake yet, watch for wakeup sequence

  ++count;
  if (c == wakeupSequence[index])
    ++index;
  else
    index = 0;
  
  if (index == (sizeof wakeupSequence)-1 && 
      count <= (sizeof wakeupSequence)+2)
    {
      awake = 1;
      uSetInterfaceMode(ChP,URS_RS232);
    }

  return false;
}

cyg_uint8 hal_unity_getc(void* __ch_data)
{
  cyg_uint8 ch;
  
  CYGARC_HAL_SAVE_GP();
  while(!hal_unity_getc_nonblock(__ch_data, &ch))
    do_us_delay(100);
  CYGARC_HAL_RESTORE_GP();
  return ch;
}

static void hal_unity_write(void* __ch_data, 
                             const cyg_uint8* __buf,
                             cyg_uint32 __len)
{
  CYGARC_HAL_SAVE_GP();
  while(__len-- > 0)
    hal_unity_putc(__ch_data, *__buf++);
  CYGARC_HAL_RESTORE_GP();
}



static void hal_unity_read(void* __ch_data, 
                            cyg_uint8* __buf, 
                            cyg_uint32 __len)
{
  CYGARC_HAL_SAVE_GP();
  while(__len-- > 0)
    *__buf++ = hal_unity_getc(__ch_data);
  CYGARC_HAL_RESTORE_GP();
}

static int msec_timeout;

static cyg_bool hal_unity_getc_timeout(void* __ch_data, cyg_uint8* ch)
{
  int delay_count;
  cyg_bool res;
  CYGARC_HAL_SAVE_GP();
  
  delay_count = msec_timeout;
  
  for(;;) 
    {
      res = hal_unity_getc_nonblock(__ch_data, ch);
      if (res || 0 == --delay_count)  // no delay for count of 1
        break;
      do_ms_tick();
    }
  
  CYGARC_HAL_RESTORE_GP();
  return res;
}


#define isr_vector 8 // TBD: is 8 right?

static int hal_unity_control(void *__ch_data, 
                              __comm_control_cmd_t __func, 
                              ...)
{
  static int irq_state = 0;
  int ret = 0;
  CYGARC_HAL_SAVE_GP();
  
  switch (__func) 
    {
     case __COMMCTL_IRQ_ENABLE:
      irq_state = 1;
      HAL_INTERRUPT_UNMASK(isr_vector);
      break;
     case __COMMCTL_IRQ_DISABLE:
      ret = irq_state;
      irq_state = 0;
      HAL_INTERRUPT_MASK(isr_vector);
      break;
     case __COMMCTL_DBG_ISR_VECTOR:
      ret = isr_vector;
      break;
     case __COMMCTL_SET_TIMEOUT:
        {
          va_list ap;
          va_start(ap, __func);
          ret = msec_timeout;
          msec_timeout = va_arg(ap, cyg_uint32);
          va_end(ap);
        }
     default:
      break;
    }
  CYGARC_HAL_RESTORE_GP();
  return ret;
}


static int hal_unity_isr_count;

static int hal_unity_isr(void *__ch_data, 
                                  int* __ctrlc,
                                  CYG_ADDRWORD __vector, 
                                  CYG_ADDRWORD __data)
{
  int res = 0;
  char c;
  CYGARC_HAL_SAVE_GP();
  ++hal_unity_isr_count;
  cyg_drv_interrupt_acknowledge(isr_vector);
  *__ctrlc = 0;
  if (hal_unity_getc_nonblock(__ch_data, &c) && cyg_hal_is_break(&c,1))
    *__ctrlc = 1;
  res = CYG_ISR_HANDLED;
  CYGARC_HAL_RESTORE_GP();
  return res;
}

//  Gets a byte from the Unity uart
//  Inputs:  __ch_data - channel struct
//           p - Gets filled in with the data read
//           cs - Gets filled with channel status before the read
//           s - Gets filled with channel status after the read
//  Returns: 1 if OK, 0 if error
static cyg_bool __test_unity_getc_nb(void* __ch_data, 
				      cyg_uint8* p, 
				      cyg_uint16 *cs, 
				      cyg_uint8 *s)
{
  unsigned short t;
  tChanReg *ChP = (tChanReg*)__ch_data;
  *cs = t = uGetChanStatus(ChP);
  
  if ((t & CHAN_STATUS_RCV_DATA_AVAIL) == 0)
    return false;
  t =  uRxByteStatus(ChP);
  *s = t>>8;
  *p = t;
  return true;
}

static void init_hal_table(int chan)
{
#if defined(CYGPKG_REDBOOT)
  int i;
  hal_virtual_comm_table_t* comm;
  
  HAL_INTERRUPT_MASK(isr_vector);
   
  CYGACC_CALL_IF_SET_CONSOLE_COMM(chan);
  comm = CYGACC_CALL_IF_CONSOLE_PROCS();
  CYGACC_COMM_IF_CH_DATA_SET(*comm, &ap->chan[0]);
  CYGACC_COMM_IF_WRITE_SET(*comm, hal_unity_write);
  CYGACC_COMM_IF_READ_SET(*comm, hal_unity_read);
  CYGACC_COMM_IF_PUTC_SET(*comm, hal_unity_putc);
  CYGACC_COMM_IF_GETC_SET(*comm, hal_unity_getc);
  CYGACC_COMM_IF_CONTROL_SET(*comm, hal_unity_control);
  CYGACC_COMM_IF_DBG_ISR_SET(*comm, hal_unity_isr);
  CYGACC_COMM_IF_GETC_TIMEOUT_SET(*comm, hal_unity_getc_timeout);
#endif
}

static void unitydiag_init(void)
{
  initChannels();
  uSetInterfaceMode(&ap->chan[0], URS_DEBUG);
  init_hal_table(2);
  diag_printf("Unity diag I/O init done.\n");
}

#if defined(CYGPKG_REDBOOT)
RedBoot_init(unitydiag_init,RedBoot_INIT_FIRST);
#endif

#define RiStatus(ch) (!(*RingIndicatorRegister & (1<<ch->chan)))
#define CtsStatus(ch) (!!(*(ch->status) & AiopChanStatus_CTS))
#define DsrStatus(ch) (!!(*(ch->status) & AiopChanStatus_DSR))
#define CdStatus(ch) (!!(*(ch->status) & AiopChanStatus_CD))

// Loopback test on DM2 Unity
//   port: 0-15
//   mode: 0=232, 1=232 RTS inv, 2=422, 3=Internal-loopback
// Return 0 for success, non-zero for failure

static void delay1ms(void)
{
  do_us_delay(1000);
}

int cyg_hal_plf_loopback(int port, int mode, int doRI, int is4530)
{
  tChanReg *ChP = &ap->chan[port];
  static const unsigned char modeBits[4] = {URS_RS232,URS_RS232,URS_RS422,URS_DISABLED};
  int i;
  unsigned char b = modeBits[mode];
  unsigned char s;   // Channel status after the read
  unsigned short cs; // Channel status before the read
  
  initChannel(ChP,9600);
  
  if (mode != 3)
    uSetInterfaceMode(ChP,b);
  
  for (i=0; i<100; ++i)
    delay1ms();

  if (mode == 3)
    uEnLoopbackMode(ChP);
  
  delay1ms();
  
  //  Put 500 bytes in the tx buffer
  for (i=0; i<500; ++i)
    __test_unity_putc(ChP,i&0xff);
  
  // wait a while
  for (i=0; i<200; ++i)
    delay1ms();

  //  Read the 500 bytes back
  for (i=0; i<500; ++i)
    {
      //  Read byte, if none, return channel status from before the read
      if (!__test_unity_getc_nb(ChP,&b,&cs,&s))
        return 0x10000 | cs;
      //  Check channel status after the read
      if (s)
        return 0x20000 | s;
      //  Compare the data
      if ((i&0xff) != b)
        return 0x30000 | ((i&0xff)<<8) | b;
    }
  
  if (__test_unity_getc_nb(ChP,&b,&cs,&s))
    return 0x40000;
  
  if (mode == 0 || mode == 1)
    {
      // test modem control lines
      int rtsState, dtrState;
      unsigned short status;
      for (rtsState = 0; rtsState <=1; ++rtsState)
        for (dtrState = 0; dtrState <=1; ++dtrState)
          {
            if (dtrState) uSetDtr(ChP); else uClrDtr(ChP);
            if (rtsState) uSetRts(ChP); else uClrRts(ChP);
            for (i=0; i<100; ++i) delay1ms();
            status = uGetChanStatus(ChP);
            if (rtsState != !!(status & CHAN_STATUS_CTS_ACTIVE))
              return 0x50000 | status;
            if (doRI && (rtsState != !!(status & CHAN_STATUS_RI_ACTIVE)))
              return 0x60000 | status;
            if (dtrState != !!(status & CHAN_STATUS_DSR_ACTIVE))
              return 0x70000 | status;
            if (dtrState != !!(status & CHAN_STATUS_CD_ACTIVE))
              return 0x80000 | status;
          }
    }
  uDisLoopbackMode(ChP);
  return 0;
}

// Port-to-port RS-485 test.
// Return 0 for success, non-zero for failure
int cyg_hal_plf_485_test(int port1, int port2, int is_1port)
{
  int i;
  tChanReg *ChP1 = &ap->chan[port1];
  tChanReg *ChP2 = &ap->chan[port2];
  unsigned char s,b;
  unsigned short cs;

  if( !is_1port )	//not 1 port
    {
      uSetInterfaceMode(ChP1,URS_RS485_HD);
      uSetInterfaceMode(ChP2,URS_RS485_HD);
      delay1ms();
  
      initChannel(ChP1,9600);
      initChannel(ChP2,9600);

      uEnRTSToggle(ChP1);
      uEnRTSToggle(ChP2);
  
      for (i=0; i<500; ++i)
        __test_unity_putc(ChP1,i&0xff);
      for (i=0; i<200; ++i)
        delay1ms();
      for (i=0; i<500; ++i)
        {
          if (!__test_unity_getc_nb(ChP2,&b,&cs,&s))
            return 0x1210000 | cs;
          if (s)
            return 0x1220000 | s;
          if ((i&0xff) != b)
            return 0x1230000 | ((i&0xff)<<8) | b;
        }
      if (__test_unity_getc_nb(ChP2,&b,&cs,&s))
        return 0x1240000;
  
      for (i=0; i<500; ++i)
        __test_unity_putc(ChP2,i&0xff);
      for (i=0; i<200; ++i)
        delay1ms();
      for (i=0; i<500; ++i)
        {
          if (!__test_unity_getc_nb(ChP1,&b,&cs,&s))
            return 0x2110000 | cs;
          if (s)
            return 0x2120000 | s;
          if ((i&0xff) != b)
            return 0x2130000 | ((i&0xff)<<8) | b;
        }
      if (__test_unity_getc_nb(ChP1,&b,&cs,&s))
        return 0x2140000;
    }
  else	//1 port 
    {
      initChannel(ChP1,115200);	//9600);

      uSetInterfaceMode(ChP1,URS_RS485_TEST);
      delay1ms();
 
      for (i=0; i<500; ++i)
        __test_unity_putc(ChP1,i&0xff);

      for (i=0; i<200; ++i)
        delay1ms();
 
      for (i=0; i<500; ++i)
        {
          if (!__test_unity_getc_nb(ChP1,&b,&cs,&s))
            return 0x1210000 | cs;		//status before read
          if (s)
            return 0x1220000 | s;		//status after read
          if ((i&0xff) != b)
            return 0x1230000 | ((i&0xff)<<8) | b;		//data read
        }
      if (__test_unity_getc_nb(ChP1,&b,&cs,&s))
        return 0x1240000;
    }

  uSetInterfaceMode(ChP1,URS_RS232);
  delay1ms();

  return 0;
}
