#include <pkgconf/hal.h>
#include <pkgconf/system.h>
#include <redboot.h>
#include <net/net.h>
#include <cyg/hal/plf_stub.h>

#if defined(CYGPKG_REDBOOT_NETWORKING)

#include "admin.h"
#include "ident.h"
#include "version.h"

//  If set, enables messages related to MAC mode communications
static int DEBUG_MACMODE = 0;

extern void startAt(unsigned addr);

static void rxMacA(pktbuf_t * pkt, eth_header_t * eth_hdr);

//  Installed to recieve ethernet packets with the protocol
//  type 0x11fe, the unofficial Comtrol protocol type.
void comtrolMacHandler(pktbuf_t * pkt, eth_header_t * eth_hdr)
{
  int byteCount;

  //  Prints out each header recieved.  Will cause binary load to fail, due to delays.
  if (0)
    {
      printf("rxMac header:\n");
//      dump_buf(pkt->eth_hdr,14);
      byteCount = pkt->pkt_bytes;
      printf("      %d data bytes:\n", byteCount);
//      dump_buf(pkt->buf,byteCount);
    }

  rxMacA(pkt, eth_hdr);
  __pktbuf_free(pkt);
}

// the "z_" is because init routines are called in
// alphabetical order and we want to make sure we are called
// after network io has been initialized.  It's a bit of a
// hack, eh?
//  Packets with the ethernet type 0x11FE are sent to the handler
static void z_comtrolMacInit(void)
{
  __eth_install_listener(ETH_P_COMTROL, comtrolMacHandler);
}

RedBoot_init(z_comtrolMacInit, RedBoot_INIT_LAST);


// The routine below was pretty much copied from the original
// ARM RPSH 2-port bootloader...

typedef unsigned char U08;
typedef unsigned short U16;
typedef unsigned long U32;

#define ETH_TYPE_COMTROL 0x11fe
#define HW_ADDR_LEN 14

int OwnershipGranted;

#if !defined(min)
# define min(a,b) ((a)<(b)?(a):(b))
#endif

//  Processes 0x11fe type packets, presumably from a controlling
//  PC (Linux or Windows w/device driver loaded).  Here's the relevent bytes:
//  data[0] = product_id, must be 0xAA (RPSHI) for this device
//  data[2] = Message class, must be 0x01 (ADMIN_CLASS)
//  data[5] = Admin packet type
static void rxMacA(pktbuf_t * pkt, eth_header_t * eth_hdr)
{
  unsigned char *ethp = (unsigned char *) eth_hdr;
  unsigned char *data = (unsigned char *) pkt->buf;
  int nowowned, i, rcnt = 0;
  U08 csum, etherdest[HW_ADDR_LEN];
  U16 uplen, upcmd;
  U32 upoff = 0, ioctlcmd, goaddr;
  unsigned long ipAddr, netMask, gateway;
  unsigned long temp_ipaddr, temp_netmask, temp_gateway;
  static int loading;

  /* Check for our product ID, discard if not */
  if (*data != 0xaa)
    {
      if (DEBUG_MACMODE)
        printf("Packet discarded, not for this device type:0x%x \n", *data);
      return;
    }

  /*  Check that message class is ADMIN, else discard */
  if (data[2] != 0x01)
    {
      if (DEBUG_MACMODE)
        printf("Packet discarded, message not ADMIN class:0x%x \n", data[2]);
      return;
    }

  /* only command we handle in secure Config mode is ID request */
  if (ident.secureConfigEnable && (data[5] != ADMIN_PRD_ID_REQ_TYPE))
    return;
  
  /*  Pressing enter on the debug port will disable response to MAC mode commands */
  if (disableDriverload)
    {
      if (DEBUG_MACMODE)
        printf("Packet discarded, driver disabled \n");
      return;                   /* ignore driver if loading is disabled */
    }

  // save return address
  memcpy(etherdest, ethp + 6, HW_ADDR_LEN);

  //  Depending on the packet type, process it
  switch (data[5])
    {

      /* boot loader request */
    case ADMIN_SEC_LOAD_TYPE:
      if (data[6] != '~')
        {
          printf("Mac: BootLoader wo ~\n");
          return;
        }
      uplen = data[8];
      uplen = (uplen << 8) + data[7];
      csum = data[7];
      csum += data[8];

      for (i = 0; i <= uplen; i++)
        csum += data[(i + 9)];

      if (csum != 0xff)
        {
          printf("Mac: BootLoader bad csum, 0x%x, len 0x%x\n", csum, uplen);
          return;
        }
      upcmd = data[10];
      upcmd = (upcmd << 8) + data[9];
      switch (upcmd)
        {
        case 0x05:

          /* IOCTL */
          ioctlcmd = data[14];
          ioctlcmd = (ioctlcmd << 8) + data[13];
          ioctlcmd = (ioctlcmd << 8) + data[12];
          ioctlcmd = (ioctlcmd << 8) + data[11];
          if (ioctlcmd == 0x0c)
            {
              /* goto command.  NOTE:  this format is 186 of 1000:0000, not what we will need.   */
              goaddr = data[16];
              goaddr = (goaddr << 8) + data[15];
              goaddr = (goaddr << 8) + data[18];
              goaddr = (goaddr << 8) + data[17];

              loadStatusShow();
              printf("\n  Starting binary at %p\n\n", goaddr);
              startAt(goaddr);
              return;
            }
          printf("RxMacA STUB boot loader IOCTL request, cmd 0x%lx\n", ioctlcmd);
          break;

        case 0x08:
          /* UPLOAD */
          disableAutoload = 1;
          upoff = data[14];
          upoff = (upoff << 8) + data[13];
          upoff = (upoff << 8) + data[12];
          upoff = (upoff << 8) + data[11];
          i = uplen - 6;
          if (!loading)
            {
              printf("\n\n  Loading binary to RAM \n");
              loading = 1;
            }
          else
            {
              printf(".");
            }
          if (ValidRam(upoff, i))
            memcpy((U08 *) upoff, &data[15], i);
          else
            printf("Mac: UPLOAD to invalid address %08lx\n", upoff);
          loadStatusUpdate(upoff, i);
          break;

        case 0x09:
          /* DOWNLOAD */
          printf("RxMacA STUB boot loader DOWNLOAD request\n");
          break;
        }

      /* send reply */
      csum = (2 + (upcmd | 0x80));
      data[6] = '|';
      data[7] = (rcnt + 2) & 0xff;
      data[8] = (rcnt + 2) >> 8;
      data[9] = upcmd | 0x80;
      data[10] = 0;
      i = 11;
      if (rcnt)
        {
          csum += rcnt & 0xff;
          csum += (rcnt >> 8) & 0xff;
          while (rcnt > 0)
            {
              csum += data[i];
              i += 1;
              rcnt -= 1;
            }
        }
      data[i] = ~csum;
      data[3] = (i - 4) & 0xff;
      data[4] = ((i - 4) >> 8) & 0xff;
      pkt->pkt_bytes = i + 1;
      __enet_send(pkt, (enet_addr_t *) (ethp + 6), ETH_TYPE_COMTROL);
      break;

      /* product ID request */
    case ADMIN_PRD_ID_REQ_TYPE:
      {
        unsigned char *p;
        int vlen,room,cnt;

        if (DEBUG_MACMODE)
          printf("ID request packet recieved \n");
        
        if (((data[12] & 0x01) == 0) && (OwnershipGranted))
          data[12] |= 1;          /* cannot take ownership */
        
        // for reasons I can not understand, this particular
        // command's data contains the host MAC addr, and we
        // will use that instead of the one in the ethernet
        // header if it looks reasonable:
        
        memcpy(etherdest, &data[6], HW_ADDR_LEN);
        if (!(((data[6] == 0) && (data[7] == 0) && (data[8] == 0) && (data[9] == 0) && (data[10] == 0) && (data[11] == 0)) || ((data[6] == 0xff) && (data[7] == 0xff) && (data[8] == 0xff) && (data[9] == 0xff) && (data[10] == 0xff) && (data[11] == 0xff))))
          {
            memcpy(etherdest, data + 6, HW_ADDR_LEN);
          }
        
        if ((data[12] & 1) == 0)
          {
            loadStatusInit();
            printf("Mac: Ownership granted\n");
            nowowned = 1;
            OwnershipGranted = 60;        /* one minute timeout */
          }
        else
          nowowned = 0;
        
        // data[0] unchanged                      // product family
        // data[1] unchanged                      // conc index
        data[2] = ADMIN_CLASS;    // packet class
        data[3] = 0;             // admin-data count lsb -- filled in later
        data[4] = 0;              // admin-data count msb
        data[5] = ADMIN_PRD_ID_REP_TYPE;  // packet type (admin cmd)
        // admin data starts here
        memcpy(&data[6], __local_enet_addr, HW_ADDR_LEN);
        data[12] = nowowned ? 0x00 : 0x02;        // flags
        data[13] = 0x01;          // Extended data avail
        data[14] = 0;             // ??
        data[15] = ident.appEnblKey;  // app enable key
        *(U32*)(data+16) = htonl(ident.modelId);  // model ID
        data[20] = ident.archId;  // architecture ID
        data[21] = ident.numPorts;        // number of ports
        *(U16 *) (data + 22) = htons(ident.programId);    // program ID 0==boot-rom
        data[24] = ident.boardRev;        // board revision
        data[25] = ident.versionMajor;    // boot rom major
        data[26] = ident.versionMinor;    // boot rom minor
        data[27] = 0;             // number of app records

        p = &data[28];

        *p++ = 1;                                  // extended data type (ver str)
        room = (pkt->bufsize - 7) - (p - (unsigned char*)(pkt->buf));
        vlen = strlen(IdentString)+1;
        cnt  = min(room,vlen);
        *p++ = cnt;                           // extended data length
        memcpy(p,IdentString,cnt);          // extended data
        p += cnt;

        *p++ = 2;                             // extended data type
        *p++ = 1;                             // extended data length
        *p++ = getFeatureBits() & 0xff;       // extended data
        *p++ = 0;                             // end of extended data
        *p++ = 0;                             // end of extended data
        data[3] = p-data;                     // number of data bytes
        pkt->pkt_bytes = p - (unsigned char*)(pkt->buf); // number of packet bytes
        __enet_send(pkt, (enet_addr_t *) & etherdest, ETH_TYPE_COMTROL);
        if (DEBUG_MACMODE)
          printf("ID response packet sent \n");
      }
      break;

    case ADMIN_LOOPBACK_TYPE:
      /* loopback request */
      printf("RxMacA STUB loopback request\n");
      break;

    case ADMIN_RESET_TYPE:
      /* reset request */
      printf("RxMacA reset\n");
      DM_BOARD_RESET();
      break;

    case 34:
      /* erase IP info */
      putIpInfo(0, 0, 0);
      getIpInfo(&temp_ipaddr, &temp_netmask, &temp_gateway);
      ipAddr = htonl(temp_ipaddr);
      netMask = htonl(temp_netmask);
      gateway = htonl(temp_gateway);
      goto sendIpReply;

    case 38:
      /* disable IP networking */
      putIpInfo(0xffffffff, 0, 0);
      getIpInfo(&temp_ipaddr, &temp_netmask, &temp_gateway);
      ipAddr = htonl(temp_ipaddr);
      netMask = htonl(temp_netmask);
      gateway = htonl(temp_gateway);
      goto sendIpReply;

    case 37:
      /* get IP address currently in use */
      memcpy(&ipAddr,__local_ip_addr,4);
      memcpy(&netMask,__local_ip_mask,4);
      memcpy(&gateway,__local_ip_gate,4);
      goto sendIpReply;

    case 32:
      /* set IP info */
      memcpy(&ipAddr,data+6,4);
      memcpy(&netMask,data+10,4);
      memcpy(&gateway,data+14,4);
      putIpInfo(ntohl(ipAddr), ntohl(netMask), ntohl(gateway));
      getIpInfo(&temp_ipaddr, &temp_netmask, &temp_gateway);
      ipAddr = htonl(temp_ipaddr);
      netMask = htonl(temp_netmask);
      gateway = htonl(temp_gateway);
      goto sendIpReply;

    case 33:
      /* IP info request */
      getIpInfo(&temp_ipaddr, &temp_netmask, &temp_gateway);
      ipAddr = htonl(temp_ipaddr);
      netMask = htonl(temp_netmask);
      gateway = htonl(temp_gateway);
      
    sendIpReply:
      data[3] = 32;             // length
      data[4] = 0;
      memcpy(data+6,&ipAddr,4);
      memcpy(data+10,&netMask,4);
      memcpy(data+14,&gateway,4);
      pkt->pkt_bytes = 18;
      __enet_send(pkt, (enet_addr_t *) & etherdest, ETH_TYPE_COMTROL);
      break;

    default:
      printf("RxMacA STUB unknown command: 0x%x\n", data[5]);
      break;
    }


  return;
}

#endif
