#include <pkgconf/hal.h>
#include <pkgconf/system.h>
#include <redboot.h>
#include <net/net.h>
#include <cyg/hal/plf_stub.h>
#include <cyg/hal/hal_endian.h>

#include "admin.h"

#if defined(CYGPKG_REDBOOT_NETWORKING)

#include "ident.h"

extern void startAt(unsigned);

static int HandleTool(tcp_socket_t * sock, unsigned char *datap, int datalen);

static tcp_socket_t rtsSocket;
static int sockState;
static unsigned char tcpRxBuf[1600];
static unsigned nextSecond;
static unsigned long goReqAddr;
static int goReq;
static int resetReq;
static int closeTimeCounter = 0;	/*  Counter used to track that TCP close's occur, and we don't wait forever */
static int loadAborted = 1;	/*  Set to re-init loading variables on the handle tool. */

void dumpHex(const char *s, void *buf, int len)
{
  unsigned char *p = buf;
  diag_printf("%s[%u]:", s, len);
  while (len--)
    diag_printf(" %02.2x", *p++);
  diag_printf("\n");
}

// "z_" is because init routines are called in alphabetical order
// and we want to make sure we are called after network io has
// been initialized.  It's a bit of a hack, eh?

static void z_comtrolTcpInit(void)
{
  __tcp_listen(&rtsSocket, 0x11fe);
  __tcp_so_reuseaddr(&rtsSocket);
  sockState = rtsSocket.state;
  nextSecond = MS_TICKS() + 1000;
}

RedBoot_init(z_comtrolTcpInit, RedBoot_INIT_LAST);

extern int OwnershipGranted;

static void CloseSocket(tcp_socket_t * sockPtr)
{
  if (sockPtr->state == _ESTABLISHED)
    {
      printf("Tcp: Closing socket\n");
      __tcp_close(sockPtr);
      closeTimeCounter = TCP_SOCKET_CLOSE_TIMEOUT;
      loadAborted = 1;
    }
}

static int looksLikeSSL(unsigned char *msg, int total)
{
  if ((total >= 6)
      && (msg[0] == 0x16)   // SSL3/TLS1 content type HANDSHAKE
      && (msg[5] == 0x01))  // CLIENT HELLO
    return 1;
  if ((total >= 3)
      && (msg[0] & 0x80)    // SSL2 two-byte length header
      && (msg[2] == 0x01))  // CLIENT HELLO
    return 1;
  return 0;
}

/*
 *  The TCP loop.  Runs continuously, untimed. 
*/
static void comtrolTcpLoop(bool is_idle)
{
  int n;
  unsigned now;
  static int exp_len;
  static int total = 0;
  int loopCount = 0;
  static int firstTime;

  now = get_ms_ticks();

  /*  Every 1 second, this loop is entered.  Decrements OwnershipGranted.  */
  if ((int) (nextSecond - now) < 0)
    {
      nextSecond += 1000;
      if (OwnershipGranted)
	{
	  --OwnershipGranted;

	  /*  If OwnershipGranted expires, other side has disappeared.  Close socket (if TCP). */
	  if (OwnershipGranted == 0)
	    {
	      printf("Tcp: Ownership timeout expired...");
	      CloseSocket(&rtsSocket);
	    }
	}

      /*  
       *  Will wait up to TCP_SOCKET_CLOSE_TIMEOUT seconds for the other side to respond to the TCP close, 
       *  then will arbitrarily close our side.
       */
      if (closeTimeCounter)
	{
	  if (rtsSocket.state == _CLOSED || rtsSocket.state == _LISTEN)
	    {
	      closeTimeCounter = 0;
	    }
	  else
	    {
	      --closeTimeCounter;

	      /*  20 second close wait has elapsed.  Check state to be sure, then close this side. */
	      if (closeTimeCounter == 0)
		{
		  if ((rtsSocket.state == _LAST_ACK) || (rtsSocket.state == _FIN_WAIT_1) || (rtsSocket.state == _FIN_WAIT_2) || (rtsSocket.state == _TIME_WAIT) || (rtsSocket.state == _CLOSING))
		    {
		      printf("Tcp: No response to close, closing local socket \n");
		      OwnershipGranted = 0;
		      do_close(&rtsSocket);
		    }
		}
	    }
	}			//  End of close time counter loop
    }				//  End of 1 second loop

  /*  If socket state changes, enter this bit */
  if (sockState != rtsSocket.state)
    {
      switch (rtsSocket.state)
	{
	  /*  
	   *  When TCP connection is established, disable autoloading of the default application.
	   *  When connection is closed, re-enable autoloading, but reset the counter back, to avoid
	   *  race with the driver, who may be trying to load nslink.
	   */
	case _ESTABLISHED:
	  printf("\nTcp: Connection established - Ownership Granted \n");
	  disableAutoload = 1;
	  OwnershipGranted = 60;
	  loadStatusInit();
	  total = 0;
          firstTime = 1;
	  break;
	case _CLOSED:
          if (sockState != _LISTEN)
            printf("Tcp: Connection closed\n");
	  disableAutoload = 0;
	  OwnershipGranted = 0;
	  loadAborted = 1;
	  if (resetReq)
            DM_BOARD_RESET();
	  if (goReq)
	    startAt(goReqAddr);
	  if (!disableDriverload)
            {
              __tcp_listen(&rtsSocket, 0x11fe);
              __tcp_so_reuseaddr(&rtsSocket);
            }
	  break;
	default:
	  break;
	}
      sockState = rtsSocket.state;
    }

  if (disableDriverload)
    {
      if (rtsSocket.state == _ESTABLISHED)
        __tcp_close(&rtsSocket);
      else if (rtsSocket.state == _LISTEN)
        __tcp_reset(&rtsSocket);
    }

  /*  
   *  If TCP socket exists, this loop does continual reads to the socket.  The only packets
   *  it recognizes are the Tool Request packets.  The first 8 bytes of those packets are
   *  a TOOL_REQUEST struct.  
   *
   *  Note that the variables total and exp_len are static, and the buffer is global, so the
   *  command builds until all bytes are received.  This handles the case where TCP fragments
   *  the message.   
   */
  if (rtsSocket.state == _ESTABLISHED || rtsSocket.state == _CLOSE_WAIT)
    {
    MORE_PACKETS:
      loopCount++;
      if (total < 8)
	{
	  /*  Read first 8 bytes so we can get the length. */
	  n = __tcp_read(&rtsSocket, tcpRxBuf + total, 8 - total);
	  if (n > 0)
	    total += n;
	}

      if (firstTime && (total >= 8) && looksLikeSSL(tcpRxBuf,total))
        {
          printf("Tcp: Received SSL Handhshake attempt.\n");
          OwnershipGranted = 0;
          CloseSocket(&rtsSocket);
        }
      else
        {
          /*  
           *  We got the 8, retrieve the expected length of the packet (bytes 2 & 3). In effect we are 
           *  casting TOOL_REQUEST to the start of the buffer, to get to the length member.
           */
          if (total == 8)
            {
              exp_len = tcpRxBuf[2] | (tcpRxBuf[3] << 8);
              if (exp_len > (sizeof(tcpRxBuf)))
                {
                  printf("Tcp: ERROR in rx packet - invalid expected length: %d\n", exp_len);
                  CloseSocket(&rtsSocket);
                  total = 0;
                  exp_len = 0;
                }
            }

          /* If we have at least 8, try processing it. */
          if (total >= 8)
            {
              /*  We have not received all the bytes.  Read the socket again. */
              if (total < exp_len)
                {
                  n = __tcp_read(&rtsSocket, tcpRxBuf + total, exp_len - total);
                  if (n > 0)
                    total += n;
                }

              /*  
               *  If we have received the entire packet now, process it.  Else, we will 
               *  try again next time through the loop.
               */
              if (total == exp_len)
                {
                  OwnershipGranted = 60;
                  if (HandleTool(&rtsSocket, tcpRxBuf, total))
                    {
                      printf("Tcp invalid cmd: 0x%02x\n", tcpRxBuf[0]);
                      CloseSocket(&rtsSocket);
                      
                    }
                  total = 0;
                }
            }
        }

      firstTime = 0;

      /* Drain the pkt list, but don't get stuck in an infinite loop, so only go around 10 times max. */
      if ((rtsSocket.rxlist != NULL) && (loopCount < 10))
	goto MORE_PACKETS;

      if (rtsSocket.state == _CLOSE_WAIT)
        __tcp_close(&rtsSocket);

    }
}

RedBoot_idle(comtrolTcpLoop, RedBoot_IDLE_LAST);

typedef unsigned char U08;
typedef unsigned short U16;
typedef unsigned long U32;

/*
 *   Tool request struct.  Each received TCP packet starts with this struct.
 */

struct tool_request_t
{
  U08 cmd;			/*  Tool command */
  U08 index;
  U16 length;			/*  Command length. */
  U32 addr;			/*  Address, used on the LOAD_TOOLREQ command */
} __attribute__ ((packed));

typedef struct tool_request_t TOOL_REQUEST;

/*
 *   TCP ID Query Response struct.
 */
struct tool_response_t
{
  U32 IpAddr;
  U32 DefaultRoute;
  U32 ServerIpAddr;
  U32 SubnetMask;
  U32 AvailMemAddr;
  U32 AvailMemLength;
  char Version[64];
  U08 MacAddr[6];
  U32 ModelId;
  U08 ArchId;
  U08 NumPorts;
  U16 ProgramId;
  U08 BoardRev;
  U08 BootRevMajor;
  U08 BootRevMinor;
  U08 NumApps;
  U08 Reserved_was_UserDataCount;
  U08 AppEnbl;
  U08 ExtendedData[5];
} __attribute__ ((packed));

typedef struct tool_response_t TOOL_RESPONSE;

#define TOOL_RESPONSE_LENGTH	(sizeof (TOOL_RESPONSE))

/* 
 * The Tool Commands 
 */
#define LOAD_TOOLREQ	1
#define GET_TOOLREQ	2
#define GOTO_TOOLREQ	3
#define MOVE_TOOLREQ	4
#define ID_TOOLREQ	5
#define SET_TOOLREQ	6
#define ENV_TOOLREQ	7
#define PING_TOOLREQ	8
#define DATA_TOOLREQ    9
#define SET_IP_TOOLREQ 10
#define ERASE_IP_TOOLREQ 11
#define DISABLE_IP_TOOLREQ 12
#define RESET_TOOLREQ   13
#define SSL_HANDSHAKE   16  // we don't implement this, but we watch
                            // for it and shut the connection down

#define HW_ADDR_LEN 14

/*
 *  Handles the Tool commands, received over the TCP socket.
 */
static int HandleTool(tcp_socket_t * sock, U08 * datap, int datalen)
{
  TOOL_REQUEST *treq;
  TOOL_RESPONSE tresp;
  U32 *lpp, reqlen, reqaddr, srcaddr;
  U32 zero = 0;
  static int loading;
  static unsigned char loadIndex;

  treq = (TOOL_REQUEST *) datap;

  //dumpHex("ToolReq",datap,16);

  if (datalen < sizeof(TOOL_REQUEST))
    {
      printf("Tcp: Bad datalen: 0x%x\n", datalen);
      return -1;
    }

  if (ident.secureConfigEnable && (treq->cmd != ID_TOOLREQ))
    return -1;

  if (loadAborted)
    {
      loading = 0;
      loadAborted = 0;
      loadIndex = 0;
    }

  /*  
   *  Switch on the Tool Command.  Redboot does not support all Tool Commands.
   */
  switch (treq->cmd)
    {
    case LOAD_TOOLREQ:
      datap += sizeof(TOOL_REQUEST);
      datalen = CYG_LE16_TO_CPU(treq->length) - sizeof(TOOL_REQUEST);
      reqaddr = CYG_LE32_TO_CPU(treq->addr);

      //diag_printf("%08x: %d\n",reqaddr, datalen);

      if (!loading)
	{
	  printf("\nTcp: Loading binary to RAM \n");
	  loading = 1;
	  loadIndex = 0;
	}
      else
	{
	  printf(".");
	}
      loadStatusUpdate(reqaddr, datalen);

      /*  Check for address outside of the RAM memory area */
      if (!ValidRam(reqaddr, datalen))
	{
	  printf("Tcp: ERROR during load - invalid RAM address: 0x%x len: 0x%x \n", (int) reqaddr, (int) datalen);
	  CloseSocket(&rtsSocket);
	}


      /*  Check for missed messages */
      else if (treq->index != loadIndex)
	{
	  printf("Tcp: ERROR during load - invalid Load Index, expected: %d  received: %d \n", loadIndex, treq->index);
	  CloseSocket(&rtsSocket);
	}

      /*  Check for oversized messages */
      else if (datalen > (sizeof(tcpRxBuf)))
	{
	  printf("Tcp: ERROR during load - invalid data length: %d", datalen);
	  CloseSocket(&rtsSocket);
	}

      /*  OK, store load data */
      else
	{
	  memcpy((U08 *) reqaddr, datap, datalen);
	}

      loadIndex++;
      break;

    case GET_TOOLREQ:
      lpp = (U32 *) (treq + 1);
      reqlen = CYG_LE32_TO_CPU(*lpp);
      reqaddr = CYG_LE32_TO_CPU(treq->addr);
      __tcp_write(sock, (U08 *) reqaddr, reqlen);
      break;

    case GOTO_TOOLREQ:
      goReq = 1;
      goReqAddr = CYG_LE32_TO_CPU(treq->addr);
      printf("Tcp: Goto %lx\n", goReqAddr);
      loadStatusShow();
      printf("\nTcp: Starting binary\n\n");
      break;

    case MOVE_TOOLREQ:
      lpp = (U32 *) (treq + 1);
      reqlen = CYG_LE32_TO_CPU(lpp[1]);
      reqaddr = CYG_LE32_TO_CPU(treq->addr);
      srcaddr = CYG_LE32_TO_CPU(lpp[0]);
      if (reqaddr < 0x800000)
	memcpy((U08 *) reqaddr, (U08 *) srcaddr, reqlen);
      else
	printf("Tcp: Move to invalid addr %lx\n", reqaddr);
      break;

    case ID_TOOLREQ:
      printf("Tcp: ID Request recieved\n");
      memcpy(&tresp.IpAddr, __local_ip_addr, 4);
      memcpy(&tresp.DefaultRoute, __local_ip_gate, 4);
      memcpy(&tresp.ServerIpAddr, &zero, 4);
      memcpy(&tresp.SubnetMask, __local_ip_mask, 4);
      tresp.AvailMemAddr = CYG_CPU_TO_LE32(LoadableStart);
      tresp.AvailMemLength = CYG_CPU_TO_LE32(LoadableEnd - LoadableStart);
      memcpy((U08 *) tresp.Version, (U08 *) IdentString, 64);
      memcpy(tresp.MacAddr, __local_enet_addr, HW_ADDR_LEN);
      tresp.ModelId = htonl(ident.modelId);
      tresp.ArchId = ident.archId;
      tresp.NumPorts = ident.numPorts;
      tresp.ProgramId = htons(ident.programId);
      tresp.BoardRev = ident.boardRev;
      tresp.BootRevMajor = ident.versionMajor;
      tresp.BootRevMinor = ident.versionMinor;
      tresp.NumApps = 0;
      tresp.Reserved_was_UserDataCount = 0;
      tresp.AppEnbl = ident.appEnblKey;
      tresp.ExtendedData[0] = 2;  // type
      tresp.ExtendedData[1] = 1;  // len
      tresp.ExtendedData[2] = getFeatureBits(); // data
      tresp.ExtendedData[3] = 0;  // end marker
      tresp.ExtendedData[4] = 0;  // end marker

      __tcp_write(sock, (U08 *) & tresp, TOOL_RESPONSE_LENGTH);
      //printf("tresp size = %ld\n", TOOL_RESPONSE_LENGTH);
      break;

    case SET_TOOLREQ:
      lpp = (U32 *) (treq + 1);
      reqlen = CYG_LE32_TO_CPU(*lpp);
      reqaddr = CYG_LE32_TO_CPU(treq->addr);
      if (ValidRam(reqaddr, reqlen))
	memset((U08 *) reqaddr, 0, reqlen);
      else
	printf("Tcp: SetTool to invalid address %lx\n", reqaddr);
      break;

    case ENV_TOOLREQ:
      datap += sizeof(TOOL_REQUEST);
      printf("Tcp: STUB setenv '%s'\n", (char *) datap);
      break;

    case PING_TOOLREQ:
      printf("Tcp: STUB PING\n");
      break;

    case RESET_TOOLREQ:
      printf("Tcp: reset\n");
      resetReq = 1;
      break;

    case ERASE_IP_TOOLREQ:
      printf("Tcp: erase IP info\n");
      putIpInfo(0, 0, 0);
      break;

    case DISABLE_IP_TOOLREQ:
      printf("Tcp: erase IP info\n");
      putIpInfo(0xffffffff, 0, 0);
      break;

    case SET_IP_TOOLREQ:
      {
	unsigned long *p = (&treq->addr) + 1;
	if (datalen < 20)
	  printf("Tcp: set IP cmd too short\n");
	else
	  {
	    printf("Tcp: set IP %08lx %08lx %08lx\n", p[0], p[1], p[2]);
	    putIpInfo(p[0], p[1], p[2]);
	  }
	break;
      }

    default:
      printf("Tcp: Invalid Command recieved - %d  \n", treq->cmd);
      return -1;
      break;
    }
  return 0;
}

#endif
