// phy functions for Agere PHY AC101L

#include <pkgconf/system.h>
#include <pkgconf/devs_eth_arm_dmrts.h>
#include <cyg/infra/cyg_type.h>
#include <cyg/infra/diag.h>
#include "phy.h"
#include "ac101l.h"

// AC101L register offsets
#define	AC101L_CNTL_REG 	0x00
#define	AC101L_STATUS_REG 	0x01
#define	AC101L_ID_REG1    	0x02
#define	AC101L_ID_REG2    	0x03
#define	AC101L_ANA_REG    	0x04
#define	AC101L_ANLPAR_REG 	0x05
#define	AC101L_ANE_REG    	0x06
#define	AC101L_INT_CNTL_REG    	0x11
#define	AC101L_DIAG_REG    	0x12
#define AC101L_RXCRC_REG        24

// AC101L Control register bit defines
#define AC101L_CNTL_RESET        0x8000
#define AC101L_CNTL_LOOPBACK     0x4000
#define AC101L_CNTL_SPEED_100    0x2000  // 1=100Meg, 0=10Meg
#define AC101L_CNTL_AN           0x1000  // 1=Enable auto negotiation, 0=disable it
#define AC101L_CNTL_PWRDN        0x0800  // 1=Enable power down
#define AC101L_CNTL_ISOLATE      0x0400  // 1=Isolate from MII
#define AC101L_CNTL_RSTRT_AN     0x0200  // 1=Restart Auto Negotioation process
#define AC101L_CNTL_FULL_DUP     0x0100  // 1=Enable full duplex mode, 0=half dup
#define AC101L_CNTL_TST_COLL     0x0080  // 1=Enable collision test

#define Bit(n) (1<<(n))

#define AC101L_ANA_PAUSE_ENA    Bit(10)
#define AC101L_ANA_100T4        Bit(9)
#define AC101L_ANA_100TX_FULL   Bit(8)
#define AC101L_ANA_100TX        Bit(7)
#define AC101L_ANA_10T_FULL     Bit(6)
#define AC101L_ANA_10T          Bit(5)
#define AC101L_ANA_SEL_802_3    Bit(0)

#define AC101L_INT_CNTL_LP_ACK_IE             Bit(11)
#define AC101L_INT_CNTL_LINK_STATUS_CHANGE_IE Bit(10)
#define AC101L_INT_CNTL_ANEN_COMP_IE          Bit(8)
#define AC101L_INT_CNTL_RXER                  Bit(6)
#define AC101L_INT_CNTL_ANEN_COMP_INT         Bit(1)

#define AC101L_STATUS_LINKUP                  Bit(2)

#define AC101L_ANLPAR_100_FD                  Bit(8)
#define AC101L_ANLPAR_100                     Bit(7)
#define AC101L_ANLPAR_10_FD                   Bit(6)
#define AC101L_ANLPAR_10                      Bit(5)

#define AC101L_DIAG_SPEED_100                 Bit(10)
#define AC101L_DIAG_FULL_DUPLEX               Bit(11)

#define ddiag_printf diag_printf

void Ac101lPhyReset(void)
{

    ddiag_printf("AC101L Reset Start\n");

    // first software reset the AC101L
    MiiStationWrite(PHY_ADDR, AC101L_CNTL_REG,  AC101L_CNTL_RESET);
    MiiStationWrite(PHY_ADDR, AC101L_CNTL_REG,  0);

    // initialize auto-negotiation capabilities
    MiiStationWrite(PHY_ADDR, AC101L_ANA_REG,
                    AC101L_ANA_PAUSE_ENA+
                    AC101L_ANA_100TX_FULL+
                    AC101L_ANA_100TX+
                    AC101L_ANA_10T_FULL+
                    AC101L_ANA_10T+
                    AC101L_ANA_SEL_802_3);

    // Now start an auto negotiation
    MiiStationWrite(PHY_ADDR, AC101L_CNTL_REG,
                    AC101L_CNTL_AN+
                    AC101L_CNTL_RSTRT_AN);

    ddiag_printf("AC101L Reset End\n");

    // clear rxerr bit
    MiiStationRead(PHY_ADDR, AC101L_INT_CNTL_REG);

}//end Ac101lPhyReset

#if !defined(CYGPKG_REDBOOT)
#include "std.h"
#include <cyg/infra/diag.h>
#include <cyg/hal/hal_arch.h>
#include <cyg/kernel/kapi.h>

#ifndef TICKS_PER_SECOND
#define TICKS_PER_SECOND 100
#endif

#ifndef cyg_thread_delay_ms
#define cyg_thread_delay_ms(n) cyg_thread_delay(n/(1000/TICKS_PER_SECOND))
#endif


static void ac101Poll(void)
{
    static unsigned wedgeDelay;
    static unsigned statusLastTime;
    static unsigned rxErrorPresent;
    static unsigned noPacketCount;

    unsigned rxCRC = MiiStationRead(PHY_ADDR, AC101L_RXCRC_REG);
    unsigned rxErr = MiiStationRead(PHY_ADDR, AC101L_INT_CNTL_REG) & AC101L_INT_CNTL_RXER;
    unsigned status = MiiStationRead(PHY_ADDR, AC101L_STATUS_REG);
    unsigned linkUp = status & AC101L_STATUS_LINKUP;;

    // *** NOTE ***
    // the above reads clear the rx CRC register and the rx Error flag

    //diag_printf("%04x, %04x, %04x %04x\n", status, linkUp, rxErr, rxCRC);

    if ((status ^ statusLastTime) & AC101L_STATUS_LINKUP) {
        // status change
        unsigned newLinkStatus = 0;
        ddiag_printf("AC101L Link ");
        if (!linkUp) {
            ddiag_printf("down\n");
        } else {
            unsigned diagReg = MiiStationRead(PHY_ADDR, AC101L_DIAG_REG);
            newLinkStatus |= PhyStatus_LinkUp;

            if (diagReg & AC101L_DIAG_SPEED_100) {
                newLinkStatus |= PhyStatus_100Mb;
                ddiag_printf("100M ");
            } else {
                ddiag_printf("10M ");
            }

            if (diagReg & AC101L_DIAG_FULL_DUPLEX) {
                set_maccon_duplex(1);
                newLinkStatus |= PhyStatus_FullDuplex;
                ddiag_printf("Full\n");
            } else {
                set_maccon_duplex(0);
                ddiag_printf("Half\n");
            }
        }
        dmrtsLinkStatus = newLinkStatus;
        statusLastTime = status;
    }

    // from here on down we're trying to detect when the PHY's Rx half
    // has locked up due to ESD.  If we see an Rx Error flag and then no
    // more packets, we reset the PHY.  Except that there's always an Rx
    // Error flag the first time after the Link comes up.  We disregard
    // the first Rx Error within 2 seconds after link up.

    if (!linkUp) {
        wedgeDelay = 2;
        return;
    }

    if (wedgeDelay) {
        if (rxErr)
            wedgeDelay = 0;
        else
            --wedgeDelay;
        return;
    }

    if (rxErr) {
        rxErrorPresent = 1;
        noPacketCount = 0;
        return;
    }

    // check for receive packet since last poll

    if (rxCRC != 0) {
        rxErrorPresent = 0;
        return;
    }

    // no rx packet since last poll

    if (rxErrorPresent) {
        ++noPacketCount;
        if (noPacketCount > 2) {
            ddiag_printf("AC101L wedged, resetting it.\n");
            ++dmrtsPhyResetCount;
            Ac101lPhyReset();
            wedgeDelay = 2;
            rxErrorPresent = 0;
        }
    }
}

static void polltask(cyg_addrword_t data)
{
    ddiag_printf("AC101L Polling Thread Started.\n");
    while (1) {
        ac101Poll();
        cyg_thread_delay_ms(1000);
    }
}
#endif

void Ac101lPhyInit(void)
{
#if !defined(CYGPKG_REDBOOT)
    static cyg_handle_t pollhandle;
    static U32 pollstack[1024];
    static cyg_thread pollthread;
    static int pollrunning = 0;
    if (!pollrunning) {
        //start polling thread
        //priority 15 appears to run well and not starve other tasks.
        cyg_thread_create(
            15, // thread priority 0-31 inclusive : 0 = highest
            polltask,
            (cyg_addrword_t) 0,
            "ac101-polling",
            (void *)pollstack,
            sizeof pollstack,
            &pollhandle,
            &pollthread
        );
        cyg_thread_resume(pollhandle);
        pollrunning = 1;
    }
#endif
}


cyg_uint32 Ac101lPhyStatus(void)
{
    cyg_uint32 statusReg, diagReg, r = 0;

    statusReg = MiiStationRead(PHY_ADDR, AC101L_STATUS_REG);
    diagReg = MiiStationRead(PHY_ADDR, AC101L_DIAG_REG);

    if (statusReg & AC101L_STATUS_LINKUP)
        r |= PHY_STATUS_LINK_UP;

    // Broadcom gave us this info about AC101L_DIAG_REG Register 18
    // bits 10 and 11 showing speed and mode.
    if (diagReg & AC101L_DIAG_SPEED_100)
        r |= PHY_STATUS_100MB;

    if (diagReg & AC101L_DIAG_FULL_DUPLEX)
        r |= PHY_STATUS_FULL_DUPLEX;

    return r;
}

void Ac101lShowPhyStatus(void)
{
    cyg_uint32 statusReg, diagReg, u, i;
    statusReg = MiiStationRead(PHY_ADDR, AC101L_STATUS_REG);
    diagReg = MiiStationRead(PHY_ADDR, AC101L_DIAG_REG);

    ddiag_printf("AC101L Status: %04x\n",statusReg);
    ddiag_printf("               link   %s\n", (statusReg & AC101L_STATUS_LINKUP) ? "up": "down");
    ddiag_printf("AC101L   Diag: %04x\n",diagReg);
    ddiag_printf("               speed  %s\n", (diagReg & AC101L_DIAG_SPEED_100) ? "100": "10");
    ddiag_printf("               duplex %s\n", (diagReg & AC101L_DIAG_FULL_DUPLEX) ? "full" : "half");
    ddiag_printf("  Reset Count: %u\n",dmrtsPhyResetCount);
    if (0)
        for (i=0; i<=31; ++i) {
            u = statusReg = MiiStationRead(PHY_ADDR, i);
            ddiag_printf("%3d: %04x\n", i, u & 0xffff);
        }
}

void Ac101lPhyLoopback(unsigned enableLoopback)
{
    unsigned u;
    u =  MiiStationRead(PHY_ADDR, AC101L_CNTL_REG);
    if (enableLoopback)
        u |= AC101L_CNTL_LOOPBACK;
    else
        u &= ~AC101L_CNTL_LOOPBACK;
    MiiStationWrite(PHY_ADDR, AC101L_CNTL_REG, u);
}

void Ac101lPhyInterruptAck(void)
{
    cyg_uint32 interruptStatusReg;

    interruptStatusReg = MiiStationRead(PHY_ADDR, AC101L_INT_CNTL_REG);
}
