// PHY functions for Micrel PHY KSZ8893MQL
#include <pkgconf/system.h>
#include <cyg/hal/iic.h>
#include <cyg/infra/diag.h>
#include <cyg/infra/cyg_type.h>

#if !defined(CYGPKG_REDBOOT)
#include <cyg/hal/hal_arch.h>
#include <cyg/kernel/kapi.h>
#endif

#if defined(CYGPKG_REDBOOT)
#include <net/net.h>
#endif

#include "phy.h"
#include "ksz8893mql.h"

static cyg_uint8 IicPhyRead(cyg_uint8 addr);
static void IicPhyWrite(cyg_uint8 addr, cyg_uint8 data);

int ksz_link_up(void)
{
    cyg_uint8 port1, port2;

    port1 = IicPhyRead(KSZ8893MQL_PORT1_STATUS0_REG);
    if (port1 & KSZ8893MQL_PORT_LINK_STAT0)	return 1;

    port2 = IicPhyRead(KSZ8893MQL_PORT2_STATUS0_REG);
    if (port2 & KSZ8893MQL_PORT_LINK_STAT0)	return 1;

    return 0;
}

static void showLink(int port, unsigned  status0, unsigned status1)
{
    diag_printf("KSZ PHY %d link ", port);
    if (status0 & KSZ8893MQL_PORT_LINK_STAT0) {
        diag_printf("up ");
        diag_printf((status1 & KSZ8893MQL_PORT_100MB_STAT1) ? "100MB " : "10 MB ");
        diag_printf((status1 & KSZ8893MQL_PORT_DUPLEX_STAT1) ? "full\n" : "half\n");
    } else
        diag_printf("down\n");
}

void ksz_phy_status(void)
{
    showLink(1,IicPhyRead(KSZ8893MQL_PORT1_STATUS0_REG),IicPhyRead(KSZ8893MQL_PORT1_STATUS1_REG));
    showLink(2,IicPhyRead(KSZ8893MQL_PORT2_STATUS0_REG),IicPhyRead(KSZ8893MQL_PORT2_STATUS1_REG));
}

#if !defined(CYGPKG_REDBOOT)
#include "std.h"
static cyg_handle_t kszpollhandle;
static U32 kszpollstack[1024];
static cyg_thread kszpollthread;
static cyg_thread_entry_t ksz_start_switch_poll;
static void ksz_check_state(void);
static void ksz_start_switch_poll(cyg_addrword_t data);
static int  numStoppedReads=0;

#define ddiag_printf diag_printf

//Handle lockup caused by ESD testing. If zapped, the chip will not
//come back in the proper state so we have to force it.
static void ksz_check_state(void)
{
    cyg_uint8 val = 0;

    //Caused by ESD testing ... DO NOT REMOVE
    //if zapped the chip will not come back in the
    //proper state so we have to force it.
    val = IicPhyRead(KSZ8893MQL_CHIP_ID1_REG);
    if (!(val & KSZ8893MQL_CHIP_ID1_START)) {
        // bump the number of stopped reads
        numStoppedReads++;
        // must have three stopped reads in a row before resetting phy
        if (numStoppedReads >= 3) {
            ddiag_printf("Ethernet switch stopped. Resetting...\n");
            PhyReset();
            numStoppedReads = 0;
        }
    } else {
        // reset number of stopped reads
        numStoppedReads = 0;
    }
}//end ksz_check_state

static void ksz_start_switch_poll(cyg_addrword_t data)
{
    cyg_thread_yield();
    cyg_thread_delay_ms(800);
    ddiag_printf("KSZ Start Switch Polling Thread Started.\n");
    while (1) {
        ksz_check_state();
        cyg_thread_delay_ms(100);
    }
}//end ksz_start_switch_poll
#endif

void ksz_init(void)
{
#if !defined(CYGPKG_REDBOOT)
    static int kszpollrunning = 0;
    if (!kszpollrunning) {
        // start up polling thread
        //priority 15 appears to run well and not starve other tasks.
        cyg_thread_create(
            15, // thread priority 0-31 inclusive : 0 = highest
            ksz_start_switch_poll,
            (cyg_addrword_t) 0,
            "ksz_start_switch_poll",
            (void *)kszpollstack,
            sizeof kszpollstack,
            &kszpollhandle,
            &kszpollthread
        );
        cyg_thread_resume(kszpollhandle);
        kszpollrunning = 1;
    }
#endif
}//end ksz_create_poll_thread

void EnableSpecialTagging(unsigned char enableBit)
{
    if ( enableBit ) {
        // Enable special tagging mode
        IicPhyWrite(KSZ8893MQL_GLOBAL_CTRL9_REG,
                    IicPhyRead(KSZ8893MQL_GLOBAL_CTRL9_REG) |
                    KSZ8893MQL_GLOBAL_CTRL9_STPID_ENABLE);

        // Insert tag for egress packets on port 3
        IicPhyWrite(KSZ8893MQL_PORT3_CTRL0_REG, KSZ8893MQL_PORT_CTRL0_TAG_INSERT);
        // Remove tag for egress packets on port 1 and port 2
        IicPhyWrite(KSZ8893MQL_PORT1_CTRL0_REG, KSZ8893MQL_PORT_CTRL0_TAG_REMOVE);
        IicPhyWrite(KSZ8893MQL_PORT2_CTRL0_REG, KSZ8893MQL_PORT_CTRL0_TAG_REMOVE);
    } else {
        // Disable special tagging mode
        IicPhyWrite(KSZ8893MQL_GLOBAL_CTRL9_REG,
                    IicPhyRead(KSZ8893MQL_GLOBAL_CTRL9_REG) &
                    ~KSZ8893MQL_GLOBAL_CTRL9_STPID_ENABLE);
    }
}

// Start the switch
void Ksz8893mqlPhyReset(void)
{
    int i;

    // Start the switch
    IicPhyWrite(KSZ8893MQL_CHIP_ID1_REG,
                IicPhyRead(KSZ8893MQL_CHIP_ID1_REG) |
                KSZ8893MQL_CHIP_ID1_START);

    // restart auto-negotiation
    for (i = 0; i < 2; i++) {
        MiiStationWrite(i, MII_ADVERTISE, ADVERTISE_ALL);
        MiiStationWrite(i, MII_BMCR, BMCR_ANENABLE | BMCR_ANRESTART);
    }

    set_maccon_duplex(1);   //always talk full duplex to this switch chip
}


// I2C Slave Serial Bus interface provides complete
// programming access to the KSZ98893MQL's 142 registers

// PHY I2C interface register read
static cyg_uint8 IicPhyRead(cyg_uint8 addr)
{
    cyg_uint8 data = 0;
#ifndef CYGPKG_REDBOOT
    iicReadRt(KSZ8893MQL_SLAVE_ADDR, addr, &data, sizeof(cyg_uint8));
#else
    iicRead(KSZ8893MQL_SLAVE_ADDR, addr, &data, sizeof(cyg_uint8));
#endif
    return data;
}

// PHY I2C interface register read
cyg_uint8 IicPhyReadRegs(cyg_uint8 addr, cyg_uint8 *buf, int count)
{
#ifndef CYGPKG_REDBOOT
    iicReadRt(KSZ8893MQL_SLAVE_ADDR, addr, buf, count);
#else
    iicRead(KSZ8893MQL_SLAVE_ADDR, addr, buf, count);
#endif
    return count;
}

// PHY I2C interface register write
static void IicPhyWrite(cyg_uint8 addr, cyg_uint8 data)
{
    iicWrite(KSZ8893MQL_SLAVE_ADDR, addr, &data, sizeof(cyg_uint8));
}

// Write static MAC address table
void WriteStaticMacTable(int addr, int fid, int use_fid, int override,
                         int valid, int forward_port, unsigned char *mac_addr)
{
    cyg_uint16 reg = ((fid & 0xf) << 6) |
                     ((use_fid & 0x1) << 5) |
                     ((override & 0x1) << 4) |
                     ((valid & 01) << 3) |
                     (forward_port & 0x7);

    IicPhyWrite(KSZ8893MQL_INDIR_DATA7_REG, reg >> 8);
    IicPhyWrite(KSZ8893MQL_INDIR_DATA6_REG, reg & 0xff);
    IicPhyWrite(KSZ8893MQL_INDIR_DATA5_REG, mac_addr[0]);
    IicPhyWrite(KSZ8893MQL_INDIR_DATA4_REG, mac_addr[1]);
    IicPhyWrite(KSZ8893MQL_INDIR_DATA3_REG, mac_addr[2]);
    IicPhyWrite(KSZ8893MQL_INDIR_DATA2_REG, mac_addr[3]);
    IicPhyWrite(KSZ8893MQL_INDIR_DATA1_REG, mac_addr[4]);
    IicPhyWrite(KSZ8893MQL_INDIR_DATA0_REG, mac_addr[5]);

    IicPhyWrite(KSZ8893MQL_INDIR_CTRL0_REG,
                KSZ8893MQL_INDIR_CTRL0_WRITE |
                KSZ8893MQL_INDIR_CTRL0_STATIC_MAC);
    IicPhyWrite(KSZ8893MQL_INDIR_CTRL1_REG, addr);
}

static char MIBCounterNames[MIB_COUNTERS_PER_PORT][24] = {
    "RxLopriorityByte    ",
    "RxHiPriorityByte    ",
    "RxUndersizePkt      ",
    "RxFragments         ",
    "RxOversize          ",
    "Rxjabbers           ",
    "RxSymbolError       ",
    "RxCRCError          ",
    "RxAlignmentError    ",
    "RxControl8808Pkts   ",
    "RxPausePkts         ",
    "RxBroadcase         ",
    "RxMulticast         ",
    "RxUnicast           ",
    "Rx64Octets          ",
    "Rx65to127Octets     ",
    "Rx128to255Octets    ",
    "Rx256to511Octets    ",
    "Rx512to1023Octets   ",
    "Rx1024to1522Octets  ",
    "TxLoPriorityByte    ",
    "TxHiPriorityByte    ",
    "TxLateCollision     ",
    "TxPausePkts         ",
    "TxBroadcastPkts     ",
    "TxMulticastPkts     ",
    "TxUnicastPkts       ",
    "TxDeferred          ",
    "TxTotalCollision    ",
    "TxExcessiveCollision",
    "TxSingleCollision   ",
    "TxMultipleCollision "
};

#define PHY_STATIC_MAC_TAB      0       // Static MAC address table
#define PHY_DYNAMIC_MAC_TAB     1       // Dynamic MAC address table
#define PHY_VLAN_TAB            2       // VLAN table
#define PHY_MIB_COUNTER         3       // MIB counter

// Print selected PHY table
void PrintPhyTable(int type)
{
    switch (type) {

    case PHY_STATIC_MAC_TAB: {
        int i;
        cyg_uint16 reg;

        diag_printf("Static MAC Address Table:\n");
        for (i = 0; i < STATIC_MAC_TABLE_ENTRIES; i++) {
            IicPhyWrite(KSZ8893MQL_INDIR_CTRL0_REG,
                        KSZ8893MQL_INDIR_CTRL0_READ |
                        KSZ8893MQL_INDIR_CTRL0_STATIC_MAC);
            IicPhyWrite(KSZ8893MQL_INDIR_CTRL1_REG, i);
            reg = (IicPhyRead(KSZ8893MQL_INDIR_DATA7_REG) << 8) |
                  IicPhyRead(KSZ8893MQL_INDIR_DATA6_REG);
            diag_printf("%d: FID=%x UseFID=%x Override=%x Valid=%x FwPort=%x "
                        "MAC=%02x:%02x:%02x:%02x:%02x:%02x\n", i,
                        (reg >> 6) & 0xf,       // Filter VLAN ID
                        (reg >> 5) & 0x1,       // Use FID
                        (reg >> 4) & 0x1,       // Override
                        (reg >> 3) & 0x1,       // Valid
                        reg & 0x7,              // Forwarding Ports
                        IicPhyRead(KSZ8893MQL_INDIR_DATA5_REG),
                        IicPhyRead(KSZ8893MQL_INDIR_DATA4_REG),
                        IicPhyRead(KSZ8893MQL_INDIR_DATA3_REG),
                        IicPhyRead(KSZ8893MQL_INDIR_DATA2_REG),
                        IicPhyRead(KSZ8893MQL_INDIR_DATA1_REG),
                        IicPhyRead(KSZ8893MQL_INDIR_DATA0_REG));
        }
        break;
    }

    case PHY_DYNAMIC_MAC_TAB: {
        unsigned int num, i = 0;
        cyg_uint8 reg;

        diag_printf("Dynamic MAC Address Table:\n");
        do {
            IicPhyWrite(KSZ8893MQL_INDIR_CTRL0_REG,
                        KSZ8893MQL_INDIR_CTRL0_READ |
                        KSZ8893MQL_INDIR_CTRL0_DYNAMIC_MAC | (i >> 8));
            IicPhyWrite(KSZ8893MQL_INDIR_CTRL1_REG, i & 0xff);
            while ((reg = IicPhyRead(KSZ8893MQL_INDIR_DATA8_REG)) & 0x80);
            if (reg & 0x04) {  // there is no valid entry in the table
                break;
            }
            // note: 0x000 means 1 entry
            num = ((reg & 0x03) << 8) | IicPhyRead(KSZ8893MQL_INDIR_DATA7_REG);
            reg = IicPhyRead(KSZ8893MQL_INDIR_DATA6_REG);
            diag_printf("%d: TimeStamp=%x SrcPort=%x FID=%x "
                        "MAC=%02x:%02x:%02x:%02x:%02x:%02x\n",
                        i, reg >> 6, (reg >> 4) & 0x03, reg & 0x0f,
                        IicPhyRead(KSZ8893MQL_INDIR_DATA5_REG),
                        IicPhyRead(KSZ8893MQL_INDIR_DATA4_REG),
                        IicPhyRead(KSZ8893MQL_INDIR_DATA3_REG),
                        IicPhyRead(KSZ8893MQL_INDIR_DATA2_REG),
                        IicPhyRead(KSZ8893MQL_INDIR_DATA1_REG),
                        IicPhyRead(KSZ8893MQL_INDIR_DATA0_REG));
        } while (i++ < num);
        break;
    }

    case PHY_VLAN_TAB: {
        int i;
        cyg_uint32 reg;

        diag_printf("VLAN Table:\n");
        for (i = 0; i < VLAN_TABLE_ENTRIES; i++) {
            IicPhyWrite(KSZ8893MQL_INDIR_CTRL0_REG,
                        KSZ8893MQL_INDIR_CTRL0_READ |
                        KSZ8893MQL_INDIR_CTRL0_VLAN_TAB);
            IicPhyWrite(KSZ8893MQL_INDIR_CTRL1_REG, i);
            reg = (IicPhyRead(KSZ8893MQL_INDIR_DATA2_REG) << 16) |
                  (IicPhyRead(KSZ8893MQL_INDIR_DATA1_REG) << 8) |
                  IicPhyRead(KSZ8893MQL_INDIR_DATA0_REG);
            diag_printf("%2d: Valid=%x Membership=%x FID=%x VID=%03x\n", i,
                        (reg >> 19) & 0xf,      // Valid
                        (reg >> 16) & 0x7,      // Membership
                        (reg >> 12) & 0xf,      // Filter ID
                        reg & 0xfff);           // VLAN ID
        }
        break;
    }

    case PHY_MIB_COUNTER: {
        int port, i, addr, overflow, change_line;
        cyg_uint8 reg;
        unsigned int value;

        diag_printf("MIB Counters:\n");

        for (port = 0; port < 3; port++) {
            diag_printf("Port %d:\n", port + 1);
            change_line = 0;
            for (i = 0; i < MIB_COUNTERS_PER_PORT; i++) {
                addr = (port * 0x20) + i;
                IicPhyWrite(KSZ8893MQL_INDIR_CTRL0_REG,
                            KSZ8893MQL_INDIR_CTRL0_READ |
                            KSZ8893MQL_INDIR_CTRL0_MIB_COUNTER);
                IicPhyWrite(KSZ8893MQL_INDIR_CTRL1_REG, addr);
                diag_printf("  %s :", MIBCounterNames[i]);
                while (!((reg = IicPhyRead(KSZ8893MQL_INDIR_DATA3_REG)) & 0x40));
                overflow = reg & 0x80;
                value = reg & 0x3f;
                value = (value << 8) | IicPhyRead(KSZ8893MQL_INDIR_DATA2_REG);
                value = (value << 8) | IicPhyRead(KSZ8893MQL_INDIR_DATA1_REG);
                value = (value << 8) | IicPhyRead(KSZ8893MQL_INDIR_DATA0_REG);
                diag_printf(" %u%s", value, overflow ? " (overflow)" : "");
                diag_printf("%s", (++change_line % 2) ? "\t" : "\n");
            }
            addr = 0x100 + (port * 2);
            IicPhyWrite(KSZ8893MQL_INDIR_CTRL0_REG, 0x1c | (addr >> 8));
            IicPhyWrite(KSZ8893MQL_INDIR_CTRL1_REG, (addr & 0xff));
            value = IicPhyRead(KSZ8893MQL_INDIR_DATA1_REG);
            value = (value << 8) | IicPhyRead(KSZ8893MQL_INDIR_DATA0_REG);
            diag_printf("  TxDropPackets        : %u", value);
            diag_printf("%s", (++change_line % 2) ? "\t" : "\n");

            addr = 0x103 + (port * 2);
            IicPhyWrite(KSZ8893MQL_INDIR_CTRL0_REG, 0x1c | (addr >> 8));
            IicPhyWrite(KSZ8893MQL_INDIR_CTRL1_REG, (addr & 0xff));
            value = IicPhyRead(KSZ8893MQL_INDIR_DATA1_REG);
            value = (value << 8) | IicPhyRead(KSZ8893MQL_INDIR_DATA0_REG);
            diag_printf("  RxDropPackets        : %u\n", value);
        }
        break;
    }

    default:
        diag_printf("Unknown type: %d\n", type);
    }
}
