
#include <pkgconf/devs_eth_arm_dmrts.h>
#include <cyg/infra/diag.h>
#include <cyg/hal/iic.h>
#include "ks5000_regs.h"
#include "ks5000_ether.h"
#include "phy.h"
#include "ac101l.h"
#include "ksz8893mql.h"
#include "ics1893.h"

unsigned int PhyId;
unsigned dmrtsLinkStatus;
unsigned dmrtsPhyResetCount;

void set_maccon_duplex(int state)
{
    if (state) {
        MACCON |= MACON_FULL_DUP;
    } else {
        MACCON &= ~MACON_FULL_DUP;
    }
}

void load_phy_id(void)
{

    PhyId = (MiiStationRead(PHY_ADDR, MII_PHYSID1) << 16) |
            MiiStationRead(PHY_ADDR, MII_PHYSID2);
}

// the specific chips reset will set the samsungs mac's duplex mode based
// their state. this will allow redboot to get one shot and being
// linked properly. the ac101 isn't touched since it doesn't work
// properly after the samsung's mode is set. don't know why but
// I am going to leave it for a later investigation.
void PhyReset(void)
{

    unsigned char chip_id =0;

    PhyId = (MiiStationRead(PHY_ADDR, MII_PHYSID1) << 16) |
            MiiStationRead(PHY_ADDR, MII_PHYSID2);

    if (PhyId == PHY_ID_AC101L) {
        diag_printf("AC101L PHY detected\n");
        Ac101lPhyReset();
        Ac101lPhyInit();
        return;
    }

    if (PhyId == PHY_ID_KSZ8893MQL) {
        unsigned long model_id;
        getModelNumber(&model_id);
        iicRead(KSZ8893MQL_SLAVE_ADDR, KSZ8893MQL_CHIP_ID0_REG, &chip_id, sizeof(unsigned char));
        // DM RTS 2P 2E uses I2C slave mode to manage the switch/PHY
        if (model_id == 5002535 || chip_id == 0x88) {
            Ksz8893mqlPhyReset();
            ksz_init();
            diag_printf("KSZ8893MQL %ul:%x Switch/PHY Reset\n", model_id, chip_id);
        } else {
            diag_printf("KSZ8893MQL %ul:%x Switch/PHY Detected\n", model_id, chip_id);
        }
        return;
    }

    // Kendin switch/PHY (KS8993) is unmanaged
    // no reset for this chip and the samsung must be half duplex
    if (PhyId == PHY_ID_KENDIN) {
        diag_printf("KS8993 Switch/PHY detected\n");
        set_maccon_duplex(0);	//in phy.c/h
        return;
    }

    if ((PhyId & PHY_ID_ICS1983_MASK) == PHY_ID_ICS1983) {
        diag_printf("ICS1893 PHY detected\n");
        ics_reset_phy();
        ics_init();
        return;
    }

    //If not identified and no model number read chip id = ksz8893mql, to enable ethernet, for production test
    iicRead(KSZ8893MQL_SLAVE_ADDR, KSZ8893MQL_CHIP_ID0_REG, &chip_id, sizeof(unsigned char));
    if (chip_id == 0x88) {
        diag_printf("KSZ8893MQL Switch/PHY detected, no model number found\n");
        Ksz8893mqlPhyReset();
        ksz_init();
        return;
    } else {
        diag_printf("Unknown PHY (ID: 0x%08x)\n", PhyId);
    }

    // Reset using standard MIIM registers
    MiiStationWrite(PHY_ADDR, MII_BMCR, BMCR_RESET);
    // Initialize auto-negotiation capabilities
    MiiStationWrite(PHY_ADDR, MII_ADVERTISE, ADVERTISE_ALL);
    // Start an auto-negotiation
    MiiStationWrite(PHY_ADDR, MII_BMCR, BMCR_ANENABLE | BMCR_ANRESTART);
}

void PhyShowStatus(void)
{
    extern void Ac101lShowPhyStatus(void);
    extern void ksz_phy_status(void);
    extern void ics_phy_status(void);
    unsigned id;

    id = (MiiStationRead(PHY_ADDR, MII_PHYSID1) << 16) | MiiStationRead(PHY_ADDR, MII_PHYSID2);

    diag_printf("Phy Id: %08x ", id);

    if (id == PhyId)
        diag_printf("OK\n");
    else
        diag_printf("BAD: expected %08x\n",PhyId);

    if (PhyId == PHY_ID_AC101L)
        Ac101lShowPhyStatus();
    else if (PhyId == PHY_ID_KSZ8893MQL)
        ksz_phy_status();
    else if (PhyId == PHY_ID_KENDIN)
        diag_printf("no status for Kendin switch\n");
    else if ((PhyId  & PHY_ID_ICS1983_MASK) == PHY_ID_ICS1983)
        ics_phy_status();
    else
        diag_printf("Invalid Phy ID: %x\n", PhyId);
}

void PhyLoopback(unsigned enableLoopback)
{
    if (PhyId == PHY_ID_AC101L)
        Ac101lPhyLoopback(enableLoopback);
    else if (PhyId == PHY_ID_KSZ8893MQL)
        diag_printf("loopback not implimented for KSZ8893MQL\n");
    else if (PhyId == PHY_ID_KENDIN)
        diag_printf("loopback not implimented for Kendin switch\n");
    else if ((PhyId  & PHY_ID_ICS1983_MASK) == PHY_ID_ICS1983)
        diag_printf("loopback not implimented for ICS1983\n");
    else
        diag_printf("Invalid Phy ID: %x\n", PhyId);
}


// MII interface
#define STACON_PHY_ADDR_MASK    0x1f
#define STACON_PHY_ADDR_OFFSET  5
#define STACON_WRITE            Bit(10)
#define STACON_BUSY             Bit(11)

// MII Interface Station Management Register Write
void MiiStationWrite(int phy_addr, int reg_addr, unsigned int data)
{
    STADATA = data;
    STACON = (reg_addr & STACON_PHY_ADDR_MASK) |
             ((phy_addr & STACON_PHY_ADDR_MASK) << STACON_PHY_ADDR_OFFSET) |
             STACON_WRITE | STACON_BUSY;
    while (STACON & STACON_BUSY);
}

// MII Interface Station Management Register Read
unsigned int MiiStationRead(int phy_addr, int reg_addr)
{
    STACON = (reg_addr & STACON_PHY_ADDR_MASK) |
             ((phy_addr & STACON_PHY_ADDR_MASK) << STACON_PHY_ADDR_OFFSET) |
             STACON_BUSY;
    while (STACON & STACON_BUSY);

    return STADATA ;
}
