//  Ethernet device driver for opencores MAC 
//    author: Fred Heyman (fred@digidescorp.com)
//    Work done for Comtrol/Altera
//
// Fixes by Grant Edwards for Comtrol, Inc.
//
//==========================================================================
//
//  Portions of this file derived from file "if_lan91cxxx.c", which bore the
//       following copyright and credit messages:
//
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
// Copyright (C) 2003 Nick Garnett 
// Copyright (C) 2004 Andrew Lunn
//
//==========================================================================
//  Raw interface portions of the code derived from file "dmrts_ether.c"
//==========================================================================

#include <cyg/infra/cyg_ass.h>
#include <pkgconf/system.h>
#include <pkgconf/devs_eth_oc.h>
#include <pkgconf/io_eth_drivers.h>
#include <cyg/infra/cyg_type.h>
#include <cyg/hal/hal_arch.h>
#include <cyg/hal/hal_intr.h>
#include <cyg/hal/hal_diag.h>
#include <cyg/infra/diag.h>
#include <cyg/hal/drv_api.h>
#include <cyg/io/eth/netdev.h>
#include <cyg/io/eth/eth_drv.h>
#ifdef CYGPKG_NET
#include <pkgconf/net.h>
#include <cyg/kernel/kapi.h>
#include <net/if.h>		/* Needed for struct ifnet */
#endif

#include "oc.h"
#include <net/etherio.h>
#include <cyg/error/codes.h>

#define UseCacheBypass 0

#if !UseCacheBypass
# include <cyg/hal/hal_cache.h>
#endif

#if 1
static void if_oc_assert_failed(void);

static void __assert(const char *file, int line, const char *cond, const char *msg)
{
  CYG_INTERRUPT_STATE old;
  HAL_DISABLE_INTERRUPTS(old);
  diag_printf("ASSERT FAILED: (%s)\n    %s\n    %s:%d\n",cond,msg,file,line);
  if_oc_assert_failed();
}
#define MY_ASSERT(e,m) ((e) ? (void)0 : __assert(__FILE__, __LINE__, #e, m))
#else
#define MY_ASSERT(e,m) /* noop */
#endif

#define OCREGS_ADDR  OPENCORES_ETHERNET_MAC_0_BASE
#define OCREGS_IRQ   OPENCORES_ETHERNET_MAC_0_IRQ

#define CYGNUM_DEVS_ETH_OC_INT_PRIO 4
#define MAC_PHY_ADDR CYGNUM_DEVS_ETH_OC_PHY_ADDR

#define DEBUG 1
#if DEBUG
#define db_printf diag_printf
#else
#define db_printf
#endif

#define MAX_PKT_SIZE      1540
#define MAX_BYTES_RX_PKT  1540

#define oc_regs   ((int*)OCREGS_ADDR)    // Base I/O address of controller
#define oc_interrupt_num OCREGS_IRQ	 // Interrupt vector used by controller

static cyg_interrupt oc_interrupt;
static cyg_handle_t  oc_interrupt_handle;

//this is not being used by the driver
// if it were necessary to make it work across multiple
// interfaces, the driver would likely be used
typedef struct
{
  int empty_struct;
} oc_priv_data;

// these aren't actually used for anything...
static struct eth_drv_mc_list the_mc_list;
static int mc_all_mode = 0;

// data pertaining to tx packets
#define TX_BUFS    CYGNUM_DEVS_ETH_OC_TXSLOTS  // must be a power of 2, and at most 64.  buffers consume 1.5kB each
static   unsigned tx_index = 0;                // index into circular queue of DMA descriptors
static   unsigned txd_index = 0;               // index into circular queue of DMA descriptors
volatile unsigned tx_pkts_queued   = 0;        // incremented as the buffer is verified to have been sent out the interface
volatile unsigned tx_pkts_complete = 0;        // incremented as a buffer is commanded to be sent out the interface

// data pertaining to rx packets
#define RX_BUFS    CYGNUM_DEVS_ETH_OC_RXSLOTS  // must be a power of 2, and at most 64.  buffers consume 1.5kB each
static unsigned rx_index = 0;	               // index into circular queue of DMA descriptors
static unsigned rx_packet_count;

static cyg_ether_info_t config;

//  this is the queue to hand off the received buffers to the normal ecos/bsd upper levels
#define RECV_QUEUE_DEPTH CYGNUM_DEVS_ETH_OC_RXQUEUE_SIZE	// must be a power of 2
tEthBuffer *recv_queue_bufs[RECV_QUEUE_DEPTH];
volatile int ix_last_pkt_into_recv_queue = 0;
volatile int ix_last_pkt_from_recv_queue = 0;

#if defined(CYGSEM_DEVS_ETH_OC_RAWAPI)
// this is the queue to hand off received buffers to the application via the raw interface
#define RAW_QUEUE_DEPTH CYGNUM_DEVS_ETH_OC_RAWQUEUE_SIZE	// must be a power of 2
tEthBuffer *raw_queue_bufs[RAW_QUEUE_DEPTH];
volatile int ix_last_pkt_into_raw_queue = 0;
volatile int ix_last_pkt_from_raw_queue = 0;
#endif

static cyg_drv_mutex_t txMutex;

static bool configDone = 0;
static bool oc_interface_up = 0;

#define EXTRA_BUFS    CYGNUM_DEVS_ETH_OC_EXTRA_BUFFERS
 // EXTRA_BUFS are used for:
 // (1) program allocation via the raw interface
 // (2) buffering of packets between when they are received by the dsr and when oc_recv() is called

#define NUM_ETH_BUFFERS  (TX_BUFS + RX_BUFS + EXTRA_BUFS)	// ~1.5kB static data used per buffer

#define EnableQueueCounts 1

typedef struct
{
  tEthBuffer *head;
  tEthBuffer *tail;
#if EnableQueueCounts
  int         count;
#endif  
} tEthBufQueue;

// queue used to hold available ethernet frame buffers
static tEthBufQueue freeList;
static int freeCount;
static tEthBuffer *allocBuffer(void);  // allocates a buffer from the freeList
static void freeBuffer(tEthBuffer * b); // returns a buffer to the freeList

void if_oc_dumpFreeList(void)
{
  tEthBuffer *p;
  
  diag_printf("freeCount %d\n",freeCount);
  diag_printf("freeList.count %d\n",freeList.count);
  diag_printf("freeList.head %p\n",freeList.head);
  diag_printf("freeList.tail %p\n",freeList.tail);
  diag_printf("freeList: \n");
  p = freeList.head;
  while (p)
    {
      diag_printf("   %p\n",p);
      p = p->next;
    }
}

// pool of ethernet buffers
static tEthBuffer __bufferPool[NUM_ETH_BUFFERS] __attribute__ ((aligned(16)));
#if UseCacheBypass
# define bufferPool ((tEthBuffer*)(  UNCACHED( __bufferPool ) ))
#else
# define bufferPool __bufferPool
#endif

// macros used to access the MAC's buffer descriptor array
#define IX_BD_TX0    0x100
#define IX_BD_RX0    0x180

#define TXBD_BUFP(i) (IX_BD_TX0+(2*i)+1)
#define TXBD_STAT(i) (IX_BD_TX0+(2*i))

#define RXBD_BUFP(i) (IX_BD_RX0+(2*i)+1)
#define RXBD_STAT(i) (IX_BD_RX0+(2*i))

#define KeepIsrStats 0

static int isrcnt = 0;
static int dsrcnt = 0;
#if KeepIsrStats
static int oc_isrcntx[8];
#endif

static void oc_poll(struct eth_drv_sc *sc);
static void print_isr_info(char *c);
static void dump_rxbufs(void);
static void dump_txbufs(void);
static void initFreeList(void);
static int bufs_free_now(void);

#if defined(CYGSEM_DEVS_ETH_OC_SWRXCRC) || defined(CYGSEM_DEVS_ETH_OC_SWTXCRC)
static cyg_uint32 crc32( unsigned char *blk_adr, unsigned blk_len);
#endif


#if DEBUG
static void bindump(char *descr, void *data, int numbytes)
{
  char tiny[4];
  unsigned char *ch;
  int ix, n, i;
  n = numbytes;
  db_printf("%s: (%d bytes)\n", descr, n);
  ch = (unsigned char *) data;
  ix = 0;
  tiny[1] = 0;
  while (ix < n)
    {
      for (i = 0; i < 16; i++)
	  if (ix + i < n)
            db_printf("%02x ", ch[ix + i]);
	  else
            db_printf("   ");
      db_printf("  ");
      for (i = 0; i < 16; i++)
	  if (ix + i < n)
	    {
	      tiny[0] = ch[ix + i];
	      if ((tiny[0] < 33) || (tiny[0] > 126))
                tiny[0] = '.';
	      db_printf(tiny);
	    }
	  else
            db_printf(" ");
      db_printf("\n");
      ix += 16;
    }
  db_printf("\n");
}
#endif //DEBUG



// Nios II specific code
// slightly modified from altera's io.h
// when accessing registers, these must be used to avoid weird caching issues
#define RD_32_DIRECT( ADDR        )  __builtin_ldwio(   (void*)(ADDR) )
#define RD_16_DIRECT( ADDR        )  __builtin_ldhuio(  (void*)(ADDR) )
#define RD_8_DIRECT(  ADDR        )  __builtin_ldbuio(  (void*)(ADDR) )

#define WR_32_DIRECT( ADDR, DATA  )  __builtin_stwio(  (void*)(ADDR), (DATA) )
#define WR_16_DIRECT( ADDR, DATA  )  __builtin_sthio(  (void*)(ADDR), (DATA) )
#define WR_8_DIRECT(  ADDR, DATA  )  __builtin_stbio(  (void*)(ADDR), (DATA) )


// Note: this uses the bit 31 to specify that the area should not
// be used for data cache.  According to the "Nios II Software Developer's
// handbook, section "Bit-31 Cache Bypass", this should work for all current
// nios II versions.  However, "In general, this feature is dependent on the
// Nios II core implementation.  Future Nios II cores may use bit 31 for other
// purposes

#if UseCacheBypass
# define UNCACHED(x) ((void*)(((int)(x))|0x80000000))
# define VALID_ADDR_COMPONENT(x) ((void*)((int)x&0x7fffffff))
#else
# define UNCACHED(x) ((void*)(x))
# define VALID_ADDR_COMPONENT(x) ((void*)(x))
#endif

#if 1
#define write_oc_reg(reg,data)  WR_32_DIRECT(&oc_regs[reg],data)
#define read_oc_reg(reg)        RD_32_DIRECT(&oc_regs[reg])
#else
int read_oc_reg(int reg)
{
  unsigned int val;
  val = RD_32_DIRECT(&oc_regs[reg]);
  return (val);
}
static void write_oc_reg(int reg, unsigned int data)
{
  WR_32_DIRECT(&oc_regs[reg],data);
}
#endif


//  MII stuff

// wait for the mii control interface to free up
// if it times out, 1 is returned
// if it is OK, 0 is returned
// note - this should probably be implemented as an absolute timer
#define MII_TIMEOUT_COUNT 1000
int wait_for_mii_to_free(void)
{
  int cnt = 0;
  // just in case it was busy, we wait for it to complete
  while (1)
    {
      if (!(read_oc_reg(OC_REG_MIISTATUS) & OC_REG_MIISTATUS_BUSY))
        return 0;
      if (++cnt >= MII_TIMEOUT_COUNT)
        return 1;
    }
}


// read_mii_reg()
//   mii reg should be 0-31, the register to read
//   this function returns the 16 bit register value
//   if the read times out, it returns -1
int mii_read(int mii_reg)
{
  int cnt, maxcnt, done, stat, i;
  int areg, creg;
  maxcnt = 5000;
  cnt = 0;
  done = 0;

  // wait for the mii control interface to free up
  if (wait_for_mii_to_free())
    {
      db_printf("to1\n");
      return (-1);
    }

  // issue the command to execute the read 
  write_oc_reg(OC_REG_MIIADDRESS, ((mii_reg & 31) << 8) | MAC_PHY_ADDR);
  write_oc_reg(OC_REG_MIICOMMAND, OC_REG_MIICOMMAND_READ);

  // wait for the mii read operation to complete
  if (wait_for_mii_to_free())
    {
      db_printf("to2\n");
      return (-1);
    }

  // return the read value
  return (read_oc_reg(OC_REG_MIIRX_DATA));
}

// mii_write()
// returns 0 if good; -1 if there was a problem
int mii_write(int mii_reg, int val)
{

  // wait for the mii control interface to free up
  if (wait_for_mii_to_free())
    return (-1);

  write_oc_reg(OC_REG_MIIADDRESS, ((mii_reg & 31) << 8) | MAC_PHY_ADDR);
  write_oc_reg(OC_REG_MIITX_DATA, val);
  write_oc_reg(OC_REG_MIICOMMAND, OC_REG_MIICOMMAND_WRITE);

  // wait for the write operation to complete mii control interface to free up
  if (wait_for_mii_to_free())
    return (-1);

  return (0);
}

// status and quickpoll registers have latching bits, so you need to
// read them twice if you want to get the current value
#define ReadPhyStatus()     (mii_read(MII_STATUS),mii_read(MII_STATUS))
#define ReadPhyQuickpoll()  (mii_read(MII_DIAGNOSTICS),mii_read(MII_DIAGNOSTICS))

// resets the mii and performs the autonegocation
static void restart_autonegotiation(void)
{
  cyg_uint16 u;
  int rd,cnt;

  // read PHY ID registers
  db_printf("PHY ID = %04x:%04x\n",mii_read(MII_PHYID_1),mii_read(MII_PHYID_2));
  
  // set autonotiate advertisement capabilities
  u = mii_read(MII_AUTONEG_ADVERTISE);
  u &= ~(Bit(5)|Bit(6)|Bit(7)|Bit(8));
  u |= (CYGSEM_DEVS_ETH_OC_ADVERTISE_10M_HD  ? Bit(5) : 0) |
       (CYGSEM_DEVS_ETH_OC_ADVERTISE_10M_FD  ? Bit(6) : 0) |
       (CYGSEM_DEVS_ETH_OC_ADVERTISE_100M_HD ? Bit(7) : 0) |
       (CYGSEM_DEVS_ETH_OC_ADVERTISE_100M_FD ? Bit(8) : 0);
  mii_write(MII_AUTONEG_ADVERTISE, u);
  
  // start autonegotiation process
  db_printf("starting autonegotiation ...\n");
  mii_write(MII_CONTROL, MII_CONTROL_ANEG_ENABLE | MII_CONTROL_RESTART_ANEG);
  cnt = 0;
  while (!(ReadPhyStatus() & MII_STATUS_ANEG_COMPLETE))
    {
      cnt++;
      if (cnt > 100000)
        {
          db_printf("autonegotiation timed out\n");
          break;
        }
    }
  db_printf("autonegotiation took %d counts\n", cnt);
}

// check PHY status and make sure MAC is configured appropriately
static void sync_with_phy(void)
{
  cyg_uint16 rd;

  db_printf("mii status  reg: 0x%04x\n", ReadPhyStatus());
  db_printf("mii control reg: 0x%04x\n", mii_read(MII_CONTROL));

  rd = ReadPhyQuickpoll();
  if (!(rd & MII_DIAGNOSTICS_LINK))
    {
      diag_printf("link down\n");
      return;
    }
  
  diag_printf("link up: ");
  
  // Set full/half duplex mode according to autonegotiation results

  unsigned r = read_oc_reg(OC_REG_MODER);
  if (rd & MII_DIAGNOSTICS_FULL_DPLX)
   {
      db_printf("full duplex ");
      write_oc_reg(OC_REG_MODER, r | OC_REG_MODER_REG_FULLD);
      write_oc_reg(OC_REG_IPGT, OC_REG_IPGT_FULL_DUPLEX_RECOMMENDED);
      write_oc_reg(OC_REG_CTRLMODER, OC_REG_CTRLMODER_TXFLOW | OC_REG_CTRLMODER_RXFLOW);
   }
  else
   {
      db_printf("half duplex ");
      write_oc_reg(OC_REG_MODER, r & ~OC_REG_MODER_REG_FULLD);
      write_oc_reg(OC_REG_IPGT, OC_REG_IPGT_HALF_DUPLEX_RECOMMENDED);
      write_oc_reg(OC_REG_CTRLMODER, 0);
   }

  if (rd & MII_DIAGNOSTICS_RATE_100)
      db_printf("100 Mbps\n");
  else
      db_printf("10 Mbps\n");
}


static void reset_mii_interface(void)
{
  restart_autonegotiation();
  sync_with_phy();
}


static void print_mii_regs(void)
{
  int i, r;

  db_printf("mii regs (new):\n");
  for (i = 0; i < 32; i++)
    {
      r = mii_read(i),mii_read(i);
      db_printf(" %2d: 0x%04x\n", i, r);
    }
}

static int process_a_received_packet(struct eth_drv_sc *sc, unsigned really);
static int process_a_transmitted_packet(struct eth_drv_sc *sc);

unsigned oc_dsr_tx_count;
unsigned oc_dsr_rx_count;
unsigned volatile oc_interrupt_src;

static void oc_dsr(cyg_uint32 vector, cyg_ucount32 count, cyg_addrword_t data)
{
  struct eth_drv_sc *sc = (struct eth_drv_sc*)data;
  cyg_uint32 cause;
  CYG_INTERRUPT_STATE old;

  // grab the status that was saved by the isr.
  HAL_DISABLE_INTERRUPTS(old);
  cause = oc_interrupt_src;
  oc_interrupt_src = 0;
  HAL_RESTORE_INTERRUPTS(old);

  dsrcnt++;

  // for some reason, if we only call process_a_transmitted_packet()
  // when the appropriate bits are set in "cause", things eventually
  // lock-up/reset.  So we'll call it every time there's an interrupt
  // -- if there's nothing to do, it will return immediately.
  if (1 || (cause & (OC_REG_INT_SOURCE_TXB|OC_REG_INT_SOURCE_TXE)))
    {
      ++oc_dsr_tx_count;
      while (process_a_transmitted_packet(sc))
        ;
    }
  if (cause & (OC_REG_INT_SOURCE_RXB|OC_REG_INT_SOURCE_RXE))
    {
      ++oc_dsr_rx_count;
      eth_drv_dsr(vector, count, data);
    }
}

static unsigned dropped_packet_count;

// This ISR is called when the ethernet interrupt occurs
static int oc_isr(cyg_vector_t vector, cyg_addrword_t data)
{
  int src;
  // get interrupt status and clear all interrupts
  src = read_oc_reg(OC_REG_INT_SOURCE);
  write_oc_reg(OC_REG_INT_SOURCE,src);
  cyg_drv_interrupt_acknowledge(oc_interrupt_num);
  isrcnt++;
  oc_interrupt_src |= src;
  // keep a few statistics
#if KeepIsrStats
  {
    unsigned  i;
    for (i = 0; i < 8; i++)
      if ((src & (1 << i)))
        oc_isrcntx[i]++;
  }
#endif
  if (src & OC_REG_INT_SOURCE_BUSY)
    ++dropped_packet_count;

  return (CYG_ISR_HANDLED | CYG_ISR_CALL_DSR);	// Run the DSR
}

// oc_deliver() is called by eth_drv_run_deliveries() which is called
// by a network alarm thread.

static unsigned deliver_count;

static void oc_deliver(struct eth_drv_sc *sc)
{
  ++deliver_count;
  if (process_a_received_packet(sc, 1))
    while (process_a_received_packet(sc, 0))
      ;
}

// this is apparently used by GDB
static int oc_int_vector(struct eth_drv_sc *sc)
{
  return (oc_interrupt_num);
}

#if DEBUG
static void dump_rxbufs(void)
{
  unsigned i,stat,bufp;
//  db_printf("read_oc_reg(OC_REG_RX_DMA_IDX) %x\n",read_oc_reg(OC_REG_RX_DMA_IDX));
  db_printf("rx_index = %d\n",rx_index);
  db_printf("rxbufs:\n");
  for (i = 0; i < RX_BUFS; i++)
    {
      bufp = read_oc_reg(RXBD_BUFP(i));
      stat = read_oc_reg(RXBD_STAT(i));
      db_printf("%c%2d: %08x %08x\n", i==rx_index ? '>':' ',i, stat, bufp);
    }
}

static void dump_txbufs(void)
{
  unsigned i,stat,bufp;
//  db_printf("read_oc_reg(OC_REG_TX_DMA_IDX) %x\n",read_oc_reg(OC_REG_TX_DMA_IDX));
  db_printf("tx_index = %d\n",tx_index);
  db_printf("txd_index = %d\n",txd_index);
  db_printf("txbufs:\n");
  for (i = 0; i < TX_BUFS; i++)
    {
      char c;
      bufp = read_oc_reg(TXBD_BUFP(i));
      stat = read_oc_reg(TXBD_STAT(i));
      if (i==txd_index && i==tx_index)
        c = '#';
      else if (i==txd_index)
        c = '+';
      else if (i==tx_index)
        c = '>';
      else
        c = ' ';
      db_printf("%c%2d: %08x %08x\n", c, i, stat, bufp);
    }
}
#endif //DEBUG

// since tx bd index in MAC doesn't get reset when we disable
// TXEN, this routine fills bd table with packet descriptors
// with "wrap" bit set so that we can force the tx index back
// to 0
static void resync_tx_bds(void)
{
  int i;
  db_printf("resync_tx_bds:");
  for (i = 0; i<TX_BUFS; ++i)
    {
      write_oc_reg(TXBD_BUFP(i), 0x8000000);
      write_oc_reg(TXBD_STAT(i), (128 << 16) | BD_TX_PKT_BEING_USED | BD_TX_PAD_EN | BD_TX_CRC_EN | BD_TX_WRAP);
    }
  // wait for slot zero to be cleared
  while (read_oc_reg(TXBD_STAT(0)) & BD_TX_PKT_BEING_USED)
    db_printf(".");
  db_printf("done\n");
  for (i = 0; i<TX_BUFS; ++i)
    {
      write_oc_reg(TXBD_STAT(i), 0);
      write_oc_reg(TXBD_BUFP(i), 0);
    }
}

static void configure_rx_bds(void)
{
  int i, stat;
  tEthBuffer *buffer;

  for (i = 0; i < RX_BUFS; i++)
    {
      stat = (MAX_BYTES_RX_PKT << 16) | BD_RX_EMPTY | BD_RX_IRQ | ((i == RX_BUFS - 1) ? BD_RX_WRAP : 0);
      buffer = allocBuffer();
      if (buffer == NULL)
        db_printf("ERROR: buffer alloc prob in configure_rx_bds()\n");
      else
	{
	  write_oc_reg(RXBD_BUFP(i), (unsigned)buffer);
	  write_oc_reg(RXBD_STAT(i), stat);
	}
    }
}

static void get_oc_mac(unsigned char *m)
{
  cyg_uint32 u;
  u = read_oc_reg(OC_REG_MAC_ADDR0);
  m[2] = u>>24;
  m[3] = u>>16;
  m[4] = u>>8;
  m[5] = u;
  u = read_oc_reg(OC_REG_MAC_ADDR1);
  m[0] = u>>8;
  m[1] = u;
}

static void set_oc_mac(unsigned char *m)
{
  // set the mac address  
  //db_printf("OC fixed mac: %02x:%02x:%02x:%02x:%02x:%02x\n", m[0], m[1], m[2], m[3], m[4], m[5]);
  write_oc_reg(OC_REG_MAC_ADDR0, (m[2] << 24) | (m[3] << 16) | (m[4] << 8) | (m[5]));
  write_oc_reg(OC_REG_MAC_ADDR1, (m[0] << 8) | (m[1]));
  //db_printf("mac address regs: 0x%08x  0x%08x\n", read_oc_reg(OC_REG_MAC_ADDR1), read_oc_reg(OC_REG_MAC_ADDR0));
}


static bool oc_init(struct cyg_netdevtab_entry *tab)
{
  int i;
  struct eth_drv_sc *sc = (struct eth_drv_sc *) tab->device_instance;

  db_printf("oc_init() begin\n");
  
  db_printf("%d ethBuffers\n", NUM_ETH_BUFFERS);
  
  get_oc_mac(config.mac_address);

  db_printf("resetting mii interface and performing autonegotiation...\n");
  reset_mii_interface();

  // set up the OS's interrupt handling
  cyg_drv_interrupt_create(oc_interrupt_num, CYGNUM_DEVS_ETH_OC_INT_PRIO, (cyg_addrword_t) sc,	//  Data item passed to interrupt handler
			   (cyg_ISR_t *) oc_isr, (cyg_DSR_t *) oc_dsr,	// The logical driver DSR
			   &oc_interrupt_handle, &oc_interrupt);
  db_printf("setting up interrupts. irq=%d\n", oc_interrupt);
  cyg_drv_interrupt_attach(oc_interrupt_handle);

  cyg_drv_mutex_init(&txMutex);
  
  // after initialization, this must be called
  sc->funs->eth_drv->init(sc, &config.mac_address[0]);

  db_printf("init complete\n");
  configDone = 1;
  return true;
}

static void oc_stop(struct eth_drv_sc *sc)
{
  volatile unsigned r;
  db_printf("oc_stop() begin\n");
  cyg_drv_interrupt_mask(oc_interrupt_num);
  write_oc_reg(OC_REG_INT_MASK,   0x0);        // disallow all ethernet interrrupts
  write_oc_reg(OC_REG_INT_SOURCE, 0x7f);       // clear any pending interrupts
  r = read_oc_reg(OC_REG_MODER);
  write_oc_reg(OC_REG_MODER, r & ~(OC_REG_MODER_REG_TXEN | OC_REG_MODER_REG_RXEN)); // disable tx and rx
  oc_interface_up = 0;
}

static void oc_start(struct eth_drv_sc *sc, unsigned char *enaddr, int flags)
{
  volatile unsigned r;
  db_printf("oc_start() begin\n");
  cyg_drv_interrupt_mask(oc_interrupt_num);
  write_oc_reg(OC_REG_INT_MASK,   0x0);        // disallow all ethernet interrrupts
  write_oc_reg(OC_REG_INT_SOURCE, 0x7f);       // clear any pending interrupts

  r = read_oc_reg(OC_REG_MODER);

  // disable tx and rx
  write_oc_reg(OC_REG_MODER, r & ~(OC_REG_MODER_REG_TXEN | OC_REG_MODER_REG_RXEN));

    
  // spec says this bit is reserved, but linux driver uses it to reset the controller
  // doesn't seem to work...
  //write_oc_reg(OC_REG_MODER, r|0x800);         
  //write_oc_reg(OC_REG_MODER, r);              

  // resync transmit and receive queues

  //db_printf("read_oc_reg(OC_REG_TX_DMA_IDX) %x\n",read_oc_reg(OC_REG_TX_DMA_IDX));
  //db_printf("read_oc_reg(OC_REG_RX_DMA_IDX) %x\n",read_oc_reg(OC_REG_RX_DMA_IDX));
  
  //tx_index = txd_index = read_oc_reg(OC_REG_TX_DMA_IDX)/2;
  tx_index = txd_index = 0;
  //db_printf("tx_index = %d\n",tx_index);
  //rx_index = (read_oc_reg(OC_REG_RX_DMA_IDX)-(IX_BD_RX0&0xff))/2;
  rx_index = 0;
  //db_printf("rx_index = %d\n",rx_index);
  
  rx_packet_count = 0;
  ix_last_pkt_into_recv_queue = 0;
  ix_last_pkt_from_recv_queue = 0;
#if defined(CYGSEM_DEVS_ETH_OC_RAWAPI)  
  ix_last_pkt_into_raw_queue = 0;
  ix_last_pkt_from_raw_queue = 0;
#endif  

  tx_pkts_queued = 0;
  tx_pkts_complete = 0;
  
  initFreeList();
  configure_rx_bds();

  // re-enable tx/rx
  write_oc_reg(OC_REG_MODER, r | (OC_REG_MODER_REG_TXEN | OC_REG_MODER_REG_RXEN));
  
  //db_printf("read_oc_reg(OC_REG_TX_DMA_IDX) %x\n",read_oc_reg(OC_REG_TX_DMA_IDX));
  //db_printf("read_oc_reg(OC_REG_RX_DMA_IDX) %x\n",read_oc_reg(OC_REG_RX_DMA_IDX));
  
  write_oc_reg(OC_REG_INT_SOURCE, 0x7f );		// clear any pending interrupts
  write_oc_reg(OC_REG_INT_MASK,   0x7f );		// allow all ethernet interrrupts
  cyg_drv_interrupt_acknowledge(oc_interrupt_num);
  cyg_drv_interrupt_unmask(oc_interrupt_num);
  oc_interface_up = 1;
  db_printf("oc_start() end\n");
}


char *str_from_mac(unsigned char *mac)
{
  static char hex[] = "0123456789abcdef";
  static char buf[18];
  unsigned char *p = buf;
  int i;
  for (i = 0; i < 6; ++i)
    {
      *p++ = hex[(mac[i] >> 4) & 0xf];
      *p++ = hex[(mac[i]) & 0xf];
      *p++ = ':';
    }
  buf[17] = '\0';
  return buf;
}


static int oc_control(struct eth_drv_sc *sc, unsigned long key, void *data, int data_length)
{
  //db_printf("oc_control() begin.  Key = 0x%x  data=0x%x  len=%d\n", key, (int) data, data_length);

  if (key == ETH_DRV_SET_MC_LIST)
    {
      // I've no clue what this control is expected to do, but
      // it currently does nothing but copy data into a buffer
      // that is never looked at
      int n, i, j;
      struct eth_drv_mc_list *mc_list;
      mc_list = (struct eth_drv_mc_list *) data;
      n = mc_list->len;
      for (i = 0; i < n; i++)
        db_printf("enabling mc addr %d: %s\n", i, str_from_mac(&mc_list->addrs[i][0]));
      memcpy(&the_mc_list, data, sizeof(struct eth_drv_mc_list));
      mc_all_mode = 0;
    }
  else if (key == ETH_DRV_SET_MC_ALL)
    {
      // again, this does nothing becuase mc_all_mode is never
      // looked at
      db_printf("all mc addresses being allowed\n");
      mc_all_mode = 1;
    }
  else if (key == ETH_DRV_SET_MAC_ADDRESS)
    {
      // this is the only control operation that does anyting.
      db_printf("setting mac address to %s\n", str_from_mac((unsigned char *) data));
      memcpy(&config.mac_address[0], data, 6);
      set_oc_mac((unsigned char *) data);
    }
  else
    {
      db_printf("oc_control() not handling key type 0x%x\n", key);
      return (-ENOSUPP);
    }
  return (ENOERR);
}


// This routine is called to see if it is possible to send another packet.
// It will return non-zero if a transmit is possible, zero otherwise.

static int num_tx_slots_free(void)
{
  return (TX_BUFS - (int)(tx_pkts_queued - tx_pkts_complete));
}

#define  min2(  a, b )  ( (a)<(b) ? (a) : (b) )

static int oc_can_send(struct eth_drv_sc *sc)
{
  int rval;

  int num_tx_free = num_tx_slots_free();	// number of tx slots free
  int num_bufs_free = bufs_free_now();

  if ((num_tx_free < 0) || (num_tx_free > TX_BUFS))
    {
      db_printf("error: bad num tx slots free (%d)\n", num_tx_free);
      num_tx_free = 0;
    }

  rval = min2(num_tx_free, num_bufs_free);
  
  //if (rval < 4) db_printf("ocs:%d\n", rval);

  return (rval);
}

#ifdef DEBUG
static void print_isr_info(char *c)
{
  int i;
  db_printf("%s: isrcnt:%d  dsrcnt:%d  ", c, isrcnt, dsrcnt);
#if KeepIsrStats
  for (i = 0; i < 8; i++)
    db_printf("%2d  ", oc_isrcntx[i]);
#endif  
  db_printf("\n");
}
#endif


// This routine is called to send data to the hardware.
static void oc_send(struct eth_drv_sc *sc, struct eth_drv_sg *sg_list, int sg_len, int total_len, unsigned long key)
{
  int i, stat, num_slots_free, nb;
  tEthBuffer *buffer;
  struct eth_drv_sg *sg;
  
  sg = sg_list;
  if (total_len >= MAX_PKT_SIZE)
    {
      db_printf("ERROR: packet too big to tx!\n");
      return;
    }
  
  if (oc_can_send(sc) <= 0)
    return;
  
  buffer = allocBuffer();
  if (!buffer)
    {
      db_printf("ERROR: cannot allocate buf for tx");
      return;
    }

  // copy data from sg list into buffer
  nb = 0;
  for (i = 0; i < sg_len; i++)
    {
      memcpy(buffer->data+nb, (void *)sg->buf, sg->len);
      nb += sg->len;
      sg++;
    }

  // tell stack we're done with sg
  sc->funs->eth_drv->tx_done(sc, key, 0);

#if !UseCacheBypass
  // flush data cache for data we just copied
  HAL_DCACHE_FLUSH(buffer->data,nb);
#endif
  
  // queue up tx descriptor

  cyg_drv_mutex_lock(&txMutex);
  i = tx_index;

  stat = (nb << 16) | BD_TX_PKT_BEING_USED | BD_TX_IRQ | BD_TX_CRC_EN | BD_TX_PAD_EN;
  if (i == (TX_BUFS - 1))
    stat |= BD_TX_WRAP;
  
  write_oc_reg(TXBD_BUFP(i), (unsigned)buffer);
  write_oc_reg(TXBD_STAT(i), stat);
  
  //db_printf("read_oc_reg(OC_REG_TX_DMA_IDX) %x\n",read_oc_reg(OC_REG_TX_DMA_IDX));
  //db_printf("read_oc_reg(OC_REG_RX_DMA_IDX) %x\n",read_oc_reg(OC_REG_RX_DMA_IDX));
  //db_printf("t%d,%d\n",i,stat>>16);

  tx_index = (i+1) & (TX_BUFS-1);
  tx_pkts_queued++;
  cyg_drv_mutex_unlock(&txMutex);
}


#define MAX_BYTES_SHOW	64
void dump_sgs(char *descr, struct eth_drv_sg *sg_list, int sg_len)
{
  int numb = 0, nc = 0, i, j;
  struct eth_drv_sg *sg;
  char buf[MAX_BYTES_SHOW];
  char newdescr[128];
  char *bufs;
  sg = sg_list;
  for (i = 0; i < sg_len; i++)
    {
      numb += sg->len;
      sg++;
    }

  if (numb > MAX_BYTES_SHOW)
    {
      sprintf(newdescr, "%s %d/%d", descr, MAX_BYTES_SHOW, numb);
      numb = MAX_BYTES_SHOW;
    }
  else
    {
      sprintf(newdescr, "%s", descr);
    }

  nc = 0;
  sg = sg_list;
  for (i = 0; i < sg_len; i++)
    {
      for (j = 0; j < sg->len; j++)
	{
	  bufs = (char *) sg->buf;
	  if (nc < numb)
            buf[nc] = bufs[j];
	  nc++;
	}
      sg++;
    }
  bindump(newdescr, buf, numb);
}


// oc_recv()
//  This is called by (sc->funs->eth_drv->recv)(sc, len);               
static void oc_recv(struct eth_drv_sc *sc, struct eth_drv_sg *sg_list, int sg_len)
{
  int i;
  struct eth_drv_sg *sg;
  unsigned char *p;
  tEthBuffer *buffer;
  int bytes = 0;

  buffer = recv_queue_bufs[(ix_last_pkt_from_recv_queue++) & (RECV_QUEUE_DEPTH - 1)];
  p = buffer->data;
  sg = sg_list;
  for (i = 0; i < sg_len; i++)
    {
      memcpy((void*)sg->buf,p,sg->len);
      p += sg->len;
      bytes += sg->len;
      sg++;
    }
  freeBuffer(buffer);
}

static unsigned rx_packet_count;


// handle a received packet, if one exists
//  returns 1 if a packet was in fact processed
//  returns 0 if there was no packet to process
static int process_a_received_packet(struct eth_drv_sc *sc, unsigned really)
{
  unsigned stat, i, len;
  unsigned short proto;
  tEthBuffer *buffer,*newbuffer;
  
  i = rx_index;
  stat = read_oc_reg(RXBD_STAT(i));

  //db_printf("r%d,%04x,%d\n",i,stat&0xffff,stat>>16);
  //db_printf("TX_DMA_IDX %x\n",read_oc_reg(OC_REG_TX_DMA_IDX));
  //db_printf("RX_DMA_IDX %x\n",read_oc_reg(OC_REG_RX_DMA_IDX)-0x80);
  
  if (really && (stat & BD_RX_EMPTY))
    {
      //diag_printf("really\n");
      // there's supposed to be a packet, so keep looking
      unsigned start = i;
      do
        {
          i = (i+1) & (RX_BUFS-1);
          stat = read_oc_reg(RXBD_STAT(i));
        } while((i != start) && (stat & BD_RX_EMPTY));
    }
  
  //db_printf(" r%d,%04x,%d\n",i,stat&0xffff,stat>>16);

  if (stat & BD_RX_EMPTY)
    return 0;

  buffer = (tEthBuffer*) read_oc_reg(RXBD_BUFP(i));
  
  newbuffer = allocBuffer();
  if (newbuffer == NULL)
    db_printf("No free Ethernet buffers!\n");  // reuse old buffer
  else
    write_oc_reg(RXBD_BUFP(i), (unsigned)newbuffer);
      
  write_oc_reg(RXBD_STAT(i), (MAX_BYTES_RX_PKT << 16) | BD_RX_EMPTY | BD_RX_IRQ | ((i == (RX_BUFS-1)) ? BD_RX_WRAP : 0));
  
  rx_index = (i+1) & (RX_BUFS-1);
  ++rx_packet_count;
  
  if (newbuffer == NULL)
    return 1;  // buffer got reused, so we're done

  if (stat & BD_RX_ANY_ERROR)
    {
      db_printf("RX packet error.  id=%d  stat=0x%08x\n", i, stat);
      freeBuffer(buffer);
      return 1;
    }
  
  len = (stat >> 16) & 0xffff;
  proto = *((cyg_uint16*)(buffer->data+12));
  //db_printf( "RXB  id:%d  len:%d  p:0x%x\n", i,  len, proto);

#if !UseCacheBypass
  // flush data cache for newly receive buffer
  HAL_DCACHE_FLUSH(buffer->data,len);;
#endif
  
#if defined(CYGSEM_DEVS_ETH_OC_RAWAPI)      
  if (config.protocol && ((config.protocol == 0xffff) || (config.protocol == proto)))
    {
      #if defined(CYGSEM_DEVS_ETH_OC_SWRXCRC)
        {
          unsigned crcOK;
          cyg_uint32 crc;
          crc = crc32(buffer->data,len-4);
          crcOK = ((cyg_uint8)(crc>>0) == buffer->data[(len-4)+0] &&
                   (cyg_uint8)(crc>>8) == buffer->data[(len-4)+1] &&
                   (cyg_uint8)(crc>>16) == buffer->data[(len-4)+2] &&
                   (cyg_uint8)(crc>>24) == buffer->data[(len-4)+3]);
          if (!crcOK)
            {
              db_printf("RX packet error.  bad CRC\n");
              freeBuffer(buffer);
              return 1;
            }
        }
      #endif      
      //db_printf( "sending packet to raw api (%d %d)\n", ix_last_pkt_into_raw_queue, ix_last_pkt_from_raw_queue );
      buffer->length = len - 4;       // apps expect length to not include the 4 CRC bytes
      if ((int)(ix_last_pkt_into_raw_queue - ix_last_pkt_from_raw_queue) < RAW_QUEUE_DEPTH)
        raw_queue_bufs[(ix_last_pkt_into_raw_queue++) & (RAW_QUEUE_DEPTH-1)] = buffer;
      else
        {
          db_printf("raw queue full! Dropping incoming pkt\n");
          freeBuffer(buffer);
        }
    }
  else
#endif
    {
      if ((int)(ix_last_pkt_into_recv_queue - ix_last_pkt_from_recv_queue) < RECV_QUEUE_DEPTH)
        {
          //db_printf( "recvd pkt %d w/ len %d\n", ix_last_pkt_into_recv_queue, len );
          recv_queue_bufs[(ix_last_pkt_into_recv_queue++) & (RECV_QUEUE_DEPTH - 1)] = buffer;
          (sc->funs->eth_drv->recv) (sc, len);	// this will call oc_recv(), which will eventually "process" the packet
        }
      else
        {
          db_printf("receive queue full!  dropping packet\n");
          freeBuffer(buffer);
        }
    }
  return 1;
}

// process_a_transmitted_packet() - handle a transmitted packet, if one exists
// returns 1 if it was able to transmit a packet
// returns 0 otherwise
static int process_a_transmitted_packet(struct eth_drv_sc *sc)
{
  unsigned int txkey;
  int len, i, stat, bufp, n;
  tEthBuffer *buffer;

  if (tx_pkts_queued == tx_pkts_complete)
    return 0;  // no pending tx packets
  
  i = txd_index;
  stat = read_oc_reg(TXBD_STAT(i));

  //db_printf("read_oc_reg(OC_REG_TX_DMA_IDX) %x\n",read_oc_reg(OC_REG_TX_DMA_IDX));
  //db_printf("read_oc_reg(OC_REG_RX_DMA_IDX) %x\n",read_oc_reg(OC_REG_RX_DMA_IDX));
  //db_printf("T%d,%d:%08x\n",i,read_oc_reg(OC_REG_TX_DMA_IDX),stat);
  
  if (stat & BD_TX_PKT_BEING_USED)
    return 0;  // this packet has not yet been sent, so we can't process it yet

  if (stat & BD_TX_ANY_ERROR)
    db_printf("TX error (stat:%08x buffer:%p)\n", stat, buffer);

  buffer = (tEthBuffer*) read_oc_reg(TXBD_BUFP(i));

  MY_ASSERT(buffer!=NULL, "tx handler found null buffer");
  
  if (!buffer)
    db_printf("TX error: null buffer\n");
  else
    {
      write_oc_reg(TXBD_BUFP(i), 0);
      freeBuffer(buffer);
    }
  txd_index = (i+1) & (TX_BUFS-1);
  tx_pkts_complete++;
  return 1;
}

// polling routine used by Redboot
static void oc_poll(struct eth_drv_sc *sc)
{
  if (process_a_received_packet(sc, 1))
    while (process_a_received_packet(sc, 0))
      ;
  while (process_a_transmitted_packet(sc))
    ;
}

static oc_priv_data ocdev_priv_data = 
{
  // not used
};

ETH_DRV_SC(ocdev_sc, &ocdev_priv_data, "eth0", oc_start, oc_stop, oc_control, oc_can_send, oc_send, oc_recv, oc_deliver, oc_poll, oc_int_vector);

NETDEVTAB_ENTRY(oc_netdev, "opencores_eth0", oc_init, &ocdev_sc);


//============================== Buffer Queue Manipulation Functions ==============================

static void ethBufQueueClear(tEthBufQueue * q)
{
  q->head = NULL;
  q->tail = NULL;
#if EnableQueueCounts
  q->count = 0;
#endif
}

//  Get the next EthBuf from the linked list q.  
static tEthBuffer *ethBufQueueGet(tEthBufQueue * q)
{
  tEthBuffer *r;
  r = q->head;
  if (r)
    {
      q->head = r->next;
#if EnableQueueCounts
      --q->count;
#endif
    }
  return r;
}

//  Add an EthBuf to the linked list q.  
static void ethBufQueuePut(tEthBufQueue * q, tEthBuffer * b)
{
  b->next = NULL;

  if (!q->head)
    {
      q->head = b;
      q->tail = b;
    }
  else
    {
      q->tail->next = b;
      q->tail = b;
    }
#if EnableQueueCounts
  ++q->count;
#endif
}

// not used at the moment
static bool ethBufQueueEmpty(tEthBufQueue * q)
{
  return q->head != NULL;
}


//  Frees an EthBuf by putting it back on the freeList
//  Do not call from ISR routine.
//  Input:  Ptr to EthBuf to be freed
static void freeBuffer(tEthBuffer * b)
{
  //db_printf("freeBuffer %p\n",b);
  cyg_drv_isr_lock();
  ++freeCount;
#if EnableAllocCounting      
  MY_ASSERT(b >= &(bufferPool[0]), "Freeing invalid buffer address");
  MY_ASSERT(b < &(bufferPool[NUM_ETH_BUFFERS]), "Freeing invalid buffer address");
  if (b->acount != 1 || b->guard != 0x12345678)
    {
      db_printf("freeBuffer %p\n",b);
      db_printf("  guard     %08x\n",b->guard);
      db_printf("  length    %u\n",b->length);
      db_printf("  userDAta  %u\n",b->userData);
      db_printf("  next      %p\n",b->next);
      db_printf("  prev      %p\n",b->prev);
      db_printf("  acount    %d\n",b->acount);
    }
  MY_ASSERT(b->guard == 0x12345678, "Freeing buffer with bad guard value");
  MY_ASSERT(b->acount == 1, "Freeing buffer with bad acount");
  --b->acount;
#endif  
  ethBufQueuePut(&freeList, b);
  cyg_drv_isr_unlock();
}

static int allocFail;

void bufferListError(void)
{
  diag_printf("%s:%d Buffer List Error!\n", __FILE__, __LINE__);
  while (1)
    ;
}

static int bufs_free_now(void)
{
  return (freeCount);
}


//  Get an EthBuf, removing it from the freeList
//  Do not call from ISR routine
//  Returns:  Ptr to the EthBuf allocated, NULL if none available
static tEthBuffer *allocBuffer(void)
{
  tEthBuffer *r;

  cyg_drv_isr_lock();
  r = ethBufQueueGet(&freeList);
  if (r)
    {
      //db_printf("allocBuffer %p\n",r);
      --freeCount;
#if EnableAllocCounting      
      MY_ASSERT(r->guard == 0x12345678, "Allocating buffer with bad guard value");
      MY_ASSERT(r->acount == 0, "Allocting buffer with bad acount");
      ++r->acount;
#endif      
    }
  else
    {
      ++allocFail;
      if (freeCount)
	bufferListError();
    }
  cyg_drv_isr_unlock();
  return r;
}

//  Frees an EthBuf by putting it back on the freeList
//  Call only from ISR routine or init.
//  Input:  Ptr to EthBuf to be freed
static void isrFreeBuffer(tEthBuffer * b)
{
  //db_printf("isrFreeBuffer %p\n",b);
  ++freeCount;
#if EnableAllocCounting
  MY_ASSERT(b->guard == 0x12345678, "Freeing buffer with bad guard value");
  MY_ASSERT(b->acount == 1, "Freeing buffer with bad acount");
  --b->acount;
#endif      
  ethBufQueuePut(&freeList,b);
}


// call only from ISR routine or init
static tEthBuffer *isrAllocBuffer(void)
{
  tEthBuffer *r;
  r = ethBufQueueGet(&freeList);
  if (r)
    {
      //db_printf("isrAllocBuffer %p\n",r);
      --freeCount;
#if EnableAllocCounting
      MY_ASSERT(r->guard == 0x12345678, "Alloc buffer with bad guard value");
      MY_ASSERT(r->acount == 0, "Alloc buffer with bad acount");
      ++r->acount;
#endif      
    }
  else
    {
      ++allocFail;
      if (freeCount)
	bufferListError();
    }
  return r;
}


//  Initialize the freeList, add all the EthBufs from the bufferPool to it.
static void initFreeList(void)
{
  int i;
  db_printf("initFreeList\n");
  ethBufQueueClear(&freeList);
  freeCount = 0;
  for (i = 0; i < NUM_ETH_BUFFERS; ++i)
    {
#if EnableAllocCounting      
      bufferPool[i].acount = 1;
      bufferPool[i].guard = 0x12345678;
#endif      
      isrFreeBuffer(bufferPool+i);
    }
}

#if  !defined(CYGPKG_REDBOOT) && defined(CYGSEM_DEVS_ETH_OC_RAWAPI) && defined(CYGPKG_IO)

//=================================  COMTROL RAW Ethernet "cyg_io" APIr =============================
//
#include <pkgconf/io.h>
#include <cyg/io/io.h>
#include <cyg/io/devtab.h>

// raw_send()
// returns 1 if it sent out
// returns 0 if could not send (tx slots full)
// buf is already an allocated packet by the time it gets here
static int raw_send(tEthBuffer * buf)
{
  int i, stat, num_slots_free;
  unsigned int nb = buf->length;

  //db_printf( "raw_send() begin (buf=0x%x)\n", (int)buf  );

  if (nb > MAX_PKT_SIZE)
    {
      db_printf("raw_send(): pkt too large\n");
      return 0;
    }
  
  if (num_tx_slots_free() <= 0)
    return 0;

  
#if defined(CYGSEM_DEVS_ETH_OC_SWTXCRC)  
  // add Ethernet CRC
  {
    cyg_uint32 crc;
    if (nb < 64)
      {
        memset(buf->data+nb,0,64-nb);
        nb = 64;
      }
    crc = crc32(buf->data,nb);
    buf->data[nb++] = crc;
    buf->data[nb++] = crc>>8;
    buf->data[nb++] = crc>>16;
    buf->data[nb++] = crc>>24;
    buf->length = nb;
  }

  stat = (nb << 16) | BD_TX_PKT_BEING_USED | BD_TX_IRQ;
#else
  stat = (nb << 16) | BD_TX_PKT_BEING_USED | BD_TX_IRQ | BD_TX_CRC_EN | BD_TX_PAD_EN;
#endif  


#if !UseCacheBypass
  // flush data cache for outbound data buffer
  HAL_DCACHE_FLUSH(buf->data,nb);
#endif
  
  cyg_drv_mutex_lock(&txMutex);
  i = tx_index;
      
<  buf->userData = 0;  // no tx packet key value
  
  if (i == (TX_BUFS - 1))
    stat |= BD_TX_WRAP;
      
  write_oc_reg(TXBD_BUFP(i), (unsigned)buf);
  write_oc_reg(TXBD_STAT(i), stat);
  tx_pkts_queued++;
  tx_index = (i+1) & (TX_BUFS-1);
  cyg_drv_mutex_unlock(&txMutex);
  
  //db_printf( "raw_send() returning 1\n" );
  return (1);
}

static bool raw_init_handler(struct cyg_devtab_entry *tab)
{
  return true;
}

//  Writes data to ethernet.  Takes as input char data, memcpy's it to a
//  tEthBuffer struct, then calls the "new" write handler to push it to 
//  the ethernet DMA.
static Cyg_ErrNo raw_write_handler(cyg_io_handle_t handle, const void *_buf, cyg_uint32 * len)
{
  tEthBuffer *txBuffer;

  while (!(configDone && oc_interface_up))
    cyg_thread_delay(10);

  if (*len < 14 || *len > MAX_ETH_FRAME_SIZE)
    {
      *len = 0;
      return -EINVAL;
    }

  txBuffer = allocBuffer();

  if (!txBuffer)
    {
      *len = 0;
      return -ENOMEM;
    }
  
  memcpy(txBuffer->data, _buf, *len);
  txBuffer->length = *len;
  txBuffer->userData = 0;
  
  while (!raw_send(txBuffer))
    {
      if (config.flags & CYG_ETHER_FLAGS_WR_NONBLOCK)
	{
	  freeBuffer(txBuffer);
	  *len = 0;
	  return -EAGAIN;
	}
      cyg_thread_delay(10);
    }

  return ENOERR;
}


//---------------------------------------------------
// raw_get_receive_buffer()
//   get a receive packet from the raw receive queue
//   returns 0 if no packet was received
static tEthBuffer *raw_get_receive_buffer(void)
{
  while (!(configDone && oc_interface_up))
    cyg_thread_delay(10);

  if ((int)(ix_last_pkt_into_raw_queue-ix_last_pkt_from_raw_queue) > 0)
    return raw_queue_bufs[ix_last_pkt_from_raw_queue++ & (RAW_QUEUE_DEPTH - 1)];
  else
    return NULL;
}

static Cyg_ErrNo raw_lookup_handler(struct cyg_devtab_entry **tab, struct cyg_devtab_entry *sub_tab, const char *name)
{
  db_printf("raw_lookup_handler() doing nothing!\n");
  return ENOERR;
}

static Cyg_ErrNo raw_read_handler(cyg_io_handle_t handle, void *_buf, cyg_uint32 * len)
{
  int size;
  tEthBuffer *rxBuffer;

  while (1)
    {
      rxBuffer = raw_get_receive_buffer();
      if (rxBuffer)
	{
	  size = rxBuffer->length;
	  if (size > *len)
            size = *len;
	  memcpy(_buf, rxBuffer->data, size);
	  freeBuffer(rxBuffer);
	  *len = size;
	  return ENOERR;
	}
      if (config.flags & CYG_ETHER_FLAGS_RD_NONBLOCK)
        {
          *len = 0;
          return -EAGAIN;
        }
      // should probably block on a event flag so that the RX DSR could wake us up
      // when a packet becomes available.
      cyg_thread_delay(10);
    }
}


static cyg_bool raw_select_handler(cyg_io_handle_t handle, cyg_uint32 which, CYG_ADDRWORD info)
{
  // not supported, so just return true
  return true;
}

#define IS_VALID_BUFFER( xbuf )  \
	((tEthBuffer*)(VALID_ADDR_COMPONENT(xbuf)) >= (tEthBuffer*)(VALID_ADDR_COMPONENT(bufferPool)) &&  \
	(tEthBuffer*)(VALID_ADDR_COMPONENT(xbuf)) < ((tEthBuffer*)(VALID_ADDR_COMPONENT(bufferPool)) + NUM_ETH_BUFFERS))

static Cyg_ErrNo raw_get_config_handler(cyg_io_handle_t handle, cyg_uint32 key, void *xbuf, cyg_uint32 * len)
{
  switch (key)
    {
    case CYG_IO_GET_CONFIG_ETHER_LINK_STATUS:       // get the link status
      {
	int reg, flags;
	if (*len < sizeof(unsigned))
          return -EINVAL;

	flags = 0;
	reg = ReadPhyQuickpoll();
	db_printf("mii diag register: 0x%04x\n", reg);

	if (reg & MII_DIAGNOSTICS_FULL_DPLX)
          flags |= CYG_ETHER_LINK_STATUS_FULL_DUPLEX;

	if (reg & MII_DIAGNOSTICS_RATE_100)
          flags |= CYG_ETHER_LINK_STATUS_100M;

	if (reg & MII_DIAGNOSTICS_LINK)
          flags |= CYG_ETHER_LINK_STATUS_UP;

	*((unsigned int *) xbuf) = flags;
	*len = sizeof(unsigned int);

	return (ENOERR);
	break;
      }

    case CYG_IO_GET_CONFIG_ETHER_INFO:                // get a copy of the ethernet info
      if (*len < sizeof(cyg_ether_info_t))
        return -EINVAL;
      *((cyg_ether_info_t *) xbuf) = config;
      *len = sizeof config;
      return ENOERR;
      break;

    case CYG_IO_GET_CONFIG_ETHER_ALLOC_BUFFER:        // perform alloc of raw buffer
      {
	tEthBuffer *b = allocBuffer();
        *(tEthBuffer **) xbuf = b;
	if (b)
          return ENOERR;
	else
          return -ENOMEM;
	break;
      }

    case CYG_IO_GET_CONFIG_ETHER_FREE_BUFFER:         // perform free of raw buffer
      if (!IS_VALID_BUFFER(xbuf))
        return -EINVAL;
      freeBuffer((tEthBuffer *) xbuf);
      return ENOERR;
      break;


    case CYG_IO_GET_CONFIG_ETHER_SEND_BUFFER:        // perform raw send    
      if (!IS_VALID_BUFFER(xbuf))
        return -EINVAL;
      while (!raw_send((tEthBuffer *) xbuf))
        {
          if (config.flags & CYG_ETHER_FLAGS_WR_NONBLOCK)
            return -EAGAIN;
          else
            cyg_thread_delay(10);
          return ENOERR;
        }
      break;

    case CYG_IO_GET_CONFIG_ETHER_RECV_BUFFER:        //  Retrieve received Ethernet data in form of a tEthBuffer struct.
      //dump_rxbufs();
      while (1)
        {
          tEthBuffer *rxBuffer = raw_get_receive_buffer();
          if (rxBuffer)
            {
              if (len)
                *len = rxBuffer->length;
              *(tEthBuffer **) xbuf = rxBuffer;
              return ENOERR;
            }
          if (config.flags & CYG_ETHER_FLAGS_RD_NONBLOCK)
            return -EAGAIN;
          // should probably block on a event flag so that the RX DSR could wake us up
          // when a packet becomes available.
          cyg_thread_delay(10);
        }
      break;

    default:
      return (-ENOSUPP);
    }
  return ENOERR;
}



static Cyg_ErrNo raw_set_config_handler(cyg_io_handle_t handle, cyg_uint32 key, const void *xbuf, cyg_uint32 * len)
{
  db_printf("begin of raw_set_config_handler()\n");

  switch (key)
    {
    case CYG_IO_SET_CONFIG_ETHER_INFO:
      if (*len < sizeof(cyg_ether_info_t))
        return -EINVAL;
      config = *((cyg_ether_info_t *) xbuf);
      memcpy(&config, xbuf, sizeof(cyg_ether_info_t));
      bindump( "new config", &config, sizeof( cyg_ether_info_t ) );
      db_printf( "protocol: 0x%x\n", ntohs(config.protocol));
      set_oc_mac(&config.mac_address[0]);
      if (!oc_interface_up)
        oc_start(NULL, NULL, 0);
      return ENOERR;
      break;

    default:
      return -ENOSUPP;
    }
}

/*-------------------------------------------------------------------------
|   structures for the raw device ("/dev/reth0") and the normal net device
-------------------------------------------------------------------------*/
DEVIO_TABLE(OC_EtherDevio, raw_write_handler, raw_read_handler, raw_select_handler, raw_get_config_handler, raw_set_config_handler)
DEVTAB_ENTRY(OC_EtherDevtab, "/dev/reth0", NULL, &OC_EtherDevio, raw_init_handler, raw_lookup_handler, NULL)
#endif  // defined(CYGSEM_DEVS_ETH_OC_RAWAPI)

void if_oc_kick(void)
{
  process_a_transmitted_packet(&ocdev_sc);
}
    
void if_oc_dump_status(void)
{
  if_oc_dumpFreeList();
  dump_rxbufs();
  diag_printf("rx_packet_count = %u\n",rx_packet_count);
  dump_txbufs();
  diag_printf("tx_pkts_queued    %u\n",tx_pkts_queued);
  diag_printf("tx_pkts_complete  %u\n",tx_pkts_complete);
  diag_printf("\n");
  diag_printf("deliver_count    %u\n",deliver_count);
  diag_printf("oc_dsr_tx_count  %u\n",oc_dsr_tx_count);
  diag_printf("oc_dsr_rx_count  %u\n",oc_dsr_rx_count);
  diag_printf("isrcnt           %u\n",isrcnt);
  diag_printf("dsrcnt           %u\n",dsrcnt);
}
    
static void if_oc_assert_failed(void)
{
  CYG_INTERRUPT_STATE old;
  HAL_DISABLE_INTERRUPTS(old);
  __builtin_stwio((cyg_uint32*)0x2001850,0); // disable watchdog
  if_oc_dump_status();
  while (1)
    ;
}

#if defined(CYGSEM_DEVS_ETH_OC_SWRXCRC) || defined(CYGSEM_DEVS_ETH_OC_SWTXCRC)
static cyg_uint32 crctable[256] =
{
0x00000000L, 0x77073096L, 0xEE0E612CL, 0x990951BAL,
0x076DC419L, 0x706AF48FL, 0xE963A535L, 0x9E6495A3L,
0x0EDB8832L, 0x79DCB8A4L, 0xE0D5E91EL, 0x97D2D988L,
0x09B64C2BL, 0x7EB17CBDL, 0xE7B82D07L, 0x90BF1D91L,
0x1DB71064L, 0x6AB020F2L, 0xF3B97148L, 0x84BE41DEL,
0x1ADAD47DL, 0x6DDDE4EBL, 0xF4D4B551L, 0x83D385C7L,
0x136C9856L, 0x646BA8C0L, 0xFD62F97AL, 0x8A65C9ECL,
0x14015C4FL, 0x63066CD9L, 0xFA0F3D63L, 0x8D080DF5L,
0x3B6E20C8L, 0x4C69105EL, 0xD56041E4L, 0xA2677172L,
0x3C03E4D1L, 0x4B04D447L, 0xD20D85FDL, 0xA50AB56BL,
0x35B5A8FAL, 0x42B2986CL, 0xDBBBC9D6L, 0xACBCF940L,
0x32D86CE3L, 0x45DF5C75L, 0xDCD60DCFL, 0xABD13D59L,
0x26D930ACL, 0x51DE003AL, 0xC8D75180L, 0xBFD06116L,
0x21B4F4B5L, 0x56B3C423L, 0xCFBA9599L, 0xB8BDA50FL,
0x2802B89EL, 0x5F058808L, 0xC60CD9B2L, 0xB10BE924L,
0x2F6F7C87L, 0x58684C11L, 0xC1611DABL, 0xB6662D3DL,
0x76DC4190L, 0x01DB7106L, 0x98D220BCL, 0xEFD5102AL,
0x71B18589L, 0x06B6B51FL, 0x9FBFE4A5L, 0xE8B8D433L,
0x7807C9A2L, 0x0F00F934L, 0x9609A88EL, 0xE10E9818L,
0x7F6A0DBBL, 0x086D3D2DL, 0x91646C97L, 0xE6635C01L,
0x6B6B51F4L, 0x1C6C6162L, 0x856530D8L, 0xF262004EL,
0x6C0695EDL, 0x1B01A57BL, 0x8208F4C1L, 0xF50FC457L,
0x65B0D9C6L, 0x12B7E950L, 0x8BBEB8EAL, 0xFCB9887CL,
0x62DD1DDFL, 0x15DA2D49L, 0x8CD37CF3L, 0xFBD44C65L,
0x4DB26158L, 0x3AB551CEL, 0xA3BC0074L, 0xD4BB30E2L,
0x4ADFA541L, 0x3DD895D7L, 0xA4D1C46DL, 0xD3D6F4FBL,
0x4369E96AL, 0x346ED9FCL, 0xAD678846L, 0xDA60B8D0L,
0x44042D73L, 0x33031DE5L, 0xAA0A4C5FL, 0xDD0D7CC9L,
0x5005713CL, 0x270241AAL, 0xBE0B1010L, 0xC90C2086L,
0x5768B525L, 0x206F85B3L, 0xB966D409L, 0xCE61E49FL,
0x5EDEF90EL, 0x29D9C998L, 0xB0D09822L, 0xC7D7A8B4L,
0x59B33D17L, 0x2EB40D81L, 0xB7BD5C3BL, 0xC0BA6CADL,
0xEDB88320L, 0x9ABFB3B6L, 0x03B6E20CL, 0x74B1D29AL,
0xEAD54739L, 0x9DD277AFL, 0x04DB2615L, 0x73DC1683L,
0xE3630B12L, 0x94643B84L, 0x0D6D6A3EL, 0x7A6A5AA8L,
0xE40ECF0BL, 0x9309FF9DL, 0x0A00AE27L, 0x7D079EB1L,
0xF00F9344L, 0x8708A3D2L, 0x1E01F268L, 0x6906C2FEL,
0xF762575DL, 0x806567CBL, 0x196C3671L, 0x6E6B06E7L,
0xFED41B76L, 0x89D32BE0L, 0x10DA7A5AL, 0x67DD4ACCL,
0xF9B9DF6FL, 0x8EBEEFF9L, 0x17B7BE43L, 0x60B08ED5L,
0xD6D6A3E8L, 0xA1D1937EL, 0x38D8C2C4L, 0x4FDFF252L,
0xD1BB67F1L, 0xA6BC5767L, 0x3FB506DDL, 0x48B2364BL,
0xD80D2BDAL, 0xAF0A1B4CL, 0x36034AF6L, 0x41047A60L,
0xDF60EFC3L, 0xA867DF55L, 0x316E8EEFL, 0x4669BE79L,
0xCB61B38CL, 0xBC66831AL, 0x256FD2A0L, 0x5268E236L,
0xCC0C7795L, 0xBB0B4703L, 0x220216B9L, 0x5505262FL,
0xC5BA3BBEL, 0xB2BD0B28L, 0x2BB45A92L, 0x5CB36A04L,
0xC2D7FFA7L, 0xB5D0CF31L, 0x2CD99E8BL, 0x5BDEAE1DL,
0x9B64C2B0L, 0xEC63F226L, 0x756AA39CL, 0x026D930AL,
0x9C0906A9L, 0xEB0E363FL, 0x72076785L, 0x05005713L,
0x95BF4A82L, 0xE2B87A14L, 0x7BB12BAEL, 0x0CB61B38L,
0x92D28E9BL, 0xE5D5BE0DL, 0x7CDCEFB7L, 0x0BDBDF21L,
0x86D3D2D4L, 0xF1D4E242L, 0x68DDB3F8L, 0x1FDA836EL,
0x81BE16CDL, 0xF6B9265BL, 0x6FB077E1L, 0x18B74777L,
0x88085AE6L, 0xFF0F6A70L, 0x66063BCAL, 0x11010B5CL,
0x8F659EFFL, 0xF862AE69L, 0x616BFFD3L, 0x166CCF45L,
0xA00AE278L, 0xD70DD2EEL, 0x4E048354L, 0x3903B3C2L,
0xA7672661L, 0xD06016F7L, 0x4969474DL, 0x3E6E77DBL,
0xAED16A4AL, 0xD9D65ADCL, 0x40DF0B66L, 0x37D83BF0L,
0xA9BCAE53L, 0xDEBB9EC5L, 0x47B2CF7FL, 0x30B5FFE9L,
0xBDBDF21CL, 0xCABAC28AL, 0x53B39330L, 0x24B4A3A6L,
0xBAD03605L, 0xCDD70693L, 0x54DE5729L, 0x23D967BFL,
0xB3667A2EL, 0xC4614AB8L, 0x5D681B02L, 0x2A6F2B94L,
0xB40BBE37L, 0xC30C8EA1L, 0x5A05DF1BL, 0x2D02EF8DL
};


static cyg_uint32 crc32( unsigned char *blk_adr, unsigned blk_len)
{
  cyg_uint32 crc = 0xffffffff;
  
  while (blk_len--)
    crc = crctable[(crc ^ *blk_adr++) & 0xffL] ^ (crc >> 8);
  
  return crc ^ 0xffffffff;
}
#endif
