//------------------------------------------------
//  oc.h  - header for the opencores MAC
//
//  written by Digital Design Corp.    
//  for Comtrol / Altera
//
//------------------------------------------------
//  registers are defined in chapter 3 of the 
//  "Ethernet IP Core Specification"
//------------------------------------------------

#define Bit(n) (1<<(n))

//---------------------------------------------------
// the indexes of the individual registers (all 32 bit, RW)
#define OC_REG_MODER         0      //@0x00
#define OC_REG_INT_SOURCE    1      //@0x04
#define OC_REG_INT_MASK      2      //@0x08
#define OC_REG_IPGT          3      //@0x0c
#define OC_REG_IPGR1         4      //@0x10
#define OC_REG_IPGR2         5      //@0x14
#define OC_REG_PACKETLEN     6      //@0x18
#define OC_REG_COLLCONF      7      //@0x1c
#define OC_REG_TX_BD_NUM     8      //@0x20
#define OC_REG_CTRLMODER     9      //@0x24
#define OC_REG_MIIMODER      0x0a   //@0x28
#define OC_REG_MIICOMMAND    0x0b   //@0x2c
#define OC_REG_MIIADDRESS    0x0c   //@0x30
#define OC_REG_MIITX_DATA    0x0d   //@0x34
#define OC_REG_MIIRX_DATA    0x0e   //@0x38
#define OC_REG_MIISTATUS     0x0f   //@0x3c
#define OC_REG_MAC_ADDR0     0x10   //@0x40
#define OC_REG_MAC_ADDR1     0x11   //@0x44
#define OC_REG_ETH_HASH0_ADR 0x12   //@0x48
#define OC_REG_ETH_HASH1_ADR 0x13   //@0x4c
#define OC_REG_ETH_TXCTRL    0x14   //@0x50
#define OC_REG_unused        0x15   //@0x54


//---------------------------------------------------
//  register OC_REG_MODER (mode register)
//  note from data sheet: "Registers should not be changed
//    after TXEN or RXEN bit is set"
// bit values:
// bits 31..16 are reserved
#define OC_REG_MODER_REG_RECSMALL      0x00010000
#define OC_REG_MODER_REG_PAD           0x00008000
#define OC_REG_MODER_REG_HUGEN         0x00004000
#define OC_REG_MODER_REG_CRCEN         0x00002000
#define OC_REG_MODER_REG_DLYCRCEN      0x00001000
// bit 11 reserved
#define OC_REG_MODER_REG_FULLD         0x00000400
#define OC_REG_MODER_REG_EXDFREN       0x00000200
#define OC_REG_MODER_REG_NOBCKOF       0x00000100
#define OC_REG_MODER_REG_LOOPBACK      0x00000080
#define OC_REG_MODER_REG_IFG           0x00000040
#define OC_REG_MODER_REG_PRO           0x00000020
#define OC_REG_MODER_REG_IAM           0x00000010
#define OC_REG_MODER_REG_BRO           0x00000008
#define OC_REG_MODER_REG_NOPRE         0x00000004
#define OC_REG_MODER_REG_TXEN          0x00000002
#define OC_REG_MODER_REG_RXEN          0x00000001
// default value: 0xa000 (add pads to short frames, enable CRC)

//---------------------------------------------------
// register OC_REG_INT_SOURCE 
//  reading 1 means the that interrupt cause happened
//  writing 1 clears the interrupt
// bit values:
// bits 31..7 are reserved
#define OC_REG_INT_SOURCE_RXC          0x00000040
#define OC_REG_INT_SOURCE_TXC          0x00000020
#define OC_REG_INT_SOURCE_BUSY         0x00000010
#define OC_REG_INT_SOURCE_RXE          0x00000008
#define OC_REG_INT_SOURCE_RXB          0x00000004      // receive buffer complete
#define OC_REG_INT_SOURCE_TXE          0x00000002
#define OC_REG_INT_SOURCE_TXB          0x00000001      // transmit buffer complete
#define OC_REG_INT_SOURCE_CLEARALL     0x0000007f
#define OC_REG_INT_SOURCE_DEFAULT      0x0000007f
// reset value: 0x0


//---------------------------------------------------
// register OC_REG_INT_MASK
//  1 means allow the interrupt
//  0 means mask the interrupt 
// bit values:
// bits 31..7 are reserved
#define OC_REG_INT_MASK_RXC_M          0x00000040
#define OC_REG_INT_MASK_TXC_M          0x00000020
#define OC_REG_INT_MASK_BUSY_M         0x00000010
#define OC_REG_INT_MASK_RXE_M          0x00000008
#define OC_REG_INT_MASK_RXF_M          0x00000004
#define OC_REG_INT_MASK_TXE_M          0x00000002
#define OC_REG_INT_MASK_TXB_M          0x00000001
// reset value: 0x0


//---------------------------------------------------
// register OC_REG_IPGT (Back to Back Inter Packet Gap Register)
//   using the recommended values cause packet gaps of
//       0.96us @ 100Mbps   or 9.6us @ 10Mbps
//
// bits 31..7 are reserved
//   reset value:  0x12
#define OC_REG_IPGT_FULL_DUPLEX_RECOMMENDED  0x15
#define OC_REG_IPGT_HALF_DUPLEX_RECOMMENDED  0x12


//---------------------------------------------------
// register OC_REG_IPGR1 (Non Back to Back Inter Packet Gap Register 1)    
//   reset value is the 0x0c
//  b31..7 are reserved;  b6-0 are used
#define OC_REG_MAC_IPGR1_RECOMMENDED       0x0c



//---------------------------------------------------
// register OC_REG_IPGR2 (Non Back to Back Inter Packet Gap Register 2)    
//   reset value is the 0x12
//  b31..7 are reserved;  b6-0 are used
#define OC_REG_MAC_IPGR2_RECOMMENDED       0x12



//---------------------------------------------------
// register OC_REG_PACKETLEN (Packet Length)
// b31-16:  Minimum frame length  (default=64)
// b15-0:   Maximum frame length  (default=1536)
#define OC_REG_PACKETLEN_DEFAULT 0x00400600




//---------------------------------------------------
// register OC_REG_COLLCONF    (Collision and Retry Configuration)
// b31-20: reserved
// b19-15: Maximum retry (default is 0xf aka 15)
// b15-6:  reserved
// b5-0:   collision valid time window.  default ix 0x3f
#define OC_REG_COLLCONF_DEFAULT 0x000f003f

//---------------------------------------------------
// register OC_REG_TX_BD_NUM    (Transmit BD Number)
// number of transmit BD's used.
// number of receive BDS = 128 - Num TX bds
// b31-8 are reserved
// b7-0: Number of tx bd regs.  max ix 0x80
#define OC_REG_TX_BD_NUM_DEFAULT 0x40

//---------------------------------------------------
// register OC_REG_CTRLMODER  (Control Module Mode Register)
// b31-3 reserved
// default value: 0
// see description in p. 15 of OC spec for description
#define OC_REG_CTRLMODER_TXFLOW  0x4		// 1 means "send pause frames"
#define OC_REG_CTRLMODER_RXFLOW  0x2        
#define OC_REG_CTRLMODER_PASSALL 0x1		// "all received frames are passed to the host"

//---------------------------------------------------
// register OC_REG_MIIMODER   (MII Mode register)
// b31-9 reserved

#define OC_REG_MIIMODER_RESET               0x200   // b9		// not clear from the docs if this is still used
#define OC_REG_MIIMODER_DONT_SEND_PREAMBLE  0x100	// b8
#define OC_REG_MIIMODER_CLKDIV_MASK                  0x0ff	// b7-0 are Mclock div (must be even)
// default value: 0x64


//---------------------------------------------------
// register OC_REG_MIICOMMAND   (MII Command)
// b31-3:  reserved
#define OC_REG_MIICOMMAND_WRITE  0x4
#define OC_REG_MIICOMMAND_READ   0x2
#define OC_REG_MIICOMMAND_SCAN   0x1
// default value: 0

//---------------------------------------------------
// register OC_REG_MIIADDRESS    (MII Address)
// b31-13: reserved
// b12-8:  register adddress within the PHY selected by FIAD[4:0]
// b7-5:   reserved
// b4-0:   FIAD-PHY address
// default value: 0


//---------------------------------------------------
// register OC_REG_MIITX_DATA    (MII Transmit Data)
// b31-16: reserved
// b15-0:  Control data to be written to the PHY
// default value: 0


//---------------------------------------------------
// register OC_REG_MIIRX_DATA    (MII Receive Data) read-only
// b31-16: reserved
// b15-0:  Control data read from the PHY 
// default value: 0

//---------------------------------------------------
// register OC_REG_MIISTATUS     0x0f   (MII Status)
// b31-3: reserved
#define OC_REG_MIISTATUS_INVALID  0x4    // 1 means data in the MSTATUS register is invalid
#define OC_REG_MIISTATUS_BUSY     0x2    // 1 means the MII is busy (operation in progress)
#define OC_REG_MIISTATUS_LINKFAIL 0x1    // 1 means the link failed

//---------------------------------------------------
// register OC_REG_MAC_ADDR0     0x10   
// for mac: aa:bb:cc:dd:ee:ff, this register would be 0xccddeeff

//---------------------------------------------------
// register OC_REG_MAC_ADDR1     0x11  
// for mac: aa:bb:cc:dd:ee:ff, this register would be 0x0000aabb


//---------------------------------------------------
// register OC_REG_ETH_HASH0_ADR 0x12   (HASH 0 Value)
// default: 0

//---------------------------------------------------
// register OC_REG_ETH_HASH1_ADR 0x13   (HASH 1 Value)
// default: 0

//---------------------------------------------------
// register OC_REG_ETH_TXCTRL    0x14   (Tx Control Register)
// b31-16: reserved
#define OC_REG_ETH_TXCTL_PAUSE_REQUEST 0x10000  // "writing 1 to this bit starts sending control frame procedure.  Bit is automatically cleared to zero"
// b15-0 pause timer value 
// default value: 0



/*-----------------------------------------------------
|  buffer descriptor bits
-----------------------------------------------------*/

//transmitter packet
// b31-16 are the length
#define BD_TX_PKT_BEING_USED     0x00008000	   // 1 if is being transmitted or on deck to be transmitted (ready probably not best name, but that is used in the documentation)
#define BD_TX_IRQ         0x00004000      // 1 means assert an interrupt
#define BD_TX_WRAP        0x00002000      // 1 means this is the last bd in the chain (and it should wrap)
#define BD_TX_PAD_EN      0x00001000      // 1 means pad short packets
#define BD_TX_CRC_EN      0x00000800      // 1 means add CRC at end of packet
// b10-9 are reserved
#define BD_TX_UNDERRUN    0x00000100      // 1 means underrun occurred while transmitting packet 
#define BD_TX_RETRY_CNT   0x000000f0      // these bits specify number of retries necessary before successful send
#define BD_TX_RETRY_LIMIT 0x00000008      // 1 means transmission failed due to retransmission limit
#define BD_TX_LATE_COLL   0x00000004      // 1 means a late collision occurred
#define BD_TX_DEFER       0x00000002      // 1 if frame was deferred
#define BD_TX_CS_LOST     0x00000001      // 1 if carrier sense was lost during frame transmission

// these are the BD transmit bits that are considered an error
#define BD_TX_ANY_ERROR \
    (BD_TX_UNDERRUN | BD_TX_RETRY_LIMIT | BD_TX_LATE_COLL | BD_TX_CS_LOST)

// RECEIVER PACKET

#define BD_RX_EMPTY     0x00008000
#define BD_RX_IRQ       0x00004000
#define BD_RX_WRAP      0x00002000
// bits 12-9 are reserved
#define BD_RX_CF        0x00000100
#define BD_RX_MISS      0x00000080
#define BD_RX_OVERRUN   0x00000040
#define BD_RX_IS        0x00000020      // invalid symbol
#define BD_RX_DRIBBLE   0x00000010
#define BD_RX_TOOLONG   0x00000008
#define BD_RX_SHORT     0x00000004
#define BD_RX_CRC_ERR   0x00000002
#define BD_RX_LATE_COLL 0x00000001

#define BD_RX_ANY_ERROR \
  ( BD_RX_OVERRUN   | \
    BD_RX_IS        | \
    BD_RX_DRIBBLE   | \
    BD_RX_TOOLONG   | \
    BD_RX_SHORT     | \
    BD_RX_CRC_ERR   | \
    BD_RX_LATE_COLL      )




/*=========================================================
|  MII registers
|
|   Information from Amd79C874 PHY data sheet
|   These are not specific to the opencores mac, and probably 
|   logically belong in their own header.
|
|   This is the set of 32 16-bit registers that control the PHY
|   Actual reading and writing of these registers is done with a
|   two wire interface
|   
|   Most PHY parts should be compatible.   Reserved registers
|   are used to support gigabit ethernet, etc.
|   
=========================================================*/

#define MII_CONTROL                       0
#define MII_STATUS                        1
#define MII_PHYID_1                       2
#define MII_PHYID_2                       3
#define MII_AUTONEG_ADVERTISE             4
#define MII_AUTONEG_LINK_PARTNER_ABILITY  5
#define MII_AUTONEG_EXPANSION             6
#define MII_NEXTPAGE_ADVERTISEMENT        7
// regs 8-15 are reserved
#define MII_MISC_FEATURES                 16
//#define MII_INTERRUPT_CONTROL_STATUS      17
//#define MII_DIAGNOSTICS                   18
#define MII_DIAGNOSTICS                   17
#define MII_POWER_MANAGEMENT_LOOPBACK     19
// reg 20 reserved
#define MII_MODE_CONTROL                  21
// reg 22 reserved
#define MII_DISCONNECT                    23
#define MII_RECEIVE_ERR_COUNTER           24
// regs 25-31 are reserved


// reg 0, MII_CONTROL
#define MII_CONTROL_RESET        0x8000
#define MII_CONTROL_LOOPBACK     0x4000
#define MII_CONTROL_SPEED_100    0x2000
#define MII_CONTROL_ANEG_ENABLE  0x1000
#define MII_CONTROL_POWER_DOWN   0x0800
#define MII_CONTROL_ISOLATE      0x0400
#define MII_CONTROL_RESTART_ANEG 0x0200
#define MII_CONTROL_FULL_DUPLEX  0x0100
#define MII_CONTROL_COLL_TEST    0x0080
// bits 6-0 unused

// reg 1, MII_STATUS
#define MII_STATUS_100BASE_T4            0x8000
#define MII_STATUS_100BASE_TX_FULL       0x4000
#define MII_STATUS_100BASE_TX_HALF       0x2000
#define MII_STATUS_10BASE_T_FULL         0x1000
#define MII_STATUS_10BASE_T_HALF         0x0800
#define MII_STATUS_PREAMBLE_SUPPRESSION  0x0040
#define MII_STATUS_ANEG_COMPLETE         0x0020
#define MII_STATUS_REMOTE_FAULT          0x0010
#define MII_STATUS_ANEGA                 0x0008		// set by anega pin
#define MII_STATUS_LINK_STATUS           0x0004		// when link fails, is set until this status register is read
#define MII_STATUS_JABBER_DETECT         0x0002
#define MII_STATUS_EXTENDED_CAPABILITY   0x0001

// reg 17, MII_DIAGNOSTICS
#define MII_DIAGNOSTICS_FULL_DPLX  0x4000       // set if full duplex
#define MII_DIAGNOSTICS_RATE_100   0x8000       // set if 100Mbps
#define MII_DIAGNOSTICS_LINK       0x0001       // set if link valid
