//      Wallclock implementation for Dallas 1501
#include <time.h>
#include <cyg/kernel/kapi.h>
#include <pkgconf/wallclock.h>	// Wallclock device config
#include <pkgconf/kernel.h>

#include <cyg/hal/hal_io.h>	// IO macros
#include <cyg/hal/hal_intr.h>	// interrupt enable/disable
#include <cyg/infra/cyg_type.h>	// Common type definitions and support

#include <cyg/io/wallclock.hxx>	// The WallClock API
#include <cyg/kernel/clock.hxx> // Clock API
#include <cyg/kernel/clock.inl> // Clock inlines
#include <cyg/io/wallclock/wallclock.inl>	// Helpers
#include <pkgconf/devices_wallclock_dallas_ds1501.h>
#include <cyg/infra/diag.h>

#define TRACE() diag_printf("TRACE -- %s:%d\n",__FILE__,__LINE__)

#define DEBUG 0

#define Bit(n) (1<<(n))

extern volatile unsigned char *DallasDS1511W;

// Registers
#define DS_SECONDS         DallasDS1511W[0x00]
#define DS_MINUTES         DallasDS1511W[0x01]
#define DS_HOURS           DallasDS1511W[0x02]
#define DS_DOW             DallasDS1511W[0x03]
#define DS_DOM             DallasDS1511W[0x04]
#define DS_MONTH           DallasDS1511W[0x05]
#define DS_YEAR            DallasDS1511W[0x06]
#define DS_CENTURY         DallasDS1511W[0x07]
#define DS_SECONDS_ALARM   DallasDS1511W[0x08]
#define DS_MINUTES_ALARM   DallasDS1511W[0x09]
#define DS_HOURS_ALARM     DallasDS1511W[0x0a]
#define DS_DATE_ALARM      DallasDS1511W[0x0b]
#define DS_WD_10MS         DallasDS1511W[0x0c]
#define DS_WD_SEC          DallasDS1511W[0x0d]
#define DS_REG_A           DallasDS1511W[0x0e]
#define DS_REG_B           DallasDS1511W[0x0f]

#define DS_RAM_ADDR        DallasDS1511W[0x10]
#define DS_RAM_DATA        DallasDS1511W[0x13]

// Control bits
#define DS_REG_A_IRQF    Bit(0)
#define DS_REG_A_WDF     Bit(1)
#define DS_REG_A_KSF     Bit(2)
#define DS_REG_A_TDF     Bit(3)
#define DS_REG_A_PAB     Bit(4)
#define DS_REG_A_PRS     Bit(5)
#define DS_REG_A_BLF2    Bit(6)
#define DS_REG_A_BLF1    Bit(7)

#define DS_REG_B_WDS     Bit(0)
#define DS_REG_B_WDE     Bit(1)
#define DS_REG_B_KIE     Bit(2)
#define DS_REG_B_TIE     Bit(3)
#define DS_REG_B_TPE     Bit(4)
#define DS_REG_B_BME     Bit(5)
#define DS_REG_B_CS      Bit(6)
#define DS_REG_B_TE      Bit(7)

#define DS_SUSPEND_XFER() DS_REG_B &= ~DS_REG_B_TE
#define DS_RESUME_XFER()  DS_REG_B |=  DS_REG_B_TE

#define BAD_BCD(b) ((((b)&0x0f)>0x09)||(((b)&0xf0)>0x90))

// don't change the name of this -- it's shared with the watchdog driver
// for the DS1501

cyg_mutex_t dallasDS1501Mutex;

//----------------------------------------------------------------------------
// Accessor functions
static void init_ds_hwclock(void)
{
#if DEBUG  
    diag_printf("init_ds_hwclock()\n");
#endif

  cyg_mutex_init(&dallasDS1501Mutex);
  
#if 0
  {
  cyg_uint8 tmp;
  int bogus = 0;

  // check for sane values

  DS_SUSPEND_XFER();
  tmp = DS_CENTURY;
  bogus |= BAD_BCD(tmp) || (tmp > 0x39);
  tmp = DS_YEAR;
  bogus |= BAD_BCD(tmp) || (tmp > 0x99);
  tmp = DS_MONTH;
  bogus |= BAD_BCD(tmp) || (tmp < 1) || (tmp > 0x12);
  tmp = DS_DOM;
  bogus |= BAD_BCD(tmp) || (tmp < 1) || (tmp > 0x31);
  tmp = DS_DOW;
  bogus |= BAD_BCD(tmp) || (tmp < 1) || (tmp > 0x07);
  tmp = DS_HOURS;
  bogus |= BAD_BCD(tmp) || (tmp > 0x23);
  tmp = DS_MINUTES;
  bogus |= BAD_BCD(tmp) || (tmp > 0x59);
  tmp = DS_SECONDS;
  bogus |= BAD_BCD(tmp) || (tmp > 0x59);
  DS_RESUME_XFER();

  if (bogus)
    {
      diag_printf("DS1501 wallclock: bogus RTC value -- resetting to epoch\n");
      DS_SUSPEND_XFER();
      DS_CENTURY = 0x19;
      DS_YEAR = 0x70;
      DS_MONTH = 1;
      DS_DOM = 1;
      DS_DOW = 5;
      DS_HOURS = 0;
      DS_MINUTES = 0;
      DS_SECONDS = 0;
      DS_RESUME_XFER();
    }
  }
#endif  
}


static inline void set_ds_hwclock(cyg_uint32 year, cyg_uint32 month, cyg_uint32 mday, cyg_uint32 hour, cyg_uint32 minute, cyg_uint32 second)
{
  cyg_mutex_lock(&dallasDS1501Mutex);
  DS_SUSPEND_XFER();
  DS_CENTURY = TO_BCD((cyg_uint8) (year / 100));
  DS_YEAR = TO_BCD((cyg_uint8) (year % 100));
  DS_MONTH = TO_BCD((cyg_uint8) month);
  DS_DOM = TO_BCD((cyg_uint8) mday);
  DS_HOURS = TO_BCD((cyg_uint8) hour);
  DS_MINUTES = TO_BCD((cyg_uint8) minute);
  DS_SECONDS = TO_BCD((cyg_uint8) second);
  DS_RESUME_XFER();
  cyg_mutex_unlock(&dallasDS1501Mutex);
#if DEBUG  
  diag_printf("DS1501 Set -------------\n");
  diag_printf("  year %04d\n", year);
  diag_printf("  month %02d\n", month);
  diag_printf("  mday %02d\n", mday);
  diag_printf("  hour %02d\n", hour);
  diag_printf("  minute %02d\n", minute);
  diag_printf("  second %02d\n", second);
#endif
}

static void get_ds_hwclock(cyg_uint32 *year, cyg_uint32 *month, cyg_uint32 *mday, cyg_uint32 *hour, cyg_uint32 *minute, cyg_uint32 *second)
{
  volatile cyg_uint8 s1, t1, t2;

  // Clearing the TE bit doesn't always stop the registers in
  // a consistent state, so it's useless when reading.  :(
  // Instead, we'll read the seconds register twice, once at
  // the start, and once at the end.  If they don't agree, we
  // try again.
  
  cyg_mutex_lock(&dallasDS1501Mutex);
  do 
    {
      s1 = DS_SECONDS;
      t1 = DS_CENTURY;
      t2 = DS_YEAR;
      *year = (cyg_uint32) TO_DEC(t1) * 100 + (cyg_uint32) TO_DEC(t2);
    
      t1 = DS_MONTH;
      *month = (cyg_uint32) TO_DEC(t1);
      
      t1 = DS_DOM;
      *mday = (cyg_uint32) TO_DEC(t1);
      
      t1 = DS_HOURS;
      *hour = (cyg_uint32) TO_DEC(t1);
      
      t1 = DS_MINUTES;
      *minute = (cyg_uint32) TO_DEC(t1);

      t1 = DS_SECONDS;
      *second = (cyg_uint32) TO_DEC(t1);

    } while (s1 != t1);
  cyg_mutex_unlock(&dallasDS1501Mutex);
  
#if DEBUG
  diag_printf("DS1501 Read: ");
  diag_printf("year %04d ", *year);
  diag_printf("month %02d ", *month);
  diag_printf("mday %02d ", *mday);
  diag_printf("hour %02d ", *hour);
  diag_printf("minute %02d ", *minute);
  diag_printf("second %02d\n", *second);
#endif  
}

//-----------------------------------------------------------------------------
// Functions required for the hardware-driver API.

// Returns the number of seconds elapsed since 1970-01-01 00:00:00.
// (Uses system time rather than Dallas RTC)

static cyg_uint64 ticksPerSecond;
static cyg_int64 timeDelta;
static int isDallas;

#if defined(CYGSEM_WALLCLOCK_DALLAS_SYNC)
static void createClockTask(void);
#define isDallasSync 1
#else
#define isDallasSync 0
#endif


cyg_uint32 Cyg_WallClock::get_hw_seconds(void)
{
  if (isDallasSync || !isDallas)
    return (Cyg_Clock::real_time_clock->current_value()+timeDelta)/ticksPerSecond;

  cyg_uint32 year, month, mday, hour, minute, second;
  get_ds_hwclock(&year, &month, &mday, &hour, &minute, &second);
  return _simple_mktime(year, month, mday, hour, minute, second);
}


// Sets the clock. Argument is seconds elapsed since 1970-01-01 00:00:00.
// Set Dallas RTC, then sync system time.

void Cyg_WallClock::set_hw_seconds(cyg_uint32 secs)
{
  cyg_uint32 year, month, mday, hour, minute, second;
  if (isDallas)
    {
      _simple_mkdate(secs, &year, &month, &mday, &hour, &minute, &second);
      set_ds_hwclock(year, month, mday, hour, minute, second);
    }
  if (isDallasSync || !isDallas)
    timeDelta = (secs * ticksPerSecond) -  Cyg_Clock::real_time_clock->current_value();
}

// Read the Dallas RTC and set the system time to match. Save
// ticks/second value for later use (assumes integer number of
// ticks per seconds -- if not, time will drift.


void Cyg_WallClock::init_hw_seconds(void)
{
#if defined(CYGSEM_WALLCLOCK_DALLAS_SYNC)  
  cyg_uint32 year, month, mday, hour, minute, second, s;
#endif
  Cyg_Clock::cyg_resolution res;
  
  res = Cyg_Clock::real_time_clock->get_resolution();
  ticksPerSecond = (res.divisor * 1000000000LL)/res.dividend;
  
  isDallas = 0;
  DS_RAM_ADDR = 0;
  DS_RAM_DATA = 0x52;
  DS_RAM_ADDR = 1;
  DS_RAM_DATA = 0x00;
  DS_RAM_ADDR = 0;
  if (DS_RAM_DATA == 0x52) 
      isDallas = 1;

  diag_printf("DS1501 Wallclock driver: ");
  
  if (!isDallas)
      {
        diag_printf("no DS1501 found\n");
        return;
      }
  
  diag_printf("DS1501 enabled\n");
  
  init_ds_hwclock();
  
#if defined(CYGSEM_WALLCLOCK_DALLAS_SYNC)  
# if defined(CYGSEM_WALLCLOCK_DALLAS_SYNC_START)
  get_ds_hwclock(&year, &month, &mday, &hour, &minute, &s);
  do 
    {
      get_ds_hwclock(&year, &month, &mday, &hour, &minute, &second);
    }while (s == second);
# else
  get_ds_hwclock(&year, &month, &mday, &hour, &minute, &second);
# endif  
  s = _simple_mktime(year, month, mday, hour, minute, second);
  timeDelta = (s * ticksPerSecond) -  Cyg_Clock::real_time_clock->current_value();
#if DEBUG  
  diag_printf("DS1501: s=%ld current=%lld timeDelta = %lld\n",s,Cyg_Clock::real_time_clock->current_value(),timeDelta);
#endif  
  createClockTask();
#endif  
}

//-----------------------------------------------------------------------------

#if defined(CYGSEM_WALLCLOCK_DALLAS_SYNC)

static cyg_uint32 getDallasSeconds(void)
{
  cyg_uint32 year, month, mday, hour, minute, second;
  get_ds_hwclock(&year, &month, &mday, &hour, &minute, &second);
  return _simple_mktime(year, month, mday, hour, minute, second);
}


// state for background clock task
static unsigned char ClockStack[2048];
static cyg_handle_t  ClockHandle;
static cyg_thread    ClockThread;
static cyg_thread_entry_t ClockTask;

static void createClockTask(void)
{
  // create clock-sync task
  cyg_thread_create(3, ClockTask, (cyg_addrword_t)0, "Clock",
                    (void *)ClockStack, sizeof ClockStack, 
                    &ClockHandle, &ClockThread);
  cyg_thread_resume(ClockHandle);
}
  
#define NumSamples 100

// sample RTC clock error and filter with FIR low-pass filter
externC cyg_uint32 cyg_hal_get_period(void);
externC void cyg_hal_set_period(cyg_uint32 period);

static void ClockTask(cyg_addrword_t data)
{
  int i;
  unsigned loops;
  unsigned delay = ticksPerSecond + ticksPerSecond/10;
  int errorTotal,error;
  static int errorSample[NumSamples];
  cyg_uint32 t1,t2;
  unsigned long period,periodMin,periodMax;
  
  diag_printf("DS1501: clock task started\n");

  period = cyg_hal_get_period();
  periodMin = period - period/1000;
  periodMax = period + period/1000;

  i = 0;
  errorTotal = 0;
  loops = 0;

  while (1)
    {
      cyg_thread_delay(delay);
      errorTotal -= errorSample[i];
      t1 = time(NULL);
      t2 = getDallasSeconds();
      errorTotal += errorSample[i] = t1-t2;
      error = errorTotal;
      // diag_printf("s %d %d %d %d %d\n",loops,t1,t2,t1-t2,error);
      ++loops;
      ++i;
      if (i >= NumSamples)
        i = 0;
      
      if ((loops % 100) == 0)
        {
          static int prevError;
          static int initDone;
          
          if (initDone)
            {
              int delta = error - prevError;
              period += 10*delta + error;
              // diag_printf("c %d %d %d\n",error,delta,period);
              if (period < periodMin)
                period = periodMin;
              if (period > periodMax)
                period = periodMax;
              cyg_hal_set_period(period);
            }
          
          initDone = 1;
          prevError = error;
        }
          
    }
}

#endif
