//==========================================================================
//
//      Watchdog implementation for Dallas DS1501
//
//==========================================================================
//####COPYRIGHTBEGIN####
//                                                                          
// Copyright 2002, Comtrol Corp.
// 
//####COPYRIGHTEND####
//==========================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):    Grant Edwards
// Date:         2002-02-07
// Purpose:      Watchdog class implementation
// Description:  Contains an implementation of the Watchdog class for use
//               with the Dallas Semi DS1501
//
//####DESCRIPTIONEND####
//
//==========================================================================

#include <pkgconf/system.h>             // system configuration file
#include <pkgconf/watchdog.h>           // configuration for this package
#include <pkgconf/kernel.h>             // kernel config

#include <cyg/infra/cyg_trac.h>         // tracing macros
#include <cyg/kernel/instrmnt.h>        // instrumentation

#include <cyg/io/watchdog.hxx>          // watchdog API
#include <cyg/kernel/kapi.h>            // needed for mutex stuff
#include <pkgconf/devices_watchdog_dallas_ds1501.h>  // our config

//  Flag to indicate if a Dallas1501 or Maxin 823 watchdog device is used.
//  The Dallas part is memory mapped, the Maxim uses a port pin.
static int isDallas;

#define Bit(n) (1<<(n))

extern volatile unsigned char *DallasDS1511W;
extern volatile unsigned PioMode;
extern volatile unsigned PioData;

// Registers
#define DS_SECONDS         DallasDS1511W[0x00]
#define DS_MINUTES         DallasDS1511W[0x01]
#define DS_HOURS           DallasDS1511W[0x02]
#define DS_DOW             DallasDS1511W[0x03]
#define DS_DOM             DallasDS1511W[0x04]
#define DS_MONTH           DallasDS1511W[0x05]
#define DS_YEAR            DallasDS1511W[0x06]
#define DS_CENTURY         DallasDS1511W[0x07]
#define DS_SECONDS_ALARM   DallasDS1511W[0x08]
#define DS_MINUTES_ALARM   DallasDS1511W[0x09]
#define DS_HOURS_ALARM     DallasDS1511W[0x0a]
#define DS_DATE_ALARM      DallasDS1511W[0x0b]
#define DS_WD_10MS         DallasDS1511W[0x0c]
#define DS_WD_SEC          DallasDS1511W[0x0d]
#define DS_REG_A           DallasDS1511W[0x0e]
#define DS_REG_B           DallasDS1511W[0x0f]
#define DS_RAM_ADDR        DallasDS1511W[0x10]
#define DS_RAM_DATA        DallasDS1511W[0x13]

// Control bits
#define DS_REG_A_IRQF    Bit(0)
#define DS_REG_A_WDF     Bit(1)
#define DS_REG_A_KSF     Bit(2)
#define DS_REG_A_TDF     Bit(3)
#define DS_REG_A_PAB     Bit(4)
#define DS_REG_A_PRS     Bit(5)
#define DS_REG_A_BLF2    Bit(6)
#define DS_REG_A_BLF1    Bit(7)

#define DS_REG_B_WDS     Bit(0)
#define DS_REG_B_WDE     Bit(1)
#define DS_REG_B_KIE     Bit(2)
#define DS_REG_B_TIE     Bit(3)
#define DS_REG_B_TPE     Bit(4)
#define DS_REG_B_BME     Bit(5)
#define DS_REG_B_CS      Bit(6)
#define DS_REG_B_TE      Bit(7)

#define DS_SUSPEND_XFER() DS_REG_B &= ~DS_REG_B_TE
#define DS_RESUME_XFER()  DS_REG_B |=  DS_REG_B_TE

#define BAD_BCD(b) ((((b)&0x0f)>0x09)||(((b)&0xf0)>0x90))
#define TO_BCD(x)  (((x/10)<<4) | (x%10))

// if wallclock driver is present, we only delcare, not define
// the mutex
#if defined(CYGPKG_DEVICES_WALLCLOCK_DALLAS_DS1501)
  extern cyg_mutex_t dallasDS1501Mutex;
#else
  cyg_mutex_t dallasDS1501Mutex;
#endif

// Constructor

void
Cyg_Watchdog::init_hw(void)
{
  CYG_REPORT_FUNCTION();
  cyg_mutex_init(&dallasDS1501Mutex);
  // configure and disable watchdog

  //  Determine which device we are by attempting to write/read Dallas SRAM
  isDallas = 0;
  DS_RAM_ADDR = 0;
  DS_RAM_DATA = 0x52;
  DS_RAM_ADDR = 1;
  DS_RAM_DATA = 0x00;
  DS_RAM_ADDR = 0;

  if (DS_RAM_DATA == 0x52) 
      isDallas = 1;

  if (isDallas) {
      // There's nothing in the docs that says the TE bit affects
      // the WD timer, but it needs to be set.
      
      DS_REG_B  |= DS_REG_B_TE;  
      DS_REG_B  &= ~(DS_REG_B_WDE+DS_REG_B_WDS);  // disable WD timer
      DS_WD_10MS = TO_BCD((CYGSEM_WATCHDOG_DALLAS_DS1501_PERIOD)%100);       
      DS_WD_SEC  = TO_BCD((CYGSEM_WATCHDOG_DALLAS_DS1501_PERIOD)/100);
      DS_REG_A   = 0;  // doesn't seem to be required, but let's force things to a known state anyway
      
      resolution = (CYGSEM_WATCHDOG_DALLAS_DS1501_PERIOD) * 10000000000LL; // timeout period in ns
  }

  CYG_REPORT_RETURN();
}

// Start watchdog timer

void
Cyg_Watchdog::start()
{    
  unsigned char c;
  CYG_REPORT_FUNCTION();
  cyg_mutex_lock(&dallasDS1501Mutex);

  if (isDallas)
      DS_REG_B |= (DS_REG_B_TE+DS_REG_B_WDS+DS_REG_B_WDE);
  else
      PioMode |= Bit(16);

  // Kick the watchdog timer
  if (isDallas) {
      c = DS_WD_10MS;
      DS_WD_10MS = c;
  }
  else {
      PioData ^= Bit(16);
  }

  cyg_mutex_unlock(&dallasDS1501Mutex);
  CYG_REPORT_RETURN();
}

// Kick the watchdog timer

void
Cyg_Watchdog::reset()
{    
  unsigned char c;
  CYG_REPORT_FUNCTION();
  cyg_mutex_lock(&dallasDS1501Mutex);

  if (isDallas) {
      c = DS_WD_10MS;
      DS_WD_10MS = c;
  }
  else {
      PioData ^= Bit(16);
  }

  cyg_mutex_unlock(&dallasDS1501Mutex);
  CYG_REPORT_RETURN();
}

