
//======================================================================
//
//   Functions to use AIOPIC channel 0 as diag channel
//   This allows #!DM on port 0 at powerup, plus loopback tests

#include "aiopic.h"

// global variable indicating whether a unit is a DM500 or not
// defined in hal/arm/dmrts/current/src/hal_diag.c
extern int IsDM500;

extern tAiopic _aiop0;
extern tAiopic _aiop1;
extern tAiopic _aiop2;
extern tAiopic _aiop3;

#define swab(x) ((((x)<<8)&0xff00)|(((x)>>8)&0x00ff))
#define NumElements(a) ((sizeof a)/(sizeof a[0]))
#define MAXPORTS 32

static tRegVal RData[] =
{
    {0x0009, 0xf682},
    {0x0209, 0x86fb},
    {0x0409, 0x000a},
    {0x0609, 0x010a},
    {0x0809, 0x8a13},
    {0x0a09, 0xc511},
    {0x0c09, 0x8685},
    {0x0e09, 0x200a},
    {0x1009, 0x210a},
    {0x1209, 0x41ff},
    {0x1409, 0x8200},
    {0x1609, 0x827b},
    {0x1809, 0x8a7d},
    {0x1a09, 0x8881},
    {0x1c09, 0x867a},
    {0x1e09, 0x8481},
    {0x2009, 0x827c},
    {0x2209, 0x0a0a},
};

static tRegVal R[] =
{
    {0x0009, 0xf682},             /* 00: Stop Rx processor */
    {0x0809, 0x8a13},             /* 04: Tx software flow control */
    {0x0a09, 0xc511},             /* 08: XON char */
    {0x0c09, 0x8685},             /* 0c: XANY */
    {0x1209, 0x41ff},             /* 10: Rx mask char */
    {0x1409, 0x8200},             /* 14: Compare/Ignore #0 */
    {0x1609, 0x827b},             /* 18: Compare #1 */
    {0x1809, 0x8a7d},             /* 1c: Compare #2 */
    {0x1a09, 0x8881},             /* 20: Interrupt #1 */
    {0x1c09, 0x867a},             /* 24: Ignore/Replace #1 */
    {0x1e09, 0x8481},             /* 28: Interrupt #2 */
    {0x2009, 0x827c},             /* 2c: Ignore/Replace #2 */
    {0x2209, 0x0a0a}              /* 30: Rx FIFO Enable */
};

static tAiopChan  chanInfo[MAXPORTS];

static void initAiopChan(tAiopic *a, int chanNum, tAiopChan *ch)
{
  int i;
  unsigned short base = chanNum*0x10;
  unsigned short baud_divisor;

  // reset UART
  a->command = AiopCommand_ResetUart | chanNum;
  a->command = chanNum;
  
  // initialize channel
  ch->base = chanNum * 0x10;

  // diag_printf("init aiop %08x chan %d\n",a,chanNum);
  
  for (i=0; i<NumElements(RData); ++i)
    AiopPutIndexedShort(a,RData[i].offset+base,RData[i].value);
  
  for (i=0; i<NumElements(ch->R); ++i)
    {
      ch->R[i].offset = R[i].offset+base;
      ch->R[i].value  = R[i].value;
    }
  
  ch->aiop = a;
  ch->chan = chanNum;
  
  ch->dataWord = &a->data[chanNum];
  ch->dataByte = (unsigned char*)&a->dataLow[chanNum];
  ch->rxFifoCnt = &a->rxFifoCnt[chanNum];
  ch->txFifoCnt = (unsigned char*)&a->txFifoCnt[chanNum];
  if (chanNum & 1)
    ch->intID = ((unsigned char*)&a->oddIntID)+1;
  else
    ch->intID = (unsigned char*)&a->evenIntID;
  ch->status = &a->chanStatus[chanNum];
  ch->statusLow = (unsigned char*)&a->chanStatusLow;
  ch->statusHigh = ((unsigned char*)&a->chanStatusHigh)+1;
  
  ch->baudDivisor.offset = base+_BAUD;
  ch->txControl.offset   = base+_TX_CTRL;
  ch->rxControl.offset   = base+_RX_CTRL;
  ch->txEnables.offset   = base+_TX_ENBLS;
  ch->txCompare.offset   = base+_TXCMP1;
  ch->txReplace1.offset  = base+_TXREP1B1;
  ch->txReplace2.offset  = base+_TXREP2;

  // Set baud rate of port 0 to 57600 to allows #!DM at powerup.
  // The formula defined in AiopBaudDivisor is only good for
  // calculating baud divisor for baud rates slower than 57600 on
  // DM500 units. The baud divisor should be set to 8 to get a
  // (close enough) 57600 baud rate on DM500 units.
  baud_divisor = IsDM500 ? 8 : AiopBaudDivisor(57600);

  PutChanReg(ch,baudDivisor,swab(baud_divisor));
  PutChanReg(ch,txControl,0);
  PutChanReg(ch,rxControl,0);
  PutChanReg(ch,txEnables,0);
  PutChanReg(ch,txCompare,0);
  PutChanReg(ch,txReplace1,0);
  PutChanReg(ch,txReplace2,0);

  
  // reset tx fifio logic
  a->command = AiopCommand_ResetTxFifo | chanNum;
  a->command = chanNum;

  // set tx fifo pointers to 0
  AiopPutIndexedShort(a,base+_TXF_OUTP,0);
  
  // reset rx fifo logic
  a->command = AiopCommand_ResetRxFifo | chanNum;
  a->command = chanNum;

  // set rx fifo pointers to 0
  AiopPutIndexedShort(a,base+_RXF_OUTP,0);
  AiopPutIndexedShort(a,base+_RXF_INP,0);

  // clear tx priority counter and pointer
  AiopPutIndexedByte(a,base+_TXP_CNT,0);
  AiopPutIndexedByte(a,base+_TXP_PNTR,0);
  
  sEnRxProcessor(ch);
  sEnTransmit(ch);
  sEnRxFifo(ch);
  sSetData8(ch);
  sEnRxStatusMode(ch);
}

static int shortDelay(void)
{
  volatile static int i;
  int j;
  for (j=0; j<100; ++j)
    ++i;
  return i;
}

void sDisRxFIFO(tAiopChan *ch)
{
  ch->R[12].value = (ch->R[12].value & 0x00ff) | 0x0a00;
  AiopPutIndexedShort(ch->aiop,ch->R[12].offset,ch->R[12].value);
}

void sStopRxProcessor(tAiopChan *ch)
{
  unsigned short v = (ch->R[0].value & 0x00ff) | 0x0a00;
  AiopPutIndexedShort(ch->aiop,ch->R[0].offset,v);
}

void sStartRxProcessor(tAiopChan *ch)
{
  AiopPutIndexedShort(ch->aiop,ch->R[0].offset,ch->R[0].value);\
}

// Flush input data
static void aiopic_flush_input(tAiopChan *ch)
{
  sDisRxFIFO(ch);
  shortDelay();
  // reset rx fifo logic
  ch->aiop->command = AiopCommand_ResetRxFifo | ch->chan;
  ch->aiop->command = ch->chan;
  // set rx fifo pointers to 0
  AiopPutIndexedShort(ch->aiop,ch->base+_RXF_OUTP,0);
  AiopPutIndexedShort(ch->aiop,ch->base+_RXF_INP,0);
  sEnRxFifo(ch);
}
      
// Flush output data
static void aiopic_flush_output(tAiopChan *ch)
{
  int txEnabled;
    
  txEnabled = GetChanReg(ch,txControl) & AiopTxCntrl_TxEnable;
  if (txEnabled)
    sDisTransmit(ch);
  sStopRxProcessor(ch);
  ch->aiop->command = AiopCommand_ResetTxFifo | ch->chan;
  ch->aiop->command = ch->chan;
  // set tx fifo pointers to 0
  AiopPutIndexedShort(ch->aiop,ch->base+_TXF_OUTP,0);
  if (txEnabled)
    sEnTransmit(ch);
  sStartRxProcessor(ch);
}

typedef struct
{
  tAiopChan *ch;
  cyg_int32 msec_timeout;
  int isr_vector;
} tAiopChanData;


// configure a UART
static void hal_aiopic_init_channel(void* __ch_data)
{
  tAiopChan* ch = ((tAiopChanData*)__ch_data)->ch;

  if ((ch-chanInfo) < 8)
    initAiopChan(&_aiop0,ch->chan,ch);
  else if ((ch-chanInfo) < 16)
    initAiopChan(&_aiop1,ch->chan,ch);
  else if ((ch-chanInfo) < 24)
    initAiopChan(&_aiop2,ch->chan,ch);
  else 
    initAiopChan(&_aiop3,ch->chan,ch);

  aiopic_flush_output(ch);
  aiopic_flush_input(ch);
}

void hal_aiopic_putc(void *__ch_data, char c)
{
  tAiopChan* ch = ((tAiopChanData*)__ch_data)->ch;

  CYGARC_HAL_SAVE_GP();
  while ((*(ch->status) & AiopChanStatus_TxFifoAvail) == 0)
    ;
  *(ch->dataByte) = c;
  CYGARC_HAL_RESTORE_GP();
}

static void __test_aiopic_putc(void *__ch_data, char c)
{
  tAiopChan* ch = ((tAiopChanData*)__ch_data)->ch;
  while ((*(ch->status) & AiopChanStatus_TxFifoAvail) == 0)
    ;
  *(ch->dataByte) = c;
}

extern volatile unsigned char EpldInterfaceSelect0;
extern volatile unsigned char EpldInterfaceSelect1;

// before we start returning data we must see a "wakeup sequence"
// that will enable the RS-232 drivers on the port and set a flag
// to allow subsequent input data to be passed to caller

static cyg_bool hal_aiopic_getc_nonblock(void* __ch_data, cyg_uint8* p)
{
  static const char wakeupSequence[] = "#!DM";
  static int awake,count,index;
  unsigned char c;
  
  tAiopChan* ch = ((tAiopChanData*)__ch_data)->ch;
  if ((*(ch->status) & AiopChanStatus_RxDataAvail) == 0)
    return false;
  
  c = *(ch->dataWord) >> 8;
  if (awake)
    {
      *p = c;
      return true;
    }
  
  // port not awake yet, watch for wakeup sequence

  ++count;
  if (c == wakeupSequence[index])
    ++index;
  else
    index = 0;
  
  if (index == (sizeof wakeupSequence)-1 && 
      count <= (sizeof wakeupSequence)+2)
    {
      awake = 1;
      EpldInterfaceSelect0 = 0;  // enable RS-232 drivers
    }
  
  return false;
}


cyg_uint8 hal_aiopic_getc(void* __ch_data)
{
  cyg_uint8 ch;
  
  CYGARC_HAL_SAVE_GP();
  while(!hal_aiopic_getc_nonblock(__ch_data, &ch))
    {
      do_us_delay(100);
      cyg_hal_rpsh_idle_stuff();
    }
  CYGARC_HAL_RESTORE_GP();
  return ch;
}

static void initAiop(tAiopic *a)
{
  a->command = Bit(1);
  shortDelay();
  a->command = 0;
  AiopPutIndexedByte(a,_CLK_PRE,AiopPrescale);
}

static void hal_aiopic_write(void* __ch_data, 
                             const cyg_uint8* __buf,
                             cyg_uint32 __len)
{
  CYGARC_HAL_SAVE_GP();
  while(__len-- > 0)
    hal_aiopic_putc(__ch_data, *__buf++);
  CYGARC_HAL_RESTORE_GP();
}

static void hal_aiopic_read(void* __ch_data, 
                            cyg_uint8* __buf, 
                            cyg_uint32 __len)
{
  CYGARC_HAL_SAVE_GP();
  while(__len-- > 0)
    *__buf++ = hal_aiopic_getc(__ch_data);
  CYGARC_HAL_RESTORE_GP();
}

static cyg_bool hal_aiopic_getc_timeout(void* __ch_data, cyg_uint8* ch)
{
  int delay_count;
  t4510ChanData* chan = (t4510ChanData*)__ch_data;
  cyg_bool res;
  CYGARC_HAL_SAVE_GP();
  
  delay_count = chan->msec_timeout;
  
  for(;;) 
    {
      res = hal_aiopic_getc_nonblock(__ch_data, ch);
      if (res || 0 == --delay_count)  // no delay for count of 1
        break;
      do_ms_tick();
      cyg_hal_rpsh_idle_stuff();
    }
  
  CYGARC_HAL_RESTORE_GP();
  return res;
}

static int hal_aiopic_control(void *__ch_data, 
                              __comm_control_cmd_t __func, 
                              ...)
{
  static int irq_state = 0;
  tAiopChanData* chan = (tAiopChanData*)__ch_data;
  int ret = 0;
  CYGARC_HAL_SAVE_GP();
  
  switch (__func) 
    {
     case __COMMCTL_IRQ_ENABLE:
      irq_state = 1;
      HAL_INTERRUPT_UNMASK(chan->isr_vector);
      break;
     case __COMMCTL_IRQ_DISABLE:
      ret = irq_state;
      irq_state = 0;
      HAL_INTERRUPT_MASK(chan->isr_vector);
      break;
     case __COMMCTL_DBG_ISR_VECTOR:
      ret = chan->isr_vector;
      break;
     case __COMMCTL_SET_TIMEOUT:
        {
          va_list ap;
          va_start(ap, __func);
          ret = chan->msec_timeout;
          chan->msec_timeout = va_arg(ap, cyg_uint32);
          va_end(ap);
        }
     default:
      break;
    }
  CYGARC_HAL_RESTORE_GP();
  return ret;
}

static int hal_aiopic_isr_count;

static int hal_aiopic_isr(void *__ch_data, 
                                  int* __ctrlc,
                                  CYG_ADDRWORD __vector, 
                                  CYG_ADDRWORD __data)
{
  int res = 0;
  tAiopChanData* chan = (tAiopChanData*)__ch_data;
  char c;
  CYGARC_HAL_SAVE_GP();
  ++hal_aiopic_isr_count;
  cyg_drv_interrupt_acknowledge(chan->isr_vector);
  *__ctrlc = 0;
  if (hal_aiopic_getc_nonblock(__ch_data, &c) && cyg_hal_is_break(&c,1))
    *__ctrlc = 1;
  res = CYG_ISR_HANDLED;
  CYGARC_HAL_RESTORE_GP();
  return res;
}

//  Gets a byte from the AIOPIC
//  Inputs:  __ch_data - AIOPIC channel struct
//           p - Gets filled in with the data read
//           cs - Gets filled with channel status before the read
//           s - Gets filled with channel status after the read
//  Returns: 1 if OK, 0 if error
static cyg_bool __test_aiopic_getc_nb(void* __ch_data, 
				      cyg_uint8* p, 
				      cyg_uint16 *cs, 
				      cyg_uint8 *s)
{
  unsigned short t;
  tAiopChan* ch = ((tAiopChanData*)__ch_data)->ch;
  *cs = t = *(ch->status);
  if ((t & AiopChanStatus_RxDataAvail) == 0)
    return false;
  t =  *(ch->dataWord);
  *p = t>>8;
  *s = t;
  return true;
}


static tAiopChanData aiop_ser_channels[MAXPORTS] =
{
    {&chanInfo[0],  1000, CYGNUM_HAL_INTERRUPT_Ext1},
    {&chanInfo[1],  1000, CYGNUM_HAL_INTERRUPT_Ext1},
    {&chanInfo[2],  1000, CYGNUM_HAL_INTERRUPT_Ext1},
    {&chanInfo[3],  1000, CYGNUM_HAL_INTERRUPT_Ext1},
    {&chanInfo[4],  1000, CYGNUM_HAL_INTERRUPT_Ext1},
    {&chanInfo[5],  1000, CYGNUM_HAL_INTERRUPT_Ext1},
    {&chanInfo[6],  1000, CYGNUM_HAL_INTERRUPT_Ext1},
    {&chanInfo[7],  1000, CYGNUM_HAL_INTERRUPT_Ext1},
    {&chanInfo[8],  1000, CYGNUM_HAL_INTERRUPT_Ext1},
    {&chanInfo[9],  1000, CYGNUM_HAL_INTERRUPT_Ext1},
    {&chanInfo[10], 1000, CYGNUM_HAL_INTERRUPT_Ext1},
    {&chanInfo[11], 1000, CYGNUM_HAL_INTERRUPT_Ext1},
    {&chanInfo[12], 1000, CYGNUM_HAL_INTERRUPT_Ext1},
    {&chanInfo[13], 1000, CYGNUM_HAL_INTERRUPT_Ext1},
    {&chanInfo[14], 1000, CYGNUM_HAL_INTERRUPT_Ext1},
    {&chanInfo[15], 1000, CYGNUM_HAL_INTERRUPT_Ext1},
    {&chanInfo[16], 1000, CYGNUM_HAL_INTERRUPT_Ext1},
    {&chanInfo[17], 1000, CYGNUM_HAL_INTERRUPT_Ext1},
    {&chanInfo[18], 1000, CYGNUM_HAL_INTERRUPT_Ext1},
    {&chanInfo[19], 1000, CYGNUM_HAL_INTERRUPT_Ext1},
    {&chanInfo[20], 1000, CYGNUM_HAL_INTERRUPT_Ext1},
    {&chanInfo[21], 1000, CYGNUM_HAL_INTERRUPT_Ext1},
    {&chanInfo[22], 1000, CYGNUM_HAL_INTERRUPT_Ext1},
    {&chanInfo[23], 1000, CYGNUM_HAL_INTERRUPT_Ext1},
    {&chanInfo[24], 1000, CYGNUM_HAL_INTERRUPT_Ext1},
    {&chanInfo[25], 1000, CYGNUM_HAL_INTERRUPT_Ext1},
    {&chanInfo[26], 1000, CYGNUM_HAL_INTERRUPT_Ext1},
    {&chanInfo[27], 1000, CYGNUM_HAL_INTERRUPT_Ext1},
    {&chanInfo[28], 1000, CYGNUM_HAL_INTERRUPT_Ext1},
    {&chanInfo[29], 1000, CYGNUM_HAL_INTERRUPT_Ext1},
    {&chanInfo[30], 1000, CYGNUM_HAL_INTERRUPT_Ext1},
    {&chanInfo[31], 1000, CYGNUM_HAL_INTERRUPT_Ext1}
};

#if defined(CYGPKG_REDBOOT)
static void hal_aiopic_init(int chan)
{
  int i;
  hal_virtual_comm_table_t* comm;
  
  HAL_INTERRUPT_MASK(aiop_ser_channels[0].isr_vector);
  
  initAiop(&_aiop0);
  initAiop(&_aiop1);
  initAiop(&_aiop2);
  initAiop(&_aiop3);
  for (i=0; i<MAXPORTS; ++i)
    {
      aiop_ser_channels[i].ch->chan = i%8;
      hal_aiopic_init_channel(&aiop_ser_channels[i]);
    }
   
  CYGACC_CALL_IF_SET_CONSOLE_COMM(chan);
  comm = CYGACC_CALL_IF_CONSOLE_PROCS();
  CYGACC_COMM_IF_CH_DATA_SET(*comm, &aiop_ser_channels[0]);
  CYGACC_COMM_IF_WRITE_SET(*comm, hal_aiopic_write);
  CYGACC_COMM_IF_READ_SET(*comm, hal_aiopic_read);
  CYGACC_COMM_IF_PUTC_SET(*comm, hal_aiopic_putc);
  CYGACC_COMM_IF_GETC_SET(*comm, hal_aiopic_getc);
  CYGACC_COMM_IF_CONTROL_SET(*comm, hal_aiopic_control);
  CYGACC_COMM_IF_DBG_ISR_SET(*comm, hal_aiopic_isr);
  CYGACC_COMM_IF_GETC_TIMEOUT_SET(*comm, hal_aiopic_getc_timeout);
}
#endif

extern volatile unsigned short *RingIndicatorRegister;
#define RiStatus(ch) (!(*RingIndicatorRegister & (1<<ch->chan)))
#define CtsStatus(ch) (!!(*(ch->status) & AiopChanStatus_CTS))
#define DsrStatus(ch) (!!(*(ch->status) & AiopChanStatus_DSR))
#define CdStatus(ch) (!!(*(ch->status) & AiopChanStatus_CD))

#define RS232_MODE  0x00
#define RS232I_MODE 0x18
#define RS422_MODE  0x08
#define RS485_MODE  0x10
#define RS485F_MODE 0x48        // Available on DM RTS 2P only
#define RSDIAG_MODE 0x50        // Available on DM RTS 2P only
#define RS485T_MODE 0x58        // Available on DM RTS 2P only

// Loopback test on devices that use AIOPICs.  
//   port: 0-31 
//   mode: 0=232, 1=232 RTS inv, 2=422, 3=Internal-loopback
// Return 0 for success, non-zero for failure
static int cyg_hal_plf_loopback_aiopic(int port, int mode, int doMC, int doRI)
{
    static unsigned char modeBits[4] = {RS232_MODE,RS232I_MODE,RS422_MODE,0xff};
    int i;
    tAiopChanData *ch = aiop_ser_channels+port;
    unsigned char b = modeBits[mode];
    unsigned char s;   // Channel status after the read
    unsigned short cs; // Channel status before the read
    
    if (mode != 3)
	SetPortMode(port,b);
    for (i=0; i<100; ++i)
	do_ms_tick();
    hal_aiopic_init_channel(ch);
    if (mode == 3)
	sEnLoopback(ch->ch);
    do_ms_tick();
  
    //  Put 900 bytes in the tx buffer
    for (i=0; i<900; ++i)
	__test_aiopic_putc(ch,i&0xff);
    for (i=0; i<200; ++i)
	do_ms_tick();

    //  Read the 900 bytes back
    for (i=0; i<900; ++i)
    {
	//  Read byte, if none, return channel status from before the read
	if (!__test_aiopic_getc_nb(ch,&b,&cs,&s))
	    return 0x10000 | cs;
	//  Check channel status after the read
	if (s)
	    return 0x20000 | s;
	//  Compare the data
	if ((i&0xff) != b)
	    return 0x30000 | ((i&0xff)<<8) | b;
    }
    if (__test_aiopic_getc_nb(ch,&b,&cs,&s))
	return 0x40000;
  
  if (doMC)
    {
      // test modem control lines
      int rtsState, dtrState;
      unsigned short status;
      for (rtsState = 0; rtsState <=1; ++rtsState)
        for (dtrState = 0; dtrState <=1; ++dtrState)
          {
            if (dtrState) sSetDTR(ch->ch); else sClrDTR(ch->ch);
            if (rtsState) sSetRTS(ch->ch); else sClrRTS(ch->ch);
            for (i=0; i<10; ++i) do_ms_tick();
            status = *(ch->ch->status);
            if ((rtsState^mode) != CtsStatus(ch->ch))
              return 0x50000 | status;
            if (doRI && ((rtsState^mode) != RiStatus(ch->ch))) 
              return 0x60000 | status;
            if (dtrState != DsrStatus(ch->ch))
              return 0x70000 | status;
            if (dtrState != CdStatus(ch->ch))
              return 0x80000 | status;
          }
    }
  sDisLoopback(ch->ch);
  return 0;
}

// Port-to-port RS-485 test
// If port1 equals port2, then use RS485 test mode (available on DM RTS 2P only)
// Otherwise, use RS485 half duplex mode
// Return 0 for success, non-zero for failure
static int cyg_hal_plf_485_test_aiopic(int port1, int port2)
{
  int i;
  tAiopChanData *ch1 = aiop_ser_channels+port1;
  tAiopChanData *ch2 = aiop_ser_channels+port2;
  unsigned char s,b;
  unsigned short cs;

  if (port1 == port2)
    {
      diag_printf("port%d->port%d, 485 test mode\n", port1, port2);

      SetPortMode(port1,RS485T_MODE);
      do_ms_tick();
      hal_aiopic_init_channel(ch1);
      sEnRTSToggle(ch1->ch);

      for (i=0; i<900; ++i)
        __test_aiopic_putc(ch1,i&0xff);
      for (i=0; i<200; ++i)
         do_ms_tick();
      for (i=0; i<900; ++i)
        {
          if (!__test_aiopic_getc_nb(ch1,&b,&cs,&s))
            return 0x1110000 | cs;
          if (s)
            return 0x1120000 | s;
          if ((i&0xff) != b)
            return 0x1130000 | ((i&0xff)<<8) | b;
        }
      if (__test_aiopic_getc_nb(ch1,&b,&cs,&s))
        return 0x1140000;

      sDisRTSToggle(ch1->ch);			// jira DMRTSBOOTLDR-58

      return 0;					// test mode success
    }

  SetPortMode(port1,RS485_MODE);
  SetPortMode(port2,RS485_MODE);
  do_ms_tick();
  hal_aiopic_init_channel(ch1);
  hal_aiopic_init_channel(ch2);
  sEnRTSToggle(ch1->ch);
  sEnRTSToggle(ch2->ch);

  diag_printf("port%d->port%d, half duplex mode\n", port1, port2);

  for (i=0; i<900; ++i)
    __test_aiopic_putc(ch1,i&0xff);
  for (i=0; i<200; ++i)
    do_ms_tick();
  for (i=0; i<900; ++i)
    {
      if (!__test_aiopic_getc_nb(ch2,&b,&cs,&s))
        return 0x1210000 | cs;
      if (s)
        return 0x1220000 | s;
      if ((i&0xff) != b)
        return 0x1230000 | ((i&0xff)<<8) | b;
    }
  if (__test_aiopic_getc_nb(ch2,&b,&cs,&s))
    return 0x1240000;

  diag_printf("port%d->port%d, half duplex mode\n", port2, port1);

  for (i=0; i<900; ++i)
    __test_aiopic_putc(ch2,i&0xff);
  for (i=0; i<200; ++i)
    do_ms_tick();
  for (i=0; i<900; ++i)
    {
      if (!__test_aiopic_getc_nb(ch1,&b,&cs,&s))
        return 0x2110000 | cs;
      if (s)
        return 0x2120000 | s;
      if ((i&0xff) != b)
        return 0x2130000 | ((i&0xff)<<8) | b;
    }
  if (__test_aiopic_getc_nb(ch1,&b,&cs,&s))
    return 0x2140000;

  sDisRTSToggle(ch1->ch);			// jira DMRTSBOOTLDR-58
  sDisRTSToggle(ch2->ch);

  return 0;					// port to port success
}


