/*********************************************************************
 *  hal_diag.c  
 * 
 *  Functions supporting the debug serial port.  This is done using the
 *  first built-in UART on the CPU.  Two CPUs are supprted - Samsung
 *  S3C4510 and S3C4530. 
 *********************************************************************/

#include <pkgconf/hal.h>
#include <pkgconf/system.h>
#include CYGBLD_HAL_PLATFORM_H
#include <cyg/infra/cyg_type.h>         // base types
#include <cyg/infra/cyg_trac.h>         // tracing macros
#include <cyg/infra/cyg_ass.h>          // assertion macros
#include <cyg/hal/hal_arch.h>           // basic machine info
#include <cyg/hal/hal_intr.h>           // interrupt macros
#include <cyg/hal/hal_io.h>             // IO macros
#include <cyg/hal/hal_if.h>             // calling interface API
#include <cyg/hal/hal_misc.h>           // helper functions
#include <cyg/hal/hal_diag.h>
#include <cyg/hal/drv_api.h>
#include "../include/ks32c5000.h"

#ifndef CYGPKG_REDBOOT
extern void diag_printf(const char *fmt, ...);
#define hal_aiopic_init(i) /* nothing */
#endif

extern void diag_printf(const char *fmt, ...);
extern void getModelNumber(unsigned long *m);

// global variable indicating whether a unit is a DM500 or not
// system clock is 48000000Hz for DM500, 44236800Hz for DM RTS
int IsDM500;
static unsigned int BaudClock;
#define BaudDivisor(b) (((BaudClock/(16*(b)))-1)<<4)

/*  Externs */
extern volatile tS3C4510Uart S3C4510_uart0;
extern volatile tS3C4510Uart S3C4510_uart1;
extern volatile tS3C4530Uart S3C4530_uart0;
extern volatile tS3C4530Uart S3C4530_uart1;
extern volatile unsigned char EpldInterfaceSelect[];
extern volatile unsigned int IOPCON1;
extern volatile unsigned int IOPMOD;
extern volatile unsigned int IOPDATA;
extern unsigned long do_ms_tick(void);

#if defined(CYGPKG_REDBOOT)
extern void do_us_delay(int);
#else
#define do_us_delay(x) CYGACC_CALL_IF_DELAY_US(x)
#endif

/*  Static Variables */
static int is4530;  // what kind of processor?
static int hal_S3C45xx_isr_count;

typedef struct
{
  volatile tS3C4510Uart *uart;
  cyg_int32 msec_timeout;
  int isr_vector;
  int awake;
  void (*enablePort)(void);
  int count;
  int index;
} t4510ChanData;


typedef struct
{
  volatile tS3C4530Uart *uart;
  cyg_int32 msec_timeout;
  int isr_vector;
  int awake;
  void (*enablePort)(void);
  int count;
  int index;
} t4530ChanData;

static void SetPortMode(int port, int mode);

static void port0enable(void)
{
  SetPortMode(0, 0);
}

static t4510ChanData S3C4510_channels[2] = 
{
    {&S3C4510_uart0, 1000, CYGNUM_HAL_INTERRUPT_Uart0Rx, 1},
    {&S3C4510_uart1, 1000, CYGNUM_HAL_INTERRUPT_Uart1Rx, 1},
};

static t4530ChanData S3C4530_channels[2] = 
{
    {&S3C4530_uart0, 1000, CYGNUM_HAL_INTERRUPT_Uart0Rx, 1},
    {&S3C4530_uart1, 1000, CYGNUM_HAL_INTERRUPT_Uart1Rx, 0, port0enable},
};

/*  Static functions */
static void hal_S3C4530_init_channel(t4530ChanData* __ch_data);
static void hal_S3C4530_putc(void *__ch_data, char c);
static cyg_bool hal_S3C4530_getc_nonblock(void* __ch_data, cyg_uint8* ch);
static cyg_uint8 hal_S3C4530_getc(void* __ch_data);
static void hal_S3C4510_putc(void *__ch_data, char c);
static cyg_bool hal_S3C4510_getc_nonblock(void* __ch_data, cyg_uint8* ch);
static cyg_uint8 hal_S3C4510_getc(void* __ch_data);
static void hal_S3C4510_init_channel(t4510ChanData* __ch_data);

// Empty function
void cyg_hal_rpsh_idle_stuff(void) { }

// pointers to put/getc routines for the two UART types
    
static void (*hal_S3C45xx_putc)(void *__ch_data, char c);
static cyg_bool (*hal_S3C45xx_getc_nonblock)(void* __ch_data, cyg_uint8* ch);
static cyg_uint8 (*hal_S3C45xx_getc)(void* __ch_data);

#if defined(CYGPKG_REDBOOT)
static int cyg_hal_plf_485_test_uart4530(void);
static int cyg_hal_plf_485_test_aiopic(int port1, int port2);
static int cyg_hal_plf_loopback_aiopic(int port, int mode, int doMC, int doRI);
static int cyg_hal_plf_loopback_4530(int mode);

#include "aiopic_redboot.c"
#include "uart4530_redboot.c"
#endif //  REDBOOT

//=====================================================================
// S3C4510 and S3C4530 common functions

static void hal_S3C45xx_write(void* __ch_data, 
                                     const cyg_uint8* __buf,
                                     cyg_uint32 __len)
{
  CYGARC_HAL_SAVE_GP();
  while(__len-- > 0)
    hal_S3C45xx_putc(__ch_data, *__buf++);
  CYGARC_HAL_RESTORE_GP();
}

static void hal_S3C45xx_read(void* __ch_data, 
                                    cyg_uint8* __buf, 
                                    cyg_uint32 __len)
{
  CYGARC_HAL_SAVE_GP();
  while(__len-- > 0)
    *__buf++ = hal_S3C45xx_getc(__ch_data);
  CYGARC_HAL_RESTORE_GP();
}

static cyg_bool hal_S3C45xx_getc_timeout(void* __ch_data, cyg_uint8* ch)
{
  int delay_count;
  cyg_bool res;
  int i;
  CYGARC_HAL_SAVE_GP();
  
  delay_count = is4530 ? ((t4530ChanData*)__ch_data)->msec_timeout :
                         ((t4510ChanData*)__ch_data)->msec_timeout;
  i = 10;
  for(;;) 
    {
      res = hal_S3C45xx_getc_nonblock(__ch_data, ch);
      if (res || (delay_count == 1))
        break;
      if (!--i)
        {
          i = 10;
          --delay_count;
        }
      cyg_hal_rpsh_idle_stuff();
      do_us_delay(100);
    }
  
  CYGARC_HAL_RESTORE_GP();
  return res;
}

static int hal_S3C45xx_control(void *__ch_data, 
                                      __comm_control_cmd_t __func, 
                                      ...)
{
  static int irq_state = 0;
  t4530ChanData* chan = (t4530ChanData*)__ch_data;
  int ret = 0;
  CYGARC_HAL_SAVE_GP();
  
  switch (__func) 
    {
     case __COMMCTL_IRQ_ENABLE:
      irq_state = 1;
      HAL_INTERRUPT_UNMASK(chan->isr_vector);
      if (is4530)
        chan->uart->inten |= (S3C4530UartIntEnaRxData | S3C4530UartIntEnaRxFifoTrig);
      break;
     case __COMMCTL_IRQ_DISABLE:
      ret = irq_state;
      irq_state = 0;
      HAL_INTERRUPT_MASK(chan->isr_vector);
      if (is4530)
        chan->uart->inten &= ~(S3C4530UartIntEnaRxData | S3C4530UartIntEnaRxFifoTrig);
      break;
     case __COMMCTL_DBG_ISR_VECTOR:
      ret = chan->isr_vector;
      break;
     case __COMMCTL_SET_TIMEOUT:
        {
          va_list ap;
          va_start(ap, __func);
          ret = chan->msec_timeout;
          chan->msec_timeout = va_arg(ap, cyg_uint32);
          va_end(ap);
        }
     default:
      break;
    }
  CYGARC_HAL_RESTORE_GP();
  return ret;
}

static int hal_S3C45xx_isr(void *__ch_data, 
                             int* __ctrlc,
                             CYG_ADDRWORD __vector, 
                             CYG_ADDRWORD __data)
{
  char c;
  t4510ChanData* chan4510 = (t4510ChanData*)__ch_data;
  t4530ChanData* chan4530 = (t4530ChanData*)__ch_data;
  CYGARC_HAL_SAVE_GP();
  ++hal_S3C45xx_isr_count;
  *__ctrlc = 0;
  if (is4530)
    {
      cyg_drv_interrupt_acknowledge(chan4530->isr_vector);
      if (chan4530->uart->ustat & S3C4530UartStatusRxDataAvail)
        {
          c = chan4530->uart->urxbuf;
          if (cyg_hal_is_break(&c,1))
            *__ctrlc = 1;
        }
      }
  else
    {
      cyg_drv_interrupt_acknowledge(chan4510->isr_vector);
      if (chan4510->uart->ustat & S3C4510UartStatusRxDataAvail)
        {
          c = chan4510->uart->urxbuf;
          if (cyg_hal_is_break(&c,1))
            *__ctrlc = 1;
        }
    }
  CYGARC_HAL_RESTORE_GP();
  return CYG_ISR_HANDLED;
}

//  Fills in the virtual vector table with the function pointer.
//  Inits the diag serial port(s).
static void hal_S3C45xx_init(void)
{
  hal_virtual_comm_table_t* comm;
  int i;

  if (is4530)
    {
      hal_S3C45xx_putc = hal_S3C4530_putc;
      hal_S3C45xx_getc = hal_S3C4530_getc;
      hal_S3C45xx_getc_nonblock = hal_S3C4530_getc_nonblock;
    }
  else
    {
      hal_S3C45xx_putc = hal_S3C4510_putc;
      hal_S3C45xx_getc = hal_S3C4510_getc;
      hal_S3C45xx_getc_nonblock = hal_S3C4510_getc_nonblock;
    }

  for (i=0; i<2; ++i)
    {
      if (is4530)
        {
	  HAL_INTERRUPT_MASK(S3C4530_channels[i].isr_vector);
          hal_S3C4530_init_channel(&S3C4530_channels[i]);
        }
      else
        {
          HAL_INTERRUPT_MASK(S3C4510_channels[i].isr_vector);
          hal_S3C4510_init_channel(&S3C4510_channels[i]);
        }

      CYGACC_CALL_IF_SET_CONSOLE_COMM(i);
      comm = CYGACC_CALL_IF_CONSOLE_PROCS();
      CYGACC_COMM_IF_CH_DATA_SET(*comm, is4530 ?(void*)(S3C4530_channels+i) : (void*)(S3C4510_channels+i));
      CYGACC_COMM_IF_WRITE_SET(*comm, hal_S3C45xx_write);
      CYGACC_COMM_IF_READ_SET(*comm, hal_S3C45xx_read);
      CYGACC_COMM_IF_PUTC_SET(*comm, hal_S3C45xx_putc);
      CYGACC_COMM_IF_GETC_SET(*comm, hal_S3C45xx_getc);
      CYGACC_COMM_IF_CONTROL_SET(*comm, hal_S3C45xx_control);
      CYGACC_COMM_IF_DBG_ISR_SET(*comm, hal_S3C45xx_isr);
      CYGACC_COMM_IF_GETC_TIMEOUT_SET(*comm, hal_S3C45xx_getc_timeout);
    }
}

void cyg_hal_plf_comms_init(void)
{
  static int initialized = 0;
  extern unsigned volatile SYSCFG;
  unsigned long model_id;

  is4530 = SYSCFG & Bit(27);

  if (initialized)
      return;
  initialized = 1;

  // set UART baud clock based on model ID
  getModelNumber(&model_id);
  IsDM500 = (model_id == 5002362 || model_id == 5002363 || model_id == 5002117);
  BaudClock = (IsDM500 ? 48000000 : 44236800) / 2;

  hal_S3C45xx_init(); //  Init the built-in CPU UARTS

  if (!is4530)         //  Init the customer data serial ports
    hal_aiopic_init(2);

  CYGACC_CALL_IF_SET_CONSOLE_COMM(0);
}

static void SetPortMode(int port, int mode)
{
    extern volatile unsigned char EpldInterfaceSelect[];

    EpldInterfaceSelect[port>>3] = (port & 0x7) | mode;
}

#if defined(CYGPKG_REDBOOT)
// Port-to-port RS-485 test.
// Return 0 for success, non-zero for failure
int cyg_hal_plf_485_test(int port1, int port2, int use4530)
{
    if (use4530) 
	return cyg_hal_plf_485_test_uart4530();
    else 
	return cyg_hal_plf_485_test_aiopic(port1, port2);
}

//  Loopback test.
//  Inputs:  port (0-31) 
//           mode: 0=232, 1=232 RTS inv, 2=422, 3=Internal-loopback
//           doMC: Check modem control lines?
//           doRI: Check ring indicator?
//           use4530: (0=devices with AIOPICS, 1=4530 devices)
//   Return: 0 for success, non-zero for failure
int cyg_hal_plf_loopback(int port, int mode, int doMC, int doRI, int use4530)
{
    if (use4530) 
	return cyg_hal_plf_loopback_4530(mode);
    else 
	return cyg_hal_plf_loopback_aiopic(port, mode, doMC, doRI);
}
#endif

//======================================================================
// S3C4510 specific UART functions

// configure a UART
static void hal_S3C4510_init_channel(t4510ChanData* __ch_data)
{
  int i;
  volatile char c;
  t4510ChanData* chan = (t4510ChanData*)__ch_data;
  chan->uart->ulcon = S3C4510UartLineWordLen8 + S3C4510UartLineStop1 + S3C4510UartLineParityNone;
  chan->uart->ucon  = S3C4510UartControlRxModeInterrupt + S3C4510UartControlTxModeInterrupt;
  chan->uart->brdiv = BaudDivisor(57600);
  // flush input data
  for (i=0 ;i<128; ++i)
    c = chan->uart->urxbuf;
}

static void hal_S3C4510_putc(void *__ch_data, char c)
{
  t4510ChanData* chan = (t4510ChanData*)__ch_data;

  CYGARC_HAL_SAVE_GP();
  while (!(chan->uart->ustat & S3C4510UartStatusTxBufEmpty))
    cyg_hal_rpsh_idle_stuff();
  chan->uart->utxbuf = c;
  CYGARC_HAL_RESTORE_GP();
}

static cyg_bool hal_S3C4510_getc_nonblock(void* __ch_data, cyg_uint8* ch)
{
  t4510ChanData* chan = (t4510ChanData*)__ch_data;
  
  if ((chan->uart->ustat & S3C4510UartStatusRxDataAvail) == 0)
    return false;
  
  *ch = chan->uart->urxbuf;
  return true;
}

static cyg_uint8 hal_S3C4510_getc(void* __ch_data)
{
  cyg_uint8 ch;
  
  CYGARC_HAL_SAVE_GP();
  while(!hal_S3C4510_getc_nonblock(__ch_data, &ch))
    {
      do_us_delay(100);
      cyg_hal_rpsh_idle_stuff();
    }
  CYGARC_HAL_RESTORE_GP();
  return ch;
}

//======================================================================
// S3C4530 specific UART functions

// Configure the UART
static void hal_S3C4530_init_channel(t4530ChanData* __ch_data)
{
    t4530ChanData* chan = (t4530ChanData*)__ch_data;

    chan->uart->ucon  = S3C4530UartControlTxModeInt +
	S3C4530UartControlRxModeInt +
	S3C4530UartControlParityNone +
	S3C4530UartControlWordLen8 +
	S3C4530UartControlTxFifoEn +
	S3C4530UartControlRxFifoEn +
	S3C4530UartControlDTR +
	S3C4530UartControlRTS;
    chan->uart->brdiv = BaudDivisor(57600);

    // flush input data
    chan->uart->ucon  |= S3C4530UartControlTxFifoRst;
    chan->uart->ucon  |= S3C4530UartControlRxFifoRst;
    chan->uart->ucon  &= ~S3C4530UartControlTxFifoRst;
    chan->uart->ucon  &= ~S3C4530UartControlRxFifoRst;
}

//  Write a byte, blocking until room in the TX fifo
static void hal_S3C4530_putc(void *__ch_data, char c)
{
    t4530ChanData* chan = (t4530ChanData*)__ch_data;

    CYGARC_HAL_SAVE_GP();
    while ((chan->uart->ustat & S3C4530UartStatusTxFifoFull)) 
	cyg_hal_rpsh_idle_stuff();
    chan->uart->utxbuf = c;
    CYGARC_HAL_RESTORE_GP();
}

//  Read a byte if available
//  Returns:  True of byte read, else false
static cyg_bool hal_S3C4530_getc_nonblock(void* __ch_data, cyg_uint8* ch)
{
  static const char wakeupSequence[] = "#!DM";
  t4530ChanData* chan = (t4530ChanData*)__ch_data;
  unsigned char c;

  // If RX Trig, there is either a byte or an RX error
  if (chan->uart->ustat & S3C4530UartStatusRxFifoTrig) 
    {
      if (chan->uart->ustat & S3C4530UartStatusRxDataAvail) 
        c = chan->uart->urxbuf; //  Good byte in read buffer
      else 
        {
        //  RX error, must set bit to clear it
          if (chan->uart->ustat & S3C4530UartStatusRxBreak)
            chan->uart->ustat |= S3C4530UartStatusRxBreak;
          if (chan->uart->ustat & S3C4530UartStatusRxFrameError)
            chan->uart->ustat |= S3C4530UartStatusRxFrameError;
          if (chan->uart->ustat & S3C4530UartStatusRxParityError)
            chan->uart->ustat |= S3C4530UartStatusRxParityError;
          if (chan->uart->ustat & S3C4530UartStatusRxOverrunError)
            chan->uart->ustat |= S3C4530UartStatusRxOverrunError;
          return false;
        }
    }
    else  
    {
      // no data avail
      return false;
    }
  
  if (chan->awake)
    {
      *ch = c;
      return true;
    }
  
  ++chan->count;
  if (c == wakeupSequence[chan->index])
    ++chan->index;
  else
    chan->index = 0;
  
  if (chan->index == (sizeof wakeupSequence)-1 &&
      chan->count <= (sizeof wakeupSequence)+2)
    {
      // !#DM has been received
      chan->awake = 1;
      if (chan->enablePort)
        chan->enablePort();
    }
  return false;
}

//  Read a byte, blocking if necessary
static cyg_uint8 hal_S3C4530_getc(void* __ch_data)
{
    cyg_uint8 ch;
  
    CYGARC_HAL_SAVE_GP();
    while(!hal_S3C4530_getc_nonblock(__ch_data, &ch))  {
	do_us_delay(100);
	
	cyg_hal_rpsh_idle_stuff();
    }
    CYGARC_HAL_RESTORE_GP();
    return ch;
}


int hal_diag_read_char_nonblocking(char *c)
{
  if (is4530 ? (S3C4530_uart0.ustat & S3C4530UartStatusRxDataAvail) :
      (S3C4510_uart0.ustat & S3C4510UartStatusRxDataAvail))
    {
      *c = S3C4510_uart0.urxbuf;
      return 1;
    }
  else
    return 0;
}
