#include <pkgconf/hal.h>

#include <cyg/infra/cyg_type.h>         // base types
#include <cyg/infra/cyg_trac.h>         // tracing macros
#include <cyg/infra/cyg_ass.h>          // assertion macros

#include <cyg/hal/hal_io.h>             // IO macros
#include <cyg/hal/hal_arch.h>           // Register state info
#include <cyg/hal/hal_diag.h>
#include <cyg/hal/hal_intr.h>           // necessary?
#include <cyg/hal/drv_api.h>

#include "../include/ks32c5000.h"

#include "version.h"
#define _Stringify(a) #a
#define Stringify(a) _Stringify(a)
char OSversionString[16] = Stringify(VERSION_MAJOR) "." Stringify(VERSION_MINOR);
char OSName[24] = "eCos-DvcMstr RTS";

#if defined(CYGPKG_REDBOOT)
#define diag_printf printf
#else
extern void diag_printf(const char *fmt, ...);
#endif

#ifndef CYGPKG_REDBOOT

#include <cyg/kernel/kapi.h>
#include <cyg/io/devtab.h>
#include <cyg/io/serial.h>
#include <cyg/io/serialio.h>
#include <cyg/hal/iic.h>

/*  Variable for the 1ms counter which uses Timer1.  Not used in Redboot. */
cyg_counter dmrtsCounterObj;
cyg_handle_t dmrtsCounterHandle;

/*  Variables used for the Timer1 interrupt */
static cyg_handle_t  dmrtsTimer1Handle;
static cyg_interrupt dmrtsTimer1Object;
static cyg_uint32 dmrtsTimer1ISR(cyg_vector_t vector, cyg_addrword_t data);
static void       dmrtsTimer1DSR(cyg_vector_t vector, cyg_ucount32 count, cyg_addrword_t data);
static void timerSetup(unsigned period);

// Variables used for the restore function interrupt
#define RESTORE_IRQ             CYGNUM_HAL_INTERRUPT_Ext3
static cyg_uint32 restoreISR(cyg_vector_t vector, cyg_addrword_t data);
static void       restoreDSR(cyg_vector_t vector, cyg_ucount32 count, cyg_addrword_t data);
static cyg_handle_t  restoreInterruptHandle;
static cyg_interrupt restoreInterruptObject;

#else

#include <pkgconf/hal.h>
#include <cyg/hal/drv_api.h>
#include <cyg/hal/hal_arch.h>
#include <cyg/hal/hal_intr.h>
 
#endif  // ifndef CYGPKG_REDBOOT


//======================================================================
// Use Timer0 for kernel clock

static cyg_uint32 _period;

extern volatile tTimers ks32c5000_timer;      // Timer registers
extern volatile unsigned PioData;
extern volatile unsigned PioMode;

volatile unsigned char *DallasDS1511W;
volatile unsigned short *RingIndicatorRegister;

static int is4530;

/* 
 *  Starts Timer0, used by the kernel as it's fundemental clock tick (10ms).
 */
void hal_clock_initialize(cyg_uint32 period)
{
  ks32c5000_timer.tmod &= ~TimerModeEnable0;      // disable timer
  ks32c5000_timer.tdata0 = _period = period;      // set period
  ks32c5000_timer.tmod |= TimerModeEnable0 + TimerModeToggle0; // enable timer
}

cyg_uint32 cyg_hal_get_period(void)
{
  return _period;
}
  
static int periodUpdate;

void cyg_hal_set_period(cyg_uint32 period)
{
  if (_period != period)
    {
      _period = period;
      periodUpdate = 1;
    }
}

void hal_clock_reset(cyg_uint32 vector, cyg_uint32 period)
{
  extern void cyg_hal_rpsh_idle_stuff(void);

  // the counter automatically reloads, so we don't need to do anything
  // unless period has changed
  if (periodUpdate)
    {
      unsigned r = ks32c5000_timer.tmod;
      ks32c5000_timer.tmod = r & ~TimerModeEnable0;
      ks32c5000_timer.tdata0 = _period;
      ks32c5000_timer.tmod = r;
      periodUpdate = 0;
    }
  cyg_hal_rpsh_idle_stuff();  // stuff we have to do periodically
}

/*
 *  Returns the number of ticks since last roll-over (timer is decrementing).
 */
void hal_clock_read(cyg_uint32 *pvalue)
{
  *pvalue = _period - ks32c5000_timer.tcnt0;
}



//======================================================================
// Interrupt controller stuff

extern volatile  tInterruptController ks32c5000_int;  // Interrupt controller registers
extern volatile  unsigned long EXTACON0;    // Extern access control reg
extern volatile  unsigned long EXTACON1;    // Extern access control reg
extern volatile  unsigned long IOPCON;      // I/O Port Control reg
extern volatile  unsigned long IOPMOD;
extern volatile unsigned long SYSCFG;
extern volatile unsigned long EXTDBWTH;

void hal_hardware_init(void)
{
  extern void hal_if_init(void);
  
  is4530 = SYSCFG & Bit(27);
  
  // Any hardware/platform initialization that needs to be done.
  ks32c5000_int.intmod = 0;           // All IRQ mode
  ks32c5000_int.intmsk = ~Bit(21);    // mask all interrupts (except global)
  ks32c5000_int.intpnd = 0xffffffff;  // clear all pending interrupts
  
  /*
   *   Setup external I/O bus timing parameters.  Sets access times for 
   *  EPLD register and UARTS 
   */
  EXTACON0 =  
       Tcos0(1)+Tacs0(0)+Tcoh0(1)+Tacc0(2)      // CPLD
      +Tcos1(4)+Tacs1(4)+Tcoh1(4)+Tacc1(1);     // AIOPICs
  
  EXTACON1 =
       Tcos2(1)+Tacs2(0)+Tcoh2(4)+Tacc2(5)      // Dallas DS1511W
      +Tcos3(0)+Tacs3(0)+Tcoh3(0)+Tacc3(2);     // Ring Indicators
  
  // Set the data width of EXT3 Ring Indicator to 16 bits
  EXTDBWTH &= ~(Bit(26) + Bit(27));
  EXTDBWTH |= Bit(27);

  if (is4530)
    {
      DallasDS1511W = (volatile unsigned char *)0x7680000;
      RingIndicatorRegister = (volatile unsigned short *)0x76c0000;
    }
  else
    {
      DallasDS1511W = (volatile unsigned char *)0x7608000;
      RingIndicatorRegister = (volatile unsigned short *)0x760c000;
    }
  
#if 0  
  // Configure extern interrupt inputs 
  // #2,3 for UARTs: Active high, rising-edge sensitive
  // #0 for PHY: Active low, falling-edge sensitive

  IOPCON &= ~((0x1f<<15) + (0x1f<<10) + (0x1f<<0));
  IOPCON |=   (0x11<<15) + (0x11<<10) + (0x12<<0);
#endif  

#ifndef CYGPKG_IO_SERIAL_16C850

#ifndef CYGPKG_REDBOOT   
  /*  
   *  Initialize and starts Timer1 for a 1ms counter.  Used to create alarms, 
   *  so tasks can poll at intervals of some multiple of 1 ms.  RPSHSI 2 Port 
   *  does not use this mechanism, it's serial driver uses Timer1 directly.
   */
  timerSetup(CYGNUM_HAL_CPUCLOCK/1000L);                /* 1ms interrupt */
  cyg_drv_interrupt_create(CYGNUM_HAL_INTERRUPT_Timer1,
                           99,                     
                           0,                      
                           dmrtsTimer1ISR,
                           dmrtsTimer1DSR,
                           &dmrtsTimer1Handle,
                           &dmrtsTimer1Object);
  cyg_drv_interrupt_attach(dmrtsTimer1Handle);
  cyg_drv_interrupt_acknowledge(CYGNUM_HAL_INTERRUPT_Timer1);
  cyg_drv_interrupt_unmask(CYGNUM_HAL_INTERRUPT_Timer1);

  // Initialize the restore function
  unsigned long model_id;
  getModelNumber(&model_id);
  // So far only DM RTS 2P models support restore function
  if ((model_id == 5002525) ||
      (model_id == 5002535) ||
      (model_id == 5002595) ||
      (model_id == 5002600))
    {
      // set up extern interrupt 3 to rising edge
      IOPCON &= ~(0x1f << 15);
      IOPCON |= (0x19 << 15);

      // set up the restore interrupt
      cyg_drv_interrupt_create(RESTORE_IRQ,
                               99,
                               0,
                               restoreISR,
                               restoreDSR,
                               &restoreInterruptHandle,
                               &restoreInterruptObject);
      cyg_drv_interrupt_attach(restoreInterruptHandle);

      // enable the restore interrupt
      cyg_drv_interrupt_acknowledge(RESTORE_IRQ);
      cyg_drv_interrupt_unmask(RESTORE_IRQ);
    }

  /*  Create the counter */
  cyg_counter_create(&dmrtsCounterHandle, &dmrtsCounterObj);

#endif // ifndef CYGPKG_REDBOOT

#endif  // ifndef CYGPKG_IO_SERIAL_16C850 


  // initialize virtual vector stuff
  hal_if_init();
}

#ifndef CYGPKG_REDBOOT

/*
 *  Timer 1 interrupt handler.  Increments counter tick.
 */
static cyg_uint32 dmrtsTimer1ISR(cyg_vector_t vector, cyg_addrword_t data)
{
    cyg_drv_interrupt_acknowledge(CYGNUM_HAL_INTERRUPT_Timer1);
    return CYG_ISR_HANDLED+CYG_ISR_CALL_DSR;
}

static void dmrtsTimer1DSR(cyg_vector_t vector, cyg_ucount32 count, cyg_addrword_t data)
{
    cyg_counter_tick(dmrtsCounterHandle);
}

// Restore function interrupt handler
static cyg_uint32 restoreISR(cyg_vector_t vector, cyg_addrword_t data)
{
    cyg_drv_interrupt_mask(RESTORE_IRQ);
    return CYG_ISR_HANDLED+CYG_ISR_CALL_DSR;
}

static void restoreDSR(cyg_vector_t vector, cyg_ucount32 count, cyg_addrword_t data)
{
    cyg_drv_isr_lock();
    restoreFactoryDefault();
    cyg_drv_isr_unlock();

    cyg_drv_interrupt_acknowledge(RESTORE_IRQ);
    cyg_drv_interrupt_unmask(RESTORE_IRQ);
    HAL_PLATFORM_RESET();
}

#endif // ifndef CYGPKT_REDBOOT

// This routine is called to respond to a hardware interrupt (IRQ).  It
// should interrogate the hardware and return the IRQ vector number.
int hal_IRQ_handler(void)
{
    //diag_init();  diag_printf("%s, status: %x\n", __PRETTY_FUNCTION__, ks32c5000_int.intoffset);
  
#if 0
  // can only use this if CPU is in IRQ or FRQ mode -- but
  // for eCOS we're in supervisor mode at this point.  Maybe
  // someday hack the assembly IRQ server to save the intoffset 
  // value?
  
  if ( ks32c5000_int.intoffset == 0x54)
    return CYGNUM_HAL_INTERRUPT_unused; // This shouldn't happen!

  return ks32c5000_int.intoffset >> 2;
#else
  {
    int pending = ks32c5000_int.intpnd;
    int v = 0;

    while (v<31)
      {
        if (pending & 1)
          return v;
        ++v;
        pending >>= 1;
      }
    return CYGNUM_HAL_INTERRUPT_unused;
  }
#endif
}


void hal_interrupt_mask(int vector)
{
  //diag_init();  diag_printf("%s(%d)\n", __PRETTY_FUNCTION__, vector);
  ks32c5000_int.intmsk |= (1<<vector);
}


void hal_interrupt_unmask(int vector)
{
  //diag_init();  diag_printf("%s(%d)\n", __PRETTY_FUNCTION__, vector);
  ks32c5000_int.intmsk &= ~(1<<vector);
}


void hal_interrupt_acknowledge(int vector)
{
  //diag_init();  diag_printf("%s(%d)\n", __PRETTY_FUNCTION__, vector);
  ks32c5000_int.intpnd = (1<<vector);
}


void hal_interrupt_configure(int vector, int level, int up)
{
  //diag_init();  diag_printf("%s(%d,%d,%d)\n", __PRETTY_FUNCTION__, vector, level, up);
}


void hal_interrupt_set_level(int vector, int level)
{
  //diag_init();  diag_printf("%s(%d,%d)\n", __PRETTY_FUNCTION__, vector, level);
}


void hal_show_IRQ(int vector, int data, int handler)
{
  //    diag_printf("IRQ - vector: %x, data: %x, handler: %x\n", vector, data, handler);
}

#define XTAL 44236800L

/*
 *  Delays a fixed number of microseconds.  
 *  
 *  In Redboot, uses timer 1.  In an application, timer1 is not available, so it does nop's.  Note 
 *  that this performs a tight loops,  it does not return control back to the scheduler.
 */
void hal_delay_us(int us)
{
#ifdef CYGPKG_REDBOOT

  extern void cyg_hal_rpsh_idle_stuff(void);
  extern unsigned long CLKCON;
  unsigned long divisor = CLKCON ? CLKCON<<1 : 1;
  unsigned long saved_tmod   = ks32c5000_timer.tmod;
  unsigned long saved_tdata1 = ks32c5000_timer.tdata1;
  
  // set up timer to count down for requested number of us 
  // we use high order bit to detect roll-over, so max usable
  // count value is 0x7fffffff (~64 seconds)

  ks32c5000_timer.tmod &= ~TimerModeEnable1;      // stop counter
  ks32c5000_timer.tdata1 = 0xffffffff;            // roll-over to all 1's
  ks32c5000_timer.tmod |= TimerModeEnable1;       // start counter
  ks32c5000_timer.tcnt1 = (us/divisor) * (XTAL/1000000L);  // set counter to proper value
  
  // wait for rollover to happen.
  
  while ((ks32c5000_timer.tcnt1 & 0x80000000) == 0)
    cyg_hal_rpsh_idle_stuff();
  
  ks32c5000_timer.tmod &= ~TimerModeEnable1;      // stop counter
  
  // restore counter settings
  ks32c5000_timer.tdata1 = saved_tdata1;
  ks32c5000_timer.tcnt1 = saved_tdata1;
  ks32c5000_timer.tmod = saved_tmod;

#else

    int i;

    /*  
     *  44Mhz clock, 44.3 ticks per us, 4 operations per for(i) loop, ~11 loops per us. 
     *  Testing showed that 2 loops per us created the correct delay
     */
    for (i = 0; i < (2 * us); i++) 
	asm("nop");
#endif
}


/*
 *   Reset the device by enabling the watchdog, then waiting...  
 */
void hal_rpsh_reset(void)
{
  CYG_INTERRUPT_STATE old;
  
  HAL_DISABLE_INTERRUPTS(old)
  PioMode |= Bit(16);
  DallasDS1511W[0x05] &= ~0x80;
  DallasDS1511W[0x0c] = 0x10;       // 0.10 second
  DallasDS1511W[0x0d] = 0x00;
  DallasDS1511W[0x0f] |= Bit(1)+Bit(0);
  
  while (1)
   ;
}

#if (0) // This conflicts with our use of UART1 for the 1 port product.  Kurt R. 

//=============================================================
// Ctrl-C handler for running under GDB stubs in bootloader.
// Watch for 0x03 receive byte on UART1.  If we receive one,
// set a breakpoint at the current location by calling the GDB
// stub routine via the virtual vector table.
//
// we only need the stuff below if we are not RedBoot and we are
// configured not to use virtual vector services.

#if !defined(CYGPKG_REDBOOT) && !defined(CYGSEM_HAL_VIRTUAL_VECTOR_DIAG)

# include <cyg/hal/drv_api.h>            // HAL ISR support

# define AbortInterrupt CYGNUM_HAL_INTERRUPT_Uart1Rx

extern volatile tS3C4510Uart S3C4510_uart1;
extern volatile tS3C4530Uart S3C4530_uart1;

// The location of the table should probably be defined by the
// linker comand file, and we should use the vector number
// definition from hal_if.h. But, that section of the inlcude
// file isn't enabled since we aren't using any other VV
// services.  Confused yet?
typedef void (*vfuncp)(void *p);
static vfuncp *redboot_virtual_vector_table = (vfuncp*)0x7c0030;
// The third param to ISR is undocumented, but it works on ARM...
static int abort_isr(cyg_vector_t vector, cyg_addrword_t data, HAL_SavedRegisters *regs)
{
  unsigned char rxData;
  int dobreak = 0;

  if (is4530)
      {
        while (S3C4530_uart1.ustat & S3C4530UartStatusRxDataAvail)
          {
            rxData = S3C4530_uart1.urxbuf;
            if (rxData == 0x03)
              dobreak = 1;
          }
      }
  else
    {
      while (S3C4510_uart1.ustat & S3C4510UartStatusRxDataAvail)
        {
          rxData = S3C4510_uart1.urxbuf;
          if (rxData == 0x03)
            dobreak = 1;
        }
    }
      
  if (dobreak)
      (redboot_virtual_vector_table[35])((void*)regs->pc);
  
  cyg_drv_interrupt_acknowledge(AbortInterrupt);
  return 0;  // Don't run DSR
}

static cyg_interrupt abort_interrupt;
static cyg_handle_t  abort_interrupt_handle;

static void init_abort_isr(void)
{
  cyg_drv_interrupt_create(AbortInterrupt,
                           99,
                           0,
                           abort_isr,
                           0,
                           &abort_interrupt_handle,
                           &abort_interrupt);
  cyg_drv_interrupt_attach(abort_interrupt_handle);
  // don't need to unmask it -- GDB stubs will do that.
}
#else
static void init_abort_isr(void)
{
  // do nothing
}
#endif

#endif  // Kurt R.

// OS information

char *getOsName(void)
{
  return OSName;
}

char *getOsVersion(void)
{
  return OSversionString;
}

#ifndef CYGPKG_REDBOOT

static void timerSetup(unsigned period)
{
    ks32c5000_timer.tmod &= ~TimerModeEnable1;                   // disable timer
    ks32c5000_timer.tdata1 = period;                             // set period
    ks32c5000_timer.tmod |= TimerModeEnable1 + TimerModeToggle1; // enable timer
}

#endif  // ifndef CYGPKG_REDBOOT
