/*********************************************************************
 *  iic.c
 *
 *  Methods to read and write iic memory
 *********************************************************************/
#include <pkgconf/system.h>
#include <cyg/hal/iic.h>
#include <string.h>

#define Bit(n) (1<<(n))

#define IIC_ACK         Bit(3)
#define IIC_Start       Bit(4)
#define IIC_Stop        Bit(5)
#define IIC_Restart     (Bit(5)|Bit(4))
#define IIC_BF          Bit(0)
#define IIC_Reset       Bit(7)

// I2C direction bit
#define IIC_WRITE       0
#define IIC_READ        1

// Log writes via diag header/port?
#define LogIicWriteEnable 0

volatile unsigned int *iicCon = (volatile unsigned int *)0x7fff000;
volatile unsigned int *iicBuf = (volatile unsigned int *)0x7fff004;
volatile unsigned int *iicPrescale = (volatile unsigned int *)0x7fff008;

extern void diag_printf(const char *fmt, ...);
extern void hal_delay_us(unsigned long usecs);

static int iicInitialReset;

//  Does an initial reset of the IIC controller on the EEPROM part
void iicInit(void) {
    *iicCon = IIC_Reset;
    *iicCon = IIC_BF;
    *iicPrescale = 100;
    *iicCon = 0;
}

//  Delay.  Each nop is 22ns, 200K of them is 4.4ms.
static void delay(void) {
    int i;
    for (i=0; i<100000; ++i) {
        asm("nop");
        asm("nop");
    }
}

//  Short delay.  Each nop is 22ns.
//  Inputs:  delay in microseconds
static void udelay(unsigned int usec) {
    int i;

    for (i = 0; i < usec * 227; ++i)  {
        asm("nop");
        asm("nop");
    }
}


/*  Wait for the BF flag to be set Returns: 1 if OK, 0 if timeout
    occurred */
static int waitForFlag(void) {
    int count = 0;

    // allow up to a msec since the scheduler should not be running
    while ((*iicCon & IIC_BF) == 0) {
        if (count++ > 100) {
            diag_printf("ERROR:  IIC timeout \n");
            iicInitialReset = 0;  // force a re-initialize of EEPROM
            return 0;
        }
        udelay(10);
    }
    return 1;
}

/*  Reads IIC data, returns count bytes at p */
/*  Not thread safe, use "Rt" functions if scheduler is running */
void iicRead(int slave_addr, int addr, void *p, int count) {
    unsigned char *buf = (unsigned char *)p;

    if (!iicInitialReset) {
        iicInit();
        iicInitialReset = 1;
        delay();
    }

    *iicCon = IIC_Start+IIC_ACK;
    *iicBuf = ((slave_addr << 1) | IIC_WRITE);
    if (!waitForFlag()) return;

    // need a delay here to avoid timing violation
    udelay(10);

    *iicBuf = addr;
    if (!waitForFlag()) return;

    // need a delay here to avoid timing violation
    udelay(10);

    *iicCon = IIC_Restart;
    *iicCon = IIC_Start+IIC_ACK;
    *iicBuf = ((slave_addr << 1) | IIC_READ);
    if (!waitForFlag()) return;

    // need a delay here to avoid timing violation
    udelay(10);

    while (count) {
        // need a delay here to avoid timing violation
        udelay(1);

        if (count == 1)
            *iicCon = 0;
        else
            *iicCon = IIC_ACK;

        if (!waitForFlag()) return;

        // need a delay here to avoid timing violation
        udelay(1);

        *buf++ = *iicBuf;

        // Do NOT remove this
        // need a delay here to avoid timing violation
        udelay(0);
        --count;
    }
    *iicCon = IIC_Stop;
}

/*  Write data p to addr */
/*  Not thread safe, use "Rt" functions if scheduler is running */
void iicWrite(int slave_addr, int addr, void *p, int count) {
    unsigned char *buf = (unsigned char *)p;
#ifdef CROSS_EIGHT_BYTE_BOUNDARIES
    unsigned long bytes_written = 0;
#endif

#if LogIicWriteEnable
#ifndef CYGPKG_REDBOOT
    {
        int ddiag_printf(char *fmt, ...);
        unsigned i;
        ddiag_printf("iicWrite(0x%04x,%p,%d)  ",addr,p,count);
        for (i=0; i<count; ++i)
            ddiag_printf("%02x", ((unsigned char *)p)[i]);
        ddiag_printf("\n");
    }
#endif
#endif

    if (!iicInitialReset) {
        iicInit();
        iicInitialReset = 1;
        delay();
    }

    *iicCon = IIC_Start+IIC_ACK;
    *iicBuf = ((slave_addr << 1) | IIC_WRITE);

    if (!waitForFlag()) return;
    *iicBuf = addr;

    while (count) {
        if (!waitForFlag()) return;
        *iicBuf = *buf++;
        --count;
#ifdef CROSS_EIGHT_BYTE_BOUNDARIES
        bytes_written++;

        // if we have crossed an eight byte boundary, we need at least 5 msec
        // wait here to allow iic part to finish writing, then restart
        // process for next block of eight bytes
        if (((((unsigned long)addr+bytes_written) % 8) == 0) && (count > 0)) {
            // get the ack for last write
            if (!waitForFlag()) return;

            // notify iic we have stopped stop sending data
            *iicCon = IIC_Stop;

            // wait 5 msec for write to complete in iic
            hal_delay_us(5000);

            // start write process again for next eight bytes
            *iicCon = IIC_Start+IIC_ACK;
            *iicBuf = ((slave_addr << 1) | IIC_WRITE);

            if (!waitForFlag()) return;

            addr += bytes_written;   // increment address to point to next block of bytes
            *iicBuf = addr;
            bytes_written = 0; // reset the bytes written count
        }
#endif // CROSS_EIGHT_BYTE_BOUNDARIES
    }
    if (!waitForFlag()) return;
    *iicCon = IIC_Stop;
    // need 5 msec wait here for write to complete
    hal_delay_us(5000);
}

void iicDelay(void) {
    int i;
    for (i=0; i<100000; ++i) {
        asm("nop");
        asm("nop");
    }
}


// EEPROM read/write routines with write-through cache.  Cache is
// initially filled by first call to iicEepromRead() -- which must be
// done before scheduler starts.  Subsequent calls to either of the
// read routines will return data from the cache.

#define EEPROM_SLAVE_ADDR       0x50
#define EEPROM_SIZE             128

static unsigned char eepromCache[EEPROM_SIZE];
static unsigned cacheValid;

void iicEepromRead(unsigned addr, void *p, int count) {
    if (!cacheValid) {
        iicRead(EEPROM_SLAVE_ADDR, 0, eepromCache, sizeof eepromCache);
        cacheValid = 1;
    }
    memcpy(p,&eepromCache[addr],count);
}

void iicEepromWrite(unsigned addr, void *p, unsigned count) {
    memcpy(&eepromCache[addr],p,count);
    iicWrite(EEPROM_SLAVE_ADDR, addr, p, count);
}

// Return EEPROM to all 0xff state (as received from distributor)
void iicEepromErase(void){
  static unsigned char eraseData[8] = {0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff};
  unsigned i = EEPROM_SIZE/(sizeof eraseData);
  unsigned p = 0;
  while (i)
    {
      iicEepromWrite(p,eraseData,sizeof eraseData);
      i -= 1;
      p += sizeof eraseData;
    }
}


void getMacAddress(unsigned char *p) {
    iicEepromRead(MAC_ADDRESS_ADDR, p, MAC_ADDRESS_SIZE);
}
void putMacAddress(unsigned char *p) {
    iicEepromWrite(MAC_ADDRESS_ADDR,     p,     4);
    delay();
    iicEepromWrite(MAC_ADDRESS_ADDR + 4, p + 4, 2);
    delay();
}

void getBoardRev(unsigned long *p) {
    iicEepromRead(BOARD_REV_ADDR, p, BOARD_REV_SIZE);
}
void putBoardRev(unsigned long p) {
    iicEepromWrite(BOARD_REV_ADDR, &p, BOARD_REV_SIZE);
    delay();
}

void getModelNumber(unsigned long *m) {
    iicEepromRead(MODEL_NUMBER_ADDR, m, MODEL_NUMBER_SIZE);
}
void putModelNumber(unsigned long m) {
    iicEepromWrite(MODEL_NUMBER_ADDR, &m, MODEL_NUMBER_SIZE);
    delay();
}

void putIpInfo(unsigned long ipAddr, unsigned long netMask, unsigned long gateway) {
    iicEepromWrite(IP_ADDRESS_ADDR, &ipAddr, IP_ADDRESS_SIZE);
    delay();
    iicEepromWrite(NETMASK_ADDR, &netMask, NETMASK_SIZE);
    delay();
    iicEepromWrite(GATEWAY_ADDR, &gateway, GATEWAY_SIZE);
    delay();
}
void getIpInfo(unsigned long *ipAddr, unsigned long *netMask, unsigned long *gateway) {
    iicEepromRead(IP_ADDRESS_ADDR, ipAddr, IP_ADDRESS_SIZE);
    iicEepromRead(NETMASK_ADDR, netMask, NETMASK_SIZE);
    iicEepromRead(GATEWAY_ADDR, gateway, GATEWAY_SIZE);
}

void getPassword(char *p) {
    iicEepromRead(PASSWORD_ADDR, p, PASSWORD_SIZE);
    p[PASSWORD_SIZE - 1] = '\0';
}
void putPassword(char *p) {
    char buf[PASSWORD_SIZE];
    strncpy(buf, p, PASSWORD_SIZE);
    buf[PASSWORD_SIZE - 1] = '\0';
    iicEepromWrite(PASSWORD_ADDR,      buf,      4);
    delay();
    iicEepromWrite(PASSWORD_ADDR + 4,  buf + 4,  4);
    delay();
    iicEepromWrite(PASSWORD_ADDR + 8,  buf + 8,  4);
    delay();
    iicEepromWrite(PASSWORD_ADDR + 12, buf + 12, 4);
    delay();
}

void getAuthMethod(unsigned char *p) {
    iicEepromRead(AUTH_METHOD_ADDR, p, AUTH_METHOD_SIZE);
}
void putAuthMethod(unsigned char m) {
    iicEepromWrite(AUTH_METHOD_ADDR, &m, AUTH_METHOD_SIZE);
    delay();
}

int getServiceEnable(unsigned char service) {
    unsigned char service_enable;
    iicEepromRead(SERVICE_ENABLE_ADDR, &service_enable, SERVICE_ENABLE_SIZE);
    if (service_enable & SERVICE_BYTE_IS_BROKEN) {
        service_enable &= SERVICE_BYTE_FIX_MASK;
        iicEepromWrite(SERVICE_ENABLE_ADDR, &service_enable, SERVICE_ENABLE_SIZE);
        delay();
    }
    return (service_enable & service) != 0;
}
void putServiceEnable(unsigned char service, unsigned char enable) {
    unsigned char service_enable;
    iicEepromRead(SERVICE_ENABLE_ADDR, &service_enable, SERVICE_ENABLE_SIZE);
    service_enable = (enable) ? (service_enable | service) :
                     (service_enable & ~service);
    iicEepromWrite(SERVICE_ENABLE_ADDR, &service_enable, SERVICE_ENABLE_SIZE);
    delay();
}

void getTelnetEnable(unsigned char *p) {
    *p = getServiceEnable(SERVICE_TELNET);
}
void putTelnetEnable(unsigned char m) {
    putServiceEnable(SERVICE_TELNET, m);
}

void getSecureDataEnable(unsigned char *p) {
    *p = getServiceEnable(SERVICE_SECURE_DATA);
}
void putSecureDataEnable(unsigned char m) {
    putServiceEnable(SERVICE_SECURE_DATA, m);
}

void getSecureConfigEnable(unsigned char *p) {
    *p = getServiceEnable(SERVICE_SECURE_CONFIG);
}
void putSecureConfigEnable(unsigned char m) {
    putServiceEnable(SERVICE_SECURE_CONFIG, m);
}

void getSecureMonitorEnable(unsigned char *p) {
    *p = getServiceEnable(SERVICE_SECURE_MONITOR);
}
void putSecureMonitorEnable(unsigned char m) {
    putServiceEnable(SERVICE_SECURE_MONITOR, m);
}

void getSnmpEnable(unsigned char *p) {
    *p = getServiceEnable(SERVICE_SNMP);
}
void putSnmpEnable(unsigned char m) {
    putServiceEnable(SERVICE_SNMP, m);
}

void getRedbootDelay(unsigned char *p) {
    iicEepromRead(TIMEOUT_ADDR, p, TIMEOUT_SIZE);
}
void putRedbootDelay(unsigned char t) {
    iicEepromWrite(TIMEOUT_ADDR, &t, TIMEOUT_SIZE);
    delay();
}

void getTelnetTimeout(unsigned long *p) {
    iicEepromRead(TELNET_TIMEOUT_ADDR, p, TELNET_TIMEOUT_SIZE);
}
void putTelnetTimeout(unsigned long t) {
    iicEepromWrite(TELNET_TIMEOUT_ADDR, &t, TELNET_TIMEOUT_SIZE);
    delay();
}

void getSerNumPrefix(unsigned short *p) {
    iicEepromRead(SERNUM_PREFIX_ADDR, p, SERNUM_PREFIX_SIZE);
}
void putSerNumPrefix(unsigned short p) {
    iicEepromWrite(SERNUM_PREFIX_ADDR, &p, SERNUM_PREFIX_SIZE);
    delay();
}

void getSerNum(unsigned short *p) {
    iicEepromRead(SERNUM_ADDR, p, SERNUM_SIZE);
}
void putSerNum(unsigned short p) {
    iicEepromWrite(SERNUM_ADDR, &p, SERNUM_SIZE);
    delay();
}

void getAppEnableKey(unsigned char *p) {
    iicEepromRead(APP_ENABLE_KEY_ADDR, p, APP_ENABLE_KEY_SIZE);
}

void getDhcpEnable(unsigned char *p) {
    // Enable DHCP (1) is stored as 0 and disable DHCP (0) is
    // stored as 1, the return value is inverted here.
    *p = (getServiceEnable(SERVICE_DHCP) == 0) ? 1: 0;
}
void putDhcpEnable(unsigned char m) {
    // Enable DHCP (1) is stored as 0 and disable DHCP (0) is
    // stored as 1, the bit to be stored is inverted here.
    putServiceEnable(SERVICE_DHCP, m ^ 0x1);
}

void getPortMacInfo(unsigned char *info, unsigned char addr, unsigned char size) {
    unsigned char port_mac_info[4];
    
    iicEepromRead(PORT_MAC_INFO_ADDR, &port_mac_info[0], PORT_MAC_INFO_SIZE);
    
	memcpy(info, &port_mac_info[addr], size);
}
void putPortMacInfo(unsigned char *info, unsigned char addr, unsigned char size) {
    unsigned char port_mac_info[4];

	iicEepromRead(PORT_MAC_INFO_ADDR, &port_mac_info[0], PORT_MAC_INFO_SIZE);
	
	memcpy(&port_mac_info[addr], info, size);

    iicEepromWrite(PORT_MAC_INFO_ADDR, &port_mac_info[0], 2);
    delay();
    iicEepromWrite(PORT_MAC_INFO_ADDR + 2, &port_mac_info[2], PORT_MAC_INFO_SIZE - 2);
    delay();
}

void getNumOfPortMacs(unsigned char *p) {
    getPortMacInfo(p, NUM_OF_PORT_MACS_ADDR, NUM_OF_PORT_MACS_SIZE);
}
void putNumOfPortMacs(unsigned char m) {
    putPortMacInfo(&m, NUM_OF_PORT_MACS_ADDR, NUM_OF_PORT_MACS_SIZE);
}

void getStartingPortMac(unsigned char *p) {
    getPortMacInfo(p, START_PORT_MAC_ADDR, START_PORT_MAC_SIZE);
}
void putStartingPortMac(unsigned char *m) {
    putPortMacInfo(m, START_PORT_MAC_ADDR, START_PORT_MAC_SIZE);
}

#ifdef CYGSEM_HAL_USER_PASSWORD_SUPPORT
void getUserPasswd(char *p) {
    iicEepromRead(USER_PASSWD_ADDR, p, USER_PASSWD_SIZE);
    p[USER_PASSWD_SIZE - 1] = '\0';
}
void putUserPasswd(char *p) {
    char buf[USER_PASSWD_SIZE];
    strncpy(buf, p, USER_PASSWD_SIZE);
    buf[USER_PASSWD_SIZE - 1] = '\0';
    iicEepromWrite(USER_PASSWD_ADDR,      buf,      2);
    delay();
    iicEepromWrite(USER_PASSWD_ADDR + 2,  buf + 2,  4);
    delay();
    iicEepromWrite(USER_PASSWD_ADDR + 6,  buf + 6,  4);
    delay();
    iicEepromWrite(USER_PASSWD_ADDR + 10, buf + 10, 4);
    delay();
    iicEepromWrite(USER_PASSWD_ADDR + 14, buf + 14, 2);
    delay();
}
#endif // If CYGSEM_HAL_USER_PASSWORD_SUPPORT


/* ============================================================== */
/*  These functions are used by eCos threads, so they need mutexs, use
    eCos delays, etc.   Use only when threads are up
    and running, not during initialization or under RedBoot. */

#if defined(CYGPKG_KERNEL)

#include <cyg/kernel/kapi.h>
static cyg_mutex_t iicMutex;
static int iicMutexInit;

/*  Waits for the BF flag to be set Returns: 1 if OK, 0 if timeout
    occurred */
static int waitForFlagRt(void) {
    unsigned long count = 0;

    while ((*iicCon & IIC_BF) == 0) {
        if (count > 20) {
            diag_printf("ERROR:  IIC timeout \n");
            iicInitialReset = 0; // force a re-initialization of iic part
            return 0;
        }
        udelay(5);
        count++;
    }
    return 1;
}

/*  Thread safe IIC read function.
    Inputs:  IIC addr to read from, void *p points to where to write data
    that is read, count is number of bytes to read*/
void iicReadRt(int slave_addr, int addr, void *p, int count) {
    unsigned char *buf = (unsigned char *)p;

    if (!iicMutexInit) {
        iicMutexInit = 1;
        cyg_mutex_init(&iicMutex);
    }
    cyg_mutex_lock(&iicMutex);

    if (!iicInitialReset) {
        iicInit();
        iicInitialReset = 1;
        cyg_thread_delay(1*CYGNUM_HAL_RTC_DENOMINATOR/100);
    }
    
    // Prevents IIC timeout errors.
    cyg_scheduler_lock();
    
    *iicCon = IIC_Start+IIC_ACK;
    *iicBuf = ((slave_addr << 1) | IIC_WRITE);

    if (!waitForFlagRt())
        goto failed;

    udelay(10);

    // set address to read from in IIC register
    *iicBuf = addr;
	
    if (!waitForFlagRt())
        goto failed;

    udelay(10);

    *iicCon = IIC_Restart;
    *iicCon = IIC_Start+IIC_ACK;
    *iicBuf = ((slave_addr << 1) | IIC_READ);

    if (!waitForFlagRt())
        goto failed;

    udelay(10);

    while (count) {
        udelay(1);

        if (count == 1)
            *iicCon = 0;
        else
            *iicCon = IIC_ACK;

        if (!waitForFlagRt())
            goto failed;

        udelay(1);

        *buf++ = *iicBuf;

        udelay(0);        // need a delay here to avoid timing violation
        --count;
    }
    *iicCon = IIC_Stop;

failed:
	cyg_scheduler_unlock();
    cyg_mutex_unlock(&iicMutex);
}


/*  Thread safe IIC write function.
    Inputs:  IIC addr to write to , void *p points to source data,
    count is number of bytes to write */
static void iicWriteRt(int slave_addr, int addr, void *p, int count) {
    unsigned char *buf = (unsigned char *)p;
#ifdef CROSS_EIGHT_BYTE_BOUNDARIES
    unsigned long bytes_written = 0;
#endif
    CYG_INTERRUPT_STATE oldstate;

#if LogIicWriteEnable
#ifndef CYGPKG_REDBOOT
    {
        int ddiag_printf(char *fmt, ...);
        unsigned i;
        ddiag_printf("iicWriteRt(0x%04x,%p,%d)  ",addr,p,count);
        for (i=0; i<count; ++i)
            ddiag_printf("%02x", ((unsigned char *)p)[i]);
        ddiag_printf("\n");
    }
#endif
#endif

    if (!iicMutexInit) {
        cyg_mutex_init(&iicMutex);
        iicMutexInit = 1;
    }

    cyg_mutex_lock(&iicMutex);

    if (!iicInitialReset) {
        iicInit();
        iicInitialReset = 1;
        cyg_thread_delay(1);
    }

    // disable interrupts
    HAL_DISABLE_INTERRUPTS(oldstate);

    *iicCon = IIC_Start+IIC_ACK;
    *iicBuf = ((slave_addr << 1) | IIC_WRITE);

    if (!waitForFlagRt()) {
        HAL_RESTORE_INTERRUPTS(oldstate);
        cyg_mutex_unlock(&iicMutex);
        return;
    }

    *iicBuf = addr;

    while (count) {
        if (!waitForFlagRt()) {
            HAL_RESTORE_INTERRUPTS(oldstate);
            cyg_mutex_unlock(&iicMutex);
            return;
        }

        *iicBuf = *buf++;
        --count;

#ifdef CROSS_EIGHT_BYTE_BOUNDARIES
        bytes_written++;

        // if we have crossed an eight byte page boundary, we need at least 5 msec
        // wait here to allow iic part to finish writing, then restart
        // process for next block of up to eight bytes
        if (((((unsigned long)addr+bytes_written) % 8) == 0) && (count > 0)) {
            // get the ack for last write
            if (!waitForFlagRt()) {
                HAL_RESTORE_INTERRUPTS(oldstate);
                cyg_mutex_unlock(&iicMutex);
                return;
            }
            // need to wait for write to complete
            *iicCon = IIC_Stop;

            // Restore interrupts while thread is delayed so other threads
            // and interrupts can run.
            HAL_RESTORE_INTERRUPTS(oldstate);
            cyg_thread_delay(1);              // 10ms - wait for write to complete
            HAL_DISABLE_INTERRUPTS(oldstate); // disable ints to continue write process

            // start write process again for next eight bytes
            *iicCon = IIC_Start+IIC_ACK;
            *iicBuf = ((slave_addr << 1) | IIC_WRITE);

            if (!waitForFlagRt()) {
                HAL_RESTORE_INTERRUPTS(oldstate);
                cyg_mutex_unlock(&iicMutex);
                return;
            }
            addr += bytes_written;   // increment address to point to next block of bytes
            *iicBuf = addr;
            bytes_written = 0; // reset the bytes written count
        }
#endif // CROSS_EIGHT_BYTE_BOUNDARIES
    }

    // wait for ack from last byte written
    if (!waitForFlagRt()) {
        HAL_RESTORE_INTERRUPTS(oldstate);
        cyg_mutex_unlock(&iicMutex);
        return;
    }

    *iicCon = IIC_Stop;
    HAL_RESTORE_INTERRUPTS(oldstate);
    cyg_thread_delay(1);  //  10ms delay to allow write to complete
    cyg_mutex_unlock(&iicMutex);
}


void iicEepromReadRt(unsigned addr, void *p, unsigned count) {
    if (!cacheValid)
        diag_printf("ERROR: IIC Eeprom cache not valid!");
    memcpy(p,&eepromCache[addr],count);
}

void iicEepromWriteRt(unsigned addr, void *p, unsigned count) {
    memcpy(&eepromCache[addr],p,count);
    iicWriteRt(EEPROM_SLAVE_ADDR, addr, p, count);
}

/*  Get/Sets the IP Address */
void getIpInfoRt(unsigned long *ipAddr, unsigned long *netMask, unsigned long *gateway) {
    iicEepromReadRt(IP_ADDRESS_ADDR, ipAddr, IP_ADDRESS_SIZE);
    iicEepromReadRt(NETMASK_ADDR, netMask, NETMASK_SIZE);
    iicEepromReadRt(GATEWAY_ADDR, gateway, GATEWAY_SIZE);
}
void putIpInfoRt(unsigned long ipAddr, unsigned long netMask, unsigned long gateway) {
    iicEepromWriteRt(IP_ADDRESS_ADDR, &ipAddr, IP_ADDRESS_SIZE);
    delay();
    iicEepromWriteRt(NETMASK_ADDR, &netMask, NETMASK_SIZE);
    delay();
    iicEepromWriteRt(GATEWAY_ADDR, &gateway, GATEWAY_SIZE);
    delay();
}

/*  Get/Sets the password */
void getPasswordRt(char *p) {
    iicEepromReadRt(PASSWORD_ADDR, p, PASSWORD_SIZE);
    p[15] = '\0';
}
void putPasswordRt(char *p) {
    p[PASSWORD_SIZE - 1] = '\0';
    iicEepromWriteRt(PASSWORD_ADDR,      p,      4);
    delay();
    iicEepromWriteRt(PASSWORD_ADDR + 4,  p + 4,  4);
    delay();
    iicEepromWriteRt(PASSWORD_ADDR + 8,  p + 8,  4);
    delay();
    iicEepromWriteRt(PASSWORD_ADDR + 12, p + 12, 4);
    delay();
}

/*  Get/Sets the service enable byte */
int getServiceEnableRt(unsigned char service) {
    unsigned char service_enable;
    iicEepromReadRt(SERVICE_ENABLE_ADDR, &service_enable, SERVICE_ENABLE_SIZE);
    return (service_enable & service) != 0;
}
void putServiceEnableRt(unsigned char service, unsigned char enable) {
    unsigned char service_enable;
    iicEepromReadRt(SERVICE_ENABLE_ADDR, &service_enable, SERVICE_ENABLE_SIZE);
    service_enable = (enable) ? (service_enable | service) :
                     (service_enable & ~service);
    iicEepromWriteRt(SERVICE_ENABLE_ADDR, &service_enable, SERVICE_ENABLE_SIZE);
    delay();
}

/*  Get/Sets the Telnet enable flag */
void getTelnetEnableRt(unsigned char *p) {
    *p = getServiceEnableRt(SERVICE_TELNET);
}
void putTelnetEnableRt(unsigned char m) {
    putServiceEnableRt(SERVICE_TELNET, m);
}

void getSecureDataEnableRt(unsigned char *p) {
    *p = getServiceEnableRt(SERVICE_SECURE_DATA);
}
void putSecureDataEnableRt(unsigned char m) {
    putServiceEnableRt(SERVICE_SECURE_DATA, m);
}

void getSecureConfigEnableRt(unsigned char *p) {
    *p = getServiceEnableRt(SERVICE_SECURE_CONFIG);
}
void putSecureConfigEnableRt(unsigned char m) {
    putServiceEnableRt(SERVICE_SECURE_CONFIG, m);
}


void getSecureMonitorEnableRt(unsigned char *p) {
    *p = getServiceEnableRt(SERVICE_SECURE_MONITOR);
}
void putSecureMonitorEnableRt(unsigned char m) {
    putServiceEnableRt(SERVICE_SECURE_MONITOR, m);
}

/*  Get/Sets the SNMP enable flag */
void getSnmpEnableRt(unsigned char *p) {
    *p = getServiceEnableRt(SERVICE_SNMP);
}
void putSnmpEnableRt(unsigned char m) {
    putServiceEnableRt(SERVICE_SNMP, m);
}

/*  Get/Sets the Dhcp enable flag */
void getDhcpEnableRt(unsigned char *p) {
    // Enable DHCP (1) is stored as 0 and disable DHCP (0) is
    // stored as 1, the return value is inverted here.
    *p = (getServiceEnableRt(SERVICE_DHCP) == 0) ? 1: 0;
}
void putDhcpEnableRt(unsigned char m) {
    // Enable DHCP (1) is stored as 0 and disable DHCP (0) is
    // stored as 1, the bit to be stored is inverted here.
    putServiceEnableRt(SERVICE_DHCP, m ^ 0x1);
}

/*  Get/Sets the "Timeout" period */
void getRedbootDelayRt(unsigned char *p) {
    iicEepromReadRt(TIMEOUT_ADDR, p, TIMEOUT_SIZE);
}
void putRedbootDelayRt(unsigned char t) {
    iicEepromWriteRt(TIMEOUT_ADDR, &t, TIMEOUT_SIZE);
    delay();
}

/*  Get/Sets the Telnet Timeout period */
void getTelnetTimeoutRt(unsigned long *p) {
    iicEepromReadRt(TELNET_TIMEOUT_ADDR, p, TELNET_TIMEOUT_SIZE);
}
void putTelnetTimeoutRt(unsigned long t) {
    iicEepromWriteRt(TELNET_TIMEOUT_ADDR, &t, TELNET_TIMEOUT_SIZE);
    delay();
}


void getPortMacInfoRt(unsigned char *info, unsigned char addr, unsigned char size) {
    unsigned char port_mac_info[4];
    
    iicEepromReadRt(PORT_MAC_INFO_ADDR, &port_mac_info[0], PORT_MAC_INFO_SIZE);
 
	memcpy(info, &port_mac_info[addr], size);
}
void putPortMacInfoRt(unsigned char *info, unsigned char addr, unsigned char size) {
    unsigned char port_mac_info[4];

	iicEepromReadRt(PORT_MAC_INFO_ADDR, &port_mac_info[0], PORT_MAC_INFO_SIZE);

	memcpy(&port_mac_info[addr], info, size);

    iicEepromWriteRt(PORT_MAC_INFO_ADDR, &port_mac_info[0], 2);
    delay();
    iicEepromWriteRt(PORT_MAC_INFO_ADDR + 2, &port_mac_info[2], PORT_MAC_INFO_SIZE - 2);
    delay();
}

void getNumOfPortMacsRt(unsigned char *p) {
    getPortMacInfoRt(p, NUM_OF_PORT_MACS_ADDR, NUM_OF_PORT_MACS_SIZE);
}
void putNumOfPortMacsRt(unsigned char m) {
    putPortMacInfoRt(&m, NUM_OF_PORT_MACS_ADDR, NUM_OF_PORT_MACS_SIZE);
}

void getStartingPortMacRt(unsigned char *p) {
    getPortMacInfoRt(p, START_PORT_MAC_ADDR, START_PORT_MAC_SIZE);
}
void putStartingPortMacRt(unsigned char *m) {
    putPortMacInfoRt(m, START_PORT_MAC_ADDR, START_PORT_MAC_SIZE);
}

void putAuthMethodRt(unsigned char m) {
    iicEepromWriteRt(AUTH_METHOD_ADDR, &m, AUTH_METHOD_SIZE);
    delay();
}

#ifdef CYGSEM_HAL_USER_PASSWORD_SUPPORT
/*  Get/Sets the user's password */
void getUserPasswdRt(char *p) {
    iicEepromReadRt(USER_PASSWD_ADDR, p, USER_PASSWD_SIZE);
    p[USER_PASSWD_SIZE - 1] = '\0';
}
void putUserPasswdRt(char *p) {
    p[USER_PASSWD_SIZE - 1] = '\0';
    iicEepromWriteRt(USER_PASSWD_ADDR,      p,      2);
    delay();
    iicEepromWriteRt(USER_PASSWD_ADDR + 2,  p + 2,  4);
    delay();
    iicEepromWriteRt(USER_PASSWD_ADDR + 6,  p + 6,  4);
    delay();
    iicEepromWriteRt(USER_PASSWD_ADDR + 10, p + 10, 4);
    delay();
    iicEepromWriteRt(USER_PASSWD_ADDR + 14, p + 14, 2);
    delay();
}
#endif // If CYGSEM_HAL_USER_PASSWORD_SUPPORT

#endif // If CYGPKG_KERNEL

// restore selected paramters to factory defaults
void restoreFactoryDefault(void) {
    diag_printf("Restoring factory defaults...\n");
    // restore default ip, mask, gateway
    putIpInfo(0xC0A8FAFA, 0xFFFF0000, 0xC0A8FA01);
    putPassword("");             // clear password
    putAuthMethod(1);            // authentication to none
    putTelnetEnable(1);          // telnet enabled
    putDhcpEnable(1);
    putSecureConfigEnable(0);
    putSecureDataEnable(0);
    putRedbootDelay(15);         // start up delay to 15 seconds
    putTelnetTimeout(300);       // telnet time out to 300 seconds
}
