
/* Loopback test routines for the 4530 UART connected to the outside world */


/*  Error states for the loopback test */
#define LOOP_PASS                 0
#define LOOP_ERROR                0x10000
#define LOOP_CHANNEL_STATUS_ERROR 0x20000
#define LOOP_DATA_VERIFY_ERROR    0x30000
#define LOOP_CTS_NOT_RTS          0x50000
#define LOOP_NO_CTS_EVENT         0x50001
#define LOOP_RI_NOT_RTS           0x60000
#define LOOP_DSR_NOT_DTR          0x70000
#define LOOP_CD_NOT_DTR           0x80000
#define LOOP_TIMEOUT              0x90000

#define NUM_BYTES 900
#define BREAK           3
#define FRAMING_ERROR   2
#define PARITY_ERROR    1
#define OVERRUN_ERROR   0

#define CD_SET (!(IOPDATA & Bit(5)))
#define RI_SET (!(IOPDATA & Bit(10)))
#define CTS_SET        (chan->uart->ustat & S3C4530UartStatusCTS)
#define DSR_SET        (chan->uart->ustat & S3C4530UartStatusDSR)
#define SET_DTR         chan->uart->ucon |= S3C4530UartControlDTR
#define CLEAR_DTR       chan->uart->ucon &= ~S3C4530UartControlDTR
#define SET_RTS         chan->uart->ucon |= S3C4530UartControlRTS
#define CLEAR_RTS       chan->uart->ucon &= ~S3C4530UartControlRTS
#define CTS_EVENT      (chan->uart->ucon &  S3C4530UartStatusCTSEvent)
#define CLEAR_CTS_EVENT chan->uart->ucon |= S3C4530UartStatusCTSEvent

//  Checks 4530 channel status
//  Inputs:  Pointer to channel, error byte to be filled in
//  Returns:  0 if OK, 1 if error
static int get4530ChannelStatus(t4530ChanData* chan, unsigned char *error)
{
    if (chan->uart->ustat & S3C4530UartStatusRxBreak) {
	chan->uart->ustat |= S3C4530UartStatusRxBreak;
	*error = BREAK;
	return 1;
    }
    if (chan->uart->ustat & S3C4530UartStatusRxFrameError) {
	chan->uart->ustat |= S3C4530UartStatusRxFrameError;
	*error = FRAMING_ERROR;
	return 1;
    }
    if (chan->uart->ustat & S3C4530UartStatusRxParityError) {
	chan->uart->ustat |= S3C4530UartStatusRxParityError;
	*error =  PARITY_ERROR;
	return 1;
    }
    if (chan->uart->ustat & S3C4530UartStatusRxOverrunError) {
	chan->uart->ustat |= S3C4530UartStatusRxOverrunError;
	*error = OVERRUN_ERROR; 
	return 1;
    }
    return 0;
}

#undef RS232_MODE
#undef RS232_INVERTED_MODE
#undef RS422_MODE
#undef INTERNAL_MODE

#define RS232_MODE          0
#define RS232_INVERTED_MODE 1
#define RS422_MODE          2
#define INTERNAL_MODE       3

// Loopback test on devices that use the 4530 CPU UART.  
//   mode: 0=232, 1=232 RTS inv, 2=422, 3=Internal-loopback
// Return:  0 for success, non-zero for failure
//          0x1000x Channel status error before write/read
//          0x2000x Channel status error after write/read
//          0x3xxyy Data compare error xx=expected yy=recieved
//          0x90000 Timeout
static int cyg_hal_plf_loopback_4530(int mode)
{
    static unsigned char modeBits[4] = {0x00,0x18,0x08,0xff};
    unsigned char c, error;
    int i, loops = 0, xmitCount = 0, rcvCount = 0;
    int xmitDone = 0, rcvDone = 0, ret, sta;
    t4530ChanData* chan = (t4530ChanData*)&S3C4530_channels[1];

    if (mode != 3)
	SetPortMode(0, modeBits[mode]);
    
    for (i=0; i<100; ++i)
	do_ms_tick();

    hal_S3C4530_init_channel(chan);       //  Init 2nd UART
    if (mode == 3)                        //  Internal loopback
	chan->uart->ucon  |= S3C4530UartControlLoopback;
    do_ms_tick();

    //  Configure UART for HW flow control, check UART status
    chan->uart->ucon |= S3C4530UartControlHwFlowEn; 
    sta = get4530ChannelStatus(chan, &error);
    if (sta) {
	ret = LOOP_CHANNEL_STATUS_ERROR | error;
	goto doneLoopback;
    }

    //  Start loopback test
    while (1) {

	//  Push in data
	if (!xmitDone) {
	    while (!(chan->uart->ustat & S3C4530UartStatusTxFifoFull)) {
		chan->uart->utxbuf = (xmitCount++ & 0xFF);
		if (xmitCount >= NUM_BYTES)
		    xmitDone = 1;
	    }
	}

	// Pull out data, verify
	if (!rcvDone) {
	    while (!(chan->uart->ustat & S3C4530UartStatusRxFifoEmpty)) {
		c = chan->uart->urxbuf;

		// Check UART status after read
		sta = get4530ChannelStatus(chan, &error);
		if (sta) {
		    ret = LOOP_CHANNEL_STATUS_ERROR | error;
		    goto doneLoopback;
		}

		//  Verify data recieved
		if (c != (rcvCount++ & 0xFF)) {
		    ret =  LOOP_DATA_VERIFY_ERROR | (((rcvCount - 1) & 0xFF) << 8) | c;
		    goto doneLoopback;
		}
		if (rcvCount >= NUM_BYTES)
		    rcvDone = 1;
	    }
	}

	//  Check for timeouts
	if (loops++ > 1000000) {
	    ret  = LOOP_TIMEOUT;
	    goto doneLoopback;
	}

	//  Check for done
	if (xmitDone && rcvDone) {
	    ret = LOOP_PASS;
	    goto doneLoopback;
	}
    }
doneLoopback:
    chan->uart->ucon &= ~S3C4530UartControlHwFlowEn; 
    chan->uart->ucon &= ~S3C4530UartControlLoopback;
    if ((ret) || mode == RS422_MODE || mode == INTERNAL_MODE)
	return ret;

    //  Enable direct control of UART1 RTS and CTS
    //  Access DCD and RI as an IO pins
    IOPCON1 = 0x3FC00C0;
    IOPMOD &= ~Bit(5);
    IOPMOD &= ~Bit(10);

    //  Modem signal tests.  Set RTS, check CTS & RI
    SET_RTS;
    for (i=0; i<200; ++i)  do_ms_tick();
    sta = get4530ChannelStatus(chan, &error);
    if (sta) {
	ret = LOOP_CHANNEL_STATUS_ERROR | error;
	goto doneModemSignalTest;
    }
    if (mode == RS232_MODE) {
	if (!CTS_SET) {
	    ret = LOOP_CTS_NOT_RTS;
	    goto doneModemSignalTest;
	}
	if (!CTS_EVENT) {
	    ret = LOOP_NO_CTS_EVENT;
	    goto doneModemSignalTest;
	}
	CLEAR_CTS_EVENT;
	if (!RI_SET) {
	    ret = LOOP_RI_NOT_RTS;
	    goto doneModemSignalTest;
	}
    }
    else {
	if (CTS_SET) {
	    ret = LOOP_CTS_NOT_RTS;
	    goto doneModemSignalTest;
	}
	if (!CTS_EVENT) {
	    ret = LOOP_NO_CTS_EVENT;
	    goto doneModemSignalTest;
	}
	CLEAR_CTS_EVENT;
	if (RI_SET) {
	    ret = LOOP_RI_NOT_RTS;
	    goto doneModemSignalTest;
	}
    }
    
    //  Clear RTS, check CTS & RI
    CLEAR_RTS;
    for (i=0; i<200; ++i)  do_ms_tick();
    sta = get4530ChannelStatus(chan, &error);
    if (sta) {
	ret = LOOP_CHANNEL_STATUS_ERROR | error;
	goto doneModemSignalTest;
    }
    if (mode == RS232_MODE) {
	if (CTS_SET) {
	    ret = LOOP_CTS_NOT_RTS;
	    goto doneModemSignalTest;
	}
	if (RI_SET) {
	    ret = LOOP_RI_NOT_RTS;
	    goto doneModemSignalTest;
	}
    }
    else {
	if (!CTS_SET) {
	    ret = LOOP_CTS_NOT_RTS;
	    goto doneModemSignalTest;
	}
	if (!RI_SET) {
	    ret = LOOP_RI_NOT_RTS;
	    goto doneModemSignalTest;
	}
    }

    //  Set DTR, check DSR & CD
    SET_DTR;
    for (i=0; i<200; ++i)  do_ms_tick();
    sta = get4530ChannelStatus(chan, &error);
    if (sta) {
	ret = LOOP_CHANNEL_STATUS_ERROR | error;
	goto doneModemSignalTest;
    }
    if (!DSR_SET) {
	    ret = LOOP_DSR_NOT_DTR;
	    goto doneModemSignalTest;
    }
    if (!CD_SET) {
	ret = LOOP_CD_NOT_DTR;
	goto doneModemSignalTest;
    }


    //  Clear DTR, check DSR  & CD
    CLEAR_DTR;
    for (i=0; i<200; ++i)  do_ms_tick();
    sta = get4530ChannelStatus(chan, &error);
    if (sta) {
	ret = LOOP_CHANNEL_STATUS_ERROR | error;
	goto doneModemSignalTest;
    }
    if (DSR_SET) {
	ret = LOOP_DSR_NOT_DTR;
	goto doneModemSignalTest;
    }
    if (CD_SET) {
	ret = LOOP_CD_NOT_DTR;
	goto doneModemSignalTest;
    }

doneModemSignalTest:
    CLEAR_DTR;
    CLEAR_RTS;
    IOPCON1 = 0x3FC0000;
    return ret;
}

#define TIMEOUT_DELAY 20000
#define CTS_BIT Bit(6)
#define RTS_BIT Bit(7)

//  Runs an internal 485 loopback test.  On the one port model,
//  there is not RS232 RTS inverted mode.  Rather, setting the 
//  CPLD to mode 3 results in the RS485 reciever being on when 
//  the transmitter is on.  Hence, we can send a byte in 485 mode
//  and recieve it back in the UART recieve buffer, testing the 485
//  tranciever.
//  Return 0 for success, non-zero for failure
static int cyg_hal_plf_485_test_uart4530(void)
{
    int xmitCount = 0, i, bytes = 1000;
    int rcvCount = 0, timeoutCounter = 0;
    unsigned char error, sendByte, rcvByte;
    t4530ChanData* chan = (t4530ChanData*)&S3C4530_channels[1];

    diag_printf("Running RS485 loopback test \n");

    //  Set RTS&CTS as UART pins, drive RTS high
    IOPCON1 = 0x03FC00C0;
    hal_S3C4530_init_channel(chan);       
    chan->uart->ucon |= S3C4530UartControlRTS;

    //  Set CPLD for RS485 loopback mode (0x18) on port0
    SetPortMode(0, 0x18);

    while (1) {

	//  Send a byte, wait a few no-ops
	sendByte = xmitCount++ & 0xFF;
	chan->uart->utxbuf = sendByte;
	for (i = 0; i < 1000; ++i)  asm("nop");

	//  Retrieve the looped byte
	while (1) {
	    if (chan->uart->ustat & S3C4530UartStatusRxFifoTrig) {

		//  Got a trigger, it's either valid data or an error
		if (chan->uart->ustat & S3C4530UartStatusRxDataAvail) {
		    rcvByte = chan->uart->urxbuf;
		    if (rcvByte != sendByte) {
			diag_printf("**  ERROR - Data verify error, expected:  %d   recieved:  %d\n\r", sendByte, rcvByte);
			goto done;
		    }
		    rcvCount++;
		    timeoutCounter = 0;
		    break;
		}
		else {  //  Not RX data avail, error
		    get4530ChannelStatus(chan, &error);
		    diag_printf("RX UART error\n");
		}
	    }
	    else {  //  Not FIFO trig
		if (timeoutCounter++ > TIMEOUT_DELAY) {
		    diag_printf("**  ERROR - Timeout \n\r");
		    goto done;
		}
	    }
	}
	if (xmitCount == bytes)
	    goto done;
    }

done:
    //  Set hardware back to RS232
    SetPortMode(0, 0);
    IOPCON1 = 0x03FC0000;
     hal_S3C4530_init_channel(chan);       

    if (xmitCount == rcvCount) {
	diag_printf("RS485 loopback test passed \n");
	return 0;
    }
    else {
	diag_printf("RS485 loopback test failed \n");
        return 0x2110000;
    }
}
