#include <pkgconf/system.h>
#include <pkgconf/io_nserial.h>
#include <pkgconf/io.h>

#include <cyg/io/io.h>
#include <cyg/hal/hal_intr.h>
#include <cyg/io/devtab.h>
#include <cyg/io/serial.h>
#include <cyg/infra/diag.h>
#include <cyg/infra/cyg_ass.h>
#include <cyg/hal/hal_io.h>
#include <string.h>

#define Trace()  diag_printf("%s:%d\n",__FILE__,__LINE__)

#define SER_REG(_x_) ((_x_)*4)

// Receive control Registers
#define REG_rhr SER_REG(0)	// Receive holding register
#define REG_isr SER_REG(2)	// Interrupt status register
#define REG_lsr SER_REG(5)	// Line status register
#define REG_msr SER_REG(6)	// Modem status register
#define REG_scr SER_REG(7)	// Scratch register

// Transmit control Registers
#define REG_thr SER_REG(0)	// Transmit holding register
#define REG_ier SER_REG(1)	// Interrupt enable register
#define REG_fcr SER_REG(2)	// FIFO control register
#define REG_lcr SER_REG(3)	// Line control register
#define REG_mcr SER_REG(4)	// Modem control register
#define REG_ldl SER_REG(0)	// LSB of baud rate
#define REG_mdl SER_REG(1)	// MSB of baud rate

// Interrupt Enable Register
#define IER_RCV 0x01
#define IER_XMT 0x02
#define IER_LS  0x04
#define IER_MS  0x08

// Line Control Register
#define LCR_WL5 0x00		// Word length
#define LCR_WL6 0x01
#define LCR_WL7 0x02
#define LCR_WL8 0x03
#define LCR_SB1 0x00		// Number of stop bits
#define LCR_SB1_5 0x04		// 1.5 -> only valid with 5 bit words
#define LCR_SB2 0x04
#define LCR_PN  0x00		// Parity mode - none
#define LCR_PE  0x18		// Parity mode - even
#define LCR_PO  0x08		// Parity mode - odd
#define LCR_PM  0x28		// Forced "mark" parity
#define LCR_PS  0x38		// Forced "space" parity
#define LCR_DL  0x80		// Enable baud rate latch

// Line Status Register
#define LSR_RSR 0x01
#define LSR_OE  0x02
#define LSR_PE  0x04
#define LSR_FE  0x08
#define LSR_BI  0x10
#define LSR_THE 0x20
#define LSR_TEMT 0x40
#define LSR_FIE 0x80

// Modem Control Register
#define MCR_DTR  0x01
#define MCR_RTS  0x02
#define MCR_INT  0x08		// Enable interrupts
#define MCR_LOOP 0x10		// Loopback mode

// Interrupt status Register
#define ISR_MS        0x00
#define ISR_nIP       0x01
#define ISR_Tx        0x02
#define ISR_Rx        0x04
#define ISR_LS        0x06
#define ISR_RxTO      0x0C
#define ISR_64BFIFO   0x20
#define ISR_FIFOworks 0x40
#define ISR_FIFOen    0x80

// Modem Status Register
#define MSR_DCTS 0x01
#define MSR_DDSR 0x02
#define MSR_TERI 0x04
#define MSR_DDCD 0x08
#define MSR_CTS  0x10
#define MSR_DSR  0x20
#define MSR_RI   0x40
#define MSR_CD   0x80

// FIFO Control Register
#define FCR_FE   0x01		// FIFO enable
#define FCR_CRF  0x02		// Clear receive FIFO
#define FCR_CTF  0x04		// Clear transmit FIFO
#define FCR_DMA  0x08		// DMA mode select
#define FCR_F64  0x20		// Enable 64 byte fifo (16750+)
#define FCR_RT14 0xC0		// Set Rx trigger at 14
#define FCR_RT8  0x80		// Set Rx trigger at 8
#define FCR_RT4  0x40		// Set Rx trigger at 4
#define FCR_RT1  0x00		// Set Rx trigger at 1

static unsigned char select_word_length[] = 
{ 
  LCR_WL5,	// 5 bits / word (char)
  LCR_WL6, LCR_WL7, LCR_WL8
};

static unsigned char select_stop_bits[] = 
{ 
  0, 
  LCR_SB1,	// 1 stop bit
  LCR_SB1_5,			// 1.5 stop bit
  LCR_SB2			// 2 stop bits
};

static unsigned char select_parity[] = 
{ 
  LCR_PN,	// No parity
  LCR_PE,			// Even parity
  LCR_PO,			// Odd parity
  LCR_PM,			// Mark parity
  LCR_PS,			// Space parity
};

static unsigned baudDivisor(baudRate)
{
  return 44236800L / (16*baudRate);
}

#define NumElements(a)  ((sizeof (a))/(sizeof (a)[0]))

void diag_pc_serial_puts(char *p);
void pc_serial_dump_regs(void);

unsigned long select_baud[] = 
{
  0,
  50,
  75,
  110,
  134,
  150,
  200,
  300,
  600,
  1200,
  1800,
  2400,
  3600,
  4800,
  7200,
  9600,
  14400,
  19200,
  38400,
  57600,
  115200,
  230400,
  460800
};

typedef struct pc_serial_info
{
  unsigned char *base;
  bool initOK;
  int int_num;
  bool fakeXoff;
  cyg_serial_error_count_t error_count;
  cyg_interrupt serial_interrupt;
  cyg_handle_t serial_interrupt_handle;
  enum {sNone=0, s8250, s16450, s16550, s16550a} deviceType;
} pc_serial_info;


static struct
  {
    cyg_mutex_t lock;
    cyg_cond_t  wait; 
  }sync;

static bool pc_serial_init(struct cyg_devtab_entry *tab);
static bool pc_serial_putc(serial_channel * chan, unsigned char c);
static Cyg_ErrNo pc_serial_lookup(struct cyg_devtab_entry **tab, struct cyg_devtab_entry *sub_tab, const char *name);
static unsigned char pc_serial_getc(serial_channel * chan);
bool pc_serial_set_config(serial_channel *chan, cyg_serial_info_t *config);
static Cyg_ErrNo pc_serial_set_config_handler(serial_channel * chan, cyg_uint32 key, const void *xbuf, cyg_uint32 * len);
static Cyg_ErrNo pc_serial_get_config_handler(serial_channel * chan, cyg_uint32 key, void *xbuf, cyg_uint32 * len);
static int pc_serial_get_rx_count(serial_channel * chan);
static int pc_serial_get_tx_count(serial_channel * chan);
static void pc_serial_flush_input(serial_channel * chan);
static void pc_serial_flush_output(serial_channel * chan);
static void pc_serial_start_xmit(serial_channel * chan);
static void pc_serial_stop_xmit(serial_channel * chan);
static void pc_serial_start_recv(serial_channel * chan);
static void pc_serial_stop_recv(serial_channel * chan);
static cyg_uint32 pc_serial_ISR(cyg_vector_t vector, cyg_addrword_t data);
static void pc_serial_DSR(cyg_vector_t vector, cyg_ucount32 count, cyg_addrword_t data);

serial_funs pc_serial_funs =
{
    putc:                pc_serial_putc,
    getc:                pc_serial_getc, 
    set_config:          pc_serial_set_config, 
    start_xmit:          pc_serial_start_xmit, 
    stop_xmit:           pc_serial_stop_xmit,
    get_config_handler:  pc_serial_get_config_handler,
    set_config_handler:  pc_serial_set_config_handler,
    start_recv:          pc_serial_start_recv,
    stop_recv:           pc_serial_stop_recv,
    flush_input:         pc_serial_flush_input,
    flush_output:        pc_serial_flush_output,
    get_tx_cnt:          pc_serial_get_tx_count,
    get_rx_cnt:          pc_serial_get_rx_count
};

#define CYG_IO_SERIAL_GENERIC_16X5X_INT_PRIORITY 4

// Internal function to actually configure the hardware to desired
// baud rate, etc.
static bool pc_serial_config_port(serial_channel * chan, cyg_serial_info_t * new_config, bool init)
{
  pc_serial_info *ser_chan = (pc_serial_info *) chan->dev_priv;
  unsigned char *base = ser_chan->base;
  unsigned short baud_divisor;
  unsigned baud;
  unsigned char _lcr, _ier;
  unsigned char _fcr_thresh;
  
  if (new_config->baud < NumElements(select_baud))
    baud = select_baud[new_config->baud];
  else
    baud = new_config->baud;

  baud_divisor = baudDivisor(baud);
  
  // Disable port interrupts while changing hardware
  HAL_READ_UINT8(base + REG_ier, _ier);
  HAL_WRITE_UINT8(base + REG_ier, 0);
  _lcr = select_word_length[new_config->word_length - CYGNUM_SERIAL_WORD_LENGTH_5] 
         | select_stop_bits[new_config->stop] 
         | select_parity[new_config->parity];
  HAL_WRITE_UINT8(base + REG_lcr, _lcr | LCR_DL);
  HAL_WRITE_UINT8(base + REG_mdl, baud_divisor >> 8);
  HAL_WRITE_UINT8(base + REG_ldl, baud_divisor & 0xFF);
  HAL_WRITE_UINT8(base + REG_lcr, _lcr);

  _fcr_thresh = FCR_RT8 |  FCR_FE | FCR_CRF | FCR_CTF;
  HAL_WRITE_UINT8(base + REG_fcr, _fcr_thresh);	// Enable and clear FIFO

  if (chan->out_cbuf.len != 0)
    _ier = IER_RCV;
  else
    _ier = 0;
  
  // Master interrupt enable
  HAL_WRITE_UINT8(base + REG_mcr, MCR_INT | MCR_DTR | MCR_RTS);

  _ier |= (IER_LS | IER_MS);

  HAL_WRITE_UINT8(base + REG_ier, _ier);
  
  if (new_config != &chan->config)
      chan->config = *new_config;

  return true;
}


// Function to initialize the device.  Called at bootstrap time.
static bool pc_serial_init(struct cyg_devtab_entry *tab)
{
  serial_channel *chan = (serial_channel *) tab->priv;
  pc_serial_info *ser_chan = (pc_serial_info *) chan->dev_priv;
  diag_printf("16x5x serial init - chan=%p  ser_chan=%p\n",chan,ser_chan);
  diag_printf("16x5x serial init - dev: %x.%d\n", ser_chan->base, ser_chan->int_num);
  
  cyg_mutex_init(&sync.lock);
  cyg_cond_init(&sync.wait,&sync.lock);
  
  // try to see if UART is there...
  if (1)
    {
      cyg_uint8 lcr, *base = ser_chan->base;
      diag_printf("16x5x serial init - checking UART at %p\n",base);
      #define HWriteReg8Value(addr,val) \
         do { \
               diag_printf("  %p <= %02x\n",addr,val); \
               HAL_WRITE_UINT8(addr,val); \
            } while (0)
      #define VerifyReg8Value(addr,val) \
         do { \
               cyg_uint8 b; \
               HAL_READ_UINT8(addr,b); \
               diag_printf("  %p => %02x, expecting %02x\n",addr,b,val); \
               if (b != (val)) \
                  goto fail; \
            } while(0)
      HWriteReg8Value(base+REG_scr, 0x55);
      VerifyReg8Value(base+REG_scr, 0x55);
      HWriteReg8Value(base+REG_scr, 0xaa);
      VerifyReg8Value(base+REG_scr, 0xaa);
      HAL_READ_UINT8(base+REG_lcr, lcr);
      HWriteReg8Value(base+REG_lcr, lcr &= ~0x80);
      HWriteReg8Value(base+REG_ier, 0);
      VerifyReg8Value(base+REG_ier, 0);
      HWriteReg8Value(base+REG_lcr, lcr |= 0x80);
      HWriteReg8Value(base+REG_ldl, 0x12);
      HWriteReg8Value(base+REG_mdl, 0x34);
      VerifyReg8Value(base+REG_ldl, 0x12);
      VerifyReg8Value(base+REG_mdl, 0x34);
      HWriteReg8Value(base+REG_lcr, lcr &= ~0x80);
      VerifyReg8Value(base+REG_ier, 0);
      HWriteReg8Value(base+REG_lcr, lcr |= 0x80);
      VerifyReg8Value(base+REG_ldl, 0x12);
      VerifyReg8Value(base+REG_mdl, 0x34);
      VerifyReg8Value(base+REG_scr, 0xaa);
      HWriteReg8Value(base+REG_lcr, lcr &= ~0x80);
      VerifyReg8Value(base+REG_scr, 0xaa);
      diag_printf("  UART OK\n");
    }
else
    {
    fail:
      diag_printf("  **** UART not detected -- driver not initialized ****\n");
      return false;
    }
  
  (chan->callbacks->serial_init) (chan);
  cyg_drv_interrupt_create(ser_chan->int_num, CYG_IO_SERIAL_GENERIC_16X5X_INT_PRIORITY, (cyg_addrword_t) chan, pc_serial_ISR, pc_serial_DSR, &ser_chan->serial_interrupt_handle, &ser_chan->serial_interrupt);
  cyg_drv_interrupt_attach(ser_chan->serial_interrupt_handle);
  cyg_drv_interrupt_unmask(ser_chan->int_num);
  pc_serial_config_port(chan, &chan->config, true);
  ser_chan->initOK = true;
  return true;
}


// This routine is called when the device is "looked" up (i.e. attached)
static Cyg_ErrNo pc_serial_lookup(struct cyg_devtab_entry **tab, struct cyg_devtab_entry *sub_tab, const char *name)
{
  serial_channel *chan = (serial_channel *) (*tab)->priv;
  pc_serial_info *ser_chan = (pc_serial_info *) chan->dev_priv;

  if (!ser_chan->initOK)
    return -ENODEV;

  (chan->callbacks->serial_init) (chan);
  return ENOERR;
}


// Send a character to the device output buffer.
// Return 'true' if character is sent to device
static bool pc_serial_putc(serial_channel * chan, unsigned char c)
{
  cyg_uint8 _lsr;
  pc_serial_info *ser_chan = (pc_serial_info *) chan->dev_priv;
  unsigned char *base = ser_chan->base;
  HAL_READ_UINT8(base + REG_lsr, _lsr);
  if (_lsr & LSR_THE)
    {
      // Transmit buffer is empty
      HAL_WRITE_UINT8(base + REG_thr, c);
      return true;
    }

  // No space
  return false;
}



// Fetch a character from the device input buffer, waiting if necessary
static unsigned char pc_serial_getc(serial_channel * chan)
{
  unsigned char c;
  cyg_uint8 _lsr;
  pc_serial_info *ser_chan = (pc_serial_info *) chan->dev_priv;
  unsigned char *base = ser_chan->base;

  // Wait for char
  do
    {
      HAL_READ_UINT8(base + REG_lsr, _lsr);
    }
  while ((_lsr & LSR_RSR) == 0);
  HAL_READ_UINT8(base + REG_rhr, c);
  return c;
}

// Set up the device characteristics; baud rate, etc.
bool pc_serial_set_config(serial_channel *chan, cyg_serial_info_t *config)
{
  return pc_serial_config_port(chan, config, false);
}

static Cyg_ErrNo pc_serial_set_config_handler(serial_channel * chan, cyg_uint32 key, const void *xbuf, cyg_uint32 * len)
{
  pc_serial_info *ser_chan = (pc_serial_info *) chan->dev_priv;
  unsigned char *base = ser_chan->base;
  Cyg_ErrNo res;

  switch (key)
    {
     case CYG_IO_SET_CONFIG_SERIAL_MODEM:
      if (*len != sizeof (cyg_uint32))
        res = -EINVAL;
      else
        {
          cyg_uint8 mcr;
          HAL_READ_UINT8(base+REG_mcr,mcr);
          
          if (*(cyg_uint32*)xbuf & CYG_SERIAL_MODEM_DTR) 
            HAL_WRITE_UINT8(base+REG_mcr, mcr |= MCR_DTR);
          else
            HAL_WRITE_UINT8(base+REG_mcr, mcr &= ~MCR_DTR);
          
          // Ignore RTS flag if it's being controlled automatically
          if (!(chan->config.flags & (CYG_SERIAL_FLAGS_RTSCTS |
                                      CYG_SERIAL_FLAGS_RTS_RXFLOW |
                                      CYG_SERIAL_FLAGS_RTS_TOGGLE)))
            {
              if (*(cyg_uint32*)xbuf & CYG_SERIAL_MODEM_RTS) 
                HAL_WRITE_UINT8(base+REG_mcr, mcr |= MCR_RTS);
              else
                HAL_WRITE_UINT8(base+REG_mcr, mcr &= ~MCR_RTS);
            }
        }
      break;

     case CYG_IO_SET_CONFIG_SERIAL_SCAN_PERIOD:
      // TBD
      break;
      
    default:
      return -EINVAL;
    }
  return ENOERR;
}

// Internal function to handle get_config keys not implimented by higher layers
Cyg_ErrNo pc_serial_get_config_handler(serial_channel *chan, 
                                       cyg_uint32 key, 
                                       void *xbuf, 
                                       cyg_uint32 *len)
{
  pc_serial_info *ser_chan = (pc_serial_info *) chan->dev_priv;
  unsigned char *base = ser_chan->base;
  Cyg_ErrNo res = ENOERR;

  //diag_printf("16x5x get_config chan=%p key=%x xbuf=%p, len=%d\n",chan,key,xbuf,*len);
  
  switch (key)
    {
     case CYG_IO_GET_CONFIG_SERIAL_MODEM:
      if (*len != sizeof (cyg_uint32))
        res = -EINVAL;
      else
        {
          cyg_uint8 msr,mcr;
          HAL_READ_UINT8(base+REG_msr,msr);
          HAL_READ_UINT8(base+REG_mcr,mcr);          
          *(cyg_uint32*)xbuf = 0;
          
          if (mcr & MCR_DTR)
            *(cyg_uint32*)xbuf |= CYG_SERIAL_MODEM_DTR;
          if (mcr & MCR_RTS)
            *(cyg_uint32*)xbuf |= CYG_SERIAL_MODEM_RTS;
          
          if (msr & MSR_DSR)
            *(cyg_uint32*)xbuf |= CYG_SERIAL_MODEM_DSR;
          if (msr & MSR_CTS)
            *(cyg_uint32*)xbuf |= CYG_SERIAL_MODEM_CTS;
          if (msr & MSR_CD)
            *(cyg_uint32*)xbuf |= CYG_SERIAL_MODEM_CD;
          if (msr & MSR_RI)
            *(cyg_uint32*)xbuf |= CYG_SERIAL_MODEM_RI;

          if (msr & MSR_DDSR)
            *(cyg_uint32*)xbuf |= CYG_SERIAL_MODEM_DSR;
          if (msr & MSR_DCTS)
            *(cyg_uint32*)xbuf |= CYG_SERIAL_MODEM_CTS;
          if (msr & MSR_DDCD)
            *(cyg_uint32*)xbuf |= CYG_SERIAL_MODEM_CD;
          if (msr & MSR_TERI)
            *(cyg_uint32*)xbuf |= CYG_SERIAL_MODEM_RI;
          
          
	  //  Set the flag indicating our transmitter is flowed off.  
	  //  First hardware flow control.  If CTS cleared, must be 
	  //  flowed off.
          if (0 /*TBD*/)
            *(cyg_uint32*)xbuf |= CYG_SERIAL_MODEM_FLOWTXOFF;
          else if ((chan->config.flags & CYG_SERIAL_FLAGS_XONXOFF_TXFLOW)) 
            {
              //  Next software flow control.  If XON/XOFF is active, check
              //  if SET_TXOFF is active.  If not, check if automatic SW flow
              //  control in the AIOPIC has our transmitter flowed off.  There
              //  is no status flag in the AIOPIC for this, so we test is by 
              //  checking (twice) that the TX FIFO is not empty but the shift 
              //  register is empty.  This is not perfect, but is OK, since the 
              //  Rocketport does it this way...
	      if (ser_chan->fakeXoff) 
                *(cyg_uint32*)xbuf |= CYG_SERIAL_MODEM_FLOWTXOFF;
	      else if (0 /*TBD*/)
                *(cyg_uint32*)xbuf |= CYG_SERIAL_MODEM_FLOWTXOFF;
            }
        }
      break;
      
     case CYG_IO_GET_CONFIG_SERIAL_ERRORS:
      if (*len < sizeof ser_chan->error_count)
        res = -EINVAL;
      else
        *((cyg_serial_error_count_t*)(xbuf)) = ser_chan->error_count;
      break;
      
     case CYG_IO_GET_CONFIG_SERIAL_CLRERRS:
      memset(&ser_chan->error_count, 0, sizeof ser_chan->error_count);
      break;
      
     case CYG_IO_GET_CONFIG_SERIAL_BREAKON:
      //TBD
      break;
      
     case CYG_IO_GET_CONFIG_SERIAL_BREAKOFF:
      //TBD
      break;

     case CYG_IO_SET_CONFIG_TX_OFF:
      if (!(chan->config.flags & CYG_SERIAL_FLAGS_XONXOFF_TXFLOW))
        break;
      pc_serial_stop_xmit(chan);
      ser_chan->fakeXoff = true;
      //TBD: need to figure out how to restart when XON is received
      break;
      
     case CYG_IO_GET_CONFIG_TX_OVERRIDE:
      if(ser_chan->fakeXoff) 
        {
          pc_serial_start_xmit(chan);
          ser_chan->fakeXoff = false;
        }
      break;
      
     default:
      res = -EINVAL;
    }
  
  return res;
}

// Enable the transmitter on the device
static void pc_serial_start_xmit(serial_channel * chan)
{
  pc_serial_info *ser_chan = (pc_serial_info *) chan->dev_priv;
  unsigned char *base = ser_chan->base;
  cyg_uint8 _ier;
  HAL_READ_UINT8(base + REG_ier, _ier);
  _ier |= IER_XMT;		// Enable xmit interrupt
  HAL_WRITE_UINT8(base + REG_ier, _ier);
}

// Disable the transmitter on the device
static void pc_serial_stop_xmit(serial_channel * chan)
{
  pc_serial_info *ser_chan = (pc_serial_info *) chan->dev_priv;
  unsigned char *base = ser_chan->base;
  cyg_uint8 _ier;
  HAL_READ_UINT8(base + REG_ier, _ier);
  _ier &= ~IER_XMT;		// Disable xmit interrupt
  HAL_WRITE_UINT8(base + REG_ier, _ier);
}

// Enable the transfer of bytes to next layer up
void pc_serial_start_recv(serial_channel *chan)
{
  pc_serial_info *ser_chan = (pc_serial_info *) chan->dev_priv;
  unsigned char *base = ser_chan->base;
  cyg_uint8 _ier;
  HAL_READ_UINT8(base + REG_ier, _ier);
  _ier |= IER_RCV;		// Enable recv interrupt
  HAL_WRITE_UINT8(base + REG_ier, _ier);
}

// Get tx FIFO count
int pc_serial_get_tx_count(serial_channel *chan)
{
  pc_serial_info *ser_chan = (pc_serial_info *) chan->dev_priv;
  unsigned char *base = ser_chan->base;
  (void)base;
  //TBD
  return 0;
}

// Get rx FIFO count
int pc_serial_get_rx_count(serial_channel *chan)
{
  pc_serial_info *ser_chan = (pc_serial_info *) chan->dev_priv;
  unsigned char *base = ser_chan->base;
  (void)base;
  //TBD
  return 0;
}

// Flush input data
void pc_serial_flush_input(serial_channel *chan)
{
  pc_serial_info *ser_chan = (pc_serial_info *) chan->dev_priv;
  unsigned char *base = ser_chan->base;
  (void)base;
  //TBD
}
  
// Flush output data
void pc_serial_flush_output(serial_channel *chan)
{
  pc_serial_info *ser_chan = (pc_serial_info *) chan->dev_priv;
  unsigned char *base = ser_chan->base;
  (void)base;
  //TBD
}
  
// Disable the transfer of bytes to next layer up -- we need to disable
// the receive interrupt so that we stop reading bytes from the Rx FIFO
// so that the UART will flow controll the other end.
void pc_serial_stop_recv(serial_channel *chan)
{
  pc_serial_info *ser_chan = (pc_serial_info *) chan->dev_priv;
  unsigned char *base = ser_chan->base;
  cyg_uint8 _ier;
  HAL_READ_UINT8(base + REG_ier, _ier);
  _ier &= ~IER_RCV;		// Disable recv interrupt
  HAL_WRITE_UINT8(base + REG_ier, _ier);
}

static unsigned isrCount, dsrCount;

// Serial I/O - low level interrupt handler (ISR)
static cyg_uint32 pc_serial_ISR(cyg_vector_t vector, cyg_addrword_t data)
{
  serial_channel *chan = (serial_channel *) data;
  pc_serial_info *ser_chan = (pc_serial_info *) chan->dev_priv;
  ++isrCount;
  cyg_drv_interrupt_mask(ser_chan->int_num);
  cyg_drv_interrupt_acknowledge(ser_chan->int_num);
  return CYG_ISR_CALL_DSR;	// Cause DSR to be run
}


// Serial I/O - high level interrupt handler (DSR)
static void pc_serial_DSR(cyg_vector_t vector, cyg_ucount32 count, cyg_addrword_t data)
{
  serial_channel *chan = (serial_channel *) data;
  pc_serial_info *ser_chan = (pc_serial_info *) chan->dev_priv;
  unsigned char *base = ser_chan->base;
  cyg_uint8 _isr;

  ++dsrCount;
  
  // Check if we have an interrupt pending - note that the interrupt
  // is pending of the low bit of the isr is *0*, not 1.
  HAL_READ_UINT8(base + REG_isr, _isr);
  
  while ((_isr & ISR_nIP) == 0)
    {
      switch (_isr & 0xE)
	{
	case ISR_Rx:
	case ISR_RxTO:

	  {
	    cyg_uint8 _lsr;
	    unsigned char c;
	    HAL_READ_UINT8(base + REG_lsr, _lsr);
	    while (_lsr & LSR_RSR)
	      {
		HAL_READ_UINT8(base + REG_rhr, c);
		(chan->callbacks->rcv_char) (chan, c);
		HAL_READ_UINT8(base + REG_lsr, _lsr);
	      }
	    break;
	  }
          
	case ISR_Tx:
	  (chan->callbacks->xmt_char)(chan);
	  break;

	case ISR_LS:
	  {
	    cyg_uint8 _lsr;
	    HAL_READ_UINT8(base + REG_lsr, _lsr);
	  }
	  break;
          
	case ISR_MS:
	  {
	    cyg_uint8 _msr;
	    HAL_READ_UINT8(base + REG_msr, _msr);
	  }
	  break;

	default:

	  // Yes, this assertion may well not be visible. *But*
	  // if debugging, we may still successfully hit a breakpoint
	  // on cyg_assert_fail, which _is_ useful
	  CYG_FAIL("unhandled serial interrupt state");
	}
      HAL_READ_UINT8(base + REG_isr, _isr);
    }
  cyg_drv_cond_broadcast(&sync.wait);
  cyg_drv_interrupt_unmask(ser_chan->int_num);
}



// define serial devices

pc_serial_info pc_serial_priv_info[1] =
{
  {
    .base = (unsigned char*)0x4000000,
    .int_num = 7
  }
};
  
static unsigned char out_cbuf_data0[4096];
static unsigned char  in_cbuf_data0[4096];

#define ChanInfo(n) \
    {  \
      .funs =         &pc_serial_funs,  \
      .callbacks =    &cyg_io_serial_callbacks,  \
      .dev_priv=      pc_serial_priv_info+n,  \
      .config.baud=   9600,  \
      .config.stop=   1,  \
      .config.parity= 0,  \
      .config.word_length= 8,  \
      .config.flags= 0,  \
      .config.interface_mode= 0,  \
      .config.rxflow_xon_char= 0x11,  \
      .config.rxflow_xoff_char= 0x13,  \
      .config.txflow_xon_char= 0x11,  \
      .config.txflow_xoff_char= 0x13,  \
      .init= false,  \
      .sync.lock= &sync.lock,  \
      .sync.wait= &sync.wait,  \
      .out_cbuf= CBUF_INIT(out_cbuf_data##n, sizeof out_cbuf_data##n), \
      .in_cbuf= CBUF_INIT(in_cbuf_data##n, sizeof in_cbuf_data##n) \
    }

struct serial_channel serialChannel[1] =
{
  ChanInfo(0)
};
  
DEVTAB_ENTRY(pc_serial_io_0, "/dev/ser0", 0, &cyg_io_serial_devio, pc_serial_init, pc_serial_lookup, &serialChannel[0]);

void diag_pc_serial_puts(char *p)
{
  while (*p)
    if (pc_serial_putc(serialChannel+0, *p))
      ++p;
}

unsigned char diag_pc_serial_getc(void)
{
  return pc_serial_getc(serialChannel+0);
}

void pc_serial_dump_regs(void)
{
  unsigned i;
  pc_serial_info *ser_chan = (pc_serial_info *) serialChannel[0].dev_priv;
  unsigned char *base = ser_chan->base;
  for (i=0; i<=7; ++i)
    {
      cyg_uint8 b;
      HAL_READ_UINT8(base + i*4, b);
      diag_printf("%d: %02x\n",i,b);
    }
}
