/***************************************************************/
/*       aiopic.c                                              */
/*                                                             */
/*       Structs and functions for the serial driver for       */
/*       Comtrol DeviceMaster RTS devices that have AIOPICs.   */
/***************************************************************/


#include <pkgconf/system.h>
#include <pkgconf/io_nserial.h>
#include <pkgconf/io.h>
#include <cyg/io/io.h>
#include <cyg/hal/hal_intr.h>
#include <cyg/io/devtab.h>
#include <cyg/io/serial.h>
#include <cyg/io/serialio.h>
#include <cyg/infra/diag.h>
#include <stdlib.h>
#include <string.h>
#include "dmrts.h"
#include "aiopic.h"

// global variable indicating whether a unit is a DM500 or not
// defined in hal/arm/dmrts/current/src/hal_diag.c
extern int IsDM500;
static unsigned int AiopClock;
static unsigned long   modelId;

extern volatile unsigned short *RingIndicatorRegister;
extern cyg_counter dmrtsCounterObj;
extern cyg_handle_t dmrtsCounterHandle;

void getModelNumber(unsigned long *m);

#define EnableDebugIcTimeout 0

#if EnableDebugIcTimeout
static void clkticks2ms(cyg_tick_count_t ctc, cyg_resolution_t *ccr, unsigned long long *n);
static cyg_resolution_t mCCR;	//call cyg_clock_get_resolution(cyg_real_time_clock()) to use
static int debugICTimeout;
#endif

#if 0
#define debug_printf(args...) diag_printf(args)
#else
#define debug_printf(args...) /* noop */
#endif

unsigned short swapshort(unsigned short x)
{
    return x<<8 | x>>8;
}

#define NumElements(a) ((sizeof a)/(sizeof a[0]))

#define NumAiop 4
#define NumChan (NumAiop*8)

static int numChanPresent = 0;
static int scanPeriod = 10;

// this is a set of flags that indicate which ports have their baud
// rates configured to be 230K or faster.  When running at "fast" baud
// rates, the rx/tx copy loops need to have nops in them to avoid
// taking up so much AIOPIC SRAM bandwidth that we end up starving the
// AIOPIC and causing overruns of the rx holding regester.
static unsigned fastBaudFlags;

static int UNIT_HAS_422_485 =1;		//gets set to 0 for SH 8 and 16

typedef struct {
    tChan          chan;
    cyg_serial_error_count_t error_count;
    bool           receive_stopped;
    bool           sentXoff;
    unsigned char  *in_buf;
    unsigned char  *out_buf;

    unsigned       interCharTimeout;
    unsigned       idleCount;
    bool           timeoutFlag;
    bool           gotData;
    unsigned       rxFifoLast;
    bool           fakeXoff;				//used with 'set xoff'
    bool           softRtsToggle;
    unsigned       preRtsHoldTime;
    unsigned       postRtsHoldTime;
    unsigned       rtsCounter;
} tAiopicChanInfo;

static cyg_uint32 aiopic_serial_aiopic_ISR(cyg_vector_t vector, cyg_addrword_t data);
static void       aiopic_serial_aiopic_DSR(cyg_vector_t vector, cyg_ucount32 count, cyg_addrword_t data);

extern tAiopic _aiop0, _aiop1, _aiop2, _aiop3;
static tAiopic *aiop[NumAiop] = {&_aiop0, &_aiop1, &_aiop2, &_aiop3};
static tAiopicChanInfo aiopChanInfo[NumChan];

#define ShowAiopicOffset(field) \
  diag_printf("%20s: %u\n", #field, ((char*)&_aiop0.field)-((char*)&_aiop0))

/*  Alarm variables */
static cyg_handle_t aiopicAlarmHandle;
static cyg_alarm aiopicAlarmObj;
static cyg_alarm_t aiopic_serial_poll;
static unsigned long alarmIndex = 0;

static struct {
    cyg_mutex_t lock;
    cyg_cond_t  wait;
} sync;

static tRegVal RData[] = {
    {0x0009, 0xf682},
    {0x0209, 0x86fb},
    {0x0409, 0x000a},
    {0x0609, 0x010a},
    {0x0809, 0x8a13},
    {0x0a09, 0xc511},
    {0x0c09, 0x8685},
    {0x0e09, 0x200a},
    {0x1009, 0x210a},
    {0x1209, 0x41ff},
    {0x1409, 0x8200},
    {0x1609, 0x827b},
    {0x1809, 0x8a7d},
    {0x1a09, 0x8881},
    {0x1c09, 0x867a},
    {0x1e09, 0x8481},
    {0x2009, 0x827c},
    {0x2209, 0x0a0a},
};

static tRegVal R[] = {
    {0x0009, 0xf682},             /* 00: Stop Rx processor */
    {0x0809, 0x8a13},             /* 04: Tx software flow control */
    {0x0a09, 0xc511},             /* 08: XON char */
    {0x0c09, 0x8685},             /* 0c: XANY */
    {0x1209, 0x41ff},             /* 10: Rx mask char */
    {0x1409, 0x8200},             /* 14: Compare/Ignore #0 */
    {0x1609, 0x827b},             /* 18: Compare #1 */
    {0x1809, 0x8a7d},             /* 1c: Compare #2 */
    {0x1a09, 0x8881},             /* 20: Interrupt #1 */
    {0x1c09, 0x867a},             /* 24: Ignore/Replace #1 */
    {0x1e09, 0x8481},             /* 28: Interrupt #2 */
    {0x2009, 0x827c},             /* 2c: Ignore/Replace #2 */
    {0x2209, 0x0a0a}              /* 30: Rx FIFO Enable */
};

void initAiopChan(tAiopic *a, int chanNum, tChan *ch, int lineNum)
{
    int i;
    unsigned short base = chanNum*0x10;

    // initialize channel

    //diag_printf("initAiopChan %08x:%d\n",a,chanNum);

    ch->base = base;
    ch->line = lineNum;

    for (i=0; i<NumElements(RData); ++i)
        AiopPutIndexedShort(a,RData[i].offset+base,RData[i].value);

    for (i=0; i<NumElements(ch->R); ++i) {
        ch->R[i].offset = R[i].offset+base;
        ch->R[i].value  = R[i].value;
    }

    ch->aiop = a;
    ch->chan = chanNum;

    ch->dataWord = &a->data[chanNum];
    ch->dataByte = (unsigned char*)&a->dataLow[chanNum];
    ch->rxFifoCnt = &a->rxFifoCnt[chanNum];
    ch->txFifoCnt = (unsigned char*)&a->txFifoCnt[chanNum];
    if (chanNum & 1)
        ch->intID = (unsigned char*)&a->oddIntID[chanNum];
    else
        ch->intID = (unsigned char*)&a->evenIntID[chanNum];
    ch->status = &a->chanStatus[chanNum];
    ch->statusLow = (unsigned char*)&a->chanStatusLow;
    ch->statusHigh = ((unsigned char*)&a->chanStatusHigh)+1;

    ch->baudDivisor.offset = base+_BAUD;
    ch->txControl.offset   = base+_TX_CTRL;
    ch->rxControl.offset   = base+_RX_CTRL;
    ch->txEnables.offset   = base+_TX_ENBLS;
    ch->txCompare.offset   = base+_TXCMP1;
    ch->txReplace1.offset  = base+_TXREP1B1;
    ch->txReplace2.offset  = base+_TXREP2;
    ch->txPriorityCntIndex = base+_TXP_CNT;
    ch->txPriorityPtrIndex = base+_TXP_PNTR;
    ch->txPriorityBufIndex = base+_TXP_BUF;

    PutChanReg(ch,baudDivisor,swapshort(AiopBaudDivisor(9600)));
    PutChanReg(ch,txControl,0);
    PutChanReg(ch,rxControl,0);
    PutChanReg(ch,txEnables,0);
    PutChanReg(ch,txCompare,0);
    PutChanReg(ch,txReplace1,0);
    PutChanReg(ch,txReplace2,0);


    // reset UART
    a->command = AiopCommand_ResetUart | chanNum;
    a->command = chanNum;

    // reset tx fifio logic
    a->command = AiopCommand_ResetTxFifo | chanNum;
    a->command = chanNum;

    // set tx fifo pointers to 0
    AiopPutIndexedShort(a,base+_TXF_OUTP,0);

    // reset rx fifo logic
    a->command = AiopCommand_ResetRxFifo | chanNum;
    a->command = chanNum;

    // set rx fifo pointers to 0
    AiopPutIndexedShort(a,base+_RXF_OUTP,0);
    AiopPutIndexedShort(a,base+_RXF_INP,0);

    // clear tx priority counter and pointer
    AiopPutIndexedByte(a,base+_TXP_CNT,0);
    AiopPutIndexedByte(a,base+_TXP_PNTR,0);

    sEnRxProcessor(ch);
    sEnTransmit(ch);
    sEnRxFifo(ch);
    sSetData8(ch);
    sSetTxXONChar(ch,0x11);
    sSetTxXOFFChar(ch,0x13);
    sDisRxStatusMode(ch);
}


void initAiop(tAiopic *a)
{
    debug_printf("Resetting Aiopic\n");
    a->command = AiopCommand_ResetAll;
    diag_printf("\n");  // need 200ns delay
    a->command = 0;
    debug_printf("aiop %08x ID: %04x\n",a,a->chanStatus[0]);
    AiopPutIndexedByte(a,_CLK_PRE,AiopPrescale);
    a->intMask = 0xff;
}

static void initAiopicsAndChannels(void)
{
    int i;

    for (i=0; i<numChanPresent/8; ++i)
        initAiop(aiop[i]);

    for (i=0; i<numChanPresent; ++i) {
        tChan *ch = &aiopChanInfo[i].chan;
        initAiopChan(aiop[i/8],i%8,ch,i);
    }
}

void sWriteTxPrioByte(tChan *ch, unsigned char c)
{
    if (sGetTxCnt(ch) == 0)
        *(ch->dataByte) = c;
    else if (!(AiopGetIndexedByte(ch->aiop,ch->txPriorityCntIndex) & 0x80)) {
        AiopPutIndexedByte(ch->aiop,ch->txPriorityBufIndex,c);
        AiopPutIndexedByte(ch->aiop,ch->txPriorityPtrIndex,0);
        AiopPutIndexedByte(ch->aiop,ch->txPriorityCntIndex,0x81);
    } else
        diag_printf("%d:wtpb %02x fail\n",ch->chan,c);
}

void sDisRxFIFO(tChan *ch)
{
    ch->R[12].value = (ch->R[12].value & 0x00ff) | 0x0a00;
    AiopPutIndexedShort(ch->aiop,ch->R[12].offset,ch->R[12].value);
}

void sStopRxProcessor(tChan *ch)
{
    unsigned short v = (ch->R[0].value & 0x00ff) | 0x0a00;
    AiopPutIndexedShort(ch->aiop,ch->R[0].offset,v);
}

void sStartRxProcessor(tChan *ch)
{
    AiopPutIndexedShort(ch->aiop,ch->R[0].offset,ch->R[0].value);
}

#define TRACE_BUFFER_ENABLED 0

#if TRACE_BUFFER_ENABLED
#define TraceBufferSize  512
static unsigned traceBuffer[TraceBufferSize] = {0};
static unsigned traceBufferIndex = 0;
void putTrace(unsigned char port, unsigned char v, unsigned short p)
{
    traceBuffer[traceBufferIndex] = (port<<24) | (v<<16) | p;
    ++traceBufferIndex;
    traceBufferIndex &= (TraceBufferSize-1);
    traceBuffer[traceBufferIndex] = 0;
}
void aiopicDumpTrace(void)
{
    unsigned count,i;
    diag_printf("fastBaudFlags = %08x\n",fastBaudFlags);
    i = (traceBufferIndex+1)&(TraceBufferSize-1);
    diag_printf("Aiopic Trace:\n");
    for (count=0; count<TraceBufferSize; ++count) {
        unsigned u = traceBuffer[i];
        diag_printf("%3d: %02x %04x (%5u)\n", (u>>24), ((u>>16)&0xff), (u&0xffff), (u&0xffff));
        i = (i+1)&(TraceBufferSize-1);
    }
}
#else
#define putTrace(port,v,p) /* noop */
void aiopicDumpTrace(void)
{
    diag_printf("fastBaudFlags = %08x\n",fastBaudFlags);
    diag_printf("Aiopic Trace not enabled.\n");
}
#endif

#define TRACE_PINS_ENABLED 0
#if TRACE_PINS_ENABLED
extern volatile unsigned PioData;
#define TracePinOn(n)  PioData |= Bit(n)
#define TracePinOff(n)  PioData &= ~Bit(n)
#else
#define TracePinOn(n)  /* noop */
#define TracePinOff(n) /* noop */
#endif


#define min(a,b) ((a)<(b)?(a):(b))

#if EnableDebugIcTimeout
# define DebugIcTimeoutMarkTransmit()           \
  do{                                           \
    if (debugICTimeout)                         \
      {                                         \
        ctc = cyg_current_time();               \
        clkticks2ms(ctc, &mCCR, &ms);                                   \
        if ((aiopic_chan->interCharTimeout < (ms - lastStamp)) & (lastStamp != 0)) \
          diag_printf("aiopic-handleTxBytes%u: (A) dataByte @ clock %llu (%llums) (last stamp = %llu) [GAP = %llu]\n",\
                      ch->chan, (unsigned long long)ctc, ms, lastStamp, (ms - lastStamp)); \
        else if (((ms - lastStamp) != 0) && ((ms - lastStamp) > aiopic_chan->interCharTimeout)) \
          diag_printf("aiopic-handleTxBytes%u: (A) dataByte @ clock %llu (%llums) (last stamp = %llu) GAP = %llu\n",\
                      ch->chan, (unsigned long long)ctc, ms, lastStamp, (ms - lastStamp)); \
        lastStamp = ms;                                                 \
      }                                                                 \
  }while(0)
#else
# define DebugIcTimeoutMarkTransmit() /*nothing*/
#endif

// Enable nops that are inserted in rx/tx loops that access AIOPIC
#if 1
#define nops(n) asm(" .rept "#n"\n nop\n .endr\n")
#else
#define nops(n) /* nothing */
#endif

static void handleTransmitBytes(serial_channel *chan)
{
    tChan *ch = &(((tAiopicChanInfo*)(chan->dev_priv))->chan);
    cbuf_t *cbuf = &chan->out_cbuf;
    int txFifoSpace;
    int avail,count,size;
    unsigned char *src;
#if EnableDebugIcTimeout
    tAiopicChanInfo *aiopic_chan = (tAiopicChanInfo*)chan->dev_priv;
    cyg_tick_count_t ctc;						//for timestamp
    unsigned long long ms = 0;					//for timestamp
    static unsigned long long lastStamp = 0;	//for timestamp
#endif


    if (cbuf->get == cbuf->put) {
        if (cbuf->waiting && sGetTxCnt(ch)==0) {
            cbuf->waiting = false;
            cyg_drv_cond_broadcast(&cbuf->wait);
        }
        return;
    }

    //putTrace((((unsigned)(ch->aiop) & 0xff)|ch->chan),0x20,cyg_current_time()&0xffff);

    while (cbuf->get != cbuf->put) {
        //putTrace((((unsigned)(ch->aiop) & 0xff)|ch->chan),0x21,cbuf->put);
        //putTrace((((unsigned)(ch->aiop) & 0xff)|ch->chan),0x22,cbuf->get);

        if (cbuf->put > cbuf->get)
            avail = cbuf->put - cbuf->get;
        else
            avail = cbuf->len - cbuf->get;

        //putTrace((((unsigned)(ch->aiop) & 0xff)|ch->chan),0x23,avail);

        txFifoSpace = 254 - sGetTxCnt(ch);

        //putTrace((((unsigned)(ch->aiop) & 0xff)|ch->chan),0x24,txFifoSpace);

        if (txFifoSpace <= 0) break;

        size = count = min(txFifoSpace,avail);
        src = cbuf->data+cbuf->get;

        //putTrace((((unsigned)(ch->aiop) & 0xff)|ch->chan),0x25,count);
        //putTrace((((unsigned)(ch->aiop) & 0xff)|ch->chan),0x26,*src);

        // need hword aligned src pointer
        if (((int)src) & 1) {
            *(ch->dataByte) = *src++;
            --count;
            DebugIcTimeoutMarkTransmit();
        }

        // xfer 16 bits at a time
        while (count > 1) {
            *(ch->dataWord) = *((unsigned short*)src);
            src += 2;
            count -= 2;
            DebugIcTimeoutMarkTransmit();
            if (fastBaudFlags)
                nops(20);                  // don't hog AIOPIC SRAM bandwidth
        }

        // finish up last byte if there is one
        if (count) {
            *(ch->dataByte) = *src++;
            --count;
            DebugIcTimeoutMarkTransmit();
        }

        //putTrace((((unsigned)(ch->aiop) & 0xff)|ch->chan),0x27,src[-1]);

        cbuf->get += size;

        if (cbuf->get == cbuf->len) cbuf->get = 0;

        if (size != avail) break;
    }


    if (cbuf->get == cbuf->put) {
        (chan->funs->stop_xmit)(chan);  // Done with transmit
        //putTrace((((unsigned)(ch->aiop) & 0xff)|ch->chan),0x28,0);
    }

    if (cbuf->waiting) {
        // See if there is now enough room to restart writer
        int space = (cbuf->len + cbuf->get) - cbuf->put;
        if (space > cbuf->len) space -= cbuf->len;
        if (space >= cbuf->low_water) {
            cbuf->waiting = false;
            cyg_drv_cond_broadcast(&cbuf->wait);
        }
    }
}//end handleTransmitBytes()

#define FlowLowWater  64
#define FlowHighWater 512

#if EnableDebugIcTimeout
# define DebugIcTimeoutMarkReceive()                                    \
  do {                                                                  \
    if (debugICTimeout)                                                 \
      {                                                                 \
        ctc = cyg_current_time();                                       \
        clkticks2ms(ctc, &mCCR, &ms);                                   \
        if (0)                                                          \
          {                                                             \
            if ((aiopic_chan->interCharTimeout < (ms - lastStamp)) & (lastStamp != 0)) \
              diag_printf("aiopic-handleRxBytes%u: (A) rxByte @ clock %llu (%llums) (last stamp = %llu) [GAP = %llu]\n", \
                          ch->chan, (unsigned long long)ctc, ms, lastStamp, (ms - lastStamp)); \
            else if (((ms - lastStamp) != 0) && ((ms - lastStamp) > aiopic_chan->interCharTimeout)) \
              diag_printf("aiopic-handleRxBytes%u: (A) rxByte @ clock %llu (%llums) (last stamp = %llu) GAP = %llu\n", \
                          ch->chan, (unsigned long long)ctc, ms, lastStamp, (ms - lastStamp)); \
          }                                                             \
        else                                                            \
          diag_printf("aiopic-handleRxBytes%u: (A) rxByte @ clock %llu (%llums) (last stamp = %llu) GAP = %llu\n", \
                      ch->chan, (unsigned long long)ctc, ms, lastStamp, (ms - lastStamp)); \
        lastStamp = ms;                                                 \
      }                                                                 \
  }while(0)
#else
# define DebugIcTimeoutMarkReceive()   /*nothing*/
#endif

static void handleReceiveBytes(serial_channel *chan)
{
    tAiopicChanInfo *aiopic_chan = (tAiopicChanInfo*)chan->dev_priv;
    tChan *ch = &aiopic_chan->chan;
    cbuf_t *cbuf = &chan->in_cbuf;
    int nextPut, rxCount;
    unsigned short chanStatus, statusAndRxByte;
    unsigned char *dst, rxByte;
    int avail,count;
    int wakeup = 0;
#if EnableDebugIcTimeout
    cyg_tick_count_t ctc;				     //for timestamp
    unsigned long long ms = 0;			     //for timestamp
    static unsigned long long lastStamp = 0;	     //for timestamp
#endif

    // is data being handled by callback instead of "normal" method?
    if (chan->config.rx_callback) return;

    rxCount = sGetRxCnt(ch);

    if (aiopic_chan->timeoutFlag) {
        chan->rxInterCharTimeoutFlag = 1;
        aiopic_chan->timeoutFlag = 0;
        wakeup = 1;
    }

    // send Xoff if Rx FIFO is getting full

    if ((chan->config.flags & CYG_SERIAL_FLAGS_XONXOFF_RXFLOW)
            && !aiopic_chan->sentXoff && (rxCount > FlowHighWater)) {
        //putTrace((((unsigned)(ch->aiop) & 0xff)|ch->chan),0x38,rxCount);
        aiopic_chan->sentXoff = 1;
        sWriteTxPrioByte(ch,chan->config.rxflow_xoff_char);
        //putTrace((((unsigned)(ch->aiop) & 0xff)|ch->chan),0x3b,chan->config.rxflow_xoff_char);
    }

    //putTrace((((unsigned)(ch->aiop) & 0xff)|ch->chan),0x30,cyg_current_time()&0xffff);
    //putTrace((((unsigned)(ch->aiop) & 0xff)|ch->chan),0x31,cbuf->put);
    //putTrace((((unsigned)(ch->aiop) & 0xff)|ch->chan),0x37,cbuf->get);

again:

    // find available contiguous space

    if (cbuf->get == 0)
        avail = (cbuf->len - cbuf->put) - 1;
    else if (cbuf->get > cbuf->put)
        avail = (cbuf->get - cbuf->put) - 1;
    else
        avail = (cbuf->len - cbuf->put);

    //putTrace((((unsigned)(ch->aiop) & 0xff)|ch->chan),0x39,avail);

    if (avail == 0) { //full
        putTrace((((unsigned)(ch->aiop) & 0xff)|ch->chan),0x33,0);
#if 0
        if (chan->funs->stop_recv) {
            cbuf->rxstopped = true;
            chan->funs->stop_recv(chan);
        }
#endif
    }

    count = min(rxCount,avail);
    //putTrace((((unsigned)(ch->aiop) & 0xff)|ch->chan),0x3a,count);

    if (count) {
        wakeup = 1;

        putTrace((((unsigned)(ch->aiop) & 0xff)|ch->chan),0x30,cyg_current_time()&0xffff);
        putTrace((((unsigned)(ch->aiop) & 0xff)|ch->chan),0x3a,count);

        nextPut = cbuf->put + count;
        if (nextPut == cbuf->len) nextPut = 0;

        dst = cbuf->data+cbuf->put;

        chanStatus = *(ch->status);

        putTrace((((unsigned)(ch->aiop) & 0xff)|ch->chan),0x3b,chanStatus);

        if ((chanStatus & AiopChanStatus_RxErr) &&
                !(chanStatus & AiopChanStatus_StatusModeEnable)) {
            sEnRxStatusMode(ch);
            chanStatus |= AiopChanStatus_StatusModeEnable;
        }

        if (chanStatus & AiopChanStatus_StatusModeEnable) {
            while (count) {
                --count;
                statusAndRxByte = *(ch->dataWord);
                putTrace((((unsigned)(ch->aiop) & 0xff)|ch->chan),0x3c,statusAndRxByte);

                rxByte = statusAndRxByte >> 8;
                if (statusAndRxByte & (AiopRxStatus_Errors|AiopRxStatus_RxBreak)) {
                    if (statusAndRxByte & AiopRxStatus_RxBreak)
                        ++aiopic_chan->error_count.rxBreak;
                    if (statusAndRxByte & AiopRxStatus_RxOverrunErr)
                        ++aiopic_chan->error_count.rxOverflow;
                    if (statusAndRxByte & AiopRxStatus_RxParityErr)
                        ++aiopic_chan->error_count.rxParity;
                    if (statusAndRxByte & AiopRxStatus_RxFrameErr)
                        ++aiopic_chan->error_count.rxFraming;
                }

                *dst++ = rxByte;

                DebugIcTimeoutMarkReceive();
            }
        } else {
            if (((int)dst) & 1) {    // need hword aligned pointer
                *dst++ = *(ch->dataByte);
                putTrace((((unsigned)(ch->aiop) & 0xff)|ch->chan),0x3d,*(dst-1));
                --count;
                DebugIcTimeoutMarkReceive();
            }

            while (count > 1) {
                *((unsigned short*)dst) = *(ch->dataWord);
                putTrace((((unsigned)(ch->aiop) & 0xff)|ch->chan),0x3e,*((unsigned short*)dst));
                dst += 2;
                count -= 2;
                DebugIcTimeoutMarkReceive();
                if (fastBaudFlags)
                    nops(20);                // don't hog AIOPIC SRAM bandwidth
            }

            if (count) {
                *dst++ = *(ch->dataByte);
                putTrace((((unsigned)(ch->aiop) & 0xff)|ch->chan),0x3d,*(dst-1));
                --count;
                DebugIcTimeoutMarkReceive();
            }
        }

        //putTrace((((unsigned)(ch->aiop) & 0xff)|ch->chan),0x34,dst[-1]);
        cbuf->put = nextPut;
        //putTrace((((unsigned)(ch->aiop) & 0xff)|ch->chan),0x35,nextPut);

        rxCount = sGetRxCnt(ch);

        // did we hit end of buffer and need to wrap around?
        if (rxCount && nextPut == 0)
            goto again;

        putTrace((((unsigned)(ch->aiop) & 0xff)|ch->chan),0x3f,rxCount);

        if (rxCount == 0 && chanStatus & AiopChanStatus_StatusModeEnable)
            sDisRxStatusMode(ch);
    }

    aiopic_chan->rxFifoLast = 0;

    // send Xon if there's room in Rx FIFO

    if ((chan->config.flags & CYG_SERIAL_FLAGS_XONXOFF_RXFLOW)
            && aiopic_chan->sentXoff && (rxCount < FlowLowWater)) {
        //putTrace((((unsigned)(ch->aiop) & 0xff)|ch->chan),0x38,rxCount);
        aiopic_chan->sentXoff = 0;
        sWriteTxPrioByte(ch,chan->config.rxflow_xon_char);
        //putTrace((((unsigned)(ch->aiop) & 0xff)|ch->chan),0x3c,chan->config.rxflow_xon_char);
    }

    if (cbuf->waiting && wakeup) {
        cbuf->waiting = false;
        cyg_drv_cond_signal(&cbuf->wait);
    }
}

// Internal function to actually configure the hardware to desired baud rate, etc.
static bool aiopic_serial_config_port(serial_channel *chan,
                                      cyg_serial_info_t *new_config,
                                      bool init)
{
    tAiopicChanInfo *aiopChanInfo = (tAiopicChanInfo*)(chan->dev_priv);
    tChan *ch = &aiopChanInfo->chan;
    unsigned long baud;
    extern volatile unsigned char EpldInterfaceSelect[];
    int mode;

    // inter char timeout

    aiopChanInfo->interCharTimeout = new_config->inter_char_timeout;
    aiopChanInfo->gotData = 0;
    aiopChanInfo->timeoutFlag = 0;

    // baud rate

    if (new_config->baud >= 50)
        baud = new_config->baud;
    else
        baud = select_baud[new_config->baud];

    if (baud == 0)
        return false;

    // map 300 baud to 500k baud for DM500
    if (IsDM500 && (baud == 300 || baud == 56000))
        baud = 500000;

    // need to know of any ports are set to 230K or faster

    if (baud >= 230400)
        fastBaudFlags |= Bit(ch->chan);
    else
        fastBaudFlags &= ~Bit(ch->chan);

    sSetBaud(ch,baud);

    // word length

    switch (new_config->word_length) {
    case CYGNUM_SERIAL_WORD_LENGTH_7:
        sSetData7(ch);
        break;

    case CYGNUM_SERIAL_WORD_LENGTH_8:
        sSetData8(ch);
        break;

    default:
        return false;
        break;
    }

    // parity

    switch (new_config->parity) {
    case CYGNUM_SERIAL_PARITY_ODD:
        sSetOddParity(ch);
        sEnParity(ch);
        break;

    case CYGNUM_SERIAL_PARITY_EVEN:
        sSetEvenParity(ch);
        sEnParity(ch);
        break;

    case  CYGNUM_SERIAL_PARITY_NONE:
        sDisParity(ch);
        break;

    default:
        return false;
        break;
    }

    // stop bits

    switch (new_config->stop) {
    case CYGNUM_SERIAL_STOP_1:
        sSetStop1(ch);
        break;

    case CYGNUM_SERIAL_STOP_2:
        sSetStop2(ch);
        break;

    default:
        return false;
        break;
    }

    // loop back

    // flow control

    if (new_config->txflow_xon_char != chan->config.txflow_xon_char)
        sSetTxXONChar(ch,new_config->txflow_xon_char);
    if (new_config->txflow_xoff_char != chan->config.txflow_xoff_char)
        sSetTxXOFFChar(ch,new_config->txflow_xoff_char);

    if (new_config->flags & CYG_SERIAL_FLAGS_XONXOFF_TXFLOW) {
        sEnTxSoftFlowCtl(ch);
        if (new_config->flags & CYG_SERIAL_FLAGS_XONXOFF_IXANY)
            sEnIXANY(ch);
        else
            sDisIXANY(ch);
    } else {
        sDisTxSoftFlowCtl(ch);
        sDisIXANY(ch);
    }

    if (new_config->flags & (CYG_SERIAL_FLAGS_RTS_RXFLOW | CYG_SERIAL_FLAGS_RTSCTS))
        sEnRTSFlowCtl(ch);
    else
        sDisRTSFlowCtl(ch);

    if (new_config->flags & (CYG_SERIAL_FLAGS_CTS_TXFLOW | CYG_SERIAL_FLAGS_RTSCTS)) {
        sClrRTS(ch);
        sEnCTSFlowCtl(ch);
    } else
        sDisCTSFlowCtl(ch);

    if (new_config->flags & CYG_SERIAL_FLAGS_RTS_TOGGLE)
        sEnRTSToggle(ch);
    else
        sDisRTSToggle(ch);

    //interface mode
    // If Serial Hub 8 or 16, reject 422 or 485 modes
    if ( (!UNIT_HAS_422_485) && (new_config->interface_mode == RS422_MODE ||
                                 new_config->interface_mode == RS485_MODE) )
        return false;

    mode = new_config->interface_mode;
    EpldInterfaceSelect[ch->line/8] = ch->chan | mode;

    //  Sanity check.  HW flow control can be active only when in RS232 mode
    if (mode != RS232_MODE) {
        sDisRTSFlowCtl(ch);
        sDisCTSFlowCtl(ch);
        new_config->flags &= ~CYG_SERIAL_FLAGS_RTS_RXFLOW;
        new_config->flags &= ~CYG_SERIAL_FLAGS_CTS_TXFLOW;
        new_config->flags &= ~CYG_SERIAL_FLAGS_RTSCTS;
    }

    // software RTS toggle

    aiopChanInfo->preRtsHoldTime = new_config->preRtsHoldTime;
    aiopChanInfo->postRtsHoldTime = new_config->postRtsHoldTime;
    if (new_config->flags & CYG_SERIAL_FLAGS_SOFT_RTS_TOGGLE) {
        sDisTransmit(ch);
        sDisRTSFlowCtl(ch);
        sDisRTSToggle(ch);
        sClrRTS(ch);
        new_config->flags &= ~CYG_SERIAL_FLAGS_RTS_RXFLOW;
        new_config->flags &= ~CYG_SERIAL_FLAGS_RTS_TOGGLE;
        aiopChanInfo->softRtsToggle = 1;
    } else {
        sEnTransmit(ch);
        aiopChanInfo->softRtsToggle = 0;
    }

    // save config

    if (new_config != &chan->config)
        chan->config = *new_config;

    // interrupts for callbacks?

    if (chan->config.modem_callback)
        sEnRxModemInt(ch);
    else
        sDisRxModemInt(ch);

    if (chan->config.rx_callback) {
        sEnRxStatusMode(ch);
        sEnRxCharInt(ch);
    } else {
        sDisRxStatusMode(ch);
        sDisRxCharInt(ch);
    }

    return true;
}

// Internal function to handle set_config keys not implimented by higher layers
Cyg_ErrNo aiopic_serial_set_config_handler(serial_channel *chan,
        cyg_uint32 key,
        const void *xbuf,
        cyg_uint32 *len)
{
    tChan *ch = &(((tAiopicChanInfo*)(chan->dev_priv))->chan);
    Cyg_ErrNo res = ENOERR;

    switch (key) {
    case CYG_IO_SET_CONFIG_SERIAL_MODEM:
        if (*len != sizeof (cyg_uint32))
            res = -EINVAL;
        else {
            if (*(cyg_uint32*)xbuf & CYG_SERIAL_MODEM_DTR)
                sSetDTR(ch);
            else
                sClrDTR(ch);
            // Ignore RTS flag if it's being controlled automatically
            if (!(chan->config.flags & (CYG_SERIAL_FLAGS_RTSCTS |
                                        CYG_SERIAL_FLAGS_RTS_RXFLOW |
                                        CYG_SERIAL_FLAGS_RTS_TOGGLE |
                                        CYG_SERIAL_FLAGS_SOFT_RTS_TOGGLE))) {
                if (*(cyg_uint32*)xbuf & CYG_SERIAL_MODEM_RTS)
                    sSetRTS(ch);
                else
                    sClrRTS(ch);
            }
        }
        break;

    case CYG_IO_SET_CONFIG_SERIAL_SCAN_PERIOD: {
        unsigned period_ms = *((unsigned*)xbuf);
        if ((*len != sizeof (cyg_uint32)) || (period_ms < 1) || (period_ms > 100))
            res = -EINVAL;
        else
            scanPeriod = period_ms;
        break;
    }

    default:
        res = -EINVAL;
    }

    return res;
}

// Internal function to handle get_config keys not implimented by higher layers
Cyg_ErrNo aiopic_serial_get_config_handler(serial_channel *chan,
        cyg_uint32 key,
        void *xbuf,
        cyg_uint32 *len)
{
    tAiopicChanInfo *serinfo = (tAiopicChanInfo *)chan->dev_priv;
    tChan *ch = &(((tAiopicChanInfo*)(chan->dev_priv))->chan);
    Cyg_ErrNo res = ENOERR;

    switch (key) {
    case CYG_IO_GET_CONFIG_SERIAL_MODEM:
        if (*len != sizeof (cyg_uint32))
            res = -EINVAL;
        else {
            unsigned short chanStatus = sGetChanStatus(ch);
            unsigned short txControl  = GetChanReg(ch,txControl);
            unsigned short chanIntID;

            // reading the interrupt ID here may clear interrupt
            // and cause edge to be missed by the callback
            // routine, so don't return deltas if callback is
            // attached.

            if (chan->config.modem_callback)
                chanIntID = 0;
            else
                chanIntID = sGetChanIntID(ch);

            *(cyg_uint32*)xbuf = 0;

            if (txControl & AiopTxCntrl_DTR)
                *(cyg_uint32*)xbuf |= CYG_SERIAL_MODEM_DTR;
            if (txControl & AiopTxCntrl_RTS)
                *(cyg_uint32*)xbuf |= CYG_SERIAL_MODEM_RTS;

            if (chanStatus & AiopChanStatus_DSR)
                *(cyg_uint32*)xbuf |= CYG_SERIAL_MODEM_DSR;
            if (chanStatus & AiopChanStatus_CTS)
                *(cyg_uint32*)xbuf |= CYG_SERIAL_MODEM_CTS;
            if (chanStatus & AiopChanStatus_CD)
                *(cyg_uint32*)xbuf |= CYG_SERIAL_MODEM_CD;
            if (!(*RingIndicatorRegister & (1 << (ch->line))))
                *(cyg_uint32*)xbuf |= CYG_SERIAL_MODEM_RI;
            if (chanIntID & AiopIntId_DSR)
                *(cyg_uint32*)xbuf |= CYG_SERIAL_MODEM_DSR_DELTA;
            if (chanIntID & AiopIntId_CTS)
                *(cyg_uint32*)xbuf |= CYG_SERIAL_MODEM_CTS_DELTA;
            if (chanIntID & AiopIntId_CD)
                *(cyg_uint32*)xbuf |= CYG_SERIAL_MODEM_CD_DELTA;
            if (0)
                *(cyg_uint32*)xbuf |= CYG_SERIAL_MODEM_RI_DELTA;

            //  Set the flag indicating our transmitter is flowed off.
            //  First hardware flow control.  If CTS cleared, must be
            //  flowed off.
            if ((chan->config.flags & (CYG_SERIAL_FLAGS_CTS_TXFLOW | CYG_SERIAL_FLAGS_RTSCTS))
                    && !(chanStatus & AiopChanStatus_CTS)) {
                *(cyg_uint32*)xbuf |= CYG_SERIAL_MODEM_FLOWTXOFF;
            }

            //  Next software flow control.  If XON/XOFF is active, check
            //  if SET_TXOFF is active.  If not, check if automatic SW flow
            //  control in the AIOPIC has our transmitter flowed off.  There
            //  is no status flag in the AIOPIC for this, so we test is by
            //  checking (twice) that the TX FIFO is not empty but the shift
            //  register is empty.  This is not perfect, but is OK, since the
            //  Rocketport does it this way...
            else if ((chan->config.flags & CYG_SERIAL_FLAGS_XONXOFF_TXFLOW)) {
                if (serinfo->fakeXoff) {
                    *(cyg_uint32*)xbuf |= CYG_SERIAL_MODEM_FLOWTXOFF;
                } else {
                    if ((chanStatus & (AiopChanStatus_TxFifoEmpty | AiopChanStatus_TxShiftEmpty)) == AiopChanStatus_TxShiftEmpty) {
                        chanStatus = sGetChanStatus(ch);
                        if ((chanStatus & (AiopChanStatus_TxFifoEmpty | AiopChanStatus_TxShiftEmpty)) == AiopChanStatus_TxShiftEmpty) {
                            *(cyg_uint32*)xbuf |= CYG_SERIAL_MODEM_FLOWTXOFF;
                        }
                    }
                }
            }
        }
        break;

    case CYG_IO_GET_CONFIG_SERIAL_ERRORS:
        if (*len < sizeof serinfo->error_count)
            res = -EINVAL;
        else
            *((cyg_serial_error_count_t*)(xbuf)) = serinfo->error_count;
        break;

    case CYG_IO_GET_CONFIG_SERIAL_CLRERRS:
        memset(&serinfo->error_count, 0, sizeof serinfo->error_count);
        break;

    case CYG_IO_GET_CONFIG_SERIAL_BREAKON:
        sSetBreak(ch);
        break;

    case CYG_IO_GET_CONFIG_SERIAL_BREAKOFF:
        sClrBreak(ch);
        break;

    case CYG_IO_SET_CONFIG_TX_OFF:
        if (!(chan->config.flags & CYG_SERIAL_FLAGS_XONXOFF_TXFLOW))
            break;
        sDisTxSoftFlowCtl(ch);
        sDisTransmit(ch);
        serinfo->fakeXoff = true;
        sEnRxIntCompare1(ch);
        sEnRxSpecInt(ch);
        break;

    case CYG_IO_GET_CONFIG_TX_OVERRIDE:
        if (serinfo->fakeXoff) {
            sEnTransmit(ch);
            sDisRxSpecInt(ch);
            sDisRxCompare1(ch);
            serinfo->fakeXoff = false;
            sEnTxSoftFlowCtl(ch);
        }
        sClrTxXOFF(ch);
        break;

    case CYG_IO_GET_CONFIG_RESTART_UART: {
        // save register states;
        unsigned short txControl  = GetChanReg(ch,txControl);
        unsigned short rxControl  = GetChanReg(ch,rxControl);
        unsigned short baudDivisor= GetChanReg(ch,baudDivisor);

        ch->aiop->command = AiopCommand_ResetUart | ch->chan;
        ch->aiop->command = ch->chan;

        PutChanReg(ch,baudDivisor,baudDivisor);
        PutChanReg(ch,txControl,txControl);
        PutChanReg(ch,rxControl,rxControl);
        break;
    }

    default:
        res = -EINVAL;
    }

    return res;
}

// Function to initialize the device.  Called at bootstrap time.
SERIAL_FUNS2(aiopic_serial_functions,
             aiopic_serial_putc,
             aiopic_serial_getc,
             aiopic_serial_set_config,
             aiopic_serial_start_xmit,
             aiopic_serial_stop_xmit,
             aiopic_serial_get_config_handler,
             aiopic_serial_set_config_handler,
             aiopic_serial_start_recv,
             aiopic_serial_stop_recv,
             aiopic_serial_flush_input,
             aiopic_serial_flush_output,
             aiopic_serial_get_tx_count,
             aiopic_serial_get_rx_count
            );


static cyg_handle_t  aiopicAiopicInterrupt0Handle;
static cyg_interrupt aiopicAiopicInterrupt0Object;
static cyg_handle_t  aiopicAiopicInterrupt1Handle;
static cyg_interrupt aiopicAiopicInterrupt1Object;

bool aiopic_serial_init(struct cyg_devtab_entry *tab)
{
    extern volatile unsigned long IOPCON;
    serial_channel *chan;
    static int init;
    int i;

    if (init) return true;

    init = 1;

    // set aiopic clock
    AiopClock = (IsDM500 ? 48000000 : 44236800) / 2;

    //tab->lookup = aiopic_serial_lookup;

    // first we need to figure out if we have one, two or four AIOPICs present
    _aiop3.intMask = 0x22;
    _aiop1.intMask = 0xaa;
    _aiop0.intMask = 0x55;

    numChanPresent = 0;
    if (_aiop0.intMask == 0x55) numChanPresent += 8;
    if (_aiop1.intMask == 0xaa) numChanPresent += 8;
    if (_aiop3.intMask == 0x22) numChanPresent += 16;

    //  If there are no AIOPICS present, this driver fails
    if (numChanPresent == 0) return false;

    // we're going to initialize all channels at once, because we don't want
    // to enable the interrupt unless all channels are initialized.

    diag_printf("AIOPIC serial driver: %d channels\n",numChanPresent);

    if (sizeof(tAiopic) != 8256) {
        diag_printf("Error, tAiopic struct size is incorrect\n");
        numChanPresent = 0;
        return false;
    }
#if 0
    diag_printf("sizeof (tAiopic)   == %u\n",sizeof (tAiopic));
    diag_printf("sizeof (tRegVal)   == %u\n",sizeof (tRegVal));
    diag_printf("sizeof (tChan)     == %u\n",sizeof (tChan));

    ShowAiopicOffset(data);
    ShowAiopicOffset(rxFifoCnt);
    ShowAiopicOffset(chanStatus);
    ShowAiopicOffset(indexAddr);
    ShowAiopicOffset(indexData);
    ShowAiopicOffset(chanStatusHigh);
    ShowAiopicOffset(oddIntID);
    ShowAiopicOffset(intChan);
    ShowAiopicOffset(indexDataHigh);
    ShowAiopicOffset(dataLow);
    ShowAiopicOffset(txFifoCnt);
    ShowAiopicOffset(chanStatusLow);
    ShowAiopicOffset(evenIntID);
    ShowAiopicOffset(command);
    ShowAiopicOffset(intMask);
    ShowAiopicOffset(indexDataLow);
#endif
    initAiopicsAndChannels();

    getModelNumber(&modelId);

    //need model id to distinguish models without 422/485 and model
    //5002240 which has floating interrupt pins from unpopulated aiops
    //to samsung so fails with level interrupts

    if (0 && modelId == 5002240) {
        // set up extern interrupt 1,2 to rising edge
        IOPCON &= ~((0x1f<<10)|(0x1f<<5));
        IOPCON |= (0x19<<10)|(0x19<<5);
    } else {
        // set up extern interrupts 1,2to level sensitive
        IOPCON &= ~((0x1f<<10)|(0x1f<<5));
        IOPCON |= (0x18<<10)|(0x18<<5);
    }

    for (chan = serialChannel; chan < serialChannel + numChanPresent; ++chan)
        aiopic_serial_config_port(chan, &chan->config, true);

    cyg_mutex_init(&sync.lock);
    cyg_cond_init(&sync.wait,&sync.lock);

    cyg_drv_interrupt_create(CYGNUM_HAL_INTERRUPT_Ext1,
                             99,
                             0,
                             aiopic_serial_aiopic_ISR,
                             aiopic_serial_aiopic_DSR,
                             &aiopicAiopicInterrupt0Handle,
                             &aiopicAiopicInterrupt0Object);
    cyg_drv_interrupt_attach(aiopicAiopicInterrupt0Handle);

    if ((numChanPresent == 16) && (modelId != 5002240)) {
        cyg_drv_interrupt_create(CYGNUM_HAL_INTERRUPT_Ext2,
                                 99,
                                 1,
                                 aiopic_serial_aiopic_ISR,
                                 aiopic_serial_aiopic_DSR,
                                 &aiopicAiopicInterrupt1Handle,
                                 &aiopicAiopicInterrupt1Object);
        cyg_drv_interrupt_attach(aiopicAiopicInterrupt1Handle);
        cyg_drv_interrupt_acknowledge(CYGNUM_HAL_INTERRUPT_Ext2);
        cyg_drv_interrupt_unmask(CYGNUM_HAL_INTERRUPT_Ext2);
    }

    //  Set up the serial channel structures.  Declared in dmrts.c, defined here.
    for (i = 0; i < numChanPresent; i++) {
        memset(&serialChannel[i], 0, sizeof(serialChannel[i]));
        serialChannel[i].funs = &aiopic_serial_functions;
        serialChannel[i].callbacks = &cyg_io_serial_callbacks;
        serialChannel[i].dev_priv = &aiopChanInfo[i];
        serialChannel[i].config.baud = 9600;
        serialChannel[i].config.stop = 1;
        serialChannel[i].config.parity = 0;
        serialChannel[i].config.word_length = 8;
        serialChannel[i].config.flags = 0;
        serialChannel[0].config.interface_mode = RS232_MODE;
        serialChannel[i].config.rxflow_xon_char = 0x11;
        serialChannel[i].config.rxflow_xoff_char = 0x13;
        serialChannel[i].config.txflow_xon_char = 0x11;
        serialChannel[i].config.txflow_xoff_char = 0x13;
        serialChannel[i].init = false;
        serialChannel[i].sync.lock = &sync.lock;
        serialChannel[i].sync.wait = &sync.wait;
    }//end for(i = 0; i < numChanPresent; i++)

    // If Serial Hub 8 or 16, set interface mode for all ports to 232
    if (modelId == 5002115 || modelId == 5002315) {
        UNIT_HAS_422_485 = 0;
        for (i = 1; i < numChanPresent; i++)
            serialChannel[i].config.interface_mode = RS232_MODE;
    } else
        UNIT_HAS_422_485 = 1;

#if EnableDebugIcTimeout
    debugICTimeout = 0; 	//set this to turn on timestamping of character xmits.

    //get the clock resolution for use with clkticks2ms()
    //these values do not change so safe to use member level vars.
    if (debugICTimeout) {
        mCCR = cyg_clock_get_resolution(cyg_real_time_clock());
        diag_printf("aiopic_serial_init: *** WARNING *** Debug IC Timeout ENABLED\n");
    }
#endif

    return true;
}//end aiopic_serial_init()


//   This routine is called when the device is "looked" up (i.e. attached)
//   The buffer areas must be malloc's here, since the malloc library is not
//   initialized when the driver init function is called.
Cyg_ErrNo aiopic_serial_lookup(struct cyg_devtab_entry **tab,
                               struct cyg_devtab_entry *sub_tab, const char *name)
{
    static int dataBuffersAllocated;
    int i;

    serial_channel *chan = (serial_channel *)(*tab)->priv;

    if (!dataBuffersAllocated) {
        for (i = 0; i < numChanPresent; i++) {
            serialChannel[i].out_cbuf.data = (unsigned char*)malloc(BufSize);
            serialChannel[i].out_cbuf.len = BufSize;
            serialChannel[i].in_cbuf.data = (unsigned char*)malloc(BufSize);
            serialChannel[i].in_cbuf.len =BufSize;
        }

        cyg_drv_interrupt_acknowledge(CYGNUM_HAL_INTERRUPT_Ext1);
        cyg_drv_interrupt_unmask(CYGNUM_HAL_INTERRUPT_Ext1);

        if (numChanPresent > 16) {
            cyg_drv_interrupt_acknowledge(CYGNUM_HAL_INTERRUPT_Ext2);
            cyg_drv_interrupt_unmask(CYGNUM_HAL_INTERRUPT_Ext2);
        }

        //When driver is first looked up, register the alarm so the poll
        //function can get called every ms.
        cyg_alarm_create(dmrtsCounterHandle, aiopic_serial_poll,
                         (cyg_addrword_t)alarmIndex, &aiopicAlarmHandle, &aiopicAlarmObj);
        cyg_alarm_initialize(aiopicAlarmHandle, 10, 1);

        dataBuffersAllocated = 1;

        diag_printf("\n\naiopic_serial_lookup called and aiopic_serial_poll alarm initialized\n\n");
    }//end if(!dataBuffersAllocated)

    if (chan >= serialChannel + numChanPresent) return -ENODEV;

    (chan->callbacks->serial_init)(chan);  // Really only required for interrupt driven devices

    return ENOERR;
}//end aiopic_serial_lookup()

// Send a character to the device output buffer.
// Return 'true' if character is sent to device
bool aiopic_serial_putc(serial_channel *chan, unsigned char c)
{
//	cyg_tick_count_t ctc;
//	unsigned long long ms = 0;

    tChan *ch = &(((tAiopicChanInfo*)(chan->dev_priv))->chan);

    //putTrace((((unsigned)(ch->aiop) & 0xff)|ch->chan),0x10,txFifoCount);

    if (sTxAvail(ch)) {
        *(ch->dataByte) = c;
        //putTrace((((unsigned)(ch->aiop) & 0xff)|ch->chan),0x11,c);
#if 0
        // I never seen this come through on the debug header so not sure
        // if it is in play. Will add timestamp to handleReceive/TransmitBytes()
        // functions instead of put/getc.
        if (debugICTimeout) {
            ctc = cyg_current_time();
            clkticks2ms(ctc, &mCCR, &ms);
            diag_printf("aiopicputc%u: [%c] @ clock %llu (%llums)\n", ch->chan, *(ch->dataByte), (unsigned long long)ctc, ms);
        }
#endif
        return true;
    } else {
        // No space
        //putTrace((((unsigned)(ch->aiop) & 0xff)|ch->chan),0x12,c);
        return false;
    }
}

// Fetch a character from the device input buffer, waiting if necessary
unsigned char aiopic_serial_getc(serial_channel *chan)
{
    //cyg_tick_count_t ctc;
    //unsigned long long ms = 0;

    tChan *ch = &(((tAiopicChanInfo*)(chan->dev_priv))->chan);

    while (!sRxAvail(ch)) {
        ;
    }

#if 0
    // KRW
    // I never seen this come through on the debug header so not sure
    // if it is in play. Will add timestamp to handleReceive/TransmitBytes()
    // functions instead of put/getc.
    if (debugICTimeout) {
        ctc = cyg_current_time();
        clkticks2ms(ctc, &mCCR, &ms);
        diag_printf("aiopicgetc%u: [%c] @ clock %llu (%llums)\n", ch->chan, *(ch->dataByte), (unsigned long long)ctc, ms);
    }
#endif

    return *(ch->dataByte);
}

// Set up the device characteristics; baud rate, etc.
bool aiopic_serial_set_config(serial_channel *chan, cyg_serial_info_t *config)
{
    return aiopic_serial_config_port(chan, config, false);
}

// Enable the transfer of bytes to next layer up
void aiopic_serial_start_recv(serial_channel *chan)
{
    tAiopicChanInfo *aiopic_chan = (tAiopicChanInfo *)chan->dev_priv;
    aiopic_chan->receive_stopped = false;
    handleReceiveBytes(chan);
}

// Get tx FIFO count
int aiopic_serial_get_tx_count(serial_channel *chan)
{
    tChan *ch = &(((tAiopicChanInfo*)(chan->dev_priv))->chan);
    return sGetTxCnt(ch) + ((*(ch->status)&AiopChanStatus_TxShiftEmpty) == 0);
}

// Get rx FIFO count
int aiopic_serial_get_rx_count(serial_channel *chan)
{
    tChan *ch = &(((tAiopicChanInfo*)(chan->dev_priv))->chan);
    return sGetRxCnt(ch);
}

// Flush input data
void aiopic_serial_flush_input(serial_channel *chan)
{
    tAiopicChanInfo *aiopic_chan = (tAiopicChanInfo *)chan->dev_priv;
    tChan *ch = &(aiopic_chan->chan);

    sDisRxFIFO(ch);
    diag_printf("\n");   // need to delay 2us
    // reset rx fifo logic
    ch->aiop->command = AiopCommand_ResetRxFifo | ch->chan;
    ch->aiop->command = ch->chan;
    // set rx fifo pointers to 0
    AiopPutIndexedShort(ch->aiop,ch->base+_RXF_OUTP,0);
    AiopPutIndexedShort(ch->aiop,ch->base+_RXF_INP,0);
    sEnRxFifo(ch);
}

// Flush output data
void aiopic_serial_flush_output(serial_channel *chan)
{
    tAiopicChanInfo *aiopic_chan = (tAiopicChanInfo *)chan->dev_priv;
    tChan *ch = &aiopic_chan->chan;
    int txEnabled;

    txEnabled = GetChanReg(ch,txControl) & AiopTxCntrl_TxEnable;
    if (txEnabled)
        sDisTransmit(ch);
    sStopRxProcessor(ch);
    ch->aiop->command = AiopCommand_ResetTxFifo | ch->chan;
    ch->aiop->command = ch->chan;
    // set tx fifo pointers to 0
    AiopPutIndexedShort(ch->aiop,ch->base+_TXF_OUTP,0);
    if (txEnabled)
        sEnTransmit(ch);
    sStartRxProcessor(ch);
}

// Disable the transfer of bytes to next layer up -- we need to disable
// the receive interrupt so that we stop reading bytes from the Rx FIFO
// so that the UART will flow controll the other end.
void aiopic_serial_stop_recv(serial_channel *chan)
{
    tAiopicChanInfo *aiopic_chan = (tAiopicChanInfo *)chan->dev_priv;
    aiopic_chan->receive_stopped = true;
}


// Enable the transmit interrupt on the device
void aiopic_serial_start_xmit(serial_channel *chan)
{
}


// Disable the transmit interrupt on the device
void aiopic_serial_stop_xmit(serial_channel *chan)
{
}


void sSetBaud(tChan *ch, int baud)
{
    long int divisor = AiopBaudDivisor(baud);
    if (divisor < 0)
        divisor = 0;
    PutChanReg(ch,baudDivisor,swapshort(divisor));
    //diag_printf("sSetBaud: baud=%d divisor=%d\n",baud,divisor);
}


static int aiopic0DsrCount;
static int aiopic1DsrCount;
static int aiopicIsrCount;
static cyg_addrword_t isrDataValue;

static cyg_uint32 aiopic_serial_aiopic_ISR(cyg_vector_t vector, cyg_addrword_t data)
{
    isrDataValue = data;
    ++aiopicIsrCount;
    if (data==0)
        cyg_drv_interrupt_mask(CYGNUM_HAL_INTERRUPT_Ext1);
    else
        cyg_drv_interrupt_mask(CYGNUM_HAL_INTERRUPT_Ext2);

    return CYG_ISR_HANDLED+CYG_ISR_CALL_DSR;
}

static void aiopic_serial_aiopic_DSR(cyg_vector_t vector, cyg_ucount32 count, cyg_addrword_t data)
{
    int i;

    for (i=0; i<numChanPresent; ++i) {
        tChan *ch = &aiopChanInfo[i].chan;
        unsigned char intID = sGetChanIntID(ch);
        if (intID) {
            unsigned short status = sGetChanStatus(ch);
            cyg_serial_info_t *config = &serialChannel[i].config;

            if ((intID&AiopIntId_Modem) && config->modem_callback) {
                unsigned current = 0;
                unsigned delta = 0;
                unsigned short txControl = GetChanReg(ch,txControl);

                if (status & AiopChanStatus_CTS )
                    current |= CYG_SERIAL_MODEM_CTS;
                if (status & AiopChanStatus_DSR )
                    current |= CYG_SERIAL_MODEM_DSR;
                if (status & AiopChanStatus_CD )
                    current |= CYG_SERIAL_MODEM_CD;
                if (*RingIndicatorRegister & (1<<(ch->chan)))
                    current |= CYG_SERIAL_MODEM_RI;
                if (txControl & AiopTxCntrl_DTR)
                    current |= CYG_SERIAL_MODEM_DTR;
                if (txControl & AiopTxCntrl_RTS)
                    current |= CYG_SERIAL_MODEM_RTS;

                if (intID & AiopIntId_CTS)
                    delta |= CYG_SERIAL_MODEM_CTS;
                if (intID & AiopIntId_DSR)
                    delta |= CYG_SERIAL_MODEM_DSR;
                if (intID & AiopIntId_CD)
                    delta |= CYG_SERIAL_MODEM_CD;
                config->modem_callback(config->modem_callback_data,current,delta);
            }

            if (config->rx_callback) {
                while (sGetChanStatus(ch) & AiopChanStatus_RxDataAvail) {
                    unsigned short rxCharAndStatus = *(ch->dataWord);
                    config->rx_callback(config->rx_callback_data,rxCharAndStatus>>8,rxCharAndStatus&0xff);
                }
            }

            if ((intID&AiopIntId_TxFifo) && 0) {
            }

            if (intID & AiopIntId_RxSpecial) {	//if it was 'set txoff'
                if (aiopChanInfo[i].fakeXoff) {
                    volatile unsigned char databyte __attribute((unused));
                    databyte = *(ch->dataByte);	//remove xon char
                    sEnTransmit(ch);
                    sDisRxSpecInt(ch);
                    sDisRxCompare1(ch);
                    sEnTxSoftFlowCtl(ch);
                    aiopChanInfo[i].fakeXoff = false;
                }
                sClrTxXOFF(ch);
            }
        }
    }

    if (isrDataValue == 0) {
        ++aiopic0DsrCount;
        cyg_drv_interrupt_acknowledge(CYGNUM_HAL_INTERRUPT_Ext1);
        cyg_drv_interrupt_unmask(CYGNUM_HAL_INTERRUPT_Ext1);
    } else {
        ++aiopic1DsrCount;
        cyg_drv_interrupt_acknowledge(CYGNUM_HAL_INTERRUPT_Ext2);
        cyg_drv_interrupt_unmask(CYGNUM_HAL_INTERRUPT_Ext2);
    }
}


#define EnableProfiling 0

#if EnableProfiling
// number of profile snapshots must be power of 2
#define NumProfileSnapshots (32*1024)
static unsigned long profileSnapshots[NumProfileSnapshots];
static int profileIndex;
#endif


// Serial polling loop, called every 1ms from the alarm.  Handles
// intercharacter timeout.  Global variable scanPeriod indicates how
// often to poll the serial ports (some multiple of 1ms).  Also
// handles soft-rts-toggle feature that allows pre/post tx RTS hold.

static void aiopic_serial_poll(cyg_handle_t alarm_handle, cyg_addrword_t data)
{
    int i, pollSerialPorts = 1;
    tAiopicChanInfo *chanInfo;
    static unsigned count;
    serial_channel *chan;
#if EnableDebugIcTimeout
    static int lastScanPeriod = 0;
    static unsigned long long tperiod = 0;
    static unsigned long long ncalls = 0;
#endif

#if EnableProfiling
    profileSnapshots[profileIndex++] = cyg_thread_self();
    profileIndex &= NumProfileSnapshots-1;
#endif

    ++count;
    if (count < scanPeriod)	pollSerialPorts = 0;

    for (i = 0, chanInfo = aiopChanInfo; i < numChanPresent; ++i, ++chanInfo) {
        tChan *ch = &chanInfo->chan;
        unsigned rxFifoCount = sGetRxCnt(ch);

        // If RX FIFO count has changes, byte(s) have been received
        if (rxFifoCount != chanInfo->rxFifoLast) {
            chanInfo->gotData = 1;
            chanInfo->rxFifoLast = rxFifoCount;
            chanInfo->idleCount = 0;
        } else {
            //No bytes received this poll
            //If intercharacter timeout is enabled, we've got data, and timeout
            //is exceeded, set the timeout flag and poll the serial ports.
            ++chanInfo->idleCount;
            if (chanInfo->gotData && chanInfo->interCharTimeout &&
                    chanInfo->idleCount >= chanInfo->interCharTimeout) {
                pollSerialPorts = 1;
                chanInfo->timeoutFlag = 1;
                chanInfo->gotData = 0;
#if EnableDebugIcTimeout
                if (debugICTimeout)
                    diag_printf("\n\naiopic_serial_poll%u: interCharTimeout (%u) triggered (idle = %u) @ clock %llu\n\n",
                                ch->chan, chanInfo->interCharTimeout, chanInfo->idleCount,
                                (unsigned long long)cyg_current_time());
#endif
            }
        }

        if (chanInfo->softRtsToggle) {
            // handle software RTS toggle with pre/post tx RTS hold options
            unsigned chanStatus = sGetChanStatus(ch);
            unsigned txControl  = GetChanReg(ch,txControl);

            if (!(chanStatus & AiopChanStatus_TxFifoEmpty) && !(txControl & AiopTxCntrl_TxEnable)) {
                // There's data waiting to be sent, but we're not transmitting

                if (!(txControl & AiopTxCntrl_RTS)) {
                    // RTS isn't on, so turn it on and start countdown 'til transmit
                    sSetRTS(ch);
                    if (chanInfo->preRtsHoldTime)
                        chanInfo->rtsCounter = chanInfo->preRtsHoldTime;
                    else
                        sEnTransmit(ch);
                } else {
                    // RTS is on, so we're counting down 'til transmit on
                    chanInfo->rtsCounter--;
                    if (chanInfo->rtsCounter == 0)
                        sEnTransmit(ch);
                }
            }

            if ((chanStatus & AiopChanStatus_TxFifoEmpty) && (chanStatus & AiopChanStatus_TxShiftEmpty) && (txControl & AiopTxCntrl_RTS)) {
                // Done sending data and RTS is still on
                if (txControl & AiopTxCntrl_TxEnable) {
                    // tx enable is still on, so turn it off and start countdown 'til RTS off
                    sDisTransmit(ch);
                    if (chanInfo->postRtsHoldTime)
                        chanInfo->rtsCounter = chanInfo->postRtsHoldTime;
                    else
                        sClrRTS(ch);
                } else {
                    // tx enable is off, so we're counting down 'til RTS goes off
                    chanInfo->rtsCounter--;
                    if (chanInfo->rtsCounter == 0)
                        sClrRTS(ch);
                }
            }
        }
    }

    if (pollSerialPorts) {
        for (i = 0, chan = serialChannel; i < numChanPresent; ++i, ++chan) {
            handleReceiveBytes(chan);
            handleTransmitBytes(chan);
        }
        cyg_drv_cond_broadcast(&sync.wait);
        count = 0;
    }

#if EnableDebugIcTimeout
    // this was to check that alarm was being processed every ms
    if (debugICTimeout) {
        ++ncalls;
        if (tperiod == 0) {
            tperiod = cyg_current_time();
            lastScanPeriod = scanPeriod;
            diag_printf("\naiopic_serial_poll call annouced @ %llu\n", tperiod);
            diag_printf("aiopic_serial_poll scan period = %d\n", lastScanPeriod);
            diag_printf("aiopic_serial_poll called %llu times (10 second report interval - %llu)\n", ncalls, tperiod);
        } else if (cyg_current_time() > (tperiod + 1000)) {
            tperiod = cyg_current_time();
            diag_printf("aiopic_serial_poll called %llu times (10 second report interval - %llu)\n", ncalls, tperiod);
            if (lastScanPeriod != scanPeriod) {
                lastScanPeriod = scanPeriod;
                diag_printf("aiopic_serial_poll scan period changed to %d @ %llu\n", lastScanPeriod, tperiod);
            }
        }
    }
#endif
}

#if EnableDebugIcTimeout
/*
 *  (nanoseconds * divisor)/dividend = clock_ticks
 *  (clock_ticks * dividend)/divisor = nanoseconds
 *  nanoseconds / 1,000,000			 = milliseconds
 */
static void clkticks2ms(cyg_tick_count_t ctc, cyg_resolution_t *ccr, unsigned long long *n)
{
    unsigned long long ms = 0, ticks = 0, divd = 0, divs = 0;

    ticks	= (unsigned long long)ctc;
    divd	= (unsigned long long)ccr->dividend;
    divs	= (unsigned long long)ccr->divisor;

    ms = 0;
    ms = ((ticks * divd)/divs)/1000000;
    *n = ms;
//	diag_printf_m("((ticks=%llu * divd=%llu)/divs=%llu)/1000000ns : ret = %llums (%llu) \n", ticks, divd, divs, ms, *n);

}//end clkticks2ms()
#endif


