#include <pkgconf/io.h>
#include <cyg/io/io.h>
#include <cyg/io/devtab.h>
#include <cyg/io/serial.h>
#include <cyg/infra/diag.h>
#include <string.h>

#if defined(CYGPKG_IO_current) || (CYGNUM_IO_VERSION_MAJOR >= 2)
#define DEVIO_HAS_SELECT 1
#else
#define DEVIO_HAS_SELECT 0
#endif

static Cyg_ErrNo serial_write(cyg_io_handle_t handle, const void *buf, cyg_uint32 * len);
static Cyg_ErrNo serial_read(cyg_io_handle_t handle, void *buf, cyg_uint32 * len);
static Cyg_ErrNo serial_get_config(cyg_io_handle_t handle, cyg_uint32 key, void *buf, cyg_uint32 * len);
static Cyg_ErrNo serial_set_config(cyg_io_handle_t handle, cyg_uint32 key, const void *buf, cyg_uint32 * len);
#if DEVIO_HAS_SELECT
static Cyg_ErrNo serial_select(cyg_io_handle_t handle, cyg_uint32 which, CYG_ADDRWORD info);
#endif
#if DEVIO_HAS_SELECT
DEVIO_TABLE(cyg_io_serial_devio, serial_write, serial_read, serial_select, serial_get_config, serial_set_config);
#else
DEVIO_TABLE(cyg_io_serial_devio, serial_write, serial_read, serial_get_config, serial_set_config);
#endif

static void serial_init(serial_channel * chan);
static void serial_xmt_char(serial_channel * chan);
static void serial_rcv_char(serial_channel * chan, unsigned char c);
SERIAL_CALLBACKS(cyg_io_serial_callbacks, serial_init, serial_xmt_char, serial_rcv_char);

static void serial_init(serial_channel * chan)
{
  if (chan->init)
    return;
  if (chan->out_cbuf.len != 0)
    {
#ifdef CYGDBG_IO_INIT
      diag_printf("Set output buffer - buf: %x len: %d\n", chan->out_cbuf.data, chan->out_cbuf.len);
#endif
      chan->out_cbuf.waiting = false;
      chan->out_cbuf.abort = false;
      chan->out_cbuf.pending = 0;
      cyg_drv_mutex_init(&chan->out_cbuf.lock);
      cyg_drv_cond_init(&chan->out_cbuf.wait, &chan->out_cbuf.lock);
      chan->out_cbuf.low_water = chan->out_cbuf.len / 4;
    }
  if (chan->in_cbuf.len != 0)
    {
#ifdef CYGDBG_IO_INIT
      diag_printf("Set input buffer - buf: %x len: %d\n", chan->in_cbuf.data, chan->in_cbuf.len);
#endif
      chan->in_cbuf.waiting = false;
      chan->in_cbuf.abort = false;
      cyg_drv_mutex_init(&chan->in_cbuf.lock);
      cyg_drv_cond_init(&chan->in_cbuf.wait, &chan->in_cbuf.lock);
    }
  chan->init = true;
}

#define min(a,b) ((a)<(b)?(a):(b))

static Cyg_ErrNo serial_write(cyg_io_handle_t handle, const void *_buf, cyg_uint32 * len)
{
  cyg_devtab_entry_t *t = (cyg_devtab_entry_t *) handle;
  serial_channel *chan = (serial_channel *) t->priv;
  serial_funs *funs = chan->funs;
  cyg_int32 size = *len;
  cyg_uint8 *buf = (cyg_uint8 *) _buf;
  int next;
  cbuf_t *cbuf = &chan->out_cbuf;
  Cyg_ErrNo res = ENOERR;
  cyg_tick_count_t timeout = 0;	// timeout in ticks
  tSerialDataLoggerFuncPtr logger;

  if (chan->config.write_timeout)
    timeout = cyg_current_time() + chan->config.write_timeout;
  cbuf->abort = false;

  cyg_drv_mutex_lock(&cbuf->lock);
  cyg_drv_dsr_lock();		// Avoid race condition testing pointers

  // If buffer is empty, write through to underlying device
  // until it is full.
  if (cbuf->put == cbuf->get)
    {
      while (size > 0)
	{
	  if (!funs->putc(chan, *buf))
	    break;
	  --size;
	  ++buf;
	}
    }

  //  Now fill buffer
  while (size > 0)
    {
      next = cbuf->put + 1;
      if (next == cbuf->len)
	next = 0;

      // Check for full buffer 
      if (next == cbuf->get)
	{
	  // Full. If non-blocking, indicate how many written, and return
	  // By default, channel is blocking
	  if (chan->config.flags & CYG_SERIAL_FLAGS_WR_NONBLOCK)
	    {
	      res = -EAGAIN;
	      break;
	    }

	  //  Full, wait for space.  Set flag, wait for driver to clear it 
	  //  when it has sent data and made buffer room.
	  cbuf->waiting = true;
	  if (!cbuf->txstopped)
	    funs->start_xmit(chan);	// Make sure xmit is running
	  if (cbuf->waiting)
	    {
	      // Note: 'start_xmit' may have obviated the need to wait :-)
	      cbuf->pending += size;	// Have this much more to send [eventually]
	      if (chan->config.write_timeout)
		{
		  if ((cyg_current_time() > timeout) || !cyg_cond_timed_wait(&cbuf->wait, timeout))
		    {
		      cbuf->pending -= size;
		      res = -ETIMEDOUT;
		      break;
		    }
		}
	      else
		{
		  cyg_drv_cond_wait(&cbuf->wait);
		}
	      cbuf->pending -= size;
	    }
	  if (cbuf->abort)
	    {
	      // Give up!
	      cbuf->abort = false;
	      cbuf->waiting = false;
	      res = -EINTR;
	      break;
	    }
	}

      //  Buffer not full, copy data into it
      else
	{
	  int avail, count;
	  // how much contiguous space is available?
	  if (cbuf->get == 0)
	    avail = (cbuf->len - cbuf->put) - 1;
	  else if (cbuf->get > cbuf->put)
	    avail = (cbuf->get - cbuf->put) - 1;
	  else
	    avail = (cbuf->len - cbuf->put);
	  // how much are we going to write?
	  count = min(size, avail);
	  // write and update
	  memcpy(cbuf->data + cbuf->put, buf, count);
	  size -= count;
	  buf += count;
	  cbuf->put += count;
	  if (cbuf->put == cbuf->len)
	    cbuf->put = 0;
	}
    }

  cyg_drv_dsr_unlock();

  if (!cbuf->txstopped)
    funs->start_xmit(chan);	// Start output as necessary
  cyg_drv_mutex_unlock(&cbuf->lock);
  *len -= size;

  if ((logger=chan->dataLogger) && *len)
    if (logger(1,_buf,*len))
      chan->dataLogger = NULL;

  return res;
}

#if DEVIO_HAS_SELECT
static cyg_bool serial_select(cyg_io_handle_t handle, cyg_uint32 which, CYG_ADDRWORD info)
{
  // not supported, so just return true
  return true;
}
#endif

static Cyg_ErrNo serial_read(cyg_io_handle_t handle, void *_buf, cyg_uint32 * len)
{
  cyg_devtab_entry_t *t = (cyg_devtab_entry_t *) handle;
  serial_channel *chan = (serial_channel *) t->priv;
  serial_funs *funs = chan->funs;
  cyg_uint8 *buf = (cyg_uint8 *) _buf;
  cyg_int32 size = 0;
  cbuf_t *cbuf = &chan->in_cbuf;
  Cyg_ErrNo res = ENOERR;
#ifdef XX_CYGDBG_DIAG_BUF
  extern int enable_diag_uart;
  int _enable = enable_diag_uart;
  int _time, _stime;
#endif // CYGDBG_DIAG_BUF
  externC cyg_tick_count_t cyg_current_time(void);
  cyg_tick_count_t timeout = 0;	// timeout in ticks
  tSerialDataLoggerFuncPtr logger;

  if (chan->config.read_timeout)
    timeout = cyg_current_time() + chan->config.read_timeout;

  cbuf->abort = false;

  cyg_drv_mutex_lock(&cbuf->lock);
  cyg_drv_dsr_lock();		// Avoid races

  chan->eolFlag = 0;
  
  while (size < *len)
    {
      if (cbuf->get != cbuf->put)
	{
	  int avail, count;

	  // figure out how many contiguous bytes are available
	  if (cbuf->put > cbuf->get)
	    avail = cbuf->put - cbuf->get;
	  else
	    avail = cbuf->len - cbuf->get;

	  // how many to copy this time around
	  count = min((*len - size), avail);

	  // Copy and update.  If IGN_NULLS set, strip 0x00 chars.
	  // If EOL flags, check for EOL characters.

	  if (chan->config.flags & (CYG_SERIAL_FLAGS_IGN_NULLS + CYG_SERIAL_FLAGS_EOL1 + CYG_SERIAL_FLAGS_EOL2))
	    {
	      int i;
              int ignnull = chan->config.flags & CYG_SERIAL_FLAGS_IGN_NULLS;
              int eol1en  = chan->config.flags & CYG_SERIAL_FLAGS_EOL1;
              int eol2en  = chan->config.flags & CYG_SERIAL_FLAGS_EOL2;

	      for (i=0; (i<count) && !chan->eolFlag; i++)
		{
		  if (!ignnull || (*(cbuf->data + cbuf->get)) != '\0')
		    {
		      *buf = (*(cbuf->data + cbuf->get));
                      if (eol2en && chan->eol1detect && *buf==chan->config.eol2)
                        chan->eolFlag = 1;
                      chan->eol1detect = *buf == chan->config.eol1;
                      if (eol1en && chan->eol1detect)
                        chan->eolFlag = 1;
		      buf++;
		      size++;
		    }
		  cbuf->get++;
		}
	    }
	  else
	    {
	      memcpy(buf, cbuf->data + cbuf->get, count);
	      buf += count;
	      size += count;
	      cbuf->get += count;
	    }

	  if (cbuf->get == cbuf->len)
            {
              cbuf->get = 0;
              continue;
            }
	}


      // done if we got the requested number of bytes

      if (*len == size)
        break;
      
      
      // done if there is an inter-char timeout or EOL detect

      if (chan->rxInterCharTimeoutFlag)
        {
          res = -EINTERCHARTIMEOUT;
          chan->rxInterCharTimeoutFlag = 0;
          break;
        }
      
      if (chan->eolFlag)
        {
          res = -EEOL;
          break;
        }
      
      // done if non-blocking, or if semi-blocking and there
      // is some data available

      if ((chan->config.flags & CYG_SERIAL_FLAGS_RD_NONBLOCK) || ((chan->config.flags & CYG_SERIAL_FLAGS_RD_SEMIBLOCK) && size))
	{
	  res = -EAGAIN;
	  break;
	}
      
      // make sure receive is running

      if (cbuf->rxstopped && funs->start_recv)
	{
	  cbuf->rxstopped = false;
	  funs->start_recv(chan);
	}

      // block while waiting for more bytes

      cbuf->waiting = true;

#ifdef XX_CYGDBG_DIAG_BUF
      enable_diag_uart = 0;
      HAL_CLOCK_READ(&_time);
      _stime = (int) cyg_current_time();
      diag_printf("READ wait - get: %d, put: %d, time: %x.%x\n", cbuf->get, cbuf->put, _stime, _time);
      enable_diag_uart = _enable;
#endif // CYGDBG_DIAG_BUF

      if (chan->config.read_timeout)
	{
	  if ((cyg_current_time() > timeout) || !cyg_cond_timed_wait(&cbuf->wait, timeout))
	    {
	      res = -ETIMEDOUT;
	      break;
	    }
	}
      else
	{
	  cyg_drv_cond_wait(&cbuf->wait);
	}

#ifdef XX_CYGDBG_DIAG_BUF
      enable_diag_uart = 0;
      HAL_CLOCK_READ(&_time);
      _stime = (int) cyg_current_time();
      diag_printf("READ continue - get: %d, put: %d, time: %x.%x\n", cbuf->get, cbuf->put, _stime, _time);
      enable_diag_uart = _enable;
#endif // CYGDBG_DIAG_BUF

      if (cbuf->abort)
	{
	  // Give up!
	  cbuf->abort = false;
	  cbuf->waiting = false;
	  res = -EINTR;
	  break;
	}
    }

  *len = size;

  // check to see if we should re-start receive
  if (cbuf->rxstopped && funs->start_recv)
    {
      int rx_room = cbuf->get - cbuf->put;
      if (rx_room < 0)
	rx_room += cbuf->len;

      if (rx_room > 128)
	{
	  cbuf->rxstopped = false;
	  funs->start_recv(chan);
	}
    }

  cyg_drv_dsr_unlock();

#ifdef XX_CYGDBG_DIAG_BUF
  enable_diag_uart = 0;
  HAL_CLOCK_READ(&_time);
  _stime = (int) cyg_current_time();
  diag_printf("READ done - size: %d, len: %d, time: %x.%x\n", size, *len, _stime, _time);
  enable_diag_uart = _enable;
#endif // CYGDBG_DIAG_BUF

  cyg_drv_mutex_unlock(&cbuf->lock);

  if ((logger=chan->dataLogger) && *len)
    if (logger(0,_buf,*len))
      chan->dataLogger = NULL;

  return res;
}

//  Retrieve serial info.  
//  Input:    key - what to retrieve
//  Returns:  xbuf - filled in with the requested data
//            len  - filled in with the length of the data
static Cyg_ErrNo serial_get_config(cyg_io_handle_t handle, cyg_uint32 key, void *xbuf, cyg_uint32 * len)
{
  cyg_devtab_entry_t *t = (cyg_devtab_entry_t *) handle;
  serial_channel *chan = (serial_channel *) t->priv;
  cyg_serial_info_t *buf = (cyg_serial_info_t *) xbuf;
  Cyg_ErrNo res = ENOERR;
  cbuf_t *cbuf = &chan->out_cbuf;
  serial_funs *funs = chan->funs;

  switch (key)
    {
    case CYG_IO_GET_CONFIG_SERIAL_INFO:
      if (*len < sizeof(cyg_serial_info_t))
	return -EINVAL;
      *buf = chan->config;
      *len = sizeof(chan->config);
      break;

    case CYG_IO_GET_CONFIG_SERIAL_OUTPUT_DRAIN:
      // Wait for any pending output to complete
      if (cbuf->len == 0)
	break;			// Nothing to do if not buffered
      cyg_drv_mutex_lock(&cbuf->lock);	// Stop any further output processing
      cyg_drv_dsr_lock();
      while (cbuf->pending || (cbuf->get != cbuf->put) || (funs->get_tx_cnt && funs->get_tx_cnt(chan)))
	{
	  cbuf->waiting = true;
	  cyg_drv_cond_wait(&cbuf->wait);
	}
      cyg_drv_dsr_unlock();
      cyg_drv_mutex_unlock(&cbuf->lock);
      break;

    case CYG_IO_GET_CONFIG_SERIAL_SYNC:
      // Synchronize with UART driver
      cyg_drv_cond_wait(chan->sync.wait);
      cyg_drv_mutex_unlock(chan->sync.lock);
      break;

    case CYG_IO_GET_CONFIG_SERIAL_INPUT_FLUSH:
      // Flush any buffered input
      cbuf = &chan->in_cbuf;
      if (cbuf->len == 0)
	break;			// Nothing to do if not buffered
      cyg_drv_mutex_lock(&cbuf->lock);	// Stop any further input processing
      cyg_drv_dsr_lock();
      if (cbuf->waiting)
	{
	  cbuf->abort = true;
	  cyg_drv_cond_signal(&cbuf->wait);
	  cbuf->waiting = false;
	}
      cbuf->get = cbuf->put;	// Flush buffered input
      if (funs->flush_input)
	funs->flush_input(chan);
      if (cbuf->rxstopped)
	{
	  cbuf->rxstopped = false;
	  funs->start_recv(chan);	// Make sure recv is running
	}
      cyg_drv_dsr_unlock();
      cyg_drv_mutex_unlock(&cbuf->lock);
      break;

    case CYG_IO_GET_CONFIG_SERIAL_ABORT:
      // Abort any outstanding I/O, including blocked reads
      // Caution - assumed to be called from 'timeout' (i.e. DSR) code
      cbuf = &chan->in_cbuf;
      if (cbuf->len != 0)
	{
	  cbuf->abort = true;
	  cyg_drv_cond_signal(&cbuf->wait);
	}
      cbuf = &chan->out_cbuf;
      if (cbuf->len != 0)
	{
	  cbuf->abort = true;
	  cyg_drv_cond_signal(&cbuf->wait);
	}
      break;

    case CYG_IO_GET_CONFIG_SERIAL_OUTPUT_FLUSH:
      // Throw away any pending output
      if (cbuf->len == 0)
	break;			// Nothing to do if not buffered
      cyg_drv_mutex_lock(&cbuf->lock);	// Stop any further output processing
      cyg_drv_dsr_lock();
      if (cbuf->get != cbuf->put)
	{
	  cbuf->get = cbuf->put;	// Empties queue!
	  funs->stop_xmit(chan);	// Done with transmit
	}
      if (funs->flush_output)
	funs->flush_output(chan);
      if (cbuf->waiting)
	{
	  cbuf->abort = true;
	  cyg_drv_cond_signal(&cbuf->wait);
	  cbuf->waiting = false;
	}
      cyg_drv_dsr_unlock();
      cyg_drv_mutex_unlock(&cbuf->lock);
      break;

    case CYG_IO_GET_CONFIG_SERIAL_TX_START:
      // start transmitting
      cyg_drv_dsr_lock();
      cbuf->txstopped = false;
      funs->start_xmit(chan);
      cyg_drv_dsr_unlock();
      break;

    case CYG_IO_GET_CONFIG_SERIAL_TX_STOP:
      // start transmitting
      cyg_drv_dsr_lock();
      cbuf->txstopped = true;
      funs->stop_xmit(chan);
      cyg_drv_dsr_unlock();
      break;

    case CYG_IO_GET_CONFIG_SERIAL_BUFFER_INFO:
      // return rx/tx buffer and FIFO sizes and counts
      {
	cyg_serial_buf_info_t *p;
	if (*len < sizeof(cyg_serial_buf_info_t))
	  return -EINVAL;

	*len = sizeof(cyg_serial_buf_info_t);
	p = (cyg_serial_buf_info_t *) xbuf;

	cyg_drv_dsr_lock();
	p->rx_bufsize = chan->in_cbuf.len;
	p->rx_count = chan->in_cbuf.put - chan->in_cbuf.get;
	if (p->rx_count < 0)
	  p->rx_count += p->rx_bufsize;

	p->tx_bufsize = chan->out_cbuf.len;
	p->tx_count = chan->out_cbuf.put - chan->out_cbuf.get;
	if (p->tx_count < 0)
	  p->tx_count += p->tx_bufsize;

	if (funs->get_tx_cnt)
	  p->tx_fifo_count = funs->get_tx_cnt(chan);
	else
	  p->tx_fifo_count = 0;

	if (funs->get_rx_cnt)
	  p->rx_fifo_count = funs->get_rx_cnt(chan);
	else
	  p->rx_fifo_count = 0;

	p->rx_inter_char_timeout_flag = chan->rxInterCharTimeoutFlag;
	cyg_drv_dsr_unlock();
      }
      break;

    default:
      // Unrecognized key.  Pass down to next layer (if there's a handler)
      if (funs->get_config_handler)
	res = funs->get_config_handler(chan, key, xbuf, len);
      else
	res = -EINVAL;
    }
  return res;
}

//  Write serial info.  
//  Input:    key - what to write
//  Returns:  xbuf - serial data
//            len  - length of the data
static Cyg_ErrNo serial_set_config(cyg_io_handle_t handle, cyg_uint32 key, const void *xbuf, cyg_uint32 * len)
{
  Cyg_ErrNo res = ENOERR;
  cyg_devtab_entry_t *t = (cyg_devtab_entry_t *) handle;
  serial_channel *chan = (serial_channel *) t->priv;
  serial_funs *funs = chan->funs;
  cyg_serial_info_t *buf = (cyg_serial_info_t *) xbuf;

  switch (key)
    {
    case CYG_IO_SET_CONFIG_SERIAL_INFO:
      if (*len != sizeof(cyg_serial_info_t))
	res = -EINVAL;
      else if (funs->set_config(chan, buf) != true)
	res = -EINVAL;		// Device is not happy with paramters or they cannot be set
      break;

    case CYG_IO_SET_CONFIG_SERIAL_WRITE_PRIORITY_BYTE:
      if (*len != 1)
	return -EINVAL;
      while (funs->putc(chan, *(char *) xbuf) == false)
	;
      break;

    case CYG_IO_SET_CONFIG_SERIAL_DATALOGGING:
      if (*len != sizeof chan->dataLogger)
	return -EINVAL;
      memcpy(&chan->dataLogger,xbuf,sizeof chan->dataLogger);
      break;

    default:
      // Unrecognized key.  Pass down to next layer (if there's a handler)
      if (funs->set_config_handler)
	res = funs->set_config_handler(chan, key, xbuf, len);
      else
	res = -EINVAL;
    }
  return res;
}

static void serial_xmt_char(serial_channel * chan)
{
  cbuf_t *cbuf = &chan->out_cbuf;
  serial_funs *funs = chan->funs;
  unsigned char c;
  int space;

  while (cbuf->get != cbuf->put)
    {
      c = cbuf->data[cbuf->get];
      if (funs->putc(chan, c))
	{
	  cbuf->get++;
	  if (cbuf->get == cbuf->len)
	    cbuf->get = 0;
	  if (cbuf->waiting)
	    {
	      // See if there is now enough room to restart writer
	      space = (cbuf->len + cbuf->get) - cbuf->put;
	      if (space > cbuf->len)
		space -= cbuf->len;
	      if (space >= cbuf->low_water)
		{
		  cbuf->waiting = false;
		  cyg_drv_cond_broadcast(&cbuf->wait);
		}
	    }
	}
      else
	return;			// Need to wait for more space
    }

  funs->stop_xmit(chan);	// Done with transmit
  if (cbuf->waiting)
    {
      cbuf->waiting = false;
      cyg_drv_cond_signal(&cbuf->wait);
    }
}

static void serial_rcv_char(serial_channel * chan, unsigned char c)
{
  cbuf_t *cbuf = &chan->in_cbuf;
  serial_funs *funs = chan->funs;
  int nextPut;

  nextPut = cbuf->put + 1;
  if (nextPut == cbuf->len)
    nextPut = 0;

  if (nextPut == cbuf->get)	//full?
    return;

  if (c == '\0' && chan->config.flags & CYG_SERIAL_FLAGS_IGN_NULLS)
    return;

  cbuf->data[cbuf->put] = c;
  cbuf->put = nextPut;

  if (cbuf->waiting)
    {
#ifdef XX_CYGDBG_DIAG_BUF
      extern int enable_diag_uart;
      int _enable = enable_diag_uart;
      int _time, _stime;
      externC cyg_tick_count_t cyg_current_time(void);
      enable_diag_uart = 0;
      HAL_CLOCK_READ(&_time);
      _stime = (int) cyg_current_time();
      diag_printf("Signal reader - time: %x.%x\n", _stime, _time);
      enable_diag_uart = _enable;
#endif // CYGDBG_DIAG_BUF
      cbuf->waiting = false;
      cyg_drv_cond_signal(&cbuf->wait);
    }

  nextPut = cbuf->put + 1;
  if (nextPut == cbuf->len)
    nextPut = 0;

  if (nextPut == cbuf->get && funs->stop_recv)	// full?
    {
      cbuf->rxstopped = true;
      funs->stop_recv(chan);
    }
}
