/***************************************************************/
/*       uart4530.c                                            */
/*                                                             */
/*       Structs and functions for the serial driver for       */
/*       Comtrol DeviceMaster RTS devices use the S3C4530 UART */
/*                                                             */
/*       See NOTES at end of this file for algorithm comments. */
/***************************************************************/
#include <pkgconf/system.h>
#include <pkgconf/io_nserial.h>
#include <pkgconf/io.h>
#include <cyg/io/io.h>
#include <cyg/hal/hal_intr.h>
#include <cyg/io/devtab.h>
#include <cyg/io/serial.h>
#include <cyg/io/serialio.h>
#include <cyg/infra/diag.h>
#include <stdlib.h>
#include <string.h>
#include "dmrts.h"
#include "uart4530.h"

/*  Globals */
extern volatile tS3C4530Uart S3C4530_uart1;
extern volatile tS3C4530Dma  S3C4530_dma0;
extern volatile unsigned int IOPCON1;
extern volatile unsigned int EpldInterfaceSelect;
extern volatile unsigned int PioMode;
extern volatile unsigned int PioData;
extern cyg_counter dmrtsCounterObj;
extern cyg_handle_t dmrtsCounterHandle;


#define BaudDivisor500K ((0x0002 << 4) | 0)

// global variable indicating whether a unit is a DM500 or not
// defined in hal/arm/dmrts/current/src/hal_diag.c
extern int IsDM500;



#define EnableDebugIcTimeout 0

#if EnableDebugIcTimeout
static void clkticks2ms(cyg_tick_count_t ctc, cyg_resolution_t *ccr, unsigned long long *n);
static cyg_resolution_t mCCR;	//call cyg_clock_get_resolution(cyg_real_time_clock()) to use
static int debugICTimeout;
#endif

/*  Static Variables */
static cyg_serial_error_count_t uartErrors;
static int defaultBaudRate = 19200;
static int hardwareFlowControlActive = 0;
static int softwareFlowControlActive = 0;
static int portMode = 0xFF;   //  Keeps track of port mode (232, 422, 485)

static int baudRate = 0;      //  Keeps track of what baud rate is set
static int dmaXferActive = 0; //  SW flowcontrol, is data being sent?
static int xoffActive = 0;    //  SW flowcontrol, has an XOFF been sent?
static int rcvdXoff = 0;
static int nextGet;           // marks the next byte to be sent after DMA completes

static int stopBits = 1;      //  Keep track of bit/stop/parity settings
static int dataBits = CYGNUM_SERIAL_WORD_LENGTH_8;
static int paritySetting = CYGNUM_SERIAL_PARITY_NONE;

static bool           softRtsToggle;
static unsigned       preRtsHoldTime;
static unsigned       postRtsHoldTime;
static unsigned       rtsCounter;

static cyg_handle_t  uartRxInterruptHandle;
static cyg_interrupt uartRxInterruptObject;
static cyg_handle_t  uartTxInterruptHandle;
static cyg_interrupt uartTxInterruptObject;
static cyg_handle_t  dmaInterruptHandle;
static cyg_interrupt dmaInterruptObject;

static int scanPeriod = 10;                //  Serial poll scan period, in ms
static int gotData = 0;                    //  Flag to start icto timer
static unsigned int intercharTimeoutValue = 0;      //  Icto value in ms
static unsigned int ictoCounter;                    //  The icto timer counter

/*  Alarm variables */
static cyg_handle_t msPollHandle;
static cyg_alarm msPollObj;
static cyg_alarm_t msPoll;
static unsigned long msPollIndex = 0;

/*  Static Prototypes */
static void SetPortMode(int mode);
static unsigned int BaudRegVal(int baud);
static void SetBaudRate(int baud);
static void StopBits2(int flag);
static void DataBits(int wordLength);
static void Parity(int parity);
static int IsRiSet(void);
static int IsCdSet(void);
static int IsDsrSet(void);
static int IsCtsSet(void);
static int IsRtsSet(void);
static int IsDtrSet(void);
static int DidCtsEventOccur(void);
static void Break(int flag);
static void Rts(int flag);
static void Dtr(int flag);
static int IsRxFifoTrig(void);
static int IsRxFifoEmpty(void);
static int IsRxTimeout(void);
static int IsRxDataAvail(void);
static unsigned char RcvByteUart(void);
static void XmitByteUart(unsigned char c);
static void GetUartStatusError(void) ;
static int RxOverrunError(unsigned int statReg);
static int RxBreak(unsigned int statReg);
static int FramingError(unsigned int statReg);
static int ParityError(unsigned int statReg);
static void FlushRxFifo(void);
static void FlushTxFifo(void);
static int IsTxHoldingRegEmpty(void);
static int TransmitBufferCount(void);
static int ReceiveBufferCount(void);
static int IsTxComplete(void);
static cyg_uint32 UartRxIsr(cyg_vector_t vector, cyg_addrword_t data);
static void UartRxDsr(cyg_vector_t vector, cyg_ucount32 count, cyg_addrword_t data);
static cyg_uint32 UartTxIsr(cyg_vector_t vector, cyg_addrword_t data);
static void UartTxDsr(cyg_vector_t vector, cyg_ucount32 count, cyg_addrword_t data);
static cyg_uint32 DmaIsr(cyg_vector_t vector, cyg_addrword_t data);
static void DmaDsr(cyg_vector_t vector, cyg_ucount32 count, cyg_addrword_t data);
static void EnableHwFlowControl(void);
static void DisableHwFlowControl(void);
static void EnableSwFlowControl(void);
static void DisableSwFlowControl(void);
static void XmitBytes(void);
static void XmitBytesDma(void);
static void SetupUart(int mode);
static void InitDmaControl(void);
static void StartDma(void);
static void StopDma(void);
static void SendControlChar(int controlChar);

static struct {
    cyg_mutex_t lock;
    cyg_cond_t  wait;
} sync;

SERIAL_FUNS2(uart4530_serial_functions,
             uart4530_putc,
             uart4530_getc,
             uart4530_set_config,
             uart4530_start_xmit,
             uart4530_stop_xmit,
             uart4530_get_config_handler,
             uart4530_set_config_handler,
             uart4530_start_recv,
             uart4530_stop_recv,
             uart4530_flush_input,
             uart4530_flush_output,
             uart4530_get_tx_count,
             uart4530_get_rx_count
            );

int RxFifoEnable = 1;


// Code that handles uart setup and port mode is so convoluted,
// mixed-together, and messed up that we've got to keep a globally
// visible copy of the the configuration flags.  :/
static unsigned configFlags;


//  Initialize the second UART on the CPU to be used for customer data.
//  Assume RS232, 19200, 8N1, no flow control, flush the FIFOs
bool uart4530_serial_init(struct cyg_devtab_entry *tab)
{
    static int init;

    //  This device only has one port, return false for all other attempts
    if (strcmp(tab->name, "/dev/ser0"))
        return false;

    //  Only init once
    if (init) return true;

    init = 1;

    SetupUart(FLOW_CONTROL_NONE);
    SetBaudRate(defaultBaudRate);
    SetPortMode(RS232_MODE);

    //  Set DTR and RTS as IO pins, all other model lines as UART pins
    IOPCON1 = 0x01FC0040;
    PioMode |= RTS_BIT;
    PioMode |= DTR_BIT;
    PioData |= RTS_BIT;            //  Sense of RTS and CTS are inverted.  Setting bit de-asserts the pins.
    PioData |= DTR_BIT;            //  Seems you can't set the data bit until after the mode bit...

    // Init the configuration struct
    memset(&serialChannel[0], 0, sizeof(serialChannel[0]));
    serialChannel[0].funs = &uart4530_serial_functions;
    serialChannel[0].config.rx_callback = NULL;
    serialChannel[0].config.modem_callback = NULL;

    serialChannel[0].config.baud = defaultBaudRate;
    serialChannel[0].config.stop = 1;
    serialChannel[0].config.parity = 0;
    serialChannel[0].config.word_length = 8;
    serialChannel[0].config.interface_mode = RS232_MODE;
    serialChannel[0].config.flags = configFlags = 0;

    serialChannel[0].config.rxflow_xon_char = 0x11;
    serialChannel[0].config.rxflow_xoff_char = 0x13;
    serialChannel[0].config.txflow_xon_char = 0x11;
    serialChannel[0].config.txflow_xoff_char = 0x13;

    serialChannel[0].init = false;

    cyg_mutex_init(&sync.lock);
    cyg_cond_init(&sync.wait,&sync.lock);
    serialChannel[0].sync.lock = &sync.lock;
    serialChannel[0].sync.wait = &sync.wait;
    serialChannel[0].rxInterCharTimeoutFlag = 0;
    serialChannel[0].eol1detect = 0;
    serialChannel[0].eolFlag = 0;

    memset(&uartErrors, 0, sizeof(uartErrors));

    // we won't unmask any interrupts until after data buffers
    // have been allocated: this happens during lookup()

    //  Create the UART Rx interrupt
    cyg_drv_interrupt_create(CYGNUM_HAL_INTERRUPT_Uart1Rx,
                             99,
                             0,
                             UartRxIsr,
                             UartRxDsr,
                             &uartRxInterruptHandle,
                             &uartRxInterruptObject);
    cyg_drv_interrupt_attach(uartRxInterruptHandle);

    //  Create the UART Tx interrupt
    cyg_drv_interrupt_create(CYGNUM_HAL_INTERRUPT_Uart1Tx,
                             99,
                             0,
                             UartTxIsr,
                             UartTxDsr,
                             &uartTxInterruptHandle,
                             &uartTxInterruptObject);
    cyg_drv_interrupt_attach(uartTxInterruptHandle);

    //  Create the DMA interrupt, don't unmask it here
    cyg_drv_interrupt_create(CYGNUM_HAL_INTERRUPT_Dma0,
                             99,
                             0,
                             DmaIsr,
                             DmaDsr,
                             &dmaInterruptHandle,
                             &dmaInterruptObject);
    cyg_drv_interrupt_attach(dmaInterruptHandle);
    cyg_drv_interrupt_acknowledge(CYGNUM_HAL_INTERRUPT_Dma0);

    /*
     *  Kluge.  Observed when converting the RTS to eCos 2.0,
     * the 1 port device would not transmit when first opened
     * by the driver in HW or SW flow control.  If the port
     * was opened in flow control NONE, and 1 or more bytes
     * was transmitted, then the port could be opened in HW or
     * SW flow control and it would transmit.  No explanation
     * for this behaviour was uncovered, eCos 1.3.1 does not
     * act this way.  Probable cause - some Samsung CPU
     * silicon bug relating to the DMA-UART1 interface. The
     * following is a work around - set baud to none, push one
     * byte to uart1 TX buf (no byte actually is transmitted),
     * set baud back to default.  Port now ready to go.
     * Kurt R. 2-8-05
     */
    S3C4530_uart1.brdiv = 0;
    XmitByteUart('z');
    SetBaudRate(defaultBaudRate);

#if EnableDebugIcTimeout
    /* krw - DO NOT REMOVE - LEAVE COMMENTED OUT */
    debugICTimeout = 0; 	//set this to turn on timestamping of character xmits.

    //get the clock resolution for use with clkticks2ms()
    //these values do not change so safe to use member level vars.
    if (debugICTimeout) {
        mCCR = cyg_clock_get_resolution(cyg_real_time_clock());
        diag_printf("uart4530_serial_init: *** WARNING *** Debug IC Timeout ENABLED\n");
    }
#endif

    return true;
}//end uart4530_serial_init()

//   This routine is called when the device is "looked" up (i.e. attached)
//   The buffer areas must be malloc's here, since the malloc library is not
//   initialized when the driver init function is called.
Cyg_ErrNo uart4530_serial_lookup(struct cyg_devtab_entry **tab,
                                 struct cyg_devtab_entry *sub_tab,
                                 const char *name)
{
    static int init;  //  Used to avoid malloc'ing more than once

    if (!init) {
        serialChannel[0].in_cbuf.data = (unsigned char *)malloc(BufSize);
        serialChannel[0].in_cbuf.put = serialChannel[0].in_cbuf.get = 0;
        serialChannel[0].in_cbuf.len = BufSize;
        serialChannel[0].out_cbuf.data = (unsigned char *)malloc(BufSize);
        serialChannel[0].out_cbuf.put = serialChannel[0].out_cbuf.get = 0;
        serialChannel[0].out_cbuf.len = BufSize;
        init = 1;

        // Register the alarm so the poll function gets called every ms.
        cyg_alarm_create(dmrtsCounterHandle, msPoll, (cyg_addrword_t)msPollIndex,
                         &msPollHandle, &msPollObj);
        cyg_alarm_initialize(msPollHandle, 10, 1);

        // initialization is now complete, so unmask interrupts
        cyg_drv_interrupt_acknowledge(CYGNUM_HAL_INTERRUPT_Uart1Rx);
        cyg_drv_interrupt_unmask(CYGNUM_HAL_INTERRUPT_Uart1Rx);
        cyg_drv_interrupt_acknowledge(CYGNUM_HAL_INTERRUPT_Uart1Tx);
        cyg_drv_interrupt_unmask(CYGNUM_HAL_INTERRUPT_Uart1Tx);
    }

    //   Detect incorrect device open calls.
    if (strcmp((*tab)->name, "/dev/ser0")) {
        diag_printf("Attempt to lookup invalid device - %s \n", (*tab)->name);
        return -ENODEV;
    }

    return ENOERR;
}

//  Send a character to the device output buffer.
//  Returns: 'true' if character is sent to device
bool uart4530_putc(serial_channel *priv, unsigned char c)
{
    //  Don't push byte directly to the UART if in flow control mode, since we may be flowed off
    if ((softwareFlowControlActive) || (hardwareFlowControlActive))
        return false;

    if (IsTxHoldingRegEmpty()) {
        XmitByteUart(c);
        return true;
    } else {
        return false;
    }
}

// Fetch a character from the device input buffer, waiting if necessary
unsigned char uart4530_getc(serial_channel *priv)
{
    while (!(IsRxFifoTrig()));
    return RcvByteUart();
}

//  Set up the device characteristics; baud rate, etc.
bool uart4530_set_config(serial_channel *priv, cyg_serial_info_t *config)
{
    static cyg_serial_info_t oldConfig;

    memcpy(&serialChannel[0].config, config, sizeof(cyg_serial_info_t));

    configFlags = config->flags;

    RxFifoEnable = !(config->flags & CYG_SERIAL_FLAGS_RXFIFO_DISABLE);


    //  If Intercharacter Timeout Value gets changed, start the ICTO counter
    if (intercharTimeoutValue != config->inter_char_timeout) {
        intercharTimeoutValue = config->inter_char_timeout;
        gotData = 0;
        ictoCounter = 0;
    }

    // Baud rate
    if (IsDM500 && (config->baud==300 || config->baud==57600))
        SetBaudRate(500000);
    else
        SetBaudRate(config->baud);

    //  Word length
    if (oldConfig.word_length != config->word_length) {
        switch (config->word_length) {
        case CYGNUM_SERIAL_WORD_LENGTH_5:
        case CYGNUM_SERIAL_WORD_LENGTH_6:
        case CYGNUM_SERIAL_WORD_LENGTH_7:
        case CYGNUM_SERIAL_WORD_LENGTH_8:
            DataBits(config->word_length);
            break;
        default:
            diag_printf(" ^^^^  Invalid word length \n ^^^^^");
            return false;
            break;
        }
    }

    // Parity
    if (oldConfig.parity != config->parity) {
        switch (config->parity) {
        case CYGNUM_SERIAL_PARITY_ODD:
        case CYGNUM_SERIAL_PARITY_EVEN:
        case CYGNUM_SERIAL_PARITY_NONE:
            Parity(config->parity);
            break;
        default:
            diag_printf(" ^^^^  Invalid parity \n ^^^^^");
            return false;
            break;
        }
    }

    // Stop bits
    if (oldConfig.stop != config->stop) {
        switch (config->stop) {
        case CYGNUM_SERIAL_STOP_1:
            StopBits2(CLEAR);
            break;
        case CYGNUM_SERIAL_STOP_2:
            StopBits2(SET);
            break;
        default:
            diag_printf(" ^^^^  Invalid stop bits \n ^^^^^");
            return false;
            break;
        }
    }

    // software RTS toggle  (force other flow control off)

    preRtsHoldTime = config->preRtsHoldTime;
    postRtsHoldTime = config->postRtsHoldTime;
    if (config->flags & CYG_SERIAL_FLAGS_SOFT_RTS_TOGGLE) {
        Rts(CLEAR);
        configFlags = config->flags &= ~(CYG_SERIAL_FLAGS_RTS_RXFLOW |
                                         CYG_SERIAL_FLAGS_RTS_TOGGLE |
                                         CYG_SERIAL_FLAGS_RTSCTS     |
                                         CYG_SERIAL_FLAGS_RTS_RXFLOW |
                                         CYG_SERIAL_FLAGS_CTS_TXFLOW |
                                         CYG_SERIAL_FLAGS_XONXOFF_TXFLOW |
                                         CYG_SERIAL_FLAGS_XONXOFF_RXFLOW);
        softRtsToggle = 1;
    } else {
        softRtsToggle = 0;
    }

    //  Setup UART based on RS mode.  After that, set up the CPLD.

    if (config->interface_mode == RS485_MODE) {
        DisableSwFlowControl();
        DisableHwFlowControl();
        SetupUart(FLOW_CONTROL_NONE);
    } else {
        //  RS232 or RS422 mode

        if (config->flags & (CYG_SERIAL_FLAGS_RTS_TOGGLE | CYG_SERIAL_FLAGS_SOFT_RTS_TOGGLE)) {
            // RTS toggle enabled
            DRV("RS232 mode with RTS Toggle\n");
            DisableSwFlowControl();
            DisableHwFlowControl();
            SetupUart(FLOW_CONTROL_NONE);
        } else if ((config->flags & (CYG_SERIAL_FLAGS_RTSCTS))     ||
                   (config->flags & (CYG_SERIAL_FLAGS_RTS_RXFLOW)) ||
                   (config->flags & (CYG_SERIAL_FLAGS_CTS_TXFLOW))) {
            // HW flow control.  This UART does not support separate
            // RTS_RXFLOW and CTS_TXFLOW settings for HW flow control.

            if (!hardwareFlowControlActive) {
                DRV("Driver call - set FC to hardware \n");
                EnableHwFlowControl();
            }
        } else {
            if (hardwareFlowControlActive) {
                DRV("Driver call - set FC to none \n");
                DisableHwFlowControl();
            }
        }

        // SW flow control
        if ((config->flags & CYG_SERIAL_FLAGS_XONXOFF_TXFLOW) ||
                (config->flags & CYG_SERIAL_FLAGS_XONXOFF_RXFLOW)) {
            S3C4530_uart1.conchar1 = (config->txflow_xon_char       |
                                      config->txflow_xon_char << 8  |
                                      config->txflow_xon_char << 16 |
                                      config->txflow_xon_char << 24);
            S3C4530_uart1.conchar2 = (config->txflow_xoff_char       |
                                      config->txflow_xoff_char << 8  |
                                      config->txflow_xoff_char << 16 |
                                      config->txflow_xoff_char << 24);

            serialChannel[0].config.rxflow_xon_char = config->rxflow_xon_char;
            serialChannel[0].config.rxflow_xoff_char = config->rxflow_xoff_char;
            serialChannel[0].config.txflow_xon_char = config->txflow_xon_char;
            serialChannel[0].config.txflow_xoff_char = config->txflow_xoff_char;

            if (!softwareFlowControlActive) {
                DRV("Driver call - set FC to software \n");
                EnableSwFlowControl();
            }
        } else {
            if (softwareFlowControlActive) {
                DRV("Driver call - set FC to none \n");
                DisableSwFlowControl();
            }
        }
    }   //  RS232 or RS422 mode


    /*  Now set the CPLD and RTS controls */
    SetPortMode(config->interface_mode);

    //  Testing has shown that using the 28 byte trigger level results
    //  in RX Overrun errors under certain circumstances.  A trigger level
    //  of 18 does not lose bytes, so it will used.
    S3C4530_uart1.ucon &= ~S3C4530UartControlRxFifoTrigMask;
    S3C4530_uart1.ucon |= S3C4530UartControlRxFifoTrig18;

    oldConfig = *config;

    return true;
}

//  Not implemented.  Note that the UART has a TX disabled mode if needed.
//  A byte is lost each time you do this, though.
void uart4530_start_xmit(serial_channel *priv)
{
}

void uart4530_stop_xmit(serial_channel *priv)
{
}

void uart4530_start_recv(serial_channel *priv)
{
}

void uart4530_stop_recv(serial_channel *priv)
{
}

//  Get Modem Signal Line conditions and UART errors
Cyg_ErrNo uart4530_get_config_handler(serial_channel *priv,
                                      cyg_uint32 key,
                                      void *xbuf,
                                      cyg_uint32 *len)
{
    Cyg_ErrNo res = ENOERR;

    switch (key) {
    case CYG_IO_GET_CONFIG_SERIAL_MODEM:
        if (*len != sizeof (cyg_uint32)) {
            res = -EINVAL;
        } else {
            *(cyg_uint32*)xbuf = 0;

            //  Check present mdmsig settings
            if (IsDtrSet())
                *(cyg_uint32*)xbuf |= CYG_SERIAL_MODEM_DTR;
            if (IsRtsSet())
                *(cyg_uint32*)xbuf |= CYG_SERIAL_MODEM_RTS;
            if (IsDsrSet())
                *(cyg_uint32*)xbuf |= CYG_SERIAL_MODEM_DSR;
            if (IsCtsSet())
                *(cyg_uint32*)xbuf |= CYG_SERIAL_MODEM_CTS;
            if (IsCdSet())
                *(cyg_uint32*)xbuf |= CYG_SERIAL_MODEM_CD;
            if (IsRiSet())
                *(cyg_uint32*)xbuf |= CYG_SERIAL_MODEM_RI;

            // Note that in HW Flow Control mode, this may not work correctly
            if (DidCtsEventOccur())
                *(cyg_uint32*)xbuf |= CYG_SERIAL_MODEM_CTS_DELTA;

            if (hardwareFlowControlActive) {
                if (IsCtsSet() == 0)
                    *(cyg_uint32*)xbuf |= CYG_SERIAL_MODEM_FLOWTXOFF;
            } else if (softwareFlowControlActive ) {
                if (rcvdXoff) {
                    *(cyg_uint32*)xbuf |= CYG_SERIAL_MODEM_FLOWTXOFF;
                }
            }
        }
        break;

    case CYG_IO_GET_CONFIG_SERIAL_ERRORS:
        if (*len < sizeof uartErrors)
            res = -EINVAL;
        else
            *((cyg_serial_error_count_t*)(xbuf)) = uartErrors;
        break;

    case CYG_IO_GET_CONFIG_SERIAL_CLRERRS:
        memset(&uartErrors, 0, sizeof(uartErrors));
        break;

    case CYG_IO_GET_CONFIG_SERIAL_BREAKON:
        Break(SET);
        break;

    case CYG_IO_GET_CONFIG_SERIAL_BREAKOFF:
        Break(CLEAR);
        break;

        /*  Set XOFF state  */
    case CYG_IO_SET_CONFIG_TX_OFF:
        if (!softwareFlowControlActive )		//not sw flow mode, bail
            break;
        rcvdXoff=1;
        if (dmaXferActive)
            StopDma();
        DFC("Set XOFF state \n ");
        break;

        /*  Clear XOFF state  */
    case CYG_IO_GET_CONFIG_TX_OVERRIDE:
        if (!softwareFlowControlActive )		//not sw flow mode, bail
            break;
        rcvdXoff=0;
        if (dmaXferActive)
            StartDma();
        DFC("Clear XOFF state \n");
        break;

    case CYG_IO_GET_CONFIG_RESTART_UART:
        // not implemented for this driver
        break;

    default:
        res = -EINVAL;
    }
    return res;
}

//  Set the modem signal lines
Cyg_ErrNo uart4530_set_config_handler(serial_channel *priv,
                                      cyg_uint32 key,
                                      const void *xbuf,
                                      cyg_uint32 *len)
{
    unsigned period_ms;
    Cyg_ErrNo res = ENOERR;

    switch (key) {
    case CYG_IO_SET_CONFIG_SERIAL_MODEM:
        if (*len != sizeof (cyg_uint32)) {
            res = -EINVAL;
        }

        // Ignore the RTS flag if hardware flow control is enabled
        if (!hardwareFlowControlActive) {
            if (*(cyg_uint32*)xbuf & CYG_SERIAL_MODEM_RTS) {
                DRV("Driver call - RTS to ON \n");
                Rts(SET);
            } else {
                DRV("Driver call - RTS to OFF \n");
                Rts(CLEAR);
            }
        }

        if (*(cyg_uint32*)xbuf & CYG_SERIAL_MODEM_DTR) {
            DRV("Driver call - DTR to ON \n");
            Dtr(SET);
        } else {
            DRV("Driver call - DTR to OFF \n");
            Dtr(CLEAR);
        }
        break;

        //  Intercharacter timeouts not supported in this driver.
        //  Therefore, there is not scan period.
    case CYG_IO_SET_CONFIG_SERIAL_SCAN_PERIOD:
        period_ms = *((unsigned*)xbuf);
        if ((*len != sizeof (cyg_uint32)) || (period_ms < 1) || (period_ms > 100))
            res = -EINVAL;
        else
            scanPeriod = period_ms;
        break;

    default:
        res = -EINVAL;
    }
    return res;
}

void uart4530_flush_input(serial_channel *priv)
{
    FlushRxFifo();
}

void uart4530_flush_output(serial_channel *priv)
{
    FlushTxFifo();
}

//  Supposed to return number of bytes in the RX and TX FIFOs.  We cannot get this count
//  directly.  These functions are used mainly during drain operations to tell when the FIFOs
//  are empty, which we can tell.
//
//  This function returns zero if there is no data buffered.  For the TX data, data could be in
//  three places:
//  1)  The DMA (if flow control enabled)
//  2)  The cbuf
//  3)  The shift and holding registers
//  Note that the shift/holding registers could be empty, yet there could be data in the DMA or
//  cbuf.  This is the case when the UART is flowed off.
//
//  So we return 0 if all three locations are empty, else 1.
//
//  Note that returning the actual count in the cbuf cause a "short write" debug message.
//  from the rpshsi thread in the devmast application.  I didn't track down what that means
//  (part of the QLanRoom stuff), but to avoid this, a boolean is returned.  This appears to
//  be sufficient for the drivers.
int uart4530_get_tx_count(serial_channel *priv)
{
    int count;

    if ((softwareFlowControlActive) || (hardwareFlowControlActive)) {
        if (dmaXferActive)
            return 1;
    }

    count = TransmitBufferCount();
    if (count) {
        return 1;
    } else {
        if (IsTxComplete())
            return 0;
        else
            return 1;
    }
}

int uart4530_get_rx_count(serial_channel *priv)
{

    int count;

    count = ReceiveBufferCount();
    if (count) {
        return 1;
    } else {
        if (IsRxFifoEmpty())
            return 0;
        else
            return 1;
    }
}

/*
 *  Sets the RS mode (RS232 = 0x00, RS422 = 0x08, RS485 = 0x10).  A CPLD (at address EpldInterfaceSelect0)
 *  connects the correct transciever hardware, and these values are what is written into the CPLD.  Note that
 *  the CPLD powers up in 0xFF mode, all transcievers disabled.  There is no way to get the CPLD into that
 *  state again once it has been set, however.  Note also that there is a fourth value  - RS232I (0x18).
 *  This is an AIOPIC only mode, and is mapped into RS232 on this platform.
 *
 *  The Samsung CPU must be also be configured based on RS mode, described below.
 *
 *  For 485 mode:
 *  -  The UART's HW Flow Control mode is enabled.
 *  -  RTS and DTR are configured as UART pins
 *  -  CTS is configured as an IO input pin
 *
 *  For 232/422 mode:
 *     if RTS toggle is enabled:
 *       - The UART's HW Flow Control mode is enabled.
 *       - RTS is configured as UART pin
 *     else
 *       - The UART's HW Flow Control mode is disabled.
 *       - RTS and DTR are configured as IO output pins
 *       - CTS is configured as a UART pin.  This allows it's use as
 *         an interrupt source during HW flow control.
 *
 *  Inputs:  mode - the RS mode
 */
static void SetPortMode(int mode)
{
    extern volatile unsigned char EpldInterfaceSelect0;

    switch (mode) {
    case RS232_MODE:
    case RS232I_MODE:
        if (portMode != RS232_MODE)
            DS("Enabling RS232 Mode \n");
        portMode = RS232_MODE;
        EpldInterfaceSelect0 = RS232_MODE;

        if (configFlags & CYG_SERIAL_FLAGS_RTS_TOGGLE) {
            DRV("Enabling HW flow control for RS232 RTS_TOGGLE\n");
            IOPCON1 |= RTS_BIT;
            IOPCON1 |= DTR_BIT;
            IOPCON1 &= ~CTS_BIT;
            S3C4530_uart1.ucon |= S3C4530UartControlHwFlowEn;
        } else {
            IOPCON1 &= ~RTS_BIT;
            PioMode |= RTS_BIT;
            IOPCON1 |= CTS_BIT;
            S3C4530_uart1.ucon &= ~S3C4530UartControlHwFlowEn;
        }

        IOPCON1 &= ~DTR_BIT;
        PioMode |= DTR_BIT;
        break;

    case RS485_MODE:
        if (portMode != RS485_MODE)
            DS("Enabling RS485 Mode \n");
        portMode = mode;
        EpldInterfaceSelect0 = mode;
        IOPCON1 |= RTS_BIT;
        IOPCON1 |= DTR_BIT;
        IOPCON1 &= ~CTS_BIT;
        PioMode &= ~RTS_BIT;
        PioMode &= ~DTR_BIT;
        S3C4530_uart1.ucon |= S3C4530UartControlHwFlowEn;
        break;

    case RS422_MODE:
        if (portMode != RS422_MODE)
            DS("Enabling RS422 Mode \n");
        portMode = mode;
        EpldInterfaceSelect0 = mode;
        IOPCON1 &= ~RTS_BIT;
        IOPCON1 &= ~DTR_BIT;
        IOPCON1 |= CTS_BIT;
        PioMode |= RTS_BIT;
        PioMode |= DTR_BIT;
        S3C4530_uart1.ucon &= ~S3C4530UartControlHwFlowEn;
        break;

    default:
        diag_printf("Invalid Port Mode entered - 0x%x \n", mode);
        return;
    }
}

//  Sets the UART baud rate
static void SetBaudRate(int baud)
{
    if (baudRate != baud) {
        baudRate = baud;
        S3C4530_uart1.brdiv = BaudRegVal(baud);
    }
}

//  Returns the baud rate divisor register value for a given baud rate
static unsigned int BaudRegVal(int baud)
{
    int Mclk2 = IsDM500 ? 48000000/2 : 44236800/2;
    int D,cnt0,cnt1;

    cnt1 = 0;
    D = ((2*Mclk2/(16*baud))+1)/2;

    while (D>4096) {
        ++cnt1;
        D >>= 4;
    }

    cnt0 = (((2*Mclk2 / (16*baud*(1<<(cnt1*4)))) + 1)/2) - 1;
    //diag_printf("uart4530: baud=%d divisor=%d,%d\n", baud, cnt0, cnt1);
    return (cnt0<<4) | cnt1;
}


static void StopBits2(int flag)
{
    int printMessages = 0;

    if (flag == SET) {
        if (printMessages) diag_printf("Setting stop bits to 2 \n");
        S3C4530_uart1.ucon |= S3C4530UartControlStop2;
        stopBits = 2;
    } else {
        if (printMessages) diag_printf("Setting stop bits to 1 \n");
        S3C4530_uart1.ucon &= ~S3C4530UartControlStop2;
        stopBits = 1;
    }
}

static void DataBits(int wordLength)
{
    int printMessages = 0;

    S3C4530_uart1.ucon &= ~S3C4530UartControlWordLenMask;
    switch (wordLength) {
    case CYGNUM_SERIAL_WORD_LENGTH_5:
        if (printMessages) diag_printf("Setting word length to 5 \n");
        S3C4530_uart1.ucon |= S3C4530UartControlWordLen5;
        break;
    case CYGNUM_SERIAL_WORD_LENGTH_6:
        if (printMessages) diag_printf("Setting word length to 6 \n");
        S3C4530_uart1.ucon |= S3C4530UartControlWordLen6;
        break;
    case CYGNUM_SERIAL_WORD_LENGTH_7:
        if (printMessages) diag_printf("Setting word length to 7 \n");
        S3C4530_uart1.ucon |= S3C4530UartControlWordLen7;
        break;
    case CYGNUM_SERIAL_WORD_LENGTH_8:
        if (printMessages) diag_printf("Setting word length to 8 \n");
        S3C4530_uart1.ucon |= S3C4530UartControlWordLen8;
        break;
    }
    dataBits = wordLength;
}

static void Parity(int parity)
{
    int printMessages = 0;

    S3C4530_uart1.ucon &= ~S3C4530UartControlParityMask;
    switch (parity) {
    case CYGNUM_SERIAL_PARITY_ODD:
        if (printMessages) diag_printf("Setting parity to odd \n");
        S3C4530_uart1.ucon |= S3C4530UartControlParityOdd;
        break;
    case CYGNUM_SERIAL_PARITY_EVEN:
        if (printMessages) diag_printf("Setting parity to even \n");
        S3C4530_uart1.ucon |= S3C4530UartControlParityEven;
        break;
    case  CYGNUM_SERIAL_PARITY_NONE:
        if (printMessages) diag_printf("Setting parity to none \n");
        S3C4530_uart1.ucon |= S3C4530UartControlParityNone;
        break;
    }
    paritySetting = parity;
}

int IsCdSet(void)
{
    return (!(PioData & DCD_BIT));
}

static int IsRiSet(void)
{
    return (!(PioData & RI_BIT));
}

static int IsDsrSet(void)
{
    return (S3C4530_uart1.ustat & S3C4530UartStatusDSR);
}

static int IsCtsSet(void)
{
    return (S3C4530_uart1.ustat & S3C4530UartStatusCTS);
}

static int IsRtsSet(void)
{
    return (!(PioData & RTS_BIT));
}

static void Rts(int flag)
{
    if (flag == SET) {
//	DMC("<RTS ON> \n");
        PioData &= ~RTS_BIT;
    } else {
//	DMC("<RTS OFF> \n");
        PioData |= RTS_BIT;
    }
}

static int IsDtrSet(void)
{
    return (!(PioData & DTR_BIT));
}

static void Dtr(int flag)
{
    if (flag == SET) {
        DMC("<DTR ON> \n");
        PioData &= ~DTR_BIT;
    } else {
        DMC("<DTR OFF> \n");
        PioData |= DTR_BIT;
    }
}

static int DidCtsEventOccur(void)
{
    if (S3C4530_uart1.ustat & S3C4530UartStatusCTSEvent) {
        S3C4530_uart1.ustat |= S3C4530UartStatusCTSEvent;
        return 1;
    } else {
        return 0;
    }
}

static void Break(int flag)
{
    if (flag == SET)
        S3C4530_uart1.ucon |= S3C4530UartControlSendBreak;
    else
        S3C4530_uart1.ucon &= ~S3C4530UartControlSendBreak;
}


static int IsRxFifoTrig(void)
{
    return (S3C4530_uart1.ustat & S3C4530UartStatusRxFifoTrig);
}

static int IsRxFifoEmpty(void)
{
    return (S3C4530_uart1.ustat & S3C4530UartStatusRxFifoEmpty);
}

static int IsRxTimeout(void)
{
    return (S3C4530_uart1.ustat & S3C4530UartStatusRxTimeout);
}

static int IsTxFifoFull(void)
{
    return (S3C4530_uart1.ustat & S3C4530UartStatusTxFifoFull);
}

static int IsTxFifoEmpty(void)
{
    return (S3C4530_uart1.ustat & S3C4530UartStatusTxFifoEmpty);
}

#define RxErrorStatus (S3C4530UartStatusRxParityError|S3C4530UartStatusRxFrameError|S3C4530UartStatusRxBreak|S3C4530UartStatusRxOverrunError)

static int IsRxErrors(void)
{
    return S3C4530_uart1.ustat & RxErrorStatus;
}

static int IsRxDataAvail(void)
{
    return (S3C4530_uart1.ustat & S3C4530UartStatusRxDataAvail);
}

static unsigned char RcvByteUart(void)
{
#if EnableDebugIcTimeout
    cyg_tick_count_t ctc;						//for timestamp
    unsigned long long ms = 0;					//for timestamp
    static unsigned long long lastStamp = 0;	//for timestamp

    if (debugICTimeout) {
        ctc = cyg_current_time();
        clkticks2ms(ctc, &mCCR, &ms);
        diag_printf("RcvByteUart: rxByte @ clock %llu (%llums) (last stamp = %llu) GAP = %llu\n", (unsigned long long)ctc, ms, lastStamp, (ms - lastStamp));
        lastStamp = ms;
    }
#endif

    return S3C4530_uart1.urxbuf;
}

static void XmitByteUart(unsigned char c)
{
    S3C4530_uart1.utxbuf = c;
}


//  In software flow control mode, checks for XON and XOFF chars received.
//  Starts/stops the DMA as needed.
//  Returns:  1 if in SW Flow Control mode and control char received, else 0
static int CheckForCtrlChars(void)
{
    unsigned int statReg;
    unsigned char c;

    if (!softwareFlowControlActive)
        return 0;

    statReg = S3C4530_uart1.ustat;
    if (statReg & S3C4530UartStatusCntrlCharDet) {
        S3C4530_uart1.ustat |= S3C4530UartStatusCntrlCharDet;
        c = (RcvByteUart());
        if (c == serialChannel[0].config.rxflow_xon_char) {
            rcvdXoff =0;
            if (dmaXferActive)
                StartDma();
            DFC(" <rcv_Xon> \n");
        } else if (c == serialChannel[0].config.rxflow_xoff_char) {
            rcvdXoff =1;
            if (dmaXferActive)
                StopDma();
            DFC(" <rcv_Xoff> \n");
        } else {
            DFC(" <rcv_other %x> \n", c);
        }
        return 1;
    }
    return 0;
}

//  Checks UART1 status.  Note that printing errors should only
//  be enabled for debug, not release code.  This function is called
//  in the RX data DSR, and sending the printf messages out the debug
//  UART takes time.  This could cause another RX FIFO full IRQ to occur.
//  This must be avoided.
static void GetUartStatusError(void)
{
    unsigned int statReg, printError = 0;

    statReg = S3C4530_uart1.ustat;

    //  Spurious break and frame errors reported when switching to
    //  422 and 485 mode, so only count them when in 232 mode
    if (RxBreak(statReg) && (portMode == RS232_MODE)) {
        uartErrors.rxBreak++;
        if (printError) diag_printf(" Rx Break ");
    }

    if (FramingError(statReg) && (portMode == RS232_MODE)) {
        uartErrors.rxFraming++;
        if (printError) diag_printf(" Rx Framing Error  - ucon:0x%x", S3C4530_uart1.ucon);
    }

    if (ParityError(statReg)) {
        uartErrors.rxParity++;
        if (printError) diag_printf(" Rx Parity Error ");
    }

    if (RxOverrunError(statReg)) {
        uartErrors.rxOverflow++;
        if (printError) diag_printf(" Rx Overrun Error ");
    }
}


//  Check for UART status errors, clears the bit if found
//  Input:  UART status register value
//  Returns:  1 if status error bits set, 0 else
static int RxOverrunError(unsigned int statReg)
{
    if (statReg & S3C4530UartStatusRxOverrunError) {
        S3C4530_uart1.ustat |= S3C4530UartStatusRxOverrunError;
        return 1;
    } else {
        return 0;
    }
}

static int RxBreak(unsigned int statReg)
{
    if (statReg & S3C4530UartStatusRxBreak) {
        S3C4530_uart1.ustat |= S3C4530UartStatusRxBreak;
        return 1;
    } else {
        return 0;
    }
}

static int FramingError(unsigned int statReg)
{
    if (statReg & S3C4530UartStatusRxFrameError) {
        S3C4530_uart1.ustat |= S3C4530UartStatusRxFrameError;
        return 1;
    } else {
        return 0;
    }
}

static int ParityError(unsigned int statReg)
{
    if (statReg & S3C4530UartStatusRxParityError) {
        S3C4530_uart1.ustat |= S3C4530UartStatusRxParityError;
        return 1;
    } else {
        return 0;
    }
}

//  Flush the UART FIFOs
//  NOTE:  Flushing (resetting) the TX FIFO can cause the transmitter
//         to stop functioning.  An alternative method is used instead.
static void FlushRxFifo(void)
{
    DRV("Flushing RX FIFO \n");
    if (S3C4530_uart1.ucon & S3C4530UartControlRxFifoEn) {
        S3C4530_uart1.ucon |= S3C4530UartControlRxFifoRst;
        S3C4530_uart1.ucon &= ~S3C4530UartControlRxFifoRst;
    }
}

static void FlushTxFifo(void)
{
    unsigned char c;

    DRV("Flushing TX FIFO \n");

    //  If flow control is active, flush by stopping DMA and clearing the active flag.
    if ((softwareFlowControlActive) || (hardwareFlowControlActive)) {
        if (dmaXferActive) {
            StopDma();
            dmaXferActive = 0;
        }
    }

    //  If no flow control is active, flush by  stopping transmitter, put UART in loopback mode, restart transmitter
    else {
        S3C4530_uart1.ucon &= ~S3C4530UartControlTxModeMask;
        S3C4530_uart1.ucon |= S3C4530UartControlLoopback;
        S3C4530_uart1.ucon |= S3C4530UartControlTxModeInt;
        S3C4530_uart1.brdiv = BaudRegVal(460800);

        //  Flush TX FIFO bytes by reading bytes
        while (!(S3C4530_uart1.ustat & S3C4530UartStatusTxFifoEmpty))
            c = RcvByteUart();

        SetupUart(FLOW_CONTROL_NONE);
        S3C4530_uart1.brdiv = BaudRegVal(baudRate);
    }
}

static int IsTxHoldingRegEmpty(void)
{
    return (S3C4530_uart1.ustat & S3C4530UartStatusTxHoldRegEmpty);
}

static int IsTxComplete(void)
{
    return (S3C4530_uart1.ustat & S3C4530UartStatusTxComplete);
}

//  Functions which calculate number of bytes or remaining room in the
//  serial drivers in/out cbuf.  Since the S3C4530 CPU UART does not provide
//  a method to count how many bytes are in the TX or RX FIFO, and we must
//  do flow control manually, we use the number of bytes in the cbuf's.
static bool IsReceiveBufferFull(void)
{
    cbuf_t *cbuf = &serialChannel[0].in_cbuf;

    return ((cbuf->put == cbuf->get - 1) ||
            ((cbuf->put == BufSize - 1) && (cbuf->get == 0)));
}

static int ReceiveBufferRoom(void)
{
    cbuf_t *cbuf = &serialChannel[0].in_cbuf;
    int room = (cbuf->get - cbuf->put)-1;
    if (room <= 0)
        room += BufSize;
    return room;
}

static int TransmitBufferCount(void)
{
    cbuf_t *cbuf = &serialChannel[0].out_cbuf;
    int count = cbuf->put - cbuf->get;
    if (count < 0)
        count += BufSize;
    return count;
}

static int ReceiveBufferCount(void)
{
    cbuf_t *cbuf = &serialChannel[0].in_cbuf;
    int count = cbuf->put - cbuf->get;
    if (count < 0)
        count += BufSize;
    return count;
}

//  Receives all bytes from the RxFIFO, puts it in the serial channel in buffer.
void ReceiveBytes(void)
{
    int wakeup = 0;  // Flag that tells us to wakeup the serial_read() function
    int syncup = 0;  // Flag that tells us to wakeup NSLink polling loop
    cbuf_t *cbuf = &serialChannel[0].in_cbuf;
    int rxroom = ReceiveBufferRoom();

    //  Raise RTS if HW flow control and we've got lots of room
    if (hardwareFlowControlActive && !IsRtsSet() && ReceiveBufferRoom() > 2000)
        Rts(SET);

    //In SW flowcontrol mode, check buffer size.  Send XON/XOFF as needed.
    if (softwareFlowControlActive) {
        if (rxroom < 200) {
            if (!xoffActive) {
                SendControlChar(XOFF);
                DFC(" <snd_XOFF> \n");
                xoffActive = 1;
            }
        } else if (rxroom > 2000) {
            if (xoffActive) {
                SendControlChar(XON);
                DFC(" <snd_XON> \n");
                xoffActive = 0;
            }
        }
    }

    if (IsRxDataAvail() || IsRxErrors()) {
        wakeup = 1;
        gotData = 1;
        ictoCounter = 0;
    }

    //  Retrieve the data bytes.  Note that if there is a UART error, we
    //  still pass the byte to the application.  This is the way SI,
    //  Windows and Linux serial drivers work, and we had a customer
    //  problem in this area.
    while (IsRxDataAvail() || IsRxErrors()) {
        cyg_uint8 rxbyte;

        //  Check for incoming XON/XOFF chars
        if (CheckForCtrlChars())
            continue;

        //  If data error, log  it
        if (IsRxErrors())
            GetUartStatusError();

        // Check for a full queue.  If full, exit
        if (rxroom == 0)
            break;

        //  go get byte

        rxbyte = RcvByteUart();
        cbuf->data[cbuf->put++] = rxbyte;


        //  Circular queue, check for wrap
        if (cbuf->put == BufSize)
            cbuf->put = 0;

        --rxroom;
    }


    //  Drop RTS if we're getting low on buffer space
    if (hardwareFlowControlActive && IsRtsSet() && ReceiveBufferRoom() < 200)
        Rts(CLEAR);

    //  Wake up rpshsi polling loop?
    if (syncup)
        cyg_drv_cond_broadcast(&sync.wait);

    //  Wake up pending read()?
    if (wakeup && cbuf->waiting) {
        cbuf->waiting = false;
        cyg_drv_cond_signal(&cbuf->wait);
    }
}


//  When no flow control is active, this function transmits bytes.  The serial layer
//  pushes serial bytes from the application into the cbuf circular buffer. This function
//  transmits those bytes by pushing them into the TX FIFO until full.  This function is
//  called initially from the 10ms poll routine.  The TX complete ISR calls this
//  function again, in case there is new data to transmit.
static void XmitBytes(void)
{
    cbuf_t *cbuf = &serialChannel[0].out_cbuf;

    //  If the put and get pointers are different, there is data to send
    while (cbuf->put != cbuf->get) {
        //  Fill FIFO
        if (!IsTxFifoFull()) {
            XmitByteUart(cbuf->data[cbuf->get++]);
            if (cbuf->get == BufSize)
                cbuf->get = 0;
        } else {
            break;
        }
    }

    if (cbuf->waiting) {
        cbuf->waiting = false;
        cyg_drv_cond_signal(&cbuf->wait);
    }
}


//  When hardware or software flow control is active, this function transmits
//  bytes.  The serial layer pushes serial bytes from the application into the
//  cbuf circular buffer.  This function uses the DMA to move data from the
//  cbuf to the UART.  This function is called initially from the 10ms poll
//  routine.  After the DMA finishes the transfer, the DMA complete ISR
//  calls this function again, in case there is new data to transmit.
static void XmitBytesDma(void)
{
    cbuf_t *cbuf = &serialChannel[0].out_cbuf;

    //  Three cases possible
    if (cbuf->put == cbuf->get) {
        // buffer empty
        return;
    } else if (cbuf->get < cbuf->put) {
        //  Data to be sent
        S3C4530_dma0.saddr = (unsigned int) &cbuf->data[cbuf->get];
        S3C4530_dma0.xfercount = (cbuf->put - cbuf->get);
        nextGet = cbuf->put;
        StartDma();
        dmaXferActive = 1;
    } else if (cbuf->get > cbuf->put) {
        //  Data to be sent, at end of buffer, wrap pointer
        S3C4530_dma0.saddr = (unsigned int) &cbuf->data[cbuf->get];
        S3C4530_dma0.xfercount = (BufSize - cbuf->get);
        nextGet = 0;
        StartDma();
        dmaXferActive = 1;
    } else {
        //  Should never get here...
        diag_printf("XmitBytesDma() ERROR \n");
    }

}


//  Serial polling loop, called every scanPeriod ms from the
//  alarm.  Used to start the TX process, and re-start the RX
//  process when HW flow control is active. In SW flowcontrol,
//  used to check for buffer room and send XON if XOFF'd. Also
//  used to kickstart the transmit process.

static void uart4530_serial_poll(void)
{
    cbuf_t *cbuf_out = &serialChannel[0].out_cbuf;
    //  If HW Flow Control active and our RTS is cleared, call
    //  ReceiveBytes.  This restarts the receive process, if we
    //  had flowed it off by dropping RTS.
    if (hardwareFlowControlActive && !IsRtsSet() && !IsReceiveBufferFull())
        ReceiveBytes();


    if (softwareFlowControlActive) {
        //  SW Flow Control
        //  Don't transmit data if we are XOFF'd off, or if a
        //  DMA transfer is already in progress
        if (!dmaXferActive && !rcvdXoff)
            XmitBytesDma();
    } else if (hardwareFlowControlActive) {
        //  HW Flow Control
        //  Don't transmit data if CTS is not set, or if a DMA
        //  transfer is already in progress
        if (!dmaXferActive && IsCtsSet())
            XmitBytesDma();
    } else {
        //  No flow control active
        if (IsTxFifoEmpty() && (cbuf_out->get != cbuf_out->put))
            XmitBytes();
    }

    // need to wake up driver if it's waiting on output drain and
    // we're done transmitting

    if (cbuf_out->waiting && IsTxComplete() && IsTxFifoEmpty() && !dmaXferActive) {
        cbuf_out->waiting = false;
        cyg_drv_cond_signal(&cbuf_out->wait);
    }


    if (softwareFlowControlActive && xoffActive && (ReceiveBufferRoom() > 300)) {
        SendControlChar(XON);
        xoffActive = 0;
    }

    //  This is required to wake up rpshi, which is waiting.  This syncs
    //  up polling here and at the application level
        cyg_drv_cond_broadcast(&sync.wait);
}

//  The 1ms Poll function.  Called every ms by the timer1 alarm:
//  -  If intercharacter timeout is enabled, acts as the icto timer
//  -  Handle soft RTS toggle with pre/post TX RTS hold times
static void msPoll(cyg_handle_t alarm_handle, cyg_addrword_t data)
{
    static int pollCounter;  //  Counts loops so serial poll routine gets called every <scanPeriod> ms

    ++pollCounter;


    if (softRtsToggle) {
        // handle software RTS toggle with pre/post tx RTS hold options
        unsigned txcount = TransmitBufferCount() + !IsTxComplete();
        unsigned rtsState = IsRtsSet();
        unsigned txEnabled = (S3C4530_uart1.ucon & S3C4530UartControlTxModeMask) == S3C4530UartControlTxModeInt;

        if (txcount && !txEnabled) {
            // There's data waiting to be sent, but we're not transmitting

            if (!rtsState) {
                // RTS isn't on, so turn it on and start countdown 'til transmit
                Rts(SET);
                if (preRtsHoldTime)
                    rtsCounter = preRtsHoldTime;
                else
                    S3C4530_uart1.ucon = (S3C4530_uart1.ucon & ~S3C4530UartControlTxModeMask) | S3C4530UartControlTxModeInt;
            } else {
                // RTS is on, so we're counting down 'til transmit on
                rtsCounter--;
                if (rtsCounter == 0)
                    S3C4530_uart1.ucon = (S3C4530_uart1.ucon & ~S3C4530UartControlTxModeMask) | S3C4530UartControlTxModeInt;
            }
        }

        if (!txcount && rtsState) {
            // Done sending data and RTS is still on
            if (txEnabled) {
                // tx enable is still on, so turn it off and start countdown 'til RTS off
                S3C4530_uart1.ucon = S3C4530_uart1.ucon & ~S3C4530UartControlTxModeMask;
                if (postRtsHoldTime)
                    rtsCounter = postRtsHoldTime;
                else
                    Rts(CLEAR);
            } else {
                // tx enable is off, so we're counting down 'til RTS goes off
                rtsCounter--;
                if (rtsCounter == 0)
                    Rts(CLEAR);
            }
        }
    }


    //  If intercharacter timeout value is non-zero, icto is
    //  active.  When gotData occurs, a byte has been
    //  received, start incrementing ictoCounter.

    if (intercharTimeoutValue) {
        ReceiveBytes();
        if (gotData) {
            ++ictoCounter;

            if (ictoCounter > intercharTimeoutValue) {
                // inter-char timeout has happened
                serialChannel[0].rxInterCharTimeoutFlag = 1;

                //  Wake up rpshsi polling loop
                cyg_drv_cond_broadcast(&sync.wait);

                //  Wake up pending read
                if (serialChannel[0].in_cbuf.waiting) {
                    serialChannel[0].in_cbuf.waiting = false;
                    cyg_drv_cond_signal(&serialChannel[0].in_cbuf.wait);
                }

                gotData = 0;
                ictoCounter = 0;

#if EnableDebugIcTimeout
                if (debugICTimeout)
                    diag_printf("\n\nmsPoll: interCharTimeout (%u) triggered (idle = %u) @ clock %llu\n\n",
                                intercharTimeoutValue, ictoCounter, (unsigned long long)cyg_current_time());
#endif
            }
        }
    }

    if (pollCounter >= scanPeriod) {
        uart4530_serial_poll();
        pollCounter = 0;
    }
}

unsigned uart4530_rxISRcount;
unsigned uart4530_rxDSRcount;

// Interrupt handler (ISR and DSR) for the UART Rx.  Triggered
// by the RX trig, timeout or CTS Event IRQs.  If read data is
// present, reads data from RX FIFO, stores in in cbuf buffer.
// If CTS Event occurred, turns DMA off/on (HW Flow Control).
static cyg_uint32 UartRxIsr(cyg_vector_t vector, cyg_addrword_t data)
{
    cyg_drv_interrupt_acknowledge(CYGNUM_HAL_INTERRUPT_Uart1Rx);
    ++uart4530_rxISRcount;
    return CYG_ISR_HANDLED + CYG_ISR_CALL_DSR;
}


static void UartRxDsr(cyg_vector_t vector, cyg_ucount32 count, cyg_addrword_t data)
{
    ++uart4530_rxDSRcount;

    if (IsRxFifoTrig() || IsRxDataAvail() || IsRxErrors() || IsRxTimeout())
        ReceiveBytes();

    S3C4530_uart1.ustat = S3C4530UartStatusRxTimeout;

    // The CTS Event interrupt is handled here, used when HW
    // Flow Control is active.
    if (DidCtsEventOccur()) {
        if (hardwareFlowControlActive && dmaXferActive) {
            if (IsCtsSet()) {
                DFC(" <CTS ON> \n");
                StartDma();
            } else {
                DFC(" <CTS OFF> \n");
                StopDma();
            }
        }
    }
}

// Interrupt handler (ISR and DSR) for the UART Tx.  Triggered
// by the TxHoldReg empty IRQ.  Fills the TxFIFO with data
// from the out buffer.
static cyg_uint32 UartTxIsr(cyg_vector_t vector, cyg_addrword_t data)
{
    cyg_drv_interrupt_acknowledge(CYGNUM_HAL_INTERRUPT_Uart1Tx);
    return CYG_ISR_HANDLED + CYG_ISR_CALL_DSR;
}

static void UartTxDsr(cyg_vector_t vector, cyg_ucount32 count, cyg_addrword_t data)
{
    XmitBytes();
}

// Interrupt handler (ISR and DSR) for the DMA.  Triggered by
// the DMA ending it's programmed transfer.  Used during SW
// flow control.
static cyg_uint32 DmaIsr(cyg_vector_t vector, cyg_addrword_t data)
{
    cyg_drv_interrupt_acknowledge(CYGNUM_HAL_INTERRUPT_Dma0);
    return CYG_ISR_HANDLED + CYG_ISR_CALL_DSR;
}

static void DmaDsr(cyg_vector_t vector, cyg_ucount32 count, cyg_addrword_t data)
{
    cbuf_t *cbuf = &serialChannel[0].out_cbuf;
    dmaXferActive = 0;
    cbuf->get = nextGet;
    if (cbuf->waiting) {
        cbuf->waiting = false;
        cyg_drv_cond_signal(&cbuf->wait);
    }
    XmitBytesDma();
}

static void EnableHwFlowControl(void)
{
    if (!hardwareFlowControlActive) {
        DFC("Enabling HW Flow Control ** \n");
        hardwareFlowControlActive = 1;
        softwareFlowControlActive = 0;
        dmaXferActive = 0;
        SetupUart(FLOW_CONTROL_HARDWARE);

        //  Enable DMA IRQ, disable UART TX IRQ
        cyg_drv_interrupt_mask(CYGNUM_HAL_INTERRUPT_Uart1Tx);
        cyg_drv_interrupt_unmask(CYGNUM_HAL_INTERRUPT_Dma0);
    }
}

static void DisableHwFlowControl(void)
{
    if (hardwareFlowControlActive) {
        DFC("Disabling HW Flow Control ** \n");
        hardwareFlowControlActive = 0;
        dmaXferActive = 0;
        StopDma();
        SetupUart(FLOW_CONTROL_NONE);
        Rts(CLEAR);

        //  Disable DMA IRQ, enable UART TX IRQ
        cyg_drv_interrupt_unmask(CYGNUM_HAL_INTERRUPT_Uart1Tx);
        cyg_drv_interrupt_mask(CYGNUM_HAL_INTERRUPT_Dma0);
    }
}

static void EnableSwFlowControl(void)
{
    if (!softwareFlowControlActive) {

        DFC("Enabling SW Flow Control ** \n");
        softwareFlowControlActive = 1;
        hardwareFlowControlActive = 0;
        dmaXferActive = 0;
        SetupUart(FLOW_CONTROL_SOFTWARE);
        Rts(CLEAR);

        //  Enable DMA IRQ, disable UART TX IRQ
        cyg_drv_interrupt_mask(CYGNUM_HAL_INTERRUPT_Uart1Tx);
        cyg_drv_interrupt_unmask(CYGNUM_HAL_INTERRUPT_Dma0);
    }
}

static void DisableSwFlowControl(void)
{
    if (softwareFlowControlActive) {
        DFC("Disabling SW Flow Control ** \n");

        softwareFlowControlActive = 0;
        dmaXferActive = 0;
        StopDma();
        SetupUart(FLOW_CONTROL_NONE);

        //  Disable DMA IRQ, enable UART TX IRQ
        cyg_drv_interrupt_unmask(CYGNUM_HAL_INTERRUPT_Uart1Tx);
        cyg_drv_interrupt_mask(CYGNUM_HAL_INTERRUPT_Dma0);
    }
}


//  Configure the UART based on the flow control mode.
static void SetupUart(int mode)
{
    static int firstTimeThrough = 1;

    unsigned InitialControlValue = S3C4530UartControlParityNone +
                                   S3C4530UartControlWordLen8 +
                                   S3C4530UartControlRxModeInt +
                                   (RxFifoEnable ? S3C4530UartControlRxFifoEn : 0);

    unsigned InitialIntEnaValue = RxFifoEnable ?
                                  S3C4530UartIntEnaRxFifoTrig +
                                  S3C4530UartIntEnaRxTimeout
                                  :
                                  S3C4530UartIntEnaRxData +
                                  S3C4530UartIntEnaParityErr +
                                  S3C4530UartIntEnaFrameErr +
                                  S3C4530UartIntEnaBreak +
                                  S3C4530UartIntEnaOverrun;

    if (mode == FLOW_CONTROL_HARDWARE) {
        S3C4530_uart1.ucon  = InitialControlValue +
                              S3C4530UartControlTxModeGDMA0;

        S3C4530_uart1.inten = InitialIntEnaValue +
                              S3C4530UartIntEnaCTS;
    } else if (mode == FLOW_CONTROL_SOFTWARE) {
        //  Disable TX FIFO, set UART1 TX source to DMA

        S3C4530_uart1.ucon  = InitialControlValue +
                              S3C4530UartControlTxModeGDMA0 +
                              S3C4530UartControlSwFlowEn;

        S3C4530_uart1.inten = InitialIntEnaValue;
    } else if (mode == FLOW_CONTROL_NONE) {
        //  First time through, start with DTR and RTS asserted.  After that, preserve what application had set.
        if (firstTimeThrough) {
            firstTimeThrough = 0;
            if (!(configFlags & CYG_SERIAL_FLAGS_SOFT_RTS_TOGGLE))
                Rts(SET);
            Dtr(SET);
        }

        if (configFlags & CYG_SERIAL_FLAGS_SOFT_RTS_TOGGLE) {
            // start with tx disabled
            Rts(CLEAR);
            S3C4530_uart1.ucon  = InitialControlValue +
                                  S3C4530UartControlTxModeDis +
                                  S3C4530UartControlTxFifoEn;
        } else {
            // start with tx enabled
            S3C4530_uart1.ucon  = InitialControlValue +
                                  S3C4530UartControlTxModeInt +
                                  S3C4530UartControlTxFifoEn;
        }

        /*  This cleans up after functions that change the IOPCON, UCON or PIOMODE registers directly, like TxFlush() */
        if (portMode == RS485_MODE)
            SetPortMode(RS485_MODE);

        if ((portMode == RS232_MODE || portMode == RS232I_MODE) && (configFlags & CYG_SERIAL_FLAGS_RTS_TOGGLE))
            S3C4530_uart1.ucon  |= S3C4530UartControlHwFlowEn;

        S3C4530_uart1.inten = InitialIntEnaValue +
                              S3C4530UartIntEnaTxEmpty;
    } else {
        diag_printf("SetupUart() - Invalid parameter \n");
        return;
    }

    //  Restore bit/stop/parity settings
    if (stopBits == 2)
        StopBits2(SET);
    else
        StopBits2(CLEAR);
    DataBits(dataBits);
    Parity(paritySetting);

    S3C4530_uart1.ucon &= ~S3C4530UartControlRxFifoTrigMask;
    S3C4530_uart1.ucon |= S3C4530UartControlRxFifoTrig18;
}

//  Configure the DMA control register.  Values used:
//  -  Disabled
//  -  Mode UART1 block
//  -  Source Address Fix (the UART)
//  -  Stop interrupt enable
//  -  Transfer direction - memory to UART
static void InitDmaControl(void)
{
    S3C4530_dma0.control = 0x54C;
}

//  Start and stop the DMA, due to reciept of XON or XOFF.
//  Note that toggling the run control bit (bit 0) should do it,
//  but it doesn't.  Must slam the entire control register to get
//  it to stop.
static void StartDma(void)
{
    S3C4530_dma0.control = 0x54D;
}

static void StopDma(void)
{
    S3C4530_dma0.control = 0;
}

//  In SW flowcontrol mode, send an XON or XOFF.  To do this,  stop the
//  DMA, switch to regular UART mode.  Push the control char into UART,
//  switch back to DMA mode.
static void SendControlChar(int controlChar)
{
    if (softwareFlowControlActive) {
        StopDma();
        SetupUart(FLOW_CONTROL_NONE);
        if (controlChar == XON)
            XmitByteUart(serialChannel[0].config.txflow_xon_char);
        else if (controlChar == XOFF)
            XmitByteUart(serialChannel[0].config.txflow_xoff_char);
        else
            diag_printf("SendControlChar - parameter error \n");
        while (!IsTxFifoEmpty())
            ;  //  Wait for byte to be sent
        SetupUart(FLOW_CONTROL_SOFTWARE);

        //  Restart DMA if it was active before
        if ((dmaXferActive) && (!xoffActive))
            StartDma();
        else
            InitDmaControl();
    } else {
        diag_printf("SendControlChar - not in Software Flow Control mode \n");
    }
}

/***************************************************************
  NOTES:

 The current implementation msPoll() appears to handle the icto mech fine
 per tests ran against a DMRTS 1 port using variable gap sizes in the
 data sent to the unit. If the pause exceeds your set threshold then
 the icto section of msPoll is triggered as you would expect. I am guessing
 that some other assertions stated in this note body may also false but
 tests of each would have to prove otherwise to be sure.

 This driver uses the S3C4530 UART configured with TX and RX FIFOs
 enabled, TX and RX IRQ used.  The part has HW flow control built-in,
 but this does not work correctly.  Specifically, when HW flow control
 is enabled in the UART: - RTS is drive high by the TX section when
 data is being transmitted, otherwise it is low.  - CTS is monitored
 by the TX section, transmittion stops when CTS is low.

 This mode is incorrect for RTS-CTS flow control, but is half correct.  The
 CTS monitoring done by the TX section is correct behaviour, but RTS should
 be controlled by the RX section.  Therefore, this driver uses the UART in
 the following manner when HW flow control is active:
 -  HW flow control in the UART enabled
 -  RTS is not controlled by the UART, but by this driver as an IO pin
 -  The receive portion of this driver controls RTS

 This half correct HW flow control mode also is used during RS485 mode.
 Since the UART drives RTS high while transmitting data, this is used
 for 485 mode.  In order to avoid it starting/stopping transmission based
 on CTS, that pin is configured as an (unused) IO pin.

 The built-in hw flow control mode is also used when RTS-toggle is enabled.

 Software flow control does not work in the UART.  Instead it is performed
 in software.

 Due to limitations in the UART, callbacks are not supported.  There are no
 interrupts for CD DSR DTR changes.

 Due to limitations in the UART, the TX FIFO flush function does not use
 the UART TX reset function.  Rather, the TX FIFO is drained using the
 loopback mode.

 The port-mode and flow-control code is messed up.  It's both mixed
 together and scattered throughout several different functions.  In
 some places flow-control is set up based on port-mode settings
 instead of flow-control settings.  That means that the two features
 are not as orthogonal as they should be and only certain combinations
 will work.

***************************************************************/


void show4530stats(void)
{
    extern unsigned uart4530_rxISRcount,uart4530_rxDSRcount;
    diag_printf("RxFifoEnable: %d\n", RxFifoEnable);
    diag_printf("rx ISR count: %d\n", uart4530_rxISRcount);
    diag_printf("rx DSR count: %d\n", uart4530_rxDSRcount);
    uart4530_rxISRcount = 0;
    uart4530_rxDSRcount = 0;
    diag_printf("ustat: %08x\n", S3C4530_uart1.ustat);
    diag_printf(" ucon: %08x\n", S3C4530_uart1.ucon);
    diag_printf("inten: %08x\n", S3C4530_uart1.inten);
    //RxFifoEnable = !RxFifoEnable;
    //diag_printf("RxFifoEnable <- %d\n", RxFifoEnable);
}


#if EnableDebugIcTimeout
/*
 *  (nanoseconds * divisor)/dividend = clock_ticks
 *  (clock_ticks * dividend)/divisor = nanoseconds
 *  nanoseconds / 1,000,000			 = milliseconds
 */
static void clkticks2ms(cyg_tick_count_t ctc, cyg_resolution_t *ccr, unsigned long long *n)
{
    unsigned long long ms = 0, ticks = 0, divd = 0, divs = 0;

    ticks	= (unsigned long long)ctc;
    divd	= (unsigned long long)ccr->dividend;
    divs	= (unsigned long long)ccr->divisor;

    ms = 0;
    ms = ((ticks * divd)/divs)/1000000;
    *n = ms;
//	diag_printf_m("((ticks=%llu * divd=%llu)/divs=%llu)/1000000ns : ret = %llums (%llu) \n", ticks, divd, divs, ms, *n);

}//end clkticks2ms()
#endif
